/*
 *
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 * Portions copyright (C) 1999 Michael Clark
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "emelfm2.h"
//#include <unistd.h>
#include <langinfo.h>
#include <dirent.h>
#include <string.h>
#include "e2_plugins.h"

// binary factor
#define BFAC 1024.0
#define BFAC2 1048576.0
#define BFAC3 1073741824.0

#ifdef E2_TREEWALK
typedef struct
{
	guint64 total;
	guint64 files;
	guint64 dirs;
} E2_Du;

/**
 * @brief update du counters
 * This is a callback for the treewalk function
 * @param local_name path of item reported by the walker, localised string
 * @param statbuf pointer to struct stat with data about @a local_name
 * @param status code from the walker, indicating what type of item is being reported
 * @param user_data pointer to data struct with total, files, dirs counters
 * @return E2TW_CONTINUE always
 **/
static E2_TwResult _e2p_du_twcb (const gchar *local_name,
	const struct stat *statbuf, E2_TwStatus status, E2_Du *user_data)
{
	switch (status)
	{
		case E2TW_F:	//not directory or link
		case E2TW_SL:	//symbolic link
		case E2TW_SLN:	//symbolic link naming non-existing file
			user_data->files++;
			user_data->total += statbuf->st_size;
			break;
		case E2TW_DL:	//directory, not opened due to tree-depth limit
		case E2TW_DM:	//directory, not opened due to different file system
		case E2TW_DNR:	//unreadable directory
			e2_output_print_errno ();
			F_DECLARE;
			gchar *utf = F_FILENAME_FROM_LOCALE ((gchar *)local_name);
			gchar *msg = g_strdup_printf ("%s %s\n",
				_("Warning: couldn't open directory"), utf);
			e2_output_print_default (msg);
			F_FREE (utf);
			g_free (msg);
		case E2TW_D:	//directory
			user_data->dirs++;
			user_data->total += statbuf->st_size;
			break;
//		case E2TW_DP:	//directory, all subdirs have been visited
//		case E2TW_NS:	//un-statable item WARNING ??
		default:
			break;
	}
	return E2TW_CONTINUE;
}
#else
//filename is localised, not necessarily utf-8
//this func largely works on localised filenames,
//so avoids glib usage unless explicitly converted to utf
static void _e2p_du_add_usage (gchar *filename, guint64 *total,
	guint64 *files, guint64 *dirs)
{
	struct stat statbuf;

	if (e2_fs_lstat (filename, &statbuf))
	{
		e2_output_print_errno();
		return;
	}
	(*total) += statbuf.st_size;
	if (S_ISDIR (statbuf.st_mode))
	{
		DIR *dp;
		struct dirent entry_store;
		struct dirent *entry;
		struct stat statbuf;
		gchar path[PATH_MAX];	//this is localised string

		(*dirs)++;
		if ((dp = opendir (filename)) == NULL) //FIXME vfs
		{
			e2_output_print_errno();
			F_DECLARE;
			gchar *utf = F_FILENAME_FROM_LOCALE (filename);
			gchar *msg = g_strdup_printf ("%s %s\n",
			_("Warning: couldn't open directory"), utf);
			e2_output_print_default (msg);
			F_FREE (utf);
			g_free (msg);
			return;
		}

		while (!readdir_r (dp, &entry_store, &entry) && entry != NULL) //FIXME vfs
		{
			if (strcmp (entry->d_name, ".") == 0 || strcmp (entry->d_name, "..") == 0)
				continue;

			snprintf (path, PATH_MAX, "%s"G_DIR_SEPARATOR_S"%s", filename, entry->d_name);
			if (e2_fs_lstat (path, &statbuf))
			{
				e2_output_print_errno();
				continue;
			}

			(*total) += statbuf.st_size;
			if (S_ISDIR (statbuf.st_mode))
				_e2p_du_add_usage (path, total, files, dirs);	//recurse dirs
			else
				(*files)++;
		}
		closedir (dp); //FIXME vfs
	}
	else
		(*files)++;
}
#endif //def E2_TREEWALK
static void
_e2p_du (void)
{
	FileInfo *info;
#ifdef E2_TREEWALK
	guint64 total, files, dirs;
#else
	guint64 total = 0, files = 0, dirs = 0;
#endif
	GString *text;
	GList *base, *tmp;
	F_DECLARE;
	gchar *local = F_FILENAME_TO_LOCALE (curr_view->dir);
	gchar *local_path;
#ifdef E2_TREEWALK
	//all counters initialized to 0
	E2_Du *cbdata = g_malloc0 (sizeof (E2_Du));
#endif
	base = tmp = e2_fileview_get_selected_local (curr_view);
	for (; tmp != NULL; tmp = tmp->next)
	{
		info = tmp->data;
//		printd (DEBUG, "fn: %s", info->filename);
		local_path = e2_utils_strcat (local, info->filename);
#ifdef E2_TREEWALK
		e2_fs_tw (local_path, _e2p_du_twcb, cbdata, -1, E2TW_PHYS);
#else
		_e2p_du_add_usage (local_path, &total, &files, &dirs);
#endif
		g_free (local_path);
	}
	F_FREE (local);
#ifdef E2_TREEWALK
	total = cbdata->total;
	files = cbdata->files;
	dirs = cbdata->dirs;
	g_free (cbdata);
#endif
	g_list_free (base);

	static gchar big[3] = { '1', ',', '\0' };
	gchar *comma = nl_langinfo (THOUSEP);
	if (comma != NULL && *comma != '\0')
		big[1] = *comma;
	text = g_string_new(_("total size: "));
	gint fwidth;

	if (total < BFAC)
	{
		gchar *b = _("bytes");
		if (total < 1000)
			g_string_append_printf (text, "%"G_GUINT64_FORMAT" %s", total, b);
		else
		{
			total -= 1000;
	    	g_string_append_printf (text, "%s%03"G_GUINT64_FORMAT" %s", big, total, b);
		}
	}
	else if (total < BFAC2)
	{
		gchar *kb = _("kilobytes");
		fwidth = (total < 10 * BFAC) ? 3 : 2;

		if ((total/BFAC) < 1000)
			g_string_append_printf(text, "%.*f %s", fwidth, (total / BFAC), kb);
		else
		{
			total -= (1000 * BFAC);
	    	g_string_append_printf (text, "%s%04.1f %s", big, (total / BFAC), kb);
		}
	}
	else if (total < BFAC3)
	{
		gchar *mb = _("Megabytes");
		fwidth = (total < 10 * BFAC2) ? 3 : 1;

		if ((total/BFAC2) < 1000)
			g_string_append_printf(text, "%.*f %s", fwidth, (total / BFAC2), mb);
		else
		{
			total -= (1000 * BFAC2);
			g_string_append_printf (text, "%s%04.1f %s", big, (total / BFAC2), mb);
		}
	}
	else
	{
		gchar *gb = _("gigabytes");
		fwidth = (total < 10 * BFAC3) ? 3 : 1;

		if ((total/BFAC3) < 1000)
			g_string_append_printf(text, "%.*f %s", fwidth, (total / BFAC3), gb);
		else
		{
			total -= (1000 * BFAC3);
			g_string_append_printf (text, "%s%04.1f %s", big, (total / BFAC3), gb);
		}
	}

	gchar *filetext = (files == 1) ? _("file") : _("files");
	gchar *dirtext = (dirs == 1) ? _("directory") : _("directories");
	g_string_append_printf (text, _("\nin %"G_GUINT64_FORMAT" %s and %"G_GUINT64_FORMAT" %s\n"),
		files, filetext, dirs, dirtext);
	status_message (text->str);
	status_message2 (_("----end-of-output----\n"), FALSE, "small", "grey", NULL);
	g_string_free (text, TRUE);
}

//aname must be confined to this module
static gchar *aname;
/**
 * @brief plugin initialization function, called by main program
 *
 * @param p ptr to plugin data struct
 *
 * @return TRUE if the initialization succeeds, else FALSE
 **/
gboolean
init_plugin (Plugin *p)
{
#define ANAME "du"
	aname = _("du");

	p->signature = ANAME VERSION;
	p->menu_name = _("_Disk usage");
	p->description = _("Calculate the 'apparent' disk usage of selected item(s)");
	p->icon = ICON_DIR"plugin_"ANAME"_"E2IP".png";  //icon file pathname

	if (p->plugin_cb == NULL)
	{
		p->plugin_cb = _e2p_du;  //set plugins menu item function
//		E2_Action* (*action_register) () = (*p->api_lookup) (E2API_ACTION_REGISTER);
		//no need to free this
		gchar *action_name = g_strconcat (PLUGIN,".",aname,NULL);
		//NULL = no data, FALSE = no argument for the function
//		(*action_register)
		e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM, _e2p_du, NULL, FALSE);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief unload this plugin
 *
 * @param p pointer to data struct for the plugin
 *
 * @return
 **/
void
unload_plugin (Plugin *p)
{
//	if (p->api_lookup != NULL)
//	{
//		gboolean (*action_unregister) () = (*p->api_lookup) (E2API_ACTION_UNREGISTER);
		gchar *action_name = g_strconcat (PLUGIN,".",aname,NULL);
//		(*action_unregister)
		e2_action_unregister (action_name);
		g_free (action_name);
//	}
}
