/*
 *
 * Copyright (C) 2005 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file plugins/e2p_rename_ext.c 
 * @brief file-rename plugin
 *
 * This file contains functions related to creation of a file-rename, and
 * execution of a find & rename task in accord with options selected in
 * that dialog. Back-end capability is provided by shell commands, bash,
 * specifically GNU find, sed, grep, mv. Some 'find' command parameters 
 * are dependant on a latish version e.g. 4.2.11 is ok
 **/

/* TODO
FIXME's
hard = detect end of shell command, and update start/stop/help btn sensitivity accordingly
make flags & entry contents persist between sessions ?
share code with the find plugin ?
*/

#include "emelfm2.h"
//#include <unistd.h>
#include <string.h>
#include "e2p_rename_ext.h"
#include "e2_dialog.h"
#include "e2_task.h"
#include "e2_filelist.h"

  /*********************/
 /***** callbacks *****/
/********************/
/**
 * @brief show some help text
 *
 * @param widget clicked button, UNUSED
 * @param rt ptr to dialog data struct
 *
 * @return
 **/
static void _e2pr_help_cb (GtkWidget *widget, E2_RenDialogRuntime *rt)
{
	gchar *msg =
	_("Full documentation on regular expressions can be found by running\n"
	" the shell command \"man grep\" and many other places.\n\n"
	"In brief: these are the special characters:\n"
	"^ match the beginning of the line\n"
	"$ match the end of line\n"
	"\\ ignore the special meaning of the next character\n"
	"\\\\ a literal \"\\\"\n"
	"- indicates a range when not in the first/last position when\n"
	"    specifying ranges with the \"[\" and \"]\" to specify a list\n"
	"| union of regular expressions\n"
	"[LIST] match a single character specified inside the brackets\n"
	"[^LIST] do not match a single character specified in the brackets\n"
	". match any single character (including non-printable ones)\n"
	"* repeat the previous regular expression 0 or more times\n"
	"+ repeat the previous regular expression 1 or more times\n"
	"? repeat the previous regular expression 0 or 1 times\n"
	"(EXP) groups the expression inside the brackets\n\n"
	"This means, for example, that \".*\" is needed where a wildcard \"*\" might\n"
	"have been used, and \"\\.\" for a literal \".\".\n"
	"Your installed software may only support 'basic' regular expressions,\n"
	"in which case any ? + {  |   (  ) needs to be preceded by a \"\\\".\n\n"
	"Groups like (EXP) may be referred to in the replacement name\n"
	"as \\1, \\2 ... in order of their occurrence.\n\n");
	e2_output_print_default (msg);
}
/**
 * @brief cleanup after cancel button is pressed
 *
 * @param widget clicked button, UNUSED
 * @param rt ptr to dialog data struct
 *
 * @return
 **/
static void _e2pr_quit_cb (GtkWidget *widget, E2_RenDialogRuntime *rt)
{
	gtk_widget_destroy (rt->dialog);
	if (rt->groups != NULL)
	{
		GSList *tmp = rt->groups;
		for (; tmp != NULL; tmp=tmp->next)
			g_slist_free (tmp->data);
	}
	g_free (rt);
	gtk_widget_grab_focus (curr_view->treeview);
}
/**
 * @brief toggle specified option flag
 *
 * @param widget clicked button, UNUSED
 * @param flagnum pointerized number of the flag to be toggled
 *
 * @return
 **/
static void _e2pr_toggle_cb (GtkWidget *widget, gpointer flagnum)
{
  flag_t flg = (flag_t) flagnum;
  gboolean newflag = ! e2pr_get_flag (flg, P_RT);
  e2pr_set_flag (flg, newflag, P_RT);
}
/**
 * @brief toggle specified option flag
 *
 * @param widget clicked button
 * @param flagnum pointerized number of the flag to be toggled
 *
 * @return
 **/
static void _e2pr_grouptoggle_cb (GtkWidget *widget, gpointer flagnum)
{
  flag_t flg = (flag_t) flagnum;
  gboolean newflag = ! e2pr_get_flag (flg, P_RT);
  e2pr_set_flag (flg, newflag, P_RT);
  if (newflag)
  {	//clear all other members of the group
	GtkWidget *tmp = g_object_get_data (G_OBJECT (widget), "group_leader");
	GSList *members = g_object_get_data (G_OBJECT (tmp), "group_members");
	for (;members != NULL; members = members->next)
	{
		tmp = members->data;
		if (tmp != widget)
		  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
	}
  }
}
/**
 * @brief check whether 'find' command is available, and if so, uses extended regex
 *
 * This tries to use the 'find' command to report '/home' using '/?ome'
 * In basic regular expressions, the metacharacters ? + {  |   (  )
 *	need to be escaped with "\"
 *
 * @param regex location to store TRUE if extended regex in use
 *
 * @return TRUE if the find command was found and run
 **/
gboolean _e2pr_check_find_cmd (gboolean *regex)
{
	//try to find /home with extended regex format (basic = /\?ome)
	E2_FILE *pipe = e2_fs_open_pipe ("find / -maxdepth 1 -regex  /?ome 2>&1");
	if (pipe == NULL)
		return FALSE;

	size_t bsize = 64;
	gchar *buf = (gchar *) g_malloc ((gulong) bsize);
#ifdef __USE_GNU
	ssize_t bytes_read = getdelim (&buf, &bsize, '\n', pipe);
	e2_fs_pipe_close (pipe);

	gboolean retval;
	if (bytes_read < 0)
		retval = FALSE;	//error
	else if (bytes_read == 0)
	{
		*regex = FALSE;
		retval = TRUE;
	}
	else //bytes_read > 0
	{
		*regex = g_str_has_prefix (buf, "/home");
		retval = *regex;  //other error message if find cmd not found will return FALSE
	}
#else //getdelim() not available
	gchar *strtmp = fgets (buf, bsize, pipe);
	e2_fs_pipe_close (pipe);

	gboolean retval;
	if (strtmp == NULL)
		retval = FALSE;	//error
	else if (*strtmp == '\n') //bytes_read = 0
	{
		*regex = FALSE;
		retval = TRUE;
	}
	else //bytes_read > 0
	{
		*regex = g_str_has_prefix (buf, "/home");
		retval = *regex;  //other error message if find cmd not found will return FALSE
	}
#endif
	g_free (buf);
	return retval;
}

#ifndef __USE_GNU
/**
 * @brief manually fill the buffer, with trailing \0
 *
 * @param buf pointer to start of buffer for storing input
 * @param bzize max. no of bytes read into @a buf
 * @param pipe stream from which to get the input
 *
 * @return no. of bytes read, excluding the trailing \0
 **/
ssize_t _e2pr_getdelim (gchar *buf, size_t bsize, E2_FILE *pipe)
{
	gint c, len = 0, chars_read = 0;
	gchar ch;
	while ((c = fgetc (pipe)) != EOF)
	{
		len = chars_read * sizeof (gchar);
		if (len >= bsize)
		{
			len -= sizeof (gchar);
			*(buf+len) = '\0';
			return ((ssize_t) len);
		}
		ch = (gchar) c;
		memcpy (buf+len, &ch, sizeof (gchar));
		if (c == 0)
			return ((ssize_t) len);
		else
			chars_read++;
	}
	return ((ssize_t) len);
}
#endif
/**
 * @brief get results of 'find' command
 * @a command will be of the form - find <absolute-root-path> [params] -print0
 * This is executed and interrogated via a pipe
 * @param command the command string to exec, utf8
 *
 * @return newly-allocated null-terminated vector of utf-8 path strings, or NULL
 **/
gchar **_e2pr_get_files_using_find_command (gchar *command)
{
	E2_FILE *pipe = e2_fs_open_pipe (command);
	if (pipe == NULL)
		return NULL;
	//accumulate array, path-by-path
	GPtrArray *allfiles = g_ptr_array_new ();
	gchar *utf;
	size_t bsize = PATH_MAX;
	gchar *buf = (gchar *) g_malloc ((gulong) bsize);
	ssize_t bytes_read;
#ifdef __USE_GNU
	while ((bytes_read = getdelim (&buf, &bsize, '\0', pipe)) != 0)
#else
	while ((bytes_read = _e2pr_getdelim (buf, bsize, pipe)) != 0)
#endif
	{
		if (bytes_read > 0)
		{
			utf = D_FILENAME_FROM_LOCALE (buf);	//always get a copy			
			g_ptr_array_add (allfiles, (gpointer) utf);
		}
		else
//			if (bytes_read == 0)
			break;
/*		else
		{
			e2_output_print_error ( _("Strange error searching for matching files\n"));
			e2_fs_pipe_close (pipe);
			g_ptr_array_free (allfiles, TRUE);
			g_free (buf);
			return NULL;
		}
*/
	}
	e2_fs_pipe_close (pipe);

	gchar **array;
	if (allfiles->len > 0)
	{
		g_ptr_array_add (allfiles, NULL);  //null-terminate the array
		array = (gchar **) allfiles->pdata;
		g_ptr_array_free (allfiles, FALSE);
	}
	else
	{
		array = NULL;
		g_ptr_array_free (allfiles, TRUE);
	}
	g_free (buf);
	return array;
}
/**
 * @brief parse replacement pattern a regex-matched file
 *
 * Carve up new tempate into chunks separated by \1, \2 .... \<E2_HUNK_LIMIT-1>
 * and store copies in P_RT->chunks[1] .... chunks[E2_HUNK_LIMIT-1]
 * chunk[0] never used
 *
 * @param newtemplate replace pattern string with regex \1 [\2 ...]
 *
 * @return
 */
static void _e2pr_parse_regexpattern (gchar *newtemplate)
{
	gint thiscount=0;
	gchar *s, *s1, *s2, *s3;
	s = s1 = g_strdup (newtemplate);	//work with a copy so original is not clobbered

	while ((s2=strchr (s1, '\\')) != NULL)	//always ascii
	{
		s2++;
		if (*s2=='\\')
		{	//ignore any escaped "\"
			s1=s2+1;
			continue;
		}
		s3 = s2;
		while (*s3 >= '0' && *s3 <= '9')
			s3++;
		if (s3 > s2)
		{
			gchar r = *s3;
			*s3 = '\0';
			thiscount = atoi (s2);
			*(s2-1) = '\0';
			if (thiscount > 0 && thiscount < E2_CHUNK_LIMIT)	//don't support \0
				P_RT->chunks[thiscount] = g_strdup (s1);
			*s3 = r;
		}
		s1 = s3;
	}
	//and get the tail of the pattern too
	P_RT->chunks[thiscount+1] = g_strdup (s1);
	g_free (s);
}
/**
 * @brief parse replacement pattern for a wildcard matched file
 *
 * Carve up new tempate into chunks separated by * or ?
 * and store copies in P_RT->chunks[1] .... chunks[E2_HUNK_LIMIT-1]
 * chunk[0] never used
 *
 * @param newtemplate replace pattern string, maybe with * or ?
 *
 * @return
 */
static void _e2pr_parse_wildpattern (gchar *newtemplate)
{
	if ( (strchr (newtemplate, '?') == NULL) && (strchr (newtemplate, '*') == NULL) ) 	//always ascii
		P_RT->chunks[1] = g_strdup (newtemplate);	//there was nothing wild about it ...
	else
	{
		gint thiscount=0; //causes bypass of [0]
		gchar **split = g_strsplit_set (newtemplate, "*?", E2_CHUNK_LIMIT);	
		gchar **tmp = split;
		while (*tmp != NULL && (thiscount < E2_CHUNK_LIMIT-1))
		{
			P_RT->chunks[++thiscount] = g_strdup (*tmp);
			tmp++;
		}
		g_strfreev (split);
	}
}
/**
 * @brief determine replacement name for a file
 * NB 'find' command uses whole path for regex searching,
 * we only process the last element of the path, here
 * for the find command, the old template may have basic regex
 * syntax eg all '(' and ')' escaped, 
 * Here we use extended regex
 * new tempate chunks[] will have been populated before this is called
 *
 * @param oldtemplate search pattern string with 'extended' regex as appropriate
 * @param oldname path string, name to be changed
 * @param ext_regex TRUE if the 'find' command used extended regex
 *
 * @return replacement name with regex matched
 **/
//FIXME detect & use the regex form, basic or extended, that the system uses
gchar *_e2pr_name_replace (gchar *oldtemplate, gchar *oldname, gboolean ext_regex)
//gchar *newtemplate, gchar *oldname)
{
	//get rid of the path, even though 'file' uses that for
	// a regex find (so we might create a not-found error)
	gchar *newbase = g_path_get_basename (oldname);	

	regex_t oldcompiled;
	gint cflags = (ext_regex) ? REG_EXTENDED : 0 ;  //not REG_ICASE
	if (regcomp (&oldcompiled, oldtemplate, cflags))
	{//handle error - eg no regex in expression
	}
	//need this many data spots
	gint matchcount = (gint)oldcompiled.re_nsub + 1;
	regmatch_t matcholdptr [matchcount];
	gint eflags = 0;	//not REG_NOTBOL | REG_NOTEOL;
	size_t error;
	if ((error = regexec (&oldcompiled, newbase, matchcount, matcholdptr, eflags)) != 0)
	{//handle nomatch
		//DEBUG
		size_t len = regerror (error,  &oldcompiled, NULL, 0);
		gchar local[len+2];
		regerror (error,  &oldcompiled,  local, len+2);
		printd (DEBUG, local);
		e2_output_print_error (local);
	}
	//transform newtemplate into new name, with corresponding matches from oldname
	gchar buf[NAME_MAX+2];
	gchar *s1;
	gchar *result = g_strdup ("");
	regmatch_t rx;
	gint i, j, L;
	//ignore hunk[0]
	j=1;
	//ignore the initial 'whole pattern' match
	for (i = 1; i < matchcount; i++)
	{
		//get the matched substring
		rx=matcholdptr[i];
		if (rx.rm_so > -1)
		{	//the substring was actually used in the match
			//get a null-terminated copy
			s1 = newbase + rx.rm_so;
			L = rx.rm_eo - rx.rm_so;
			memcpy (buf, s1, L);
			buf[L] = '\0';
		}
		else
		{
			//FIXME
			continue;
		}
		//progressively join up fixed and variable
		for ( ; j <= i ; j++)
		{
			if (P_RT->chunks[j] == NULL || *(P_RT->chunks[j]) == '\0')
				continue;
			s1=result;
			result = g_strconcat (s1, P_RT->chunks[j], NULL);
			g_free (s1);
		}
		s1=result;
		result = g_strconcat (s1, buf, NULL);
		g_free (s1);
	}
	//and add any left-over chunks
	for (;j<E2_CHUNK_LIMIT;j++)
	{
		if (P_RT->chunks[j] == NULL || *(P_RT->chunks[j]) == '\0')
			continue;
		s1=result;
		result = g_strconcat (s1, P_RT->chunks[j], NULL);
		g_free (s1);
	}
	//cleanup
	g_free (newbase);
	regfree (&oldcompiled);

	return result;
}
/**
 * @brief callback for rename button click
 *
 * Parses the selected options, converts to corresponding
 * shell 'find' command, and executes it
 * Then processes the results
 *
 * @param button clicked widget
 * @param rt ptr to dialog data struct
 *
 * @return
 **/
static void _e2pr_rename_cb (GtkWidget *button, E2_RenDialogRuntime *rt)
{
  const gchar *old = gtk_entry_get_text (GTK_ENTRY (rt->pattern));
  if (*old == '\0')
  {
	  e2_output_print (_("No current filename pattern is specified"),
		  NULL, TRUE, FALSE, TRUE, NULL);
	  return;
  }
  const gchar *new = gtk_entry_get_text (GTK_ENTRY (rt->newpattern));
  if (*new == '\0')
  {
	  e2_output_print (_("No replacement filename pattern is specified"),
		  NULL, TRUE, FALSE, TRUE, NULL);
	  return;
  }
  //update dialog button sensitivities while we search
  gtk_widget_set_sensitive (rt->help_button, FALSE);
  gtk_widget_set_sensitive (rt->start_button, FALSE);
  gtk_widget_set_sensitive (rt->stop_button, TRUE);

//  const gchar *tmp;
  gchar *tmp2, *tmp3;

  //construct shell 'find' command string, from the dialog data
  GString *command = g_string_sized_new (256);
  /* find [-[HLP] [path...]
	 -P  never follow symbolic links (=default)
	 -L  follow symbolic links
	 -H  do not follow symbolic links, except those appearing on the command line
*/
  // pipe(s) in command, need to exec in a separate shell, prepend %
//  command = g_string_assign (command, "%find ");
  command = g_string_assign (command, "find ");
  gboolean freeflg = FALSE;
  if (e2pr_get_flag (SEARCH_ALL_P, rt))
	  tmp2 = G_DIR_SEPARATOR_S;	//root of filesystem tree
  else if (e2pr_get_flag (SEARCH_CURRENT_P, rt))
	  tmp2 = curr_view->dir;
  else if (e2pr_get_flag (SEARCH_OTHER_P, rt))
	  tmp2 = other_view->dir;
  else
  {
	  tmp2 = (gchar *) gtk_entry_get_text (GTK_ENTRY (rt->directory));
	  if (! g_str_has_suffix (tmp2, G_DIR_SEPARATOR_S))
	  {
		  tmp2 = g_strconcat (tmp2, G_DIR_SEPARATOR_S, NULL);
		  freeflg = TRUE;
	  }
  }
  command = g_string_append (command, tmp2);
  if (freeflg)
	  g_free (tmp2);
/*
	[expression]
	! expr
	expr1 -a expr2 = expr1 expr2
	expr1 -o expr2
	\( expr \) force precedence, esp. for -o expressions

	-xdev = don't descend directories on other filesystems
*/
  if ( ! e2pr_get_flag (SEARCH_SUBDIRS_P, rt))
	command = g_string_append (command, " -maxdepth 1");  //or -prune

/*
	-regex pattern
              Filename matches regular expression pattern. (this is a match
              on the whole path, not just the name)
  
  NB %f or %F result in quoted (space-separated if > 1 selected) names,
  _after_ the command is executed FIXME = how to wildcard those ??
//  FIXME find whether the system supports basic or extended regex,
//   escape brackets only if needed
*/
  tmp2 = g_strdup (old);

  if (e2pr_get_flag (SEARCH_WILD_P, rt))
	  //include a separator, in case the user didn't append one
//		g_string_append_printf (command, " -name %s", tmp3);
  //for piped command, need to quote wildcards
		g_string_append_printf (command, " -name '%s'", tmp2);
  else
  {
	//to avoid shell parsing, any '(' or ')' needs to be escaped
	  gchar **split = g_strsplit (tmp2, "(", -1);
	  g_free (tmp2);
	  tmp2 = g_strjoinv ("\\(", split);
	  g_strfreev (split);
	  split = g_strsplit (tmp2, ")", -1);
	  g_free (tmp2);
	  tmp2 = g_strjoinv ("\\)",split);
	  g_strfreev (split);
	// and then regex needs surrounding '  ' to get any \(   \) to work
	 //regex searches on whole path
	  g_string_append_printf (command, " -regex '%s", tmp2);
	  if (!g_str_has_suffix (command->str, G_DIR_SEPARATOR_S))
		  command = g_string_append_c (command, G_DIR_SEPARATOR);
	  g_string_append_printf (command, "%s'", tmp2);
  }
  g_free (tmp2);

/*  
     -exec command ;
              Execute  command;  true  if 0 status is returned.  All following
              arguments to find are taken to be arguments to the command until
              an  argument  consisting of `;' is encountered.  The string `{}'
              is replaced by the current file name being processed  everywhere
              it occurs in the arguments to the command, not just in arguments
              where it is alone, as in some versions of find.  Both  of  these
              constructions might need to be escaped (with a `\') or quoted to
              protect them from expansion by the shell.  The command  is  exe-
              cuted in the starting directory.
*/
  //setup for returning matching files one-at-a-time, via pipe
  command = g_string_append (command, " -print0");
  
  gchar **candidates = _e2pr_get_files_using_find_command (command->str);
  if (candidates == NULL)
  {
	//couldn't find anything
	tmp2 = g_strdup_printf (_("Unable to find anything which matches %s\n"), old);
	e2_output_print_error (tmp2);
	g_free (tmp2);
	g_string_free (command, TRUE);
	gtk_widget_set_sensitive (rt->help_button, TRUE);
	gtk_widget_set_sensitive (rt->start_button, TRUE);
	gtk_widget_set_sensitive (rt->stop_button, FALSE);
	return;
  }

  tmp2 = g_strdup (old);
  tmp3 = g_strdup (new);
//  tmp3 = e2_utf8_from_locale (new);

  gboolean check = e2_option_bool_get ("confirm-overwrite");
  e2_filelist_disable_refresh ();
  e2_widget_set_cursor (GDK_WATCH);
 
  gchar **thisone = candidates;
  gboolean extended = TRUE;
  //walk the vector of matching files
  while (*thisone != NULL)
  {
	gchar *base = g_path_get_basename (*thisone);
	gchar *dir = g_path_get_dirname (*thisone);
	gchar *newbase;
	if (!P_RT->parsed)
	{	//we haven't got the replacement chunks yet, get them now
		gchar **split;
		if (e2pr_get_flag (SEARCH_WILD_P, rt))
		{	//wildcard or specific names used
			//adjust the find pattern
			//replace all '.', '*' and '?' in old template with extended regex equivalents
			split = g_strsplit (tmp2, ".", -1);
			g_free (tmp2);
			tmp2 = g_strjoinv ("\\.", split);
			g_strfreev (split);
			split = g_strsplit (tmp2,	"*", -1);
			g_free (tmp2);
			tmp2 = g_strjoinv ("(.*)", split);
			g_strfreev (split);
			split = g_strsplit (tmp2, "?", -1);
			g_free (tmp2);
			tmp2 = g_strjoinv ("(.)", split);
			g_strfreev (split);
	//	printd (DEBUG, "wildcard rename pattern is %s", pattern);
			//get the chunks of the replacement pattern
			_e2pr_parse_wildpattern (tmp3);
			//we will convert the name using extended regex
			extended = TRUE;
		}
		else
		{	//regex names (actually, paths) used
/*			//cleanup the find pattern
			//FIXME = should only do this if 'find' needed basic regex,
			// and in that case, also fix more things !!
			split = g_strsplit (tmp2, "\\(", -1);
			pattern = g_strjoinv ("(", split);
			g_strfreev (split);
			split = g_strsplit (pattern, "\\)", -1);
			g_free (pattern);
			pattern = g_strjoinv (")", split);
			g_strfreev (split);
*/
			//get the chunks of the replacement pattern
			_e2pr_parse_regexpattern (tmp3);
			//we will convert the name using the same regex as 'find' used
			extended = P_RT->extended_rgx;
		}
		P_RT->parsed = TRUE;
	}
	//get the replacement basename
	newbase = _e2pr_name_replace (tmp2, *thisone, extended);
	
	DialogButtons doit;
	//ask, if the confirm option is is force, and the parent's stop btn not pressed
	if (e2pr_get_flag (CONFIRM_P, rt) && ! P_RT->abort)
	{
		GString *dialog_prompt = g_string_sized_new (NAME_MAX + PATH_MAX);	//pick a reasonable initial size
		g_string_printf (dialog_prompt, "%s\n<b>%s</b>\n%s\n<b>%s</b>\n%s %s",
			_("Rename"), base, _("to"), newbase, _("in"), dir);
		GtkWidget *dialog = e2_dialog_create (GTK_STOCK_DIALOG_QUESTION,
			dialog_prompt->str, _("confirm"), e2_dialog_response_decode_cb, &doit);
		g_string_free (dialog_prompt, TRUE);
		
		//set default button to 'no'
		E2_BUTTON_NO.showflags |= E2_BTN_DEFAULT;
		E2_BUTTON_YES.showflags &= !E2_BTN_DEFAULT;
		//non-modal dialog
		e2_dialog_show (dialog, app.main_window,
			0,  //no  _RUN (hence window stays sensitive)  no _FREE, in case we want to hide it
				&E2_BUTTON_NOTOALL, &E2_BUTTON_NO, &E2_BUTTON_YES, NULL);
		gtk_main ();

		gtk_widget_destroy (dialog);
	}
	else	//no confirmation
		doit = OK;

	if (doit == OK)
	{
		gchar *newpath = g_build_filename (dir, newbase, NULL);

		F_DECLARE;
		gchar *local = F_FILENAME_TO_LOCALE (newpath);

		if (check && e2_fs_access2 (local) == 0)
		{	//get o/w confirmation if that option is in force
			e2_filelist_enable_refresh ();  //allow updates while we wait
			doit = e2_dialog_ow_check (newpath, NOALL);
			e2_filelist_disable_refresh ();
		}
		if (doit == OK)
			e2_task_backend_rename (*thisone, local);

		g_free (newpath);
		F_FREE (local);
	}

	g_free (base);
	g_free (dir);
	g_free (newbase);

	if (doit == NO_TO_ALL || P_RT->abort)
	{	//confirm dialog or o/w dialog or main dialog stop btn pressed
		P_RT->abort=FALSE;  //make sure this is now clear
		break;
	}

	thisone++;
  }
  e2_widget_set_cursor (GDK_LEFT_PTR);
  e2_filelist_enable_refresh();

  //cleanups
  g_strfreev (candidates);
  g_free (tmp2);
  g_free (tmp3);
  g_string_free (command, TRUE);
  gint j;
  for (j=0;j<E2_CHUNK_LIMIT;j++)  //[0] never used, actually
  {
	if (P_RT->chunks[j] != NULL)
	{
	  g_free (P_RT->chunks[j]);
	  P_RT->chunks[j] = NULL;
	}
  }
  P_RT->parsed = FALSE;
//revert the buttons
  gtk_widget_set_sensitive (rt->help_button, TRUE);
  gtk_widget_set_sensitive (rt->start_button, TRUE);
  gtk_widget_set_sensitive (rt->stop_button, FALSE);
}
/**
 * @brief callback for stop button click
 *
 * @param w activated widget, UNUSED
 * @param rt ptr to dialog data struct
 *
 * @return
 **/
static void _e2pr_abort_cb (GtkWidget *w, E2_RenDialogRuntime *rt)
{
	P_RT->abort = TRUE;
}

  /******************/
 /***** utilities *****/
/******************/

/**
 * @brief set specified flag to T/F
 *
 * The relevant array value is set
 *
 * @param f enumerated value of flag to be set
 * @param value new value for the flag
 * @param rt ptr to dialog data struct
 *
 * @return
 **/
void e2pr_set_flag (flag_t f, gboolean value, E2_RenDialogRuntime *rt)
{
  if (f < MAX_FLAGS)
    rt->flags[ (gint) f] = value;
}
/**
 * @brief return the value of a specified flag
 *
 * @param f enumerated value of flag to be interrogated
 * @param rt
 *
 * @return flag value, T/F, or FALSE if the value is not recognised
 **/
gboolean e2pr_get_flag (flag_t f, E2_RenDialogRuntime *rt)
{
   if (f < MAX_FLAGS)
	  return (rt->flags[(gint) f]);
   else
     return FALSE;
}
/**
 * @brief set all flags to FALSE
 *
 * @param rt ptr to dialog data struct
 *
 * @return
 **/
void e2pr_reset_flags (E2_RenDialogRuntime *rt)
{
  gint i;
  for (i = 0; i < MAX_FLAGS; i++)
      rt->flags[i] = FALSE;
}

  /***************************/
 /***** widget creation *****/
/**************************/
/**
 * @brief make widgets with filename search/replace options
 *
 * @param container the widget into which the things will be packed
 * @param rt ptr to dialog data struct
 *
 * @return
 **/
void e2pr_make_widgets (GtkWidget *container, E2_RenDialogRuntime *rt)
{
//  GtkWidget *hbox = e2pr_create_hbox (container);
  e2pr_create_label (container, _("Search for files:"));
  GtkWidget *radio = e2pr_create_radio_button (container, 
		SEARCH_ALL_P, FALSE, _("anywhere"), rt);
  e2pr_create_radio_grouped_button (container, radio,
		SEARCH_CURRENT_P, TRUE, _("in active directory"), rt);
  e2pr_create_radio_grouped_button (container, radio,
		SEARCH_OTHER_P, FALSE, _("in other directory"), rt);
  e2pr_create_radio_grouped_button (container, radio,
		SEARCH_THIS_P, FALSE, _("here"), rt);
  rt->directory = e2pr_create_entry (container);
//  g_object_set_data (G_OBJECT (rt->directory), "reset_yourself", reset_entry);
  e2pr_create_toggle_button (container,
		SEARCH_SUBDIRS_P, FALSE, _("Recurse subdirectories"), rt);
  e2pr_create_hseparator (container);

  e2pr_create_toggle_button (container,
		CONFIRM_P, TRUE, _("Confirm before rename"), rt);
  e2pr_create_hseparator (container);

  GtkWidget *hbox = e2pr_create_hbox (container);
  radio = e2pr_create_radio_button (hbox, 
		SEARCH_WILD_P, TRUE, _("Match exact/wildcard"), rt);
  e2pr_create_radio_grouped_button (hbox, radio,
		SEARCH_REGEX_P, FALSE, _("Match regular expression"), rt);
  e2pr_create_hseparator (container);
  
  hbox = e2pr_create_hbox (container);
  e2pr_create_label (hbox, _("Current name is like this:"));
  rt->pattern = e2pr_create_entry (hbox);
//  gtk_entry_set_text (GTK_ENTRY (rt->pattern), "(.*)");  //CHECKME ok without utf-8 ??
//  g_object_set_data (G_OBJECT (rt->pattern), "reset_yourself", reset_entry);
  e2pr_create_hseparator (container);

  hbox = e2pr_create_hbox (container);
  e2pr_create_label (hbox, _("New name is like this:"));
  rt->newpattern = e2pr_create_entry (hbox);
//  gtk_entry_set_text (GTK_ENTRY (rt->newpattern), "\\1");  //FIXME utf-8
//  g_object_set_data (G_OBJECT (rt->newpattern), "reset_yourself", reset_entry);
}

  /**********************************/
 /***** lesser widget creation *****/
/*********************************/

/**
 * @brief create and show a vbox widget in a specified container widget
 *
 * @param container widget into which the vbox is to be added
 *
 * @return the vbox widget
 **/
GtkWidget *e2pr_create_vbox (GtkWidget *container)
{
  GtkWidget *rv = gtk_vbox_new (FALSE, E2_PADDING);
  gtk_container_add (GTK_CONTAINER (container), rv);
  gtk_widget_show (rv);
  return (rv);
}
/**
 * @brief create and show a hbox widget in a specified container widget
 *
 * @param container widget into which the hbox is to be added
 *
 * @return the hbox widget
 **/
GtkWidget *e2pr_create_hbox (GtkWidget *container)
{
  GtkWidget *rv = gtk_hbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (container), rv);
  gtk_widget_show (rv);
  return(rv);
}
/**
 * @brief create and show a horiz separator widget in a specified container widget
 *
 * @param container widget into which the separator is to be added
 *
 * @return the separator widget
 **/
GtkWidget *e2pr_create_hseparator (GtkWidget *container)
{
  GtkWidget *rv = gtk_hseparator_new ();
  gtk_container_add (GTK_CONTAINER (container), rv);
  gtk_widget_show (rv);
  return (rv);
}
/**
 * @brief create and show a label widget in a specified container widget
 *
 * @param container widget into which the label is to be added
 * @param text utf8 string with text to be added to the label
 *
 * @return the label widget
 **/
GtkWidget *e2pr_create_label (GtkWidget *container, gchar *text)
{
//  gchar *utf =  e2_utf8_from_locale (text);
  GtkWidget *rv = gtk_label_new (text);	//utf);
  gtk_container_add (GTK_CONTAINER (container), rv);
  gtk_widget_show (rv);
  return (rv);
}
/**
 * @brief create and show an entry widget in a specified container widget
 *
 * @param container widget into which the entry is to be added
 *
 * @return the entry widget
 **/
GtkWidget *e2pr_create_entry (GtkWidget *container)
{
  GtkWidget *rv = gtk_entry_new ();
  gtk_container_add (GTK_CONTAINER (container), rv);
  gtk_widget_show (rv);
  return (rv);
}
/**
 * @brief create and show a button in a specified container
 *
 * @param container the widget into which the button is to be placed
 * @param callback callback function to be connected
 * @param callback_data data to be provided to the callback fn
 *
 * @return the button widget
 **/
/* UNUSED
GtkWidget *e2pr_create_button (GtkWidget *container, 
	void (*callback) (), gpointer callback_data)
{
  GtkWidget *rv = gtk_button_new ();
  g_signal_connect (G_OBJECT(rv), "clicked", G_CALLBACK (callback),
		     callback_data);
  gtk_container_add (GTK_CONTAINER (container), rv);
  gtk_widget_show (rv);
  return (rv);
} */
static GtkWidget *_e2pr_create_toggle_button (GtkWidget *container, 
	flag_t f, gboolean state, gchar *label, gpointer callback_fn, 
	E2_RenDialogRuntime *rt)
{
  //cached flags default to FALSE
  if (state)
	 e2pr_set_flag (f, TRUE, rt);
  GtkWidget *rv = gtk_check_button_new_with_label (label);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (rv), state);
  g_signal_connect (G_OBJECT (rv), "toggled", G_CALLBACK (callback_fn),
		(gpointer) f);  //CHECKME = correct value?
//  g_object_set_data (G_OBJECT (rv), "reset_yourself",
//		state ? e2pr_set_toggle_button_on : e2pr_set_toggle_button_off );
  gtk_container_add (GTK_CONTAINER (container), rv);
  gtk_widget_show (rv);
  return (rv);
}
/**
 * @brief create and show a grouped toggle in a specified container
 *
 * @param container the widget into which the button is to be placed
 * @param f enumerated value of flag to be associated with the button
 * @param state T/F initial state of the toggle
 * @param label  translated string for the button label
 * @param leader widget for the 'leader' of the group, or NULL if this is the leader
 * @param rt ptr to dialog data struct
 *
 * @return the button widget (UNUSED, now)
 **/
GtkWidget *e2pr_create_toggle_grouped_button (GtkWidget *container, 
	flag_t f, gboolean state, gchar *label, GtkWidget *leader, E2_RenDialogRuntime *rt)
{
	GtkWidget *rv = _e2pr_create_toggle_button (container, f, state, label, _e2pr_grouptoggle_cb, rt);
	GtkWidget *ptr;
	GSList *members;
	if (leader == NULL)
	{  //this is the leader of a new group
		ptr =  rv;  //point to self
		members = NULL;
		rt->groups = g_slist_append (rt->groups, members);  //remember it, for cleaning up
	}
	else
	{  //this is a group member
		ptr = leader;  //point to group leader, which has list
		members = g_object_get_data (G_OBJECT (leader), "group_members");
	}
	g_object_set_data (G_OBJECT (rv), "group_leader", ptr);
	members = g_slist_append (members, rv);
	g_object_set_data (G_OBJECT (ptr), "group_members", members);
	return (rv);
}
/**
 * @brief create and show a toggle in a specified container
 *
 * @param container the widget into which the button is to be placed
 * @param f enumerated value of flag to be associated with the button
 * @param state T/F initial state of the toggle
 * @param label  translated string for the button label
 *
 * @return the button widget (UNUSED, now)
 **/
GtkWidget *e2pr_create_toggle_button (GtkWidget *container, 
	flag_t f, gboolean state, gchar *label, E2_RenDialogRuntime *rt)
{
	GtkWidget *rv = _e2pr_create_toggle_button (container, f, state, label, _e2pr_toggle_cb, rt);
	return (rv);
}
/**
 * @brief create and show a radio btn in a specified container
 *
 * @param container the widget into which the button is to be placed
 * @param f enumerated value of flag to be associated with the button
 * @param state T/F initial state of the toggle (UNUSED)
 * @param label  translated string for the button label
 * @param rt
 *
 * @return the button widget
 **/
GtkWidget *e2pr_create_radio_button (GtkWidget *container, 
	flag_t f, gboolean state, gchar *label, E2_RenDialogRuntime *rt)
{
  //cached flags default to FALSE, leader of radio-button group needs to start off SET
//  if (state)
	 e2pr_set_flag (f, TRUE, rt);
  GtkWidget *rv = gtk_radio_button_new_with_label (NULL, label);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (rv), TRUE);
  g_signal_connect (G_OBJECT (rv), "toggled", G_CALLBACK (_e2pr_toggle_cb),
		(gpointer) f);
//  g_object_set_data (G_OBJECT (rv), "reset_yourself",
//		state ? e2pr_set_toggle_button_on : e2pr_set_toggle_button_off );
  gtk_container_add (GTK_CONTAINER (container), rv);
  gtk_widget_show (rv);
  return (rv);
}
/**
 * @brief create and show a radio btn in a specified container
 *
 * @param container the widget into which the button is to be placed
 * @param group the radio button widget that 'leads' the group
 * @param f enumerated value of flag to be associated with the button
 * @param state T/F initial state of the toggle
 * @param label  translated string for the button label
 * @param rt
 *
 * @return the button widget
 **/
GtkWidget *e2pr_create_radio_grouped_button (GtkWidget *container, GtkWidget *group, 
	flag_t f, gboolean state, gchar *label, E2_RenDialogRuntime *rt)
{
  //cached flags default to FALSE
  if (state)
	 e2pr_set_flag (f, TRUE, rt);
  GSList *list = gtk_radio_button_get_group (GTK_RADIO_BUTTON (group));
  GtkWidget *rv = gtk_radio_button_new_with_label (list, label);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (rv), state);
  g_signal_connect (G_OBJECT (rv), "toggled", G_CALLBACK (_e2pr_toggle_cb),
		(gpointer) f);
  //CHECKME = is this automatically done for toggled groups ?
//  g_object_set_data (G_OBJECT (rv), "reset_yourself",
//		state ? e2pr_set_toggle_button_on : e2pr_set_toggle_button_off );
  gtk_container_add (GTK_CONTAINER (container), rv);
  gtk_widget_show (rv);
  return (rv);
}
/**
 * @brief  create and show a file selection widget
 *
 * @param title string used for file selection widget title
 * @param ok ptr to void ok-callback for the widget
 * @param cancel  ptr to void cancel-callback for the widget
 * @param rt ptr to find data struct
 *
 * @return the file selection widget
 **/
/* deprecated
GtkWidget *e2pr_create_filesel (gchar *title, void (*ok)(GtkWidget *, E2_FindDialogRuntime *),
	       void(*cancel)(GtkWidget *, E2_FindDialogRuntime *), E2_FindDialogRuntime *rt )
{
  GtkWidget *rv = gtk_file_selection_new (title);
  g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (rv)->ok_button),
		     "clicked", G_CALLBACK (ok), rt);
  g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (rv)->cancel_button),
		     "clicked", G_CALLBACK (cancel), rt);
  gtk_widget_show (rv);
  return (rv);
} */
/**
 * @brief establish and show the dialog
 *
 * @param
 *
 * @return
 **/
void e2p_rename_dialog_create (void)
{
  //init runtime object
  P_RT = g_malloc0 (sizeof (E2_RenDialogRuntime));
//  e2pr_reset_flags (P_RT);  //initialise all flags (all are 0 from malloc)
  //check for 'find' command, and whether it can use extended regex
  if (! _e2pr_check_find_cmd(&P_RT->extended_rgx))
  {
	  e2_output_print_error (_("Cannot open 'find' command\n"));
	  g_free (P_RT);
	  return;
  }	
  //create dialog
  P_RT->dialog = gtk_dialog_new ();
  E2_DIALOG_TITLE (GTK_WINDOW (P_RT->dialog), _("rename files"));
  GtkWidget *dialog_vbox = GTK_DIALOG (P_RT->dialog)->vbox;
//  GtkWidget *action_area = GTK_DIALOG (rt->dialog)->action_area; //this is for the buttons
  gtk_container_set_border_width (GTK_CONTAINER (dialog_vbox), E2_PADDING);
  //populate it
  e2pr_make_widgets (dialog_vbox, P_RT);
  // add buttons in the order that they will appear
  P_RT->help_button = e2_dialog_add_undefined_button_custom
		(P_RT->dialog, FALSE, E2_RESPONSE_USER2, _("_Help"), GTK_STOCK_HELP,
		_("get advice on rename options"), _e2pr_help_cb, P_RT);
/*  //omit this
  button = e2pr_create_button (hbox, _e2pr_save_search_cb, NULL);
  label = e2pr_create_label (button, _("_Save"));
*/
 P_RT->stop_button = e2_dialog_add_button_custom (P_RT->dialog, FALSE,
		&E2_BUTTON_NOTOALL, _("stop the current search"), _e2pr_abort_cb, P_RT);
  //de-sensitize stop btn, at this stage
  gtk_widget_set_sensitive (P_RT->stop_button, FALSE);

  P_RT->start_button = e2_dialog_add_undefined_button_custom (P_RT->dialog, FALSE, 
		E2_RESPONSE_USER1, _("_Rename"), GTK_STOCK_CONVERT, _("begin renaming"), _e2pr_rename_cb, P_RT);
  e2_dialog_add_button_custom (P_RT->dialog, TRUE, &E2_BUTTON_CLOSE, NULL, _e2pr_quit_cb, P_RT);

  e2_dialog_setup (P_RT->dialog, app.main_window);
  gtk_widget_show (P_RT->dialog);
  gtk_main ();
}

//aname must be confined to this module
static gchar *aname;
/**
 * @brief plugin initialization function, called by main program
 *
 * @param p ptr to plugin data struct
 *
 * @return TRUE if the initialization succeeds, else FALSE
 **/
gboolean init_plugin (Plugin *p)
{
#define ANAME "renext"
  aname = _("renext");

  p->signature = ANAME VERSION;
  p->menu_name = _("_Rename..");
  p->description = _("Rename item(s), using wildcards or regular-expressions");
  p->icon = ICON_DIR"plugin_rename_"E2IP".png";

  if (p->plugin_cb == NULL)
  {
	p->plugin_cb = e2p_rename_dialog_create;  //set plugins menu item function
//	E2_Action* (*action_register) () = (*p->api_lookup) (E2API_ACTION_REGISTER);
	//no need to free this
	gchar *action_name = g_strconcat (PLUGIN,".",aname,NULL);
	//NULL = no data, FALSE = no argument for the function
//	(*action_register)
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM, e2p_rename_dialog_create, 
		NULL, FALSE);
	return TRUE;
  }
  return FALSE;
}
/**
 * @brief unload this plugin
 *
 * Most cleanup work is done in the main program, 
 * here the associated actions are de-registered
 *
 * @param p ptr to plugin data struct
 *
 * @return
 **/
void unload_plugin (Plugin *p)
{
//  if (p->api_lookup != NULL)
//  {
//	gboolean (*action_unregister) () = (*p->api_lookup) (E2API_ACTION_UNREGISTER);
	gchar *action_name = g_strconcat (PLUGIN,".",aname,NULL);
//	(*action_unregister)
	e2_action_unregister (action_name);
	g_free (action_name);
//  }
}
