/*
 *
 * Copyright (C) 2005 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
/**
 * @file plugins/e2p_upgrade.c 
 * @brief configuration upgrade functions
 *
 * This file contains functions that help upgrading the 
 * configuration file to the current version.
 **/

#include "emelfm2.h"
#include <string.h>
#include "e2_plugins.h"
#include "e2_option.h"
#include "e2_output.h"
#include "e2_dialog.h"
#include "e2_task.h"
#include "e2_complete.h"

static void _e2p_upgrade_reconfig (void)
{
	e2_option_tree_stores_clear ();
	//some (imperfect) cleanups
	gint i;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
//		printd (DEBUG, "destroying set %s", set->name);
		E2_OptionFlags frees = set->flags;

		if (frees & E2_OPTION_FLAG_FREENAME)
			g_free (set->name);
		if (frees & E2_OPTION_FLAG_FREEGROUP)
			g_free (set->group);
		if (frees & E2_OPTION_FLAG_FREEDESC)
			g_free (set->desc);
		if (frees & E2_OPTION_FLAG_FREETIP)
			g_free (set->tip);
		if (frees &
			(E2_OPTION_FLAG_INT | E2_OPTION_FLAG_STR
			 | E2_OPTION_FLAG_SEL | E2_OPTION_FLAG_FONT
			 | E2_OPTION_FLAG_COLOR))
			g_free (set->sval);

		//extras
		if (frees & E2_OPTION_FLAG_TREE)
		{
			if (set->ex.tree.def != NULL)
				g_strfreev (set->ex.tree.def);
			e2_list_free_with_data (&set->ex.tree.columns);	//LEAK - each column has freeable things
		}
		//sel-option choices are all constant strings ??
		
		//CHECKME free hooks ??
	}
	g_ptr_array_free (options_array, TRUE);
	options_array = g_ptr_array_new ();

	g_hash_table_destroy (options_hash);
	options_hash = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);

//	g_hash_table_destroy (options_queue);
//	options_queue = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

	e2_option_default_register ();
	e2_option_file_read ();
}

static void _e2p_upgrade_backup (gchar *file)
{
	gchar *config_file = g_strdup_printf ("%s"G_DIR_SEPARATOR_S"%s",
		e2_cl_options.config_dir, file);
	F_DECLARE;
	gchar *local = F_FILENAME_TO_LOCALE (config_file);
	if (e2_fs_access (local, F_OK) == 0)	//traverse link, if any
	{
		gchar *saved_file = g_strconcat (local, ".save", NULL);
		e2_task_backend_rename (local, saved_file);
		g_free (saved_file);
	}
	g_free (config_file);
	F_FREE (local);
}

gint _e2p_upgrade_dialog (gchar *msg)
{
	//main button structs not created yet
	E2_Button E2P_BUTTON_OK =
		{ N_("_Ok"),"",
#ifdef E2_IMAGECACHE
	//FIXME gtk images not yet available
		NULL,
#else
		GTK_STOCK_OK,
#endif
		E2_BTN_DEFAULT,E2_BTN_DEFAULT,GTK_RESPONSE_OK};
	E2_Button E2P_BUTTON_CANCEL =
		{ N_("_Cancel"),"",
#ifdef E2_IMAGECACHE
	//FIXME gtk images not yet available
		NULL,
#else
		GTK_STOCK_CANCEL,
#endif
		E2_BTN_DEFAULT,E2_BTN_DEFAULT,GTK_RESPONSE_CANCEL};

	GtkWidget *dialog = e2_dialog_create (
#ifdef E2_IMAGECACHE
	//FIXME gtk images not yet available
		NULL,
#else
		GTK_STOCK_DIALOG_INFO,
#endif
		msg, _("update information"), NULL, NULL);
	e2_dialog_show (dialog, app.main_window, 0,
		&E2P_BUTTON_CANCEL, &E2P_BUTTON_OK, NULL);
	gint choice = gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
	return choice;
}

void _e2p_upgrade_pre_0_1 (void)
{
	gchar *msg = g_strdup_printf (
	_("Configuration arrangements for this version %s of %s are considerably "
	"different from those of earlier versions. To reliably ensure access to the "
	"program's current features, it is best to start with fresh settings.\n"
	"If you proceed, the superseded configuration files in\n %s will have '.save' "
	"appended to their names.\nFeel free to delete them."
	),
	VERSION, PROGNAME, e2_cl_options.config_dir);
	gint choice = _e2p_upgrade_dialog (msg);
	g_free (msg);

	if (choice == GTK_RESPONSE_OK)
	{
		_e2p_upgrade_backup ("config");
//		_e2p_upgrade_backup ("cache");  this one is ok to leave as is
		_e2p_upgrade_backup ("filetypes");
		_e2p_upgrade_backup ("plugins");
		_e2p_upgrade_backup ("settings");

		e2_option_tree_stores_clear ();
		g_ptr_array_free (options_array, TRUE);	//this leaks !!!
		options_array = g_ptr_array_new ();
		g_hash_table_destroy (options_hash);
		options_hash = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
//		g_hash_table_destroy (options_queue);
//		options_queue = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
		e2_option_default_register ();
	}
	else
		exit (1);
}

void _e2p_upgrade_0_1 (void)
{
	gchar *msg = g_strdup_printf (
	_("Several default configuration settings for this version %s of %s are "
	"different from those of recent versions (see changelog).\n"
	"If you click Ok, those settings will be updated where possible.\n"
	"Or else you can Cancel, and manually change settings, or even "
	"revert all settings to current defaults, via the configuration dialog."
	),
	VERSION, PROGNAME);
	gint choice = _e2p_upgrade_dialog (msg);
	g_free (msg);
	
	if (choice == GTK_RESPONSE_OK)
	{
		gchar *sed = g_find_program_in_path ("sed");
		if (sed == NULL)
			printd (ERROR, "can't find 'sed' so i can't upgrade the config file");
		else
		{
			
	//pseudo action to rename
	gchar *oldstr1 = _("<file actions>");
	gchar *newstr1 = g_strconcat (_A(5),".",_A(20),NULL);
	//keybinding to add
	gchar *oldstr2 = g_strconcat ("\t\t|",_("Delete"),"|false|",_A(5),".",_A(35),"|",NULL); //"delete"
//	gchar *newstr2 = g_strconcat (oldstr2,"\\n\t\t|<Shift>",_("Delete"),"||",_A(5),".",_A(82),"|",NULL); //"trashempty"
	gchar *newstr2 = g_strconcat ("\t\t|<Shift>",_("Delete"),"|false|",_A(5),".",_A(82),"|",NULL); //"trashempty"
	//filetype string to revise
	gchar *oldstr3 = g_strconcat ("<", _A(11), ".", _A(67),";<", _A(11), ".", _A(78), ";>bzip2 -d -c %F [\\]| tar xf -;<", _A(11), ".", _A(79), ";<", _A(11), ".", _A(67),NULL);
	gchar *newstr3 = g_strconcat ("<", _A(11), ".", _A(78), ";>bzip2 -d -c %f ","\\\\","| tar -C %D -xf -;<", _A(11), ".", _A(79), NULL);

			gchar *cfg_file = g_build_filename (e2_cl_options.config_dir,
				default_config_file, NULL);
			gchar *command = g_strconcat ("cp ", cfg_file, " ", cfg_file, ".save;",
				sed, " -e '1s/", cfg.version, "/", VERSION, "/'",
				" -e 's/",oldstr1,"/",newstr1,"/'",
				" -e '/",oldstr2,"$/a\\\n",newstr2,"'",
				" -e 's/",oldstr3,"/",newstr3,"/' ",
				cfg_file,".save >",cfg_file,NULL);
//	printd (DEBUG, command);
//			e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
			system (command);
			g_free (command);
			g_free (cfg_file);

	g_free (newstr1);
	g_free (newstr2);
	g_free (oldstr2);
	g_free (newstr3);
	g_free (oldstr3);

		}
		g_free (sed);
		//implement the update
		_e2p_upgrade_reconfig ();
	}
}

void _e2p_upgrade_0_1_1 (void)
{
	gchar *msg = g_strdup_printf (
	_("Several default keybindings and other aspects of this version "
	"%s of %s are different from those of recent versions (see changelog).\n"
	"If you click Ok, those settings will be updated where possible.\n"
	"Or else you can Cancel, and manually change settings, or even "
	"revert all settings to current defaults, via the configuration dialog."
	),
	VERSION, PROGNAME);
	gint choice = _e2p_upgrade_dialog (msg);
	g_free (msg);
	
	if (choice == GTK_RESPONSE_OK)
	{
		gchar *sed = g_find_program_in_path ("sed");
		if (sed == NULL)
			printd (ERROR, "can't find 'sed' so i can't upgrade the config file");
		else
		{
	//keybinding to replace
	//NOTE needs |false|, not ||
			//dir line
		gchar *oldstr1 = g_strconcat("\t\t|",_("Insert"),"|false|",_A(1),".",_A(29),"|",NULL); //"clear"
		gchar *newstr1 = g_strconcat("\t\t|<Ctrl>",_("Delete"),"|false|",_A(1),".",_A(29),"|",  //"clear"
			"\\\n\t\t|<Alt>",_("Delete"),"|false|",_A(1),".",_A(30),"|",NULL);	//clear-history

	//keybindings to add
			//general
		gchar *oldstr2 = g_strconcat("\t|<Ctrl>",_("r"),"|false|",_A(11),".",_A(59),"|",NULL); //"refresh"
		gchar *newstr2 = g_strconcat
			("\t|<Ctrl>","1","|false|",_A(4),".",_A(39),"|1\\\n""\t|<Ctrl>","2","|false|",_A(4),".",_A(39),"|2",NULL);

		//command line
		gchar *oldstr3 = g_strconcat("\t\t|<Ctrl>",_("Delete"),"|false|",_A(1),".",_A(29),"|",NULL);	//clear
		gchar *newstr3 = g_strconcat("\t\t|<Alt>",_("Delete"),"|false|",_A(1),".",_A(30),"|",NULL); //clear-history

		//helpdoc locations maybe version-specific
		//e.g. usage-help-doc=/usr/share/doc/emelfm2-0.1.2/USAGE
		gchar *oldstr4 = g_strconcat ("\\(usage-help-doc=.*-\\)",cfg.version,"\\(\\/.*\\)$", NULL);
		gchar *newstr4 = g_strconcat ("\\","1",VERSION,"\\","2",NULL);
		//config-help-doc=/usr/share/doc/emelfm2-0.1.2/CONFIGURATION
		gchar *oldstr5 = g_strconcat ("\\(config-help-doc=.*-\\)",cfg.version,"\\(\\/.*\\)$", NULL);
		
		//commandbar button to add (NOTE no / in match-item path, sed hates it)
		gchar *oldstr6 = g_strconcat
			(_("su.."),"|.*.png|",_("Run something as root"),"|",_A(16),"|xterm -e 'su -c \"%{.*read'",NULL);
		gchar *newstr6 = g_strconcat
			(_("mts.."),"|"ICON_DIR"mounts_"E2IP".png|",_("Mount or unmount a device"),"|",_A(94),".",_A(23),"|",_C(6), NULL);

			gchar *cfg_file = g_build_filename (e2_cl_options.config_dir,
				default_config_file, NULL);
			gchar *command = g_strconcat ("cp ", cfg_file, " ", cfg_file, ".save;",
				sed, " -e '1s/", cfg.version, "/", VERSION, "/'",
				" -e 's/",oldstr4,"/",newstr4,"/'",
				" -e 's/",oldstr5,"/",newstr4,"/'",
				" -e 's/",oldstr1,"/",newstr1,"/'",
				" -e '/",oldstr2,"$/a\\\n",newstr2,"'",
				" -e '/",oldstr3,"$/a\\\n",newstr3,"' ",
#ifdef E2_FS_MOUNTABLE
				"-e '/",oldstr6,"$/a\\\n",newstr6,"' ",	//~ to avoid problems with / in paths
#endif
				cfg_file,".save >",cfg_file,NULL);
//	printd (DEBUG, command);
//			e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
			system (command);
			g_free (command);
			g_free (cfg_file);

	g_free (newstr1);
	g_free (oldstr1);
	g_free (newstr2);
	g_free (oldstr2);
	g_free (newstr3);
	g_free (oldstr3);
	g_free (newstr4);
	g_free (oldstr4);
	g_free (oldstr5);
#ifdef E2_FS_MOUNTABLE
	g_free (newstr6);
	g_free (oldstr6);
#endif
		}
		g_free (sed);
		//implement the update
		_e2p_upgrade_reconfig ();
	}
}

gboolean init_plugin (Plugin *p)
{
	if (strcmp (cfg.version,"0.1") < 0)
		_e2p_upgrade_pre_0_1();
	else if (strcmp (cfg.version,"0.1.1") < 0)
		_e2p_upgrade_0_1();
	else if (strcmp (cfg.version,"0.1.2") < 0)
	_e2p_upgrade_0_1_1();

	return TRUE;
}

//void unload_plugin (Plugin *p)
//{
//}
