/* $Id: e2_action.c 1093 2005-08-28 07:15:24Z tpgww $
 * 
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/actions/e2_action.c 
 * @brief action system
 *
 * This file contains functions for the action system.
 **/

#include "emelfm2.h"
#include <string.h>
#include "e2_action.h"
#include "e2_action_option.h"
#include "e2_task.h"
#include "e2_plugins.h"
#include "e2_filetype.h"
#include "e2_dialog.h"
#include "e2_complete.h"

static gboolean _e2_action_unregister (E2_Action *action);

static GHashTable *actions_hash;
static GPtrArray *actions_array;
static GtkTreeStore *actions_store;

  /*********************/
 /***** callbacks *****/
/********************/

static void _e2_action_custom_command_cb (GtkWidget *widget, gpointer data)
{
	gchar *command = g_strdup (data);
	e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
	g_free (command);
//	e2_command_run ((gchar *)data, E2_COMMAND_RANGE_DEFAULT);
}

static gint _e2_action_tree_compare_cb (GtkTreeModel *model,
	GtkTreeIter *a, GtkTreeIter *b, gpointer userdata)
{
	gint ret = 0;
	gchar *name1, *name2;
	gtk_tree_model_get(model, a, 1, &name1, -1);
	gtk_tree_model_get(model, b, 1, &name2, -1);
	if (name1 == NULL || name2 == NULL)
	{
		if (name1 == NULL && name2 == NULL)
		return ret;
		ret = (name1 == NULL) ? -1 : 1;
	} else
		ret = g_utf8_collate (name1,name2);
	g_free(name1);
	g_free(name2);
	return ret;
}

static gboolean _e2_action_visible_cb (GtkTreeModel *model,
	GtkTreeIter *iter, gpointer data)
{
	E2_ACTION_EXCLUDE ex1 = GPOINTER_TO_INT (data);
	E2_ACTION_EXCLUDE ex2;
	gtk_tree_model_get (model, iter, 2, &ex2, -1);
	return !(ex1 & ex2);
}

  /******************/
 /***** public *****/
/******************/

/**
 * @brief option helper for action columns
 *
 * This function can be used as source function for @a E2_OPTION_TREE_TYPE_SEL
 * columns of E2_OPTION_TYPE_TREE options.
 *
 * @param ex pointerised exclude flags which specify action(s) that should not be visible
 *
 * @return GtkTreeFilterModel of all actions
 **/
GtkTreeModel *e2_actions_fill_store (gpointer ex)
{
	GtkTreeModel *filter = gtk_tree_model_filter_new (GTK_TREE_MODEL
		(actions_store), NULL);
	gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (filter),
		(GtkTreeModelFilterVisibleFunc) _e2_action_visible_cb, ex, NULL);
	return filter;
}

/** @brief short-form register action function
 *
 * See e2_action_register function description. This one has no exclude flags or data2 
*
 * @return registered action E2_Action
 **/
E2_Action *e2_action_register_simple (gchar *name, E2_ACTION_TYPE type,
	void *func, gpointer data, gboolean has_arg)
{
	return e2_action_register (name, type, func, data, has_arg, 0, NULL);
}

gint entry_level = 0;

/**
 * @brief register action names
 *
 * This function registers 'public' action names, and corresponding parameters, for later use as needed.
 * It replaces any previously-registered action with the same name.
 * Also adds the action name etc to the options store, as a convenience for the option system
 * Action names need to be freed (= action removed from the hash) when the session ends, if not before.
 * Potentially re-entrant.
 * As another convenience, the non-constant name string used to register the action is freed at exit.
 *
 * @param name non-constant string with possible '.' separating 'parent/ and 'child' parts of the action
 * @param type flags for one of: item (?) submenu, file_actions, custom_command, plugins, separator,
	tear_off_menu, toggle, command_line, bookmarks, dummy
 * @param func the function to be called when the action is run
 * @param data pointer to data to be supplied to @a func, or NULL
 * @param has_arg boolean flag signalling whether ??WHAT has argument(s)
 * @param exclude flags indicating whether to exclude this action from toolbars, menus &| keybindings
 * @param data2 additional data needed by the action, or NULL
*
 * @return registered action E2_Action
 **/
E2_Action *e2_action_register (gchar *name, E2_ACTION_TYPE type,
	void *func, gpointer data, gboolean has_arg, E2_ACTION_EXCLUDE exclude,
	gpointer data2)
{
	//unregister any previously registered actions with that name
	//(for convenience)

	entry_level++; //keep note of whether re-entrant

	e2_action_unregister (name);
	//now allocate the action structure and initialize it from the
	//function parameters
	E2_Action *action = g_malloc (sizeof (E2_Action));
	//freed when the action is removed from the actions_hash table
	//because the names are used as keys
	action->name = g_strdup (name);  //retain a copy that is freeable
	action->type = type;
	action->func = func;
	action->data = data;
	action->has_arg = has_arg;
	action->exclude = exclude;
	if ((type == E2_ACTION_TYPE_TOGGLE)
		&& entry_level == 1)
	{  //toggle action, and not re-entrant
		E2_ToggleActionEx *ex = g_malloc0 (sizeof (E2_ToggleActionEx));
		action->data2 = ex;
		g_ptr_array_add (toggles_array, ex);
	}
	else
		action->data2 = data2;

	//add the action to the static hash table and pointer array register
	g_hash_table_insert (actions_hash, action->name, action);
	g_ptr_array_add (actions_array, action);

	//to get structured lists, find the name of the 'parent' action, if any
	gboolean registered_parent = FALSE;
	E2_Action *parent = NULL;
	gchar *search = g_strdup (name);
	gchar *child = strrchr (search, '.');	//always ascii '.', don't need g_utf8_strrchr()
	if (child == NULL)
		child = strrchr (search, ':'); //ditto
	if (child != NULL)
	{ //also register a new parent, if need be
		*child = '\0';
		parent = e2_action_get (search);
		if (parent == NULL)
		{
			//watch out !! re-entry
			parent = e2_action_register_simple (search, E2_ACTION_TYPE_DUMMY,
				NULL, NULL, FALSE);
			registered_parent = TRUE; 
		}
	}

	//now add the action to the static tree store
	GtkTreeIter iter;
	if (parent == NULL)
		gtk_tree_store_append (actions_store, &iter, NULL);
	else
	{
		GtkTreeIter iter_parent;
		if (e2_tree_ref_to_iter (actions_store, parent->ref, &iter_parent))
			gtk_tree_store_append (actions_store, &iter, &iter_parent);
		else
			gtk_tree_store_append (actions_store, &iter, NULL);
	}
	//setup store; that's also convenience for the option system
	gtk_tree_store_set (actions_store, &iter, 0, action->name,
		1, child == NULL ? action->name : g_strstrip (child + 1),
		2, action->exclude, 3, action->type != E2_ACTION_TYPE_DUMMY, 4, TRUE, -1);

	//keep reference around to be able to to find out the parent iter
	//for possible children later on
	action->ref = e2_tree_iter_to_ref (actions_store, &iter);

	//cleanup
	if (!registered_parent)
		g_free (search);  //registered actions are kept, but this one was not needed
	else if (child != NULL)
		*child = '.';
	if (--entry_level == 0)  //revert re-entrant level, for next time
		g_free (name);  //and this is not re-entrant, so free the original name-space
	return action;
}

gboolean e2_action_unregister (const gchar *name)
{
	E2_Action *action = g_hash_table_lookup (actions_hash, name);
	if (action == NULL)
		return FALSE;
	return (_e2_action_unregister (action));
}

static gboolean _e2_action_unregister (E2_Action *action)
{
	if (!g_ptr_array_remove (actions_array, action))
		return FALSE;

	g_hash_table_remove (actions_hash, action->name);

	//free any stored strings
//	g_free (action->name); this one cleared in the hash table remove

	//free extra data for toggles
	if (action->type == E2_ACTION_TYPE_TOGGLE)
		g_free (action->data2);

	GtkTreeIter iter;
	if (e2_tree_ref_to_iter (actions_store, action->ref, &iter))
		gtk_tree_store_remove (actions_store, &iter);
	gtk_tree_row_reference_free (action->ref);

	g_free (action);
	return TRUE;
}

/**
 * @brief find an action by name
 *
 * The action with the name @a name is looked up in the internal hash table
 * and a pointer to the action object is returned. The name should be the
 * complete name of the action including any parent categories. If the action
 * cannot be found, NULL is returned.
 *
 * @param name  the name of the action
 *
 * @return  pointer to an action object or NULL
 **/
E2_Action *e2_action_get (gchar *name)
{
	E2_Action *action = g_hash_table_lookup (actions_hash, name);
/*	if (action == NULL)
	{
		gchar *name_try;
		//maybe there is a 'parameter' appended to the command, after a space
		if ((name_try = strchr (name, ' ')) != NULL)
		{
			name_try = g_strndup (name, name_try - name);
// superseded stuff
//			else if ((strchr (name, '.')) == NULL)
//			name_try = e2_utils_str_replace (name, ": ", ".");
//		else
//			name_try = e2_utils_str_replace (name, ".", ": ");
			action = g_hash_table_lookup (actions_hash, name_try);
			g_free (name_try);
		}
	} */
	return action;
}

/**
 * @brief run a certain action by name
 *
 * This function is to run a certain action and wraps around
 * e2_action_run_simple_from(). It sets the @a from widget
 * of e2_action_run_simple_from to NULL.
 *
 * @param name  the name of the action to run
 * @param arg  additional user data
 *
 * @return  TRUE on success, FALSE on failure
 **/
gboolean e2_action_run_simple (gchar *name, gpointer arg)
{
	return e2_action_run_simple_from (name, arg, NULL);
}

/**
 * @brief run a certain action by name
 *
 * This function is to run a certain action. First, the action specified by
 * @a name is looked up. Then a runtime object is created and the action is
 * run. Afterwards the runtime object is freed again. If the action could not
 * be found, an error message is printed to the output pane.
 * 
 * @param name  the name of the action to run
 * @param arg  additional user data
 * @param from  the widget the action is run from
 *
 * @return  TRUE on success, FALSE on failure
 **/
gboolean e2_action_run_simple_from (gchar *name, gpointer arg,
	GtkWidget *from)
{
	E2_Action *action = e2_action_get (name);
	if (action != NULL)
	{
		E2_ActionRuntime *actionrt = e2_action_pack_runtime (action,
			arg != NULL ? arg : NULL, NULL);
		e2_action_run_from (actionrt, from);
		e2_action_free_runtime (actionrt);
		return TRUE;
	} else
	{
		gchar *message = g_strdup_printf (_("Could not find action '%s'.\n"), name);
		e2_output_print_error (message);
		g_free (message);
		return FALSE;
	}
}

/**
 * @brief  run a certain action; may be used as callback
 *
 * This function is to run a certain action and wraps e2_action_run_from().
 * It may be used as callback to Gtk+ signals.
 * 
 * @param widget  this parameter is ignored
 * @param rt  the runtime object of the action
 *
 * @return  TRUE on success, FALSE on failure
 **/
gboolean e2_action_run (GtkWidget *widget, E2_ActionRuntime *rt)
{
	return e2_action_run_from (rt, widget);
}

/**
 * @brief  run a certain action
 *
 * This function is the real action run function. There are several wrappers
 * to simlify access. A runtime object has to be supplied that may contain
 * additional user data. The @a from widget is optional and may be used by
 * the action callback to determine the widget on which to operate.
 *
 * @param rt  the runtime object from this action run
 * @param from  the widget the action is run from or NULL
 *
 * @return  TRUE on success, FALSE on failure
 **/
gboolean e2_action_run_from (E2_ActionRuntime *rt, GtkWidget *from)
{
	//some quick checks
	if ((rt == NULL) || (rt->action == NULL) || (rt->action->func == NULL))
	{
		printd (DEBUG, "cannot run action, it doesn't exist");
		return TRUE;
	}
	printd (DEBUG, "e2_action_run_from (rt:%s,from:)", rt->action->name);
	gboolean (*fun) (gpointer, gpointer, GtkWidget *) = rt->action->func;
	return fun (rt->action->data, rt->data, from);
}

/**
 * @brief create an action runtime object
 *
 * This function creates an action runtime object and initializes it with
 * supplied user data.
 *
 * @param action  the action to create the runtime object for
 * @param data  action user data
 * @param data_free  function to free data when the runtime object is freed
 *
 * @return newly allocated action runtime object that should be freed
 **/
E2_ActionRuntime *e2_action_pack_runtime (E2_Action *action, gpointer data,
	gpointer data_free)
{
//	printd (DEBUG, "action pack runtime for %s", action->name);
	E2_ActionRuntime *rt = g_malloc (sizeof (E2_ActionRuntime));
	rt->action = action;
	rt->data = data;
	rt->data_free = data_free;
	return rt;
}

/**
 * @brief free action runtime object
 *
 * The runtime data is freed if a data free function has been supplied.
 * In addition the runtime object is freed and must not be used anymore after
 * calling this function.
 *
 * @param rt the action runtime object to free
 *
 * @return 
 **/
void e2_action_free_runtime (E2_ActionRuntime *rt)
{
	printd (DEBUG, "e2_action_free_runtime");
	if (rt == NULL)
		return;

	if (rt->data_free != NULL)
	{
		void (*fun) (gpointer) = rt->data_free;
		fun (rt->data);
	}
	g_free (rt);
}

/**
 * @brief internal action system memory cleanup
 *
 * Frees memory allocated to non-constant strings in each structure
 * recorded in the actions hash
 *
 * @return 
 **/
/* presently unused
static void free_name (gchar *key, E2_Action *value, gpointer data)
{
	g_free (key); //=value->name
	return;
} */

/**
 * @brief internal action system cleanup
 *
 * After calling the function, the action system may not be used anymore.
 * It is only called internally just before exit.
 *
 * @return 
 **/
/* this done only at session end, don't bother
void e2_actions_clean ()
{
//CHECKME = should this be able to handle ->data &| ->data2 ?
	g_object_unref (actions_store);
	//clean up non-constant action name strings in the hash
	g_hash_table_foreach (actions_hash, (GFunc) free_name, NULL);
	g_hash_table_destroy (actions_hash);
	//CHECKME = array cleanup ok for non-constant strings? could this replace the hash foreach?
	g_ptr_array_free (actions_array, TRUE);
} */

//static array of actions & pseudo-actions, same order as the
//name-generation statments in e2_actions_init()
//the first few of these like "<>" have no other logical home ?
static E2_Action e2_actions[] =
{
	{ "", E2_ACTION_TYPE_ITEM, _e2_action_custom_command_cb, TRUE, 0 },
	{ "", E2_ACTION_TYPE_FILE_ACTIONS, NULL, FALSE, E2_ACTION_EXCLUDE_ACCEL | E2_ACTION_EXCLUDE_TOOLBAR },
	{ "", E2_ACTION_TYPE_PLUGINS, NULL, FALSE, E2_ACTION_EXCLUDE_ACCEL },
	{ "", E2_ACTION_TYPE_SEPARATOR, NULL, FALSE, E2_ACTION_EXCLUDE_ACCEL  },
	{ "", E2_ACTION_TYPE_SUBMENU, NULL, FALSE, E2_ACTION_EXCLUDE_ACCEL  },
	{ "", E2_ACTION_TYPE_ITEM, e2_main_user_shutdown, FALSE, 0 },

	{ "", E2_ACTION_TYPE_ITEM, e2_task_configure, TRUE, 0 },
	{ "", E2_ACTION_TYPE_ITEM, e2_task_configure_default, FALSE, 0 },	
	{ "", E2_ACTION_TYPE_ITEM, e2_plugins_configure, FALSE, 0 },	
};

/**
 * @brief internal action system initialization
 *
 * This function is only called once at startup to initialize the action
 * system. Afterwards, it has no effect.
 * 
 * @return 
 **/
void e2_actions_init (void)
{
	//ensure that this function is only performed once
	RUN_ONCE_CHECK ();

	//initialize local data structures that are used to keep track of
	//the action parameters
	actions_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
	actions_array = g_ptr_array_new ();
	toggles_array = g_ptr_array_new ();
	actions_store = gtk_tree_store_new (5, G_TYPE_STRING, G_TYPE_STRING,
		G_TYPE_INT, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
	//FIXME: is this still necessary?
	g_object_ref (actions_store);
	//store actions sorted in the tree store
	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (actions_store);
    gtk_tree_sortable_set_sort_func (sortable, 0, _e2_action_tree_compare_cb,
		NULL, NULL);
    gtk_tree_sortable_set_sort_column_id (sortable, 0, GTK_SORT_ASCENDING);
	//note - the above need to be done before any plugin is loaded
	//note action labels setup in options initialiser, as names needed before config loaded
	//FIXME: this should be removed at some point if _all_ locally-defined
	//actions have been moved to their modules
	//all action names converted to discardable forms, for consistent freeing
	e2_actions[0].name = g_strdup(_A(16));
	e2_actions[1].name = g_strconcat(_A(5),".",_A(20),NULL);
	e2_actions[2].name = g_strconcat(_A(12),".",_A(23),NULL);
	e2_actions[3].name = g_strdup(_A(17));
	e2_actions[4].name = g_strdup(_A(18));

	e2_actions[5].name = g_strconcat(_A(1),".",_A(83),NULL);
	//configuration
	e2_actions[6].name = g_strconcat(_A(2),".",_A(32),NULL);
	e2_actions[7].name = g_strconcat(_A(2),".",_A(84),NULL);
	e2_actions[8].name = g_strconcat(_A(2),".",_C(31),NULL);

	gint count = sizeof (e2_actions) / sizeof (e2_actions[0]);
	gint i;
	for (i = 0; i < count; i++)
	{
		e2_action_register (e2_actions[i].name, e2_actions[i].type,
			e2_actions[i].func, NULL, e2_actions[i].has_arg,
			e2_actions[i].exclude, NULL);
	}
	//register most other actions
	// (some others are in pane create fn)
	e2_task_actions_register ();
	e2_output_actions_register();
//	e2_context_menu_actions_register();
	e2_bookmark_actions_register ();
	e2_filetype_actions_register ();
	e2_command_line_actions_register ();
	e2_command_actions_register ();
	e2_window_actions_register ();
	e2_pane_actions_register ();
	e2_toolbar_actions_register ();
	e2_dialog_actions_register ();
	e2_view_dialog_actions_register ();
	e2_search_dialog_actions_register ();
	e2_mkdir_dialog_actions_register ();
	e2_action_option_actions_register ();
#ifdef E2_FS_MOUNTABLE
	e2_complete_actions_register ();
#endif
}

/**
 * @brief setup array of i18n action labels
 *
 * No spaces in names, so action arguments can be separated
 * by a space
 * Some of these are the same as config dialog labels
 * Array size defined in e2_action.h
 * 
 * @return 
 **/
void e2_action_setup_labels (void)
{   //CHECKME= which of these names can be rationalised ?
	//'parent' names
	action_labels[0] = _("bookmark");
	action_labels[1] = _("command");
	action_labels[2] = _("configure");
	action_labels[3] = _("dialog");
	action_labels[4] = _("dirline");
	action_labels[5] = _("file");
	action_labels[6] = _("find");
	action_labels[7] = _("option");
	action_labels[8] = _("output");  // = _C(25)
	action_labels[9] = _("pane1");  // ~ _C(43)
	action_labels[10] = _("pane2");  // ~ C(44)
	action_labels[11] = _("panes");  // = _C(30)
	action_labels[12] = _("plugin");  //this is also used in menus - see PLUGIN
//	action_labels[13] = _("quit"); UNUSED

	action_labels[15] = _("separator");  //not really an action, but not a config either
	action_labels[16] = _("<custom command>");
	action_labels[17] = _("<separator>");
	action_labels[18] = _("<submenu>");

	//'child' names
	action_labels[20] = _("<actions>");
	action_labels[21] = _("<bookmarks>");
	action_labels[22] = _("<filter>");
	action_labels[23] = _("<list>");
	action_labels[24] = _("<line>");

	action_labels[26] = _("about");
	action_labels[27] = _("add");
	action_labels[28] = _("adjust_ratio");
	action_labels[29] = _("clear");
	action_labels[30] = _("clear_history");
	action_labels[31] = _("complete");
	action_labels[32] = _("application");
	action_labels[33] = _("copy");
	action_labels[34] = _("copy_as");
	action_labels[35] = _("delete");
	action_labels[36] = _("filetype");
	action_labels[37] = _("info");	//out of order, was "file info"
	action_labels[38] = _("focus");
	action_labels[39] = _("focus_toggle");
	action_labels[40] = _("go_back");
	action_labels[41] = _("go_forward");
	action_labels[42] = _("go_up");
	action_labels[43] = _("goto_bottom");
	action_labels[44] = _("goto_top");
	action_labels[45] = _("help");
	action_labels[46] = _("insert_selection");
	action_labels[47] = _("list_children");
	action_labels[48] = _("move");
	action_labels[49] = _("move_as");
	action_labels[50] = _("mkdir");
	action_labels[51] = _("open");
	action_labels[52] = _("open_in_other");
	action_labels[53] = _("open_with");
	action_labels[54] = _("owners"); //(ex user/group) = _C()
	action_labels[55] = _("page_down");
	action_labels[56] = _("page_up");
	action_labels[57] = _("permissions");  // = _C()
	action_labels[58] = _("print");
	action_labels[59] = _("refresh");
	action_labels[60] = _("rename");
	action_labels[61] = _("scroll_down");
	action_labels[62] = _("scroll_up");
	action_labels[63] = _("search");  // =  _C()
	action_labels[64] = _("select_all");
	action_labels[65] = _("set");
	action_labels[66] = _("show_menu");
	action_labels[67] = _("switch");
	action_labels[68] = _("symlink");
	action_labels[69] = _("symlink_as");
	action_labels[70] = _("sync");
	action_labels[71] = _("toggle_direction");
	action_labels[72] = _("toggle_hidden");
	action_labels[73] = _("toggle_select_all");
	action_labels[74] = _("toggle_full");
	action_labels[75] = _("view");  // =  _C()
	action_labels[76] = _("view_at");
	//late entries, out of order
	action_labels[77] = _("mirror");
	action_labels[78] = _("refreshsuspend");
	action_labels[79] = _("refreshresume");
//	action_labels[80] = _("toggle"); used for deprecated action only
	action_labels[81] = _("trash");
	action_labels[82] = _("trashempty");
	action_labels[83] = _("quit");
	action_labels[84] = _("default");

	//these are action _parameter_ strings
	action_labels[85] = _("child");
	action_labels[86] = _("ctrl");
	action_labels[87] = _("dirs");
	action_labels[88] = _("escape");
	action_labels[89] = _("expand");
	action_labels[90] = _("files");
	action_labels[91] = _("quote");
	action_labels[92] = _("shift");
	action_labels[93] = _("top");
	//maybe more space here, see array definition in header file
	action_labels[94] = _("mountpoints");
}

/*
//copy all actions, for help document
void build_list (gpointer key, gpointer value, GString *text)
{
	E2_Action *action = (E2_Action *) value;
	if (action->type != E2_ACTION_TYPE_DUMMY)
	{
		text = g_string_append (text, action->name);
		text = g_string_append_c (text, '\t');
		text = g_string_append_c (text, (action->has_arg) ? 'Y' : 'N');
		text = g_string_append_c (text, '\n');
	}
	return;
}
void e2_action_list_all (void)
{
	GString *cmds_list = g_string_new ("");
	g_hash_table_foreach (actions_hash, (GHFunc) build_list, cmds_list);

	GtkClipboard* clipper = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
	gtk_clipboard_set_text (clipper, cmds_list->str, cmds_list->len);
	ls e2*
	
	g_string_free (cmds_list, TRUE);
} */
