/* $Id: e2_cache.c 1118 2005-09-12 22:57:53Z tpgww $
 * 
 * Copyright (C) 2004 Florian Zaehringer (flo.zaehringer@web.de)
 * Copyright (C) 2004-2005 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file e2_cache.c
 * @brief cache system
 *
 * This file contains functions for e2's cache system: eg. to register and
 * unregister caches, and to write and read them to/from disk.
 **/

#include "e2_cache.h"
#include <string.h>
//#include <unistd.h>
//#include <glib/gprintf.h>
#include "e2_fs.h"
#include "e2_cl_option.h"

static gchar *default_cache_file;
static GList *cache_list;
static GList *unknown_list;

  /*****************/
 /***** utils *****/
/*****************/

static gint str_compare (E2_Cache *cache_element, gchar *name)
{
	return !g_str_equal (name, cache_element->name);
}

  /******************/
 /***** public *****/
/******************/

#ifdef E2_IMAGECACHE
E2_Image *e2_cache_image_cache (gchar *name, GtkIconSize size)
{
	if (name == NULL)
	{
		//default icon image name
		name = GTK_STOCK_MISSING_IMAGE;
	}
	E2_Image *cached = g_hash_table_lookup (cfg.icons, name);
	gint psize = e2_utils_get_icon_size (size);
	if (cached != NULL && cached->size == psize)
	{
//		g_object_ref (G_OBJECT (cached->pixbuf)); don't need to change this
//		cached->refcount++; not used, at present
	}
	else
	{
		cached = g_malloc (sizeof (E2_Image));
		if (e2_utils_check_stock_icon (name))
		{	//stock image
			cached->pixbuf = gtk_icon_set_render_icon (
				gtk_icon_factory_lookup_default (name),
				gtk_rc_get_style (app.main_window),
				gtk_widget_get_default_direction (),
				GTK_STATE_NORMAL, size, NULL, NULL);

//			cached->type = E2_IMAGE_BMP;
		}
		else
		{	//try to sniff the icon file
			gint width, height;
			GdkPixbufFormat *pixinfo = gdk_pixbuf_get_file_info
				(name, &width, &height);
			if (pixinfo != NULL)
			{
/*				gchar *fmt = gdk_pixbuf_format_get_name(pixinfo);
				if (g_str_equal (fmt, "svg"))	//CHECKME
					cached->type = E2_IMAGE_SVG;
				else
					cached->type = E2_IMAGE_BMP;
*/
				cached->pixbuf = gdk_pixbuf_new_from_file_at_size
					(name, psize, psize, NULL);
			}
			else
			{
				cached->pixbuf = gtk_icon_set_render_icon (
					gtk_icon_factory_lookup_default (GTK_STOCK_MISSING_IMAGE),
					gtk_rc_get_style (app.main_window),
					gtk_widget_get_default_direction (),
					GTK_STATE_NORMAL, size, NULL, NULL);

//				cached->type = E2_IMAGE_BMP;
			}
		}
		cached->size = psize;
//		cached->refcount = 1;
		g_hash_table_insert (cfg.icons, name, cached);
	}
	return cached;
}

E2_Image *e2_cache_image_get (gchar *name, GtkIconSize size)
{
	E2_Image *cached = g_hash_table_lookup (cfg.icons, name);
	gint psize = e2_utils_get_icon_size (size);
	if (cached == NULL || cached->size != psize)
		//cache this icon
		cached = e2_cache_image_cache (name, size);
//	else
//		cached->refcount++;

	return cached;
}
#endif

void e2_cache_int_register (gchar *name, gint *value, gint def)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register int cache '%s'", name);
		return;
	}
	//first-time processing
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_INT;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{ //not in the queue, either
		printd (DEBUG, "int cache '%s' not found, creating default with %d", name, def);
		cache->name = g_strdup (name);
		*value = def;
	} else
	{  //process the queued item
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		gchar *end = NULL;
		gchar *str = (gchar *) unknown->data;
		*value = g_ascii_strtoull (str, &end, 10);
		if (end == str)
		{
			printd (WARN, "int cache '%s' data conversion failed from '%s", name, str);
			*value = def;
		} else
			g_free (str);
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = GINT_TO_POINTER (value);
	cache_list = g_list_append (cache_list, cache);
	return;
}

void e2_cache_double_register (gchar *name, gdouble *value, gdouble def)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register double cache '%s'", name);
		return;
	}
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_DOUBLE;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{  //not in queue either
		printd (DEBUG, "double cache '%s' not found, creating default with %f", name, def);
		cache->name = g_strdup (name);
		*value = def;
	} else
	{  //process queued item
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		gchar *end = NULL;
		gchar *str = (gchar *) unknown->data;
		*value = g_ascii_strtod (str, &end);
		if (end == str)
		{
			printd (WARN, "double cache '%s' data conversion failed from '%s", name, str);
			*value = def;
		} else
			g_free (str);
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer *) value;
	cache_list = g_list_append (cache_list, cache);
	return;
}

void e2_cache_str_register (gchar *name, gchar **str, gchar *def)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register str cache '%s'", name);
		return;
	}
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_STR;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{
		printd (DEBUG, "str cache '%s' not found, creating dedault with %s", name, def);
		cache->name = g_strdup (name);
		*str = strdup(def);
	} else
	{
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		*str = (gchar *) unknown->data;
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer *) str;
	cache_list = g_list_append (cache_list, cache);
	return;
}

E2_Cache *e2_cache_list_register (gchar *name, GList **list)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register list cache '%s'", name);
		cache = found->data;
		return cache;
	}
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_LIST;
	cache->sync_func = NULL;
	cache->sync_data = NULL;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{
		printd (DEBUG, "list cache '%s' not found, creating NULL entry", name);
		cache->name = g_strdup (name);
		*list = NULL;
	} else
	{
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		*list = (GList *) unknown->data;
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer *) list;
	cache_list = g_list_append (cache_list, cache);
	return cache;
}

void e2_cache_list_unregister (gchar *name)
{
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{
		printd (WARN, "trying to unregister cache '%s' which doesn't exist", name);
		return;
	}
	cache_list = g_list_remove_link (cache_list, found);
	E2_Cache *cache = found->data;
	E2_Cache *unknown = malloc (sizeof (E2_Cache));
	unknown->name = cache->name;
	unknown->type = cache->type;
	unknown->data = *cache->data;
	unknown_list = g_list_append (unknown_list, unknown);
	g_free (cache);
}

void e2_cache_file_write (void)
{
	if (cache_list == NULL)
	{
		printd (DEBUG, "no caches registered");
		return;
	}

	gchar *filename = g_build_filename
		(e2_cl_options.config_dir, default_cache_file, NULL);

	E2_FILE *f = e2_fs_open_file_to_write (filename);
	if (f == NULL)
	{
		printd (WARN, "cannot write cache file: %s (%s)", filename, g_strerror (errno));
		g_free (filename);
		return;
	}
	printd (DEBUG, "write cache file: %s", filename);
	g_free (filename);

	e2_fs_file_write (f,
	    //first line is language-independent, for version checking
		"# "PROGNAME" (v "VERSION")\n\n");
	e2_fs_file_write (f,
		_("%sThis is the %s cache file.\n"
		"%sIt will be overwritten each time the program is shut down!\n\n"),
		 "# ", PROGNAME, "# ");

	GList *list;
	for (list = cache_list; list != NULL; list = g_list_next (list))
	{
		E2_Cache *cache = list->data;
		switch (cache->type)
		{
			case E2_CACHE_TYPE_BOOL:
				e2_fs_file_write (f, "%s=%s\n", cache->name,
					*cache->data ? "true" : "false");
				break;
			case E2_CACHE_TYPE_DOUBLE:
			{
				gchar doubl[G_ASCII_DTOSTR_BUF_SIZE];
				g_ascii_dtostr (doubl, G_ASCII_DTOSTR_BUF_SIZE, *((gdouble *) cache->data));
				e2_fs_file_write (f, "%s=%s\n", cache->name, doubl);
			}
				break;
			case E2_CACHE_TYPE_INT:
				e2_fs_file_write (f, "%s=%d\n", cache->name, GPOINTER_TO_INT (*cache->data));
				break;
			case E2_CACHE_TYPE_STR:
				e2_fs_file_write (f, "%s=%s\n", cache->name, (gchar *) *cache->data);
				break;
			case E2_CACHE_TYPE_LIST:
				if (cache->sync_func != NULL)
				{
					void (*fun) (gpointer) = cache->sync_func;
					fun (cache->sync_data);
				}
				if ((cache->data != NULL) && (*cache->data != NULL))
				{
					e2_fs_file_write (f, "<%s\n", cache->name);
					GList *tmp;
					for (tmp = (GList *) *cache->data; tmp != NULL; tmp = g_list_next (tmp))
					{
						e2_fs_file_write (f, "%s\n", (gchar *) tmp->data);
					}
					e2_fs_file_write (f, ">\n");
					//CHECKME = leaks when this is done before session-end ??
				}
				break;
			default:
				printd (WARN, "don't know how to write '%s' to cache file",
					cache->name);
				break;
		}
	}
	//FIXME: write unknown cache options to file
	e2_fs_file_close (f);
}

static void _e2_cache_read_from_string (gchar *f[])
{
	gboolean list_mode = FALSE;
	GList *list = NULL;
	E2_Cache *cache = NULL;
	
	gint i = 0;
	gchar *line;
	
	while ((line = f[i++]) != NULL)
	{
		g_strchomp (line);
		//ignore empty lines and comments
		if (*line == '\0')
			continue;
		if (line[0] == '#')
			continue;
		
		if (line[0] == '<')
		{
			list_mode = TRUE;
			list = NULL;
			cache = malloc (sizeof (E2_Cache));
			cache->name = g_strdup (line + 1);
			cache->type = E2_CACHE_TYPE_LIST;
			continue;
		}
		
		if (line[0] == '>')
		{
			list_mode = FALSE;
			if (list == NULL)
				continue;
			cache->data = (gpointer *) list;
			unknown_list = g_list_append (unknown_list, cache);
			continue;
		}
		
		if (list_mode)
		{
			list = g_list_append (list, g_strdup (line));
		} else
		{
			cache = malloc (sizeof (E2_Cache));
			gchar **split = g_strsplit (line, "=", 2);
			cache->name = g_strdup (split[0]);
			cache->data = (gpointer *) g_strdup (split[1]);
			cache->type = E2_CACHE_TYPE_STR;
			g_strfreev (split);
			unknown_list = g_list_append (unknown_list, cache);
		}
	} 
}
static void _e2_cache_file_read (gchar *fn)
{
	//find absolute path to config file
	gchar *filepath = g_build_filename (e2_cl_options.config_dir, fn, NULL);
	gchar *cont;
	if (! e2_fs_get_file_contents (filepath, &cont))
	{
		printd (WARN, "could not open cache file '%s'", filepath);
	}
	else
	{
		printd (DEBUG, "cache read from file '%s'", filepath);
		gchar **split = g_strsplit (cont, "\n", -1);
		_e2_cache_read_from_string (split);
		g_strfreev (split);
		g_free (cont);
	}
	g_free (filepath);
}
/**
 * @brief initialize the cache
 *
 * This function will initilize the cache system of e2. It will automatically
 * be called at startup.
 *
 * @return 
 **/
void e2_cache_init (void)
{
	cache_list = NULL;
	unknown_list = NULL;
	default_cache_file = "cache"; //DO NOT TRANSLATE THIS
	_e2_cache_file_read (default_cache_file);
}
