/* $Id: e2_cl_option.c 1113M 2005-09-13 11:09:38Z (local) $
 * 
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "emelfm2.h"
#include <getopt.h>
#include <string.h>
#include <glib/gprintf.h>
#include "e2_cl_option.h"

E2_CommandLineOptions e2_cl_options;

static void print_usage ()
{
	g_printf ("usage: %s [option]\n", BINNAME);
}

static void print_help ()
{
	print_usage ();
	g_printf
	(
		_("Program options:\n"
		"-1,--one=DIR           set 1st pane's start directory to DIR\n"
		"-2,--two=DIR           set 2nd pane's start directory to DIR\n"
		"-c,--config=DIR        set config directory to DIR (default: ~/."BINNAME")\n"
		"-d,--debug=[1-5]       set debug level from 1/error to 5/debug\n"
		"-e,--encoding=TYPE     set filesystem character encoding to TYPE\n"
		"-f,--fallback-encoding set fallback encoding (default: ISO-8859-1)\n"
		"-i,--ignore-problems   ignore encoding/locale problems (at your own risk!)\n"
		"-l,--leave-alone       don't suppress gtk output on console\n"
		"-n,--no-detach         don't detach from terminal\n"
		"-r,--run-at-start      command to run after session start\n"
		"-s,--set-option        set one-line gui options in config file format\n"
		"-t,--trash=DIR         set trash directory to DIR (default: ~/."BINNAME"/Trash)\n"
		"-x,--verbose           display time/location info on debug output\n"
		"\n"
		"Help options:\n"
		"-h,--help              show this help message\n"
		"-u,--usage             display brief usage messsage\n"
		"-v,--version           display version and build info\n")
	);
}

static void print_version ()
{
	g_printf
	(
		_("%s v. %s\n"
		"Licensed under the GPL\n"
		"Copyright (C) %s\n"
		"Build date: %s\n"
		"Build platform: GTK+ %d.%d.%d %s \n"),
		PROGNAME, VERSION, COPYRIGHT, BUILDDATE,
		GTK_MAJOR_VERSION, GTK_MINOR_VERSION, GTK_MICRO_VERSION, BUILDINFO
	);
}

void e2_cl_option_process (int argc, char *argv[])
{
	e2_cl_options.pane1_path = NULL;
	e2_cl_options.pane2_path = NULL;
	e2_cl_options.config_dir = NULL;
	e2_cl_options.encoding = NULL;
	e2_cl_options.fallback_encoding = g_strdup ("ISO-8859-1");
	e2_cl_options.overrides = NULL;
	e2_cl_options.debug_level = DEBUG_LEVEL;
	e2_cl_options.no_detach = DEBUG_LEVEL > 4 ? TRUE : FALSE;
	e2_cl_options.verbose = FALSE;
	e2_cl_options.suppress_gtk_output = DEBUG_LEVEL > 4 ? FALSE : TRUE;
	e2_cl_options.ignore_problems = FALSE;

	gint c, d = 0;
	while (1)
	{
		gint option_index = 0;
		static struct option long_options[] =
		{
			{"one", 1, 0, '1'},
			{"two", 1, 0, '2'},
			{"config", 1, 0, 'c'},
			{"debug", 1, 0, 'd'},
			{"ignore-problems", 0, 0, 'i'},
			{"encoding", 0, 0, 'e'},
			{"fallback-encoding", 0, 0, 'f'},
			{"leave-alone", 0, 0, 'l'},
			{"no-detach", 0, 0, 'n'},
			{"run-at-start", 1, 0, 'r'},
			{"set-option", 1, 0, 's'},
			{"verbose", 0, 0, 'x'},
			{"help", 0, 0, 'h'},
			{"trash", 1, 0, 't'},
			{"usage", 0, 0, 'u'},
			{"version", 0, 0, 'v'},
			{0, 0, 0, 0}
		};

		c = getopt_long (argc, argv, "1:2:c:d:e:f:ilnr:s:xhuv", long_options, &option_index);
		if (c == -1) break;

		d++;
		
		switch (c)
		{
			case '1':
				printd (DEBUG, "setting pane 1 startup path '%s'", optarg);
				e2_cl_options.pane1_path = g_strdup (optarg);
				break;
			case '2':
				printd (DEBUG, "setting pane 2 startup path '%s'", optarg);
				e2_cl_options.pane2_path = g_strdup (optarg);
				break;
			case 'c':
				printd (DEBUG, "setting config directory '%s'", optarg);
				e2_cl_options.config_dir = g_strdup (optarg);
				break;
			case 'd':
				printd (DEBUG, "setting debug level '%s'", optarg);
				{
					gchar *end;
					e2_cl_options.debug_level = g_ascii_strtoull (optarg, &end, 10);
					if (end == optarg)
					{
						e2_cl_options.debug_level = DEBUG_LEVEL;
						printd (WARN, "failed setting debug level '%s'", optarg);
					}
					if (e2_cl_options.debug_level > 4)
						e2_cl_options.suppress_gtk_output = FALSE;
				}
				break;
			case 'i':
				e2_cl_options.ignore_problems = TRUE;
				break;
			case 'e':
				e2_cl_options.encoding = g_strdup (optarg);
				break;
			case 'f':
				g_free (e2_cl_options.fallback_encoding);
				e2_cl_options.fallback_encoding = g_strdup (optarg);
				break;
			case 'l':
				e2_cl_options.suppress_gtk_output = FALSE;
			case 'n':
				e2_cl_options.no_detach = TRUE;
				break;
			case 'r':
				e2_cl_options.startup_commands = g_list_append 
					(e2_cl_options.startup_commands, g_strdup (optarg));
				break;
			case 's':
				e2_cl_options.option_overrides = g_list_append
					(e2_cl_options.option_overrides, g_strdup (optarg));
				break;
			case 't':
				printd (DEBUG, "setting trash directory '%s'", optarg);
				e2_cl_options.trash_dir = g_strdup (optarg);
				break;
			case 'x':
				e2_cl_options.verbose = TRUE;
				break;
			case 'u':
				print_usage ();
				exit (0);
				break;
			case 'h':
				print_help ();
				exit (0);
				break;
			case 'v':
				print_version ();
				exit (0);
				break;
			case '?':
				exit (0);
				break;
			default:
				printd (ERROR, "unknown option with character code: 0%o\n", c);
		}
	}

	if (optind < argc)
	{
		g_printf ("You may only supply options starting with - or --.\n");
		exit (0);
	}
	printd (DEBUG, "parsed %d command line options", d);
}
