/* $Id: e2_option.c 1067 2005-08-22 23:13:36Z tpgww $
 *
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file e2_option.c
 * @brief general option handling
 *
 * This file contains general functions to register, unregister and set
 * options.
 **/

#include "emelfm2.h"
#include <string.h>
#include "e2_fs.h"
#include "e2_option.h"
#include "e2_option_tree.h"
#include "e2_action.h"
#include "e2_task.h"
#include "e2_filetype.h"
#include "e2_plugins.h"
#include "e2_cl_option.h"
#include "e2_filelist.h"

typedef enum
{
	E2_TREE_STARTED,
	E2_TREE_COMPLETED,
	E2_TREE_ABORTED
} E2_TreeStatus;

static gboolean _e2_option_set_from_string (gchar *option, gchar *str);
static gboolean _e2_option_set_value_from_string_ (E2_OptionSet *set, gchar *str);

  /******************/
 /***** public *****/
/******************/

/**
 * @brief check if config dir exists
 *
 * Initialize the config directory internal pointer and check if it really exists. If not,
 * create it or print an appropriate error message through the debug system.
 * If it exists or is successfully created, log its creation date/time.
 * 
 * @return TRUE if dir exists or is successfully created, else FALSE
 **/
gboolean e2_option_check_config_dir (void)
{
	struct stat stat_buf;
	//check if config dir is already set through a command line option
	if (e2_cl_options.config_dir == NULL || *e2_cl_options.config_dir == '\0')
	{
		//if not, build it
		const gchar *cfg = g_getenv ("XDG_CONFIG_HOME");
		if (cfg != NULL)
			e2_cl_options.config_dir = g_build_filename (cfg, BINNAME, NULL);
		else
		{
			const gchar *home = g_get_home_dir ();
			if (home != NULL)
				e2_cl_options.config_dir = g_build_filename (E2_CONFIG_DIR, NULL);
			else
			{
				e2_output_print_error(_("Cannot identify where to get config data. Please set the XDG_CONFIG_HOME or HOME environment variable."));
				return FALSE;	//exit (1);
			}
		}
	}
	else  //config dir was specified in the startup command
	{
		//if the command line config dir path is not absolute,
		//correct this to prevent problems later on
		if (!g_path_is_absolute (e2_cl_options.config_dir))
		{
			gchar *current_dir = g_get_current_dir ();
			gchar *config_dir = g_build_filename (current_dir, e2_cl_options.config_dir, NULL);
			g_free (current_dir);
			g_free (e2_cl_options.config_dir);
			e2_cl_options.config_dir = config_dir;
		}
	}
	//check if the config dir is really there
	F_DECLARE;
	gchar *local = F_FILENAME_TO_LOCALE (e2_cl_options.config_dir);
	if (e2_fs_access (local, F_OK))
	{	//if not, create it
		printd (INFO, "creating config directory '%s'", e2_cl_options.config_dir);
		if (e2_fs_mkdir (local, 0755))
		{
			printd (WARN, "could not create config directory '%s'",
				e2_cl_options.config_dir);
			F_FREE (local);
			return FALSE;
		}
		else	//succeeded
		{
//#ifndef E2_FAM
			//save config dir timestamp
			e2_fs_stat (local, &stat_buf);
			cfg.config_mtime = stat_buf.st_mtime;
//#endif
		}
	}
	else if (!e2_fs_is_dir3 (local))
	{
		printd (WARN, "could not create config directory '%s', "
			"something is in the way", e2_cl_options.config_dir);
		F_FREE (local);
		return FALSE;
	}
	else
	{
//#ifndef E2_FAM
		e2_fs_stat (local, &stat_buf);
		cfg.config_mtime = stat_buf.st_mtime;
//#endif
	}
	F_FREE (local);
	return TRUE;
}
/**
 * @brief Initialize the default trash directory and a pointer to it
 *
 * Uses command-line option if that was provided. Otherwise,
 * tries to use $HOME/.Trash. If that is not available, uses the 
 * e2 default. Creates dir if doesn't exist, or print error message.
 * Sets cfg.trashpath to the path, or to NULL if dir not available.
 * Path has a trailing '/'.
 *
 * @return
 **/
void e2_option_set_trash_dir (void)
{
	gchar *local;
	F_DECLARE;
	//if trash dir was not specified by a command line option
	if (e2_cl_options.trash_dir == NULL || *e2_cl_options.trash_dir == '\0')
	{
		const gchar *home = g_get_home_dir ();
		if (home == NULL)
		{
			e2_output_print_error(_("Cannot identify your home directory. Please set the HOME environment variable.\n"));
			cfg.trashpath = NULL;
			return;
		}
		cfg.trashpath = g_build_filename (SYSTEM_TRASH_DIR, G_DIR_SEPARATOR_S, NULL);
		local = F_FILENAME_TO_LOCALE (cfg.trashpath);
		if (!e2_fs_is_dir3 (local))
		{
			g_free (cfg.trashpath);
			cfg.trashpath = g_build_filename (E2_TRASH_DIR, G_DIR_SEPARATOR_S, NULL);
		}
		F_FREE (local);
	}
	else //use the dir specified on the command line
		cfg.trashpath = g_strconcat (e2_cl_options.trash_dir, G_DIR_SEPARATOR_S, NULL);
	//make the trash dir if it doesn't exist
	local = F_FILENAME_TO_LOCALE (cfg.trashpath);
	if (!e2_fs_is_dir3 (local))
	{
		if (e2_fs_mkdir (local, 0755))
		{
			gchar *message = g_strdup_printf (_("Cannot create trash directory %s\n"), cfg.trashpath);
			e2_output_print_error(message);
			g_free (message);
			g_free (cfg.trashpath);
			cfg.trashpath = NULL;
		}
	}
	F_FREE (local);
}
/**
 * @brief dump all current configuration data and recreate it, with window content updates as appropriate
 *
 * @param reload TRUE to reload the config file before recreating config data
 * @param recreate TRUE to recreate main window in accord with updated data
 *
 * @return
 **/
void e2_option_refresh (gboolean reload, gboolean recreate)
{
	cfg.completed = FALSE;	//block recursion
	//this is the one option that we want to keep ...
	gboolean advanced = e2_option_bool_get ("advanced-config");
	e2_keybinding_clean ();  //??no need to destroy the categories list, even though actual bindings may change
	e2_plugins_unload_all ();
	g_hash_table_destroy (cfg.filetypes);
	e2_option_tree_stores_clear ();
	//some (imperfect) cleanups
	gint i;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
//		printd (DEBUG, "destroying set %s", set->name);
		E2_OptionFlags frees = set->flags;

		if (frees & E2_OPTION_FLAG_FREENAME)
			g_free (set->name);
		if (frees & E2_OPTION_FLAG_FREEGROUP)
			g_free (set->group);
		if (frees & E2_OPTION_FLAG_FREEDESC)
			g_free (set->desc);
		if (frees & E2_OPTION_FLAG_FREETIP)
			g_free (set->tip);
		if (frees &
			(E2_OPTION_FLAG_INT | E2_OPTION_FLAG_STR
			 | E2_OPTION_FLAG_SEL | E2_OPTION_FLAG_FONT
			 | E2_OPTION_FLAG_COLOR))
			g_free (set->sval);

		//extras
		if (frees & E2_OPTION_FLAG_TREE)
		{
			if (set->ex.tree.def != NULL)
				g_strfreev (set->ex.tree.def);
			e2_list_free_with_data (&set->ex.tree.columns);	//LEAK - each column has freeable things
		}
		//sel-option choices are all constant strings ??
		
		//CHECKME free hooks ??
	}
	g_ptr_array_free (options_array, TRUE);
	options_array = g_ptr_array_new ();

	g_hash_table_destroy (options_hash);
	options_hash = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);

//	g_hash_table_destroy (options_queue);
//	options_queue = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

	//now re-create things, in the same order as at session-start

	e2_option_default_register ();

	if (reload)
	{
		printd (INFO, "reloading config due to external change");
		e2_option_file_read ();
	}

	e2_option_trees_confirm ();

	e2_plugins_add_all (); // load plugins (if any)

	//re-initialise things that are not done in normal 'recreate' functions
	e2_pane_create_option_data (&app.pane1);
	e2_pane_create_option_data (&app.pane2);

	e2_toolbar_initialise (E2_BAR_PANE1);
	e2_toolbar_initialise (E2_BAR_PANE2);
	e2_toolbar_initialise (E2_BAR_TASK);
	e2_toolbar_initialise (E2_BAR_COMMAND);

	if (recreate)
		e2_window_recreate (&app.window); //this also recreates key bindings
	else
		e2_keybinding_register_all ();

	e2_filetype_add_all ();

	e2_option_bool_set ("advanced-config", advanced);

	cfg.completed = TRUE;	//unblock
}
/**
 * @brief disable checking for changed config file, if that is in effect
 *
 * @return
 **/
void e2_option_disable_config_checks (void)
{
	if (e2_option_bool_get ("auto-refresh-config"))
	{
#ifdef E2_FAM
		if (cfg.monitor_type != E2_MONITOR_DEFAULT)
			e2_fs_FAM_config_ignore ();	//don't trigger a config file reload due to this write
		else
#endif
//		if (cfg.check_config_id != 0)
			g_source_remove (cfg.check_config_id);
	}
}
/**
 * @brief enable checking for changed config file, if that is in effect
 *
 * @return
 **/
void e2_option_enable_config_checks (void)
{
	if (e2_option_bool_get ("auto-refresh-config"))
	{
#ifdef E2_FAM
		if (cfg.monitor_type != E2_MONITOR_DEFAULT)
			e2_fs_FAM_config_monitor ();
#endif
		cfg.check_config_id = g_timeout_add 
			(E2_CONFIGCHECK_INTERVAL, (GSourceFunc) e2_option_check_config_files, NULL);
	}
}

gboolean cfgdirty;

/**
 * @brief check for and respond to changed config file
 *
 * This is the fn called by the check-config-files timer
 * With FAM use, update is normally detected and flagged as part
 * of the filepane monitoring, as all change-data arrives via the
 * same stream
 * Otherwise, here we check config file mtim.tvsec against a stored
 * value.
 * If update is indicated, recreate window after re-reading config data
 *
 * @return TRUE, always, so timer continues
 **/
gboolean e2_option_check_config_files (void)
{
#ifdef E2_FAM
	if (cfg.monitor_type == E2_MONITOR_DEFAULT)
	{
#endif
		struct stat stat_buf;
		F_DECLARE;

//		gchar *local = F_FILENAME_TO_LOCALE (e2_cl_options.config_dir);
//		if (e2_option_check_config_dir ())
//		{
			gchar *filename = g_build_filename (e2_cl_options.config_dir, default_config_file, NULL);
			gchar *local = F_FILENAME_TO_LOCALE (filename);
			cfgdirty = (e2_fs_stat (local, &stat_buf) == 0
				&& stat_buf.st_mtime != cfg.config_mtime);
			if (cfgdirty && cfg.completed)
				cfg.config_mtime = stat_buf.st_mtime;
			F_FREE (local);
			g_free (filename);
//		}
#ifdef E2_FAM
	}
#endif
	if (cfgdirty && cfg.completed)
	{
		e2_option_refresh (TRUE, TRUE);
		status_message (_("Configuration data re-loaded\n"));
		cfgdirty = FALSE;
	}
	return TRUE;
}
/**
 * @brief revert all options to default values
 *
 *
 * @return
 **/
void e2_option_restore_defaults (void)
{
	//dump old data, don't re-read the config file, recreate screen
	e2_option_refresh (FALSE, TRUE);
}
/**
 * @brief register config option
 *
 * @param set pointer to structure E2_OptionSet to be registered
 *
 * @return was TRUE if option was converted from unknown queue to registered
 * @return  FALSE otherwise, ie the option is already registered, or is not in the unknown options queue
**/
//gboolean 
void e2_option_register (E2_OptionSet *set)
{
	set->widget = NULL;

	if ((g_hash_table_lookup (options_hash, set)) == NULL)
	{
		g_ptr_array_add (options_array, set);
		g_hash_table_insert (options_hash, set->name, set);
		g_hook_list_init (&set->hook_value_changed, sizeof (GHook));
		set->hook_freezed = FALSE;
/*
		gchar *value;
		if ((value = g_hash_table_lookup (options_queue,
			set->name)) != NULL)
		{
			if (set->type == E2_OPTION_TYPE_TREE)
				set->ex.tree.unknown = (gchar **) value;
			else
			{
				_e2_option_set_value_from_string_ (set, value);	//set default value, no checks
				g_free (value);
			}
			//FIXME = free any space allocated to members, before doing this
			g_hash_table_remove (options_queue, set->name);
			return TRUE;
		}
*/
	}
//	return FALSE;
}

/**
 * @brief get E2_OptionSet with warning
 *
 * @param option name of the option
 *
 * @return option or NULL in case i cannot be found
 **/
E2_OptionSet *e2_option_get (gchar *option)
{
	E2_OptionSet *set;
	if ((set = (E2_OptionSet *) g_hash_table_lookup (options_hash, option)) == NULL)
	{
		printd (WARN, "trying to get option '%s' which doesn't exist", option);
		return NULL;
	}
	return set;
}
/**
 * @brief get E2_OptionSet without warning
 *
 * @param option name of the option
 *
 * @return option or NULL if the option cannot be found
 **/
E2_OptionSet *e2_option_get_simple (gchar *option)
{
	return (E2_OptionSet *) g_hash_table_lookup (options_hash, option);
}

/* UNUSED
void *e2_option_void_get (E2_OptionSet *set, gchar *option)
{
	switch (set->type)
	{
		case E2_OPTION_TYPE_BOOL:
			if (g_str_equal (option, "true")) return (void *) TRUE;
			else return (void *) FALSE;
			break;
		case E2_OPTION_TYPE_STR:
		case E2_OPTION_TYPE_FONT:
		case E2_OPTION_TYPE_ICON:
		case E2_OPTION_TYPE_SEL:
			return (void *) option;
		default:
			return NULL;
			break;
	}
	return NULL;
}
*/

E2_OptionSet *e2_option_attach_value_changed (gchar *option, GtkWidget *widget,
	gpointer func, gpointer data)
{
	E2_OptionSet *set = e2_option_get (option);
	if (set != NULL)
		e2_option_attach_value_changed_simple (set, widget, func, data);
	return set;
}

void e2_option_attach_value_changed_simple (E2_OptionSet *set, GtkWidget *widget,
	gpointer func, gpointer data)
{
	GHook *hook = e2_hook_register (&set->hook_value_changed, func, data);
	E2_Duo *duo = g_malloc (sizeof (E2_Duo));
	duo->a = hook;
	duo->b = &set->hook_value_changed;
	g_object_set_data_full (G_OBJECT (widget), "e2-option-attach-value-changed",
		duo, (GDestroyNotify) e2_hook_unattach_cb);
}

void e2_option_tree_stores_clear (void)
{
	gint i;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
		if (set->type == E2_OPTION_TYPE_TREE)
			gtk_tree_store_clear (set->ex.tree.model);
	}
}
/* this was used only at session end, so don't bother to cleanup
void destroy_config ()
{
	g_hash_table_destroy (options_hash);
	g_hash_table_destroy (options_queue);
	g_ptr_array_free (options_array, TRUE);
} */

/**
 * @brief cleanup helper for icon hash
 *
 * @param data pointer to E2_Image struct for an icon
 */
#ifdef E2_IMAGECACHE
// this is for icon-caching, post-0.1
void _e2_option_iconhash_remove (E2_Image *data)
{
	g_object_unref (G_OBJECT (data->pixbuf));
	g_free (data);
}
/**
 * @brief setup for icon caching
 *
 */
void e2_option_icons_init (void)
{
	cfg.icons = g_hash_table_new_full (g_str_hash, g_str_equal,
		NULL, (GDestroyNotify) _e2_option_iconhash_remove);
	e2_utils_init_icon_sizes ();
}
#endif
/**
 * @brief write current configuration to file
 *
 * The current configuration will be written to the default config file 'config'.
 */
void e2_option_file_write (void)
{
	gchar *filepath = g_build_filename (e2_cl_options.config_dir,
		default_config_file, NULL);

	E2_FILE *f = e2_fs_open_file_to_write (filepath);
	if (f == NULL)
	{
		printd (WARN, "Cannot write config file: %s (%s)", filepath, g_strerror (errno));
		g_free (filepath);
		return;
	}
	printd (DEBUG, "write config file: %s", filepath);
	g_free (filepath);

	e2_fs_file_write (f,
		//the 1st line is language-independent, for version verification
		"# "PROGNAME" (v "VERSION")\n\n" );
	e2_fs_file_write (f,
	 //FIXME = remind translators not to remove # from line starts
	  _("# This is the %s configuration data file.\n"
		"# It will be overwritten each time the program is run!\n\n"
		"# If you're inclined to edit the file between program sessions, note this:\n"
		"# for tree options, you have to use \\| to escape | and you have to use \\< to escape <,\n"
		"#  if that is the first non-space character on a line.\n\n"),
		PROGNAME
	);

	gint i;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
		switch (set->type)
		{
			case E2_OPTION_TYPE_INT:
				e2_fs_file_write (f, "# %s\n%s=%d\n",
					set->desc, set->name, set->ival);
				break;
			case E2_OPTION_TYPE_BOOL:
//				e2_fs_file_write (f, "# %s\n%s=%s\n",
//					set->desc, set->name, set->sval);
//				break;
			case E2_OPTION_TYPE_STR:
			case E2_OPTION_TYPE_FONT:
			case E2_OPTION_TYPE_COLOR:
//				e2_fs_file_write (f, "# %s\n%s=%s\n",
//					set->desc, set->name, set->sval);
//				break;
			case E2_OPTION_TYPE_SEL:
				e2_fs_file_write (f, "# %s\n%s=%s\n",
					set->desc, set->name, set->sval);
				break;
			case E2_OPTION_TYPE_TREE:
				e2_fs_file_write (f, "# %s\n<%s\n",
					set->name, set->name);
				GtkTreeIter iter;
				if (gtk_tree_model_get_iter_first (set->ex.tree.model, &iter))
					e2_option_tree_write_to_file (f, set, &iter, 0);
				e2_fs_file_write (f, ">\n");
				break;
			default:
				printd (WARN, "don't know how to write option '%s' to config file",
					set->name);
				break;
		}
	}

	e2_fs_file_close (f);
}
/**
 * @brief Set value of option named @a option
 *
 * This is called from the config file line parser.
 * If the set data exists already, its value is set according to @a str.
 * If the set data does not exist already, the set is registered in the 
 * 'unknown' queue
 *
 * @param option name of option to be set
 * @param str string with value to be stored for option named @a option
 *
 * @return TRUE if @a option was set successfully, even if in the unknown options queue
 **/
static gboolean _e2_option_set_from_string (gchar *option, gchar *str)
{
	E2_OptionSet *set = e2_option_get_simple (option);
	if (set == NULL)
	{
		//CHECKME = discardable group name ?
//		e2_option_unknown_register (option, str);
//		return TRUE;
		return FALSE;
	}
	return _e2_option_set_value_from_string_ (set, str);
}
/**
 * @brief Setup the default value for @a set in accord with parameter @a str 
 * 
 * This sets the appropriate set->sval, depending on the type of option
 * There is no check for existence of @a set
 *
 * @param set pointer to data struct for the option being processed
 * @param str string form of value to be stored for @a set
 *
 * @return 
 **/
static gboolean _e2_option_set_value_from_string_
	(E2_OptionSet *set, gchar *str)
{
	gboolean retval = TRUE;
	switch (set->type)
	{
		case E2_OPTION_TYPE_BOOL:
			_e2_option_bool_set (set, g_str_equal (str, "true") ? TRUE : FALSE);
			break;
		case E2_OPTION_TYPE_INT:
		{
			//FIXME: check for conversion success
			gint i = (gint) g_ascii_strtoull (str, NULL, 10);
			_e2_option_int_set (set, i);
			break;
		}
		case E2_OPTION_TYPE_STR:
		case E2_OPTION_TYPE_FONT:
			_e2_option_str_set (set, str);
			break;
		case E2_OPTION_TYPE_COLOR:
			return _e2_option_color_set_str (set, str);
			break;
		case E2_OPTION_TYPE_SEL:
		{
			gchar *val;
			gint i = 0;
			set->ival = -1;
			while ((val = set->ex.sel.def[i]) != NULL)
			{
				if (g_str_equal (str, val))
				{
					set->ival = i;
					set->sval = val;
					break;
				}
				i++;
			}
			if (set->ival == -1)
			{
				printd (WARN, "bad value for sel option '%s'", set->name);
				retval = FALSE;
				set->ival = 0;
				set->sval = set->ex.sel.def[0];
			}
			if (!set->hook_freezed)
				e2_hook_list_run (&set->hook_value_changed, GINT_TO_POINTER (set->ival));
		}
			break;
		default:
			break;
	}
	return retval;
}
/**
 * @brief Read config options from @a f and parse them
 *
 * Contents of f[] are assumed to be ascii
 *
 * @param f NULL-terminated array of strings in config file format
 *
 * @return 
 **/
void e2_option_read_from_string (gchar *f[])
{
	//this flag tells whether a tree option (>1 line) is being processed 
	E2_TreeStatus tree_mode = E2_TREE_COMPLETED;
	//helper GList for tree options used to process parent/child dependencies
	GList *parents = g_list_append (NULL, NULL);
	//line counter
	gint i = 0;
	//pointer to the current line
	gchar *line;
	//helper for tree option processing to save pointer to the tree option
	E2_OptionSet *set = NULL;

	while ((line = f[i++]) != NULL)
	{
		g_strchomp (line);
		//ignore empty lines and comments
		if (*line == '\0') continue;
		if (line[0] == '#') continue;
		
		//tree option begins
		if (line[0] == '<')
		{
			tree_mode = E2_TREE_STARTED;
			//try to find pointer to option
			set = e2_option_get_simple (line + 1);
			//found config for this tree option, so don't init it with the
			//default config later on
			if (set != NULL)
			{
//				printd (DEBUG, "found config tree for %s>", line);
				set->ex.tree.found_in_config_file = TRUE;
			}
//			else
				//unknown tree option, save for later processing
				//(when the option is actually added)
////CHECKME = discardable group name
//				i = e2_option_tree_register_unknown (line + 1, i - 1, f);

			continue;
		}
		
		//tree option ends
		if (line[0] == '>')
		{
			//make sure the set is recognised as clean
			set->ex.tree.flags &= ~E2_OPTION_TREE_SET_EDITED;
			tree_mode = E2_TREE_COMPLETED;
			continue;
		}

		if (tree_mode == E2_TREE_STARTED)
		{
			//be on the safe side for non-existing list options
			if (set == NULL)
				continue;

			//depth of the current tree option row
			gint level = 0;
			while (line[level] == '\t')
				level++;
			
			//skip the tabs at the start of the line
			line += level;

			//"unescape" the first character if need be
			// CHECKME compiler requires 2 \'s to avoid warning ?
			if ((line[0] == '\\') && ((line[1] == '<') || (line[1] == '>')))
				line++;

			//FIXME: strsplit isn't a gtreat approach because it makes
			//unescaping too complicated
			gchar **split = g_strsplit (line, "|", -1);
			void *tmp[set->ex.tree.columns_num];
			//split counter, increased when stepping through **split
			gint i = 0;
			//tree column counter
			gint j = 0;
			//tree column pointer
			GList *column;
			//unescape helper for concatting, later on
			GList *free = NULL;
			E2_OptionTreeColumn *opt;
			//transform splitted values to void pointers and
			//add default values in case of a missing value/field
			for (column = set->ex.tree.columns; column != NULL; column = g_list_next (column))
			{
				//not enough separators in the line = bad config data
				if (split[i] == NULL)
				{
					tree_mode = E2_TREE_ABORTED;
					break;
				}
				//current tree column pointer
				//E2_OptionTreeSet *
				opt = column->data;
				if ( *(split[i]) != '\0')
				{
					gchar *value = split[i];
					//unescape | and concat values
					while (split[i + 1] != NULL)  //was '\0'
					{
						gint len = strlen (value);
						if ((len > 0) && (value[len - 1] == '\\'))
						{
							value[len - 1] = '\0';
							value = g_strconcat (value, "|", split[i + 1], NULL);
							//save pointer to free it after
							//knowing how long this really gets
							//(there could be several escaped | in one line)
							free = g_list_append (free, value);
							i++;
						} else
							break;
					}
					//get void pointer to add it to the model
					tmp[j] = e2_option_tree_get_void_simple (opt, value);
				} else
					//empty string found found, use default value
					tmp[j] = e2_option_tree_get_void_simple (opt, opt->sval);
				j++;
				i++;
			}

			//too many separators in the line = bad config data, BUT
			//lines with any escaped '|' will seem to have left-overs
/*			if (split[j] != NULL)
			{
				printd (DEBUG, "LONG %s in %s", line, set->name);
				tree_mode = E2_TREE_ABORTED;
			}
*/
			if (tree_mode == E2_TREE_ABORTED)
			{
				//clear any data recorded already
				gtk_tree_store_clear (set->ex.tree.model);
				set->ex.tree.found_in_config_file = FALSE;
				printd (WARN, "bad config data for %s, using default", set->name);
			}
			else
			{
				GtkTreeIter iter;
				//add tree option row
				_e2_option_tree_add (set, &iter, g_list_nth_data (parents, level),
					FALSE, TRUE, set->ex.tree.columns_num, tmp);
				//add iter pointer to be able to add a child to this row later on
				GList *list = g_list_nth (parents, level + 1);
				if (list == NULL)
					parents = g_list_append (parents, gtk_tree_iter_copy (&iter));
				else
				{
					gtk_tree_iter_free (list->data);
					list->data = gtk_tree_iter_copy (&iter);
				}
			}
			//cleanup
			g_strfreev (split);
			g_list_foreach (free, (GFunc) g_free, NULL);
			g_list_free (free);
		}
		else if (tree_mode == E2_TREE_ABORTED)
		{
			continue;
		}
		else
		//normal option (not a tree option)
		{
			gchar **split = g_strsplit (line, "=", 2);
			//store value for the set, if the set's 'known', otherwise place
			//the set&value in the unknown queue (CHECKME why bother ??)
			if (! _e2_option_set_from_string (split[0], split[1]))
				printd (WARN, "could not set option '%s'", split[0]);
			g_strfreev (split);
		}
	}

	//cleanup
	parents = g_list_remove (parents, NULL);
	g_list_foreach (parents, (GFunc) gtk_tree_iter_free, NULL);
	g_list_free (parents);
}
/**
 * @brief read config file, store results at @a contents
 *
 * @param contents pointer to place to store the loaded file content
 *
 * @return TRUE if the file was read
 */
static gboolean _e2_option_config_file_read (gchar **contents)
{
	gboolean retval = TRUE;
	//find absolute path to config file
	gchar *filepath = g_build_filename (e2_cl_options.config_dir, default_config_file, NULL);
	//get file contents
	if (! e2_fs_get_file_contents (filepath, contents))
	{
		printd (WARN, "could not open config file '%s'", filepath);
		retval = FALSE;
	} 
	g_free (filepath);
	return retval;
}
/**
 * @brief read config file
 *
 * First, the configuration is read from @a fn and parsed with
 * e2_option_read_from_string. If tree options have no configuration,
 * they are initialized with a default one.
 * (All other options have been intialized when they were registered.)
 *
 * @param fn filename inside the config directory
 * @param cont pointer to place to store the loaded file content
 */
gboolean e2_option_file_read (void)
{
	gchar *contents;
	if (_e2_option_config_file_read (&contents))
	{
		printd (DEBUG, "config options read from file '%s'", default_config_file);
		//everything goes into an array
		gchar **split = g_strsplit (contents, "\n", -1);
		//and now parse it
		//get the config file version (always ascii) from 1st line of file
		//(done here to avoid problems when line(s) not from the file are processed
		if (strstr (split[0], "(v") != NULL )
		{
			gchar *sp = g_strrstr (split[0], ")");
			if (sp != NULL)
				*sp = '\0';
			sp = g_strstrip (strstr (split[0], "(v") + 2);
			g_strlcpy (cfg.version, sp, sizeof(cfg.version));
		}
		e2_option_read_from_string (split);
		g_strfreev (split);
		g_free (contents);
		return TRUE;
	}
	else
		return FALSE;
}
/**
 * @brief initialize option system
 *
 * internal data structures of the option system will be initialized. 
 * this includes creation of a hash table for fast option lookup.
 * this function is called on startup, by the main function
 *
 * @return 
 **/
void e2_option_init (void)
{
	default_config_file = "config";  //DO NOT TRANSLATE
	options_hash = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
//	options_queue = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
	options_array = g_ptr_array_new ();
	//moved to main loop
//	_e2_option_setup_labels ();
//	e2_action_setup_labels();  //need this before option default register, to avoid null registrations when no config file
	//before config loaded, setup all option types, default values, tips etc
	//tree option data is not registered yet, 
	e2_option_default_register ();
#ifndef E2_FAM
	e2_fs_FAM_config_stamp ();	//set baseline time, for config refreshing
#endif
	cfg.completed = TRUE;	//set signal which allows config refreshes
	
#ifdef E2_IMAGECACHE
	e2_option_icons_init ();
#endif
}
/**
 * @brief setup array of i18n config dialog labels
 *
 * Some of these are the same as action labels
 * Array size defined in e2_config_dialog.h
 * 
 * @return 
 **/
void e2_option_setup_labels (void)
{
	config_labels[0]=_("aliases");
	config_labels[1]=_("bookmarks");
	config_labels[2]=_("colors 1");
	config_labels[3]=_("colors 2");
//	config_labels[4]=_("colors 3"); unused
	config_labels[5]=_("columns");
	config_labels[6]=_("command bar");
	config_labels[7]=_("command line");
	config_labels[8]=_("commands");
	config_labels[9]=_("confirmation");
	config_labels[10]=_("context menu");
	config_labels[11]=_("delays");
	config_labels[12]=_("dialogs");
	config_labels[13]=_("directory lines");
	config_labels[14]=_("file operations");
	config_labels[15]=_("filetypes");
	config_labels[16]=_("fonts");
	config_labels[17]=_("general");
	config_labels[18]=_("history");
	config_labels[19]=_("interface");
	config_labels[20]=_("key bindings");
	config_labels[21]=_("make directory");
	config_labels[22]=_("menus");
	config_labels[23]=_("miscellaneous");
	config_labels[24]=_("options");
	config_labels[25]=_("output");
	config_labels[26]=_("pane1 bar");
	config_labels[27]=_("pane2 bar");
//	config_labels[28]=NULL;  //_("pane1 bar 2") is coming ??
//	config_labels[29]=NULL;  //_("pane2 bar 2") ??
	config_labels[30]=_("panes");
	config_labels[31]=_("plugins");
	config_labels[32]=_("position");
	config_labels[33]=_("search");
	config_labels[34]=_("startup");
	config_labels[35]=_("style");
	config_labels[36]=_("tab completion");
	config_labels[37] =_("task bar");
	config_labels[38]=_("view");
//capacity for more
	config_labels[42]=_("extensions");
	config_labels[43]=_("pane 1");
	config_labels[44]=_("pane 2");
//maybe more space here, see array definition in header file
}
