/* $Id: e2_option_tree.c 1118 2005-09-12 22:57:53Z tpgww $
 * 
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/config/e2_option_tree.c 
 * @brief functions to handle tree options
 *
 * Functions to handle tree options
 *
 **/

#include "emelfm2.h"
#include <string.h>
#include "e2_option_tree.h"
#include "e2_dialog.h"
#include "mg-cell-renderer-list.h"
#include "eggcellrendererkeys.h"

static GtkWidget *config_dialog = NULL;
static gboolean _e2_option_tree_add_below_cb (GtkWidget *wid, E2_OptionSet *set);
static gboolean _e2_option_tree_add_child_cb (GtkWidget *wid, E2_OptionSet *set);
static gboolean _e2_option_tree_move_up_cb (GtkWidget *widget, E2_OptionSet *set);
static gboolean _e2_option_tree_move_down_cb (GtkWidget *widget, E2_OptionSet *set);

  /*************************/
 /***** context menu *****/
/*************************/

#define INCLUDED_IN_PARENT
#include "e2_option_tree_context_menu.c"
#undef INCLUDED_IN_PARENT

/**
 * @brief create flag for the config dialog, to signal that the set's tree data has been modified
 *
 * @param set pointer to data struct for the option that the dialog configures
 *
 * @return
 **/
void e2_option_tree_flag_change (E2_OptionSet *set)
{
	set->ex.tree.flags |= E2_OPTION_TREE_SET_EDITED;
}

  /**************************/
 /***** data functions *****/
/**************************/

static void _e2_option_tree_bool_data_func (GtkTreeViewColumn *tree_column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	E2_OptionTreeColumn *col)
{
	gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));
	
	gboolean b;
	gtk_tree_model_get (model, iter, column, &b, -1);
	
	gboolean visible = TRUE;
	if (col->visible_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeModel *, GtkTreeIter *, GtkCellRenderer *,
			gpointer) = col->visible_check_func;
		visible = fun (model, iter, cell,  col->visible_check_func_data);
	}
	g_object_set (G_OBJECT (cell), "visible", visible, "active", b, NULL);
}

static void _e2_option_tree_str_data_func (GtkTreeViewColumn *tree_column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	E2_OptionTreeColumn *col)
{
	gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));
	
	gchar *str;
	gtk_tree_model_get (model, iter, column, &str, -1);
	
	gboolean visible = TRUE;
	if (col->visible_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeModel *, GtkTreeIter *, GtkCellRenderer *,
			gpointer) = col->visible_check_func;
		visible = fun (model, iter, cell,  col->visible_check_func_data);
	}
	g_object_set (G_OBJECT (cell), "visible", visible, "text", str, NULL);
	g_free (str);
}

static void _e2_option_tree_sel_data_func (GtkTreeViewColumn *tree_column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	E2_OptionTreeColumn *col)
{
	GtkTreePath *path = gtk_tree_model_get_path (model, iter);
	gchar *name;
	gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));
	gtk_tree_model_get (model, iter, column, &name, -1);
	gboolean visible = TRUE;
	if (col->visible_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeModel *, GtkTreeIter *, GtkCellRenderer *,
			gpointer) = col->visible_check_func;
		visible = fun (model, iter, cell,  col->visible_check_func_data);
	}
	g_object_set (cell, "text", name, "visible", visible, NULL);
	g_free (name);
	gtk_tree_path_free (path);
}

static void _e2_option_tree_key_data_func (GtkTreeViewColumn *tree_column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	E2_OptionTreeColumn *col)
{
	gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));
	
	gchar *key;
	gtk_tree_model_get (model, iter, (int) column, &key, -1);
	
	GdkModifierType mask = 0;
	guint keyval = 0;
	if (key != NULL)
	{
		gtk_accelerator_parse (key, &keyval, &mask);
	}
	gboolean visible = TRUE;
	if (col->visible_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeModel *, GtkTreeIter *, GtkCellRenderer *,
			gpointer) = col->visible_check_func;
		visible = fun (model, iter, cell,  col->visible_check_func_data);
	}
	g_object_set (G_OBJECT (cell), "visible", visible, "accel_key", keyval,
		"accel_mask", mask, NULL);
}

  /*********************/
 /***** callbacks *****/
/*********************/

static void _e2_option_tree_page_opened_cb (GtkWidget *widget, E2_OptionSet *set)
{
	//backup set data when the page is opened (if it's not done already)
	e2_option_tree_backup (set);
}

//this is the row_draggable function for the treeview's GtkTreeDragSourceIface
static gboolean _e2_option_tree_draggable_check_cb (GtkTreeDragSource *drag_source,
	GtkTreePath *path)
{
	if (!GTK_IS_TREE_MODEL (drag_source))
		return TRUE;
	GtkTreeModel *model = GTK_TREE_MODEL (drag_source);
	E2_OptionSet *set = g_object_get_data (G_OBJECT (model), "e2-option-set");
	if (set == NULL)
		return TRUE;
	if (set->ex.tree.draggable_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeDragSource *, GtkTreePath *) =
			set->ex.tree.draggable_check_func;
		return fun (drag_source, path);
	}
	else
		return TRUE;
}
/**
 * @brief delete the row at @a path, because it was moved somewhere else via drag-and-drop.
 *
 * @param drag_source drag source data struct
 * @param path	gtk tree path to the row to be deleted
 *
 * @return FALSE if the deletion fails because path no longer exists, or for some model-specific reason
 **/
static gboolean _e2_option_tree_drag_delete_cb (GtkTreeDragSource *drag_source,
	GtkTreePath *path)
{
	if (!GTK_IS_TREE_MODEL (drag_source))
		return FALSE;
	GtkTreeModel *model = GTK_TREE_MODEL (drag_source);
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter (model, &iter, path))
	{
		gtk_tree_store_remove (GTK_TREE_STORE (model), &iter);
		E2_OptionSet *set = g_object_get_data (G_OBJECT (model), "e2-option-set");
		e2_option_tree_flag_change (set);
		return TRUE;
	}
	else
		return FALSE;
}

static void _e2_option_tree_key_edited_cb (GtkCellRendererText *cell,
	const char *path_string, guint keyval, GdkModifierType mask,
	guint hardware_keycode, E2_OptionSet *set)
{
	GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter (set->ex.tree.model, &iter, path))
	{
		gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));
		gchar *key = gtk_accelerator_name (keyval, mask);
		_e2_option_tree_set (set, &iter, column, key);
		g_free (key);
	}
	gtk_tree_path_free (path);
}

/*static gboolean _e2_option_tree_key_press_cb (GtkWidget *widget,
	GdkEventKey *event, gpointer data)
{
	if (event->keyval == GDK_Return)
		return e2_confdlg_ok_cb (NULL, FALSE);
	if (event->keyval == GDK_Escape)
		return e2_confdlg_cancel_cb (NULL);
	return FALSE;
} */

static gboolean _e2_option_tree_move_up_cb (GtkWidget *widget,
	E2_OptionSet *set)
{
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (set->widget)), &model, &iter))
	{
		GtkTreePath *path;
		path = gtk_tree_model_get_path (model, &iter);
		if (gtk_tree_path_prev (path))
		{
			_e2_option_tree_move (set, &iter, path, TRUE, TRUE);	//this sets the dirty-flag
		}
		if (path)
	    	gtk_tree_path_free (path);
	}
	return TRUE;
}

static gboolean _e2_option_tree_move_down_cb (GtkWidget *widgget,
	E2_OptionSet *set)
{
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (set->widget)), &model, &iter))
	{
		GtkTreePath *path;
		path = gtk_tree_model_get_path (model, &iter);
		gtk_tree_path_next (path);
		_e2_option_tree_move (set, &iter, path, TRUE, FALSE);	//this sets the dirty-flag
		e2_option_tree_flag_change (set);
		if (path)
	    	gtk_tree_path_free (path);
	}
	return TRUE;
}

static void _e2_option_tree_cell_pixbuf_data_func (GtkTreeViewColumn *tree_column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tree_column), "column"));
	gchar *icon;
	gtk_tree_model_get (model, iter, column, &icon, -1);
#ifdef E2_IMAGECACHE
	GdkPixbuf *pb;
#endif
//	if (icon != NULL)
	if (*icon != '\0')
	{
#ifdef E2_IMAGECACHE
// FIXME = adjust img->refcount ??
		GtkIconSize sz = GPOINTER_TO_INT (data);
		E2_Image *img = e2_cache_image_get (icon, sz);
		pb = img->pixbuf;
#else
		if (e2_utils_check_stock_icon (icon))
		{
			g_object_set (GTK_CELL_RENDERER (cell), "stock-id", icon, NULL);
			g_object_set (GTK_CELL_RENDERER (cell), "pixbuf", NULL, NULL);
		}
		else	//not a stock item
		{
			GdkPixbuf *pixbuf;
			GdkPixbuf *pixbuf2;
			if ((pixbuf = gdk_pixbuf_new_from_file (icon, NULL)) != NULL)
			{
				pixbuf2 = pixbuf;
				pixbuf = gdk_pixbuf_scale_simple (pixbuf2, 16, 16, GDK_INTERP_BILINEAR);
				g_object_unref (pixbuf2);
			}
			g_object_set (GTK_CELL_RENDERER (cell), "stock-id", NULL, NULL);
			g_object_set (GTK_CELL_RENDERER (cell), "pixbuf", pixbuf, NULL);
			if (pixbuf) g_object_unref (pixbuf);
		} 
#endif
	}
	else
#ifdef E2_IMAGECACHE
		pb = NULL;
	g_object_set (GTK_CELL_RENDERER (cell), "pixbuf", pb, NULL);
#else
	{
		g_object_set (GTK_CELL_RENDERER (cell), "stock-id", NULL, NULL);
		g_object_set (GTK_CELL_RENDERER (cell), "pixbuf", NULL, NULL);
	}
#endif

	g_free (icon);
}

static void _e2_option_tree_icon_column_set_cb (GtkWidget *dialog,
	gint response, E2_OptionSet *set) 
{
	if ((response == GTK_RESPONSE_APPLY) || (response == GTK_RESPONSE_OK) ||
		(response == E2_RESPONSE_REMOVE))
	{
		GtkTreeIter iter;
		if ((gtk_tree_selection_get_selected (gtk_tree_view_get_selection
			(GTK_TREE_VIEW (set->widget)), NULL, &iter)) == TRUE)
		{
			gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (dialog), "column"));
			gchar *icon = g_object_get_data (G_OBJECT (dialog), "image");
			_e2_option_tree_set (set, &iter, column, icon);	//this sets the dirty-flag
		}
	}
	if ((response != GTK_RESPONSE_APPLY) && (response != E2_RESPONSE_REMOVE))
		gtk_widget_destroy (dialog);
}

static gboolean _e2_option_tree_popup_menu_cb (GtkWidget *treeview,
	GtkWidget *menu)
{
	//set appropriate item sensitivities
	e2_option_tree_menu_set_sensitive (menu, treeview);

	gint event_time = gtk_get_current_event_time ();
	gtk_menu_popup (GTK_MENU (menu), NULL, NULL,
		(GtkMenuPositionFunc) _e2_option_tree_menu_set_position,
		treeview, 0, event_time);	//button code = 0 as this was a menu-button press
	return TRUE;
}
/**
 * @brief Create context menu
 *
 * Context menu is created when mouse button-3 is pressed
 *
 * @param treeview widget where the button was pressed
 * @param event gdk event data
 * @param menu the context-menu widget
 *
 * @return TRUE (to prevent further handlers) for a button=3 press, else FALSE
 */
static gboolean _e2_option_tree_button_press_cb2 (GtkWidget *treeview,
	GdkEventButton *event, GtkWidget *menu)
{
	if (event->button != 3)
		return FALSE;
	//set appropriate item sensitivities
	e2_option_tree_menu_set_sensitive (menu, treeview);

	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, event->button, event->time);
	return TRUE;
}

static gboolean _e2_option_tree_button_press_cb (GtkWidget *treeview,
	GdkEventButton *event, E2_OptionSet *set)
{
	if (event->button != 1)
		return FALSE;
	GtkTreePath *path;
	GtkTreeViewColumn *tree_column;
	if (!gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
		event->x, event->y, &path, &tree_column, NULL, NULL))
			return FALSE;
	GtkTreeSelection *selection = gtk_tree_view_get_selection 
		(GTK_TREE_VIEW (treeview));
	if (!gtk_tree_selection_path_is_selected (selection, path))
		return FALSE;
	GtkWidget *dialog = NULL;
	gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tree_column), "column"));
	GList *list = g_list_nth (set->ex.tree.columns, column);
	E2_OptionTreeColumn *opt = list->data;
	if (opt->type == E2_OPTION_TREE_TYPE_ICON)
	{
		//get the current value
		gchar *icon;
		GtkTreeIter iter;
		gtk_tree_model_get_iter (GTK_TREE_MODEL (set->ex.tree.model), &iter, path);
		gtk_tree_model_get (set->ex.tree.model, &iter, column, &icon, -1);
		gchar *dialog_name = g_strdup_printf (_("select icon for %s"), set->name);
		dialog = e2_sid_create (config_dialog, dialog_name, icon,
			_e2_option_tree_icon_column_set_cb, set);
		g_free (dialog_name);
		g_object_set_data (G_OBJECT (dialog), "column", GINT_TO_POINTER (column));
		return TRUE;
	}
	return FALSE;
}

static gboolean _e2_option_tree_edited_cb (GtkCellRendererText *cell, 
	const gchar *path_string, const gchar *new_text, E2_OptionSet *set)
{
	//find out where
	GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
	gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));
	GtkTreeIter iter;
	gtk_tree_model_get_iter (GTK_TREE_MODEL (set->ex.tree.model), &iter, path);

	//change model
	gtk_tree_store_set (GTK_TREE_STORE (set->ex.tree.model), &iter, column, new_text, -1);

	e2_option_tree_flag_change (set);
	//clean up
	gtk_tree_path_free (path);
	return TRUE;
}

static gboolean _e2_option_tree_list_edited_cb (MgCellRendererList *cell, 
	gchar *path_string, gchar *new_text, E2_OptionSet *set)
{
	//find out where
	GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
	gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));
	GtkTreeIter iter;
	gtk_tree_model_get_iter (GTK_TREE_MODEL (set->ex.tree.model), &iter, path);

	GtkTreeIter iter_sel;
	GtkTreeModel *model_sel;
	if (gtk_tree_selection_get_selected (cell->selection,
		&model_sel, &iter_sel))
	{
		gchar *cur;
		gtk_tree_model_get (model_sel, &iter_sel, 0, &cur, -1);
	
		//change model
		gtk_tree_store_set (GTK_TREE_STORE (set->ex.tree.model), &iter, column, cur, -1);
	}
	e2_option_tree_flag_change (set);

	gtk_widget_grab_focus (set->widget);

	//clean up
	gtk_tree_path_free (path);
	return TRUE;
}

//callback after tree view cell was toggled
static gboolean _e2_option_tree_toggle_cb (GtkCellRendererToggle *cell,
	const gchar *path_str, E2_OptionSet *set)
{
	//find out where
	GtkTreePath *path = gtk_tree_path_new_from_string (path_str);
	GtkTreeIter iter;
	gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));
	gtk_tree_model_get_iter (set->ex.tree.model, &iter, path);
	
	//reverse current value
	gboolean toggle_item;
	gtk_tree_model_get (set->ex.tree.model, &iter, column, &toggle_item, -1);
	toggle_item ^= 1;

	//change model
	gtk_tree_store_set (GTK_TREE_STORE (set->ex.tree.model), &iter, column, toggle_item, -1);

	e2_option_tree_flag_change (set);
	//clean up
	gtk_tree_path_free (path);
	return TRUE;
}

static void e2_cd_option_show_popup_cb (MgCellRendererList *cell,
	const gchar *path_string, gint x1, gint y1, gint x2, gint y2, E2_OptionSet *set)
{
	if (cell->model == NULL)
	{
		gint column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));
		GList *list = g_list_nth (set->ex.tree.columns, column);
		E2_OptionTreeColumn *opt = list->data;
		GtkTreeModel *(*fun) (gpointer) = opt->extra;
		cell->model = fun (opt->extra_data);
	}
}

static void _e2_option_tree_hide_popup_cb (MgCellRendererList *cell,
	GtkWidget *widget)
{
	gtk_widget_grab_focus (widget);
}

static gboolean _e2_option_tree_help_cb (GtkWidget *button, E2_OptionSet *set)
{
	//get the config dialog page label, from set->group
	// (after a parent separator '.' if any)
	gchar *page, *realpage;
	page = set->group;
	if ((realpage = strchr (page, '.')) != NULL)	//if always ascii ',', don't need g_utf8_strchr()
	{
		NCHR(realpage); page = realpage;
	}
	//relevant section label is "[page]"
	gchar *command = g_strconcat("<",_A(5),".",_A(76)," ", 
		e2_option_str_get ("config-help-doc")," [",page,"]",NULL);	//file.view_at;
	e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
	g_free (command);
	return TRUE;
}

static gboolean _e2_option_tree_del_cb (GtkWidget *button, E2_OptionSet *set)
{
	GtkTreeIter iter;
	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection
			(GTK_TREE_VIEW (set->widget)), NULL, &iter))
	{
		GtkTreePath *path = gtk_tree_model_get_path (set->ex.tree.model, &iter);

		gint response = GTK_RESPONSE_YES;
		gint children = gtk_tree_model_iter_n_children (set->ex.tree.model, &iter);
		if (children > 0)
		{
			gchar *question = g_strdup_printf (
				_("Are you sure that you want to delete this row and <b>%d</b> %s?"),
				children, children == 1 ? _("child") : _("children"));
			
			GtkWidget *dialog = e2_dialog_create (GTK_STOCK_DIALOG_QUESTION, question,
			_("confirm row delete"), NULL, NULL);
			response = e2_dialog_show (dialog, config_dialog,
				E2_DIALOG_RUN | E2_DIALOG_FREE, &E2_BUTTON_NO, &E2_BUTTON_YES, NULL);
			g_free (question);
		}

		if (response == GTK_RESPONSE_YES)
			_e2_option_tree_del (set, &iter);

		if (iter.user_data == NULL)
		{
			if (!gtk_tree_path_prev (path))
				if (!gtk_tree_path_up (path))
					path = NULL;
		} else
			path = gtk_tree_model_get_path
				(GTK_TREE_MODEL (set->ex.tree.model), &iter);
		if ((path != NULL) &&
			(gtk_tree_model_get_iter_first (set->ex.tree.model, &iter)))
		{
			gtk_tree_view_set_cursor (GTK_TREE_VIEW (set->widget), path,
				gtk_tree_view_get_column (GTK_TREE_VIEW (set->widget), 0),
				FALSE);
		}
	}
	return TRUE;
}

static gboolean _e2_option_tree_add_child_cb (GtkWidget *wid, E2_OptionSet *set)
{
	GtkTreeIter iter;
	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (set->widget)), NULL, &iter))
	{
		if (iter.user_data != NULL)
		{
			GtkTreeIter iter2;
			_e2_option_tree_add_default (set, &iter2, &iter, FALSE);
		}
	} else
	{
		_e2_option_tree_add_default (set, &iter, NULL, FALSE);
	}

	return TRUE;
}

static gboolean _e2_option_tree_add_below_cb (GtkWidget *wid, E2_OptionSet *set)
{
	GtkTreeIter iter;
	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (set->widget)), NULL, &iter))
	{
		if (iter.user_data != NULL)
		{
			GtkTreeIter iter2;
			_e2_option_tree_add_default (set, &iter2, &iter, TRUE);
		}
	} else
	{
		_e2_option_tree_add_default (set, &iter, NULL, TRUE);
	}

	return TRUE;
}

  /******************/
 /***** public *****/
/******************/

/**
 * @brief (de)sensitize config dialog buttons that are not relevant to 'category' lines
 * @param view
 * @param value TRUE to senstitize the buttons, FALSE to desensitize
 * @return
 **/
void e2_option_tree_adjust_buttons (GtkTreeView *view, gboolean value)
{
	E2_Sextet *q = g_object_get_data (G_OBJECT (view), "e2-config-buttons");
	if (q == NULL)
		return;
	if (q->b != NULL)
		gtk_widget_set_sensitive (q->b, value);
	if (q->c != NULL)
		gtk_widget_set_sensitive (q->c, value);
	if (q->d != NULL)
		gtk_widget_set_sensitive (q->d, value);
	if (q->e != NULL)
		gtk_widget_set_sensitive (q->e, value);
}

void e2_option_tree_add_widget (GtkWidget *parent, GtkWidget *box,
	E2_OptionSet *set)
{
	config_dialog = parent;
	GtkWidget *vbox = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (box), vbox, TRUE, TRUE, 0);
	GtkWidget *sw = e2_widget_add_sw (vbox, GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC, TRUE, E2_PADDING);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw), GTK_SHADOW_IN);
	
	//setup treeview
	GtkTreeView *treeview = GTK_TREE_VIEW (gtk_tree_view_new_with_model
		((GtkTreeModel *) set->ex.tree.model));
	gtk_container_add (GTK_CONTAINER (sw), GTK_WIDGET (treeview));
	gtk_tree_view_set_reorderable (treeview, TRUE);
//	gtk_tree_view_set_headers_clickable (treeview, FALSE);  =default
//	gtk_tree_view_set_rules_hint (treeview, FALSE);  =default

	set->widget = GTK_WIDGET (treeview);

#ifdef E2_IMAGECACHE
	//setup for appropriate icon size
	gint pwide, phigh;
	e2_widget_get_font_pixels (set->widget, &pwide, &phigh);
	GtkIconSize iconsizes = e2_utils_get_best_icon_size (phigh + 2);
#endif

	GtkTreeSelection *selection = gtk_tree_view_get_selection (treeview);
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);
//	gtk_tree_selection_set_mode (selection, GTK_SELECTION_MULTIPLE);
	g_object_set_data (G_OBJECT (set->ex.tree.model), "e2-option-set", set);
//	if (set->ex.tree.draggable_check_func != NULL)
//	{
//		printd (INFO, "drag func: %d", set->ex.tree.draggable_check_func);
		GtkTreeDragSourceIface *iface = GTK_TREE_DRAG_SOURCE_GET_IFACE (set->ex.tree.model);
//		iface->row_draggable = set->ex.tree.draggable_check_func;
		iface->row_draggable = _e2_option_tree_draggable_check_cb;
//	}
	//we need a way to set the 'dirty-flag' when any row is dragged
	iface->drag_data_delete = _e2_option_tree_drag_delete_cb;

	if (set->ex.tree.selection_check_func != NULL)
		gtk_tree_selection_set_select_function (selection, 
			(GtkTreeSelectionFunc) set->ex.tree.selection_check_func,
			set, NULL);

	//add columns
	int j = 0;
	GList *collist;
	for (collist = set->ex.tree.columns; collist != NULL; collist = g_list_next (collist))
	{
		GtkCellRenderer *renderer;
		GtkTreeViewColumn *column;
		E2_OptionTreeColumn *opt = collist->data;
		gboolean editable = !(opt->flags & E2_OPTION_TREE_COL_NOT_EDITABLE);
		switch (opt->type)
		{
			case E2_OPTION_TREE_TYPE_BOOL:
				renderer = gtk_cell_renderer_toggle_new ();
				g_object_set (renderer, "activatable", TRUE, NULL);
				g_signal_connect (renderer, "toggled",
					G_CALLBACK (_e2_option_tree_toggle_cb), set);
				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, "active", j, NULL);
				g_object_set (column, "alignment", 0.5, NULL);
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_column_set_cell_data_func (column, renderer,
					(GtkTreeCellDataFunc) _e2_option_tree_bool_data_func, opt, NULL);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			case E2_OPTION_TREE_TYPE_STR:
				renderer = gtk_cell_renderer_text_new ();
				g_object_set (renderer, "editable", editable, NULL);
				g_signal_connect (renderer, "edited",
					G_CALLBACK (_e2_option_tree_edited_cb), set);
				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, "text", j, NULL);
				gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
					GTK_TREE_VIEW_COLUMN_GROW_ONLY);
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_column_set_cell_data_func (column, renderer,
					(GtkTreeCellDataFunc) _e2_option_tree_str_data_func, opt, NULL);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			case E2_OPTION_TREE_TYPE_ICON:
				renderer = gtk_cell_renderer_pixbuf_new ();
				g_object_set (renderer, "stock-size", GTK_ICON_SIZE_MENU, NULL);
				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, NULL);
				gtk_tree_view_column_set_cell_data_func (column, renderer,
#ifdef E2_IMAGECACHE
					_e2_option_tree_cell_pixbuf_data_func, GINT_TO_POINTER (iconsizes), NULL);
#else
					_e2_option_tree_cell_pixbuf_data_func, NULL, NULL);
#endif
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
					GTK_TREE_VIEW_COLUMN_GROW_ONLY);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			case E2_OPTION_TREE_TYPE_SEL:
				renderer = mg_cell_renderer_list_new ();
				g_object_set (renderer, "editable", editable, NULL);
				//FIXME: bad hack for the cell renderer to be able to calculate
				//the correct pop-up postion if activated through a key-press
				g_object_set_data (G_OBJECT (renderer), "scrolled-window", sw);
				g_signal_connect (renderer, "edited",
					G_CALLBACK (_e2_option_tree_list_edited_cb), set);
				g_signal_connect (renderer, "show-popup",
					G_CALLBACK (e2_cd_option_show_popup_cb), set);
				g_signal_connect (renderer, "hide-popup",
					G_CALLBACK (_e2_option_tree_hide_popup_cb), treeview);
				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, NULL);
				gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
					GTK_TREE_VIEW_COLUMN_GROW_ONLY);
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_column_set_cell_data_func (column, renderer, 
					(GtkTreeCellDataFunc) _e2_option_tree_sel_data_func, opt, NULL);
//				g_object_set_data (G_OBJECT (column),
//					"filter-func", tree_view_sel_data_func);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			case E2_OPTION_TREE_TYPE_KEY:
				renderer = egg_cell_renderer_keys_new ();
				g_object_set (renderer, "editable", TRUE, NULL);
				if (!editable)
					g_object_set (renderer, "visible", FALSE, NULL);
				g_signal_connect (G_OBJECT (renderer), "keys_edited",
					G_CALLBACK (_e2_option_tree_key_edited_cb), set);

				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, NULL);
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_column_set_cell_data_func (column, renderer,
					(GtkTreeCellDataFunc) _e2_option_tree_key_data_func, opt, NULL);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			default:
				renderer = NULL;
				column = NULL;  //warning prevention
				printd (WARN, "don't know how to render column %d of %s", j, set->name);
				break;
		}
		if (renderer != NULL)
		{
			g_object_set_data (G_OBJECT (renderer), "column", GINT_TO_POINTER (j));
			g_object_set_data (G_OBJECT (column), "column", GINT_TO_POINTER (j));
		}
		j++;
	}

	//spacer column
	if ((int) set->ex.tree.flags & E2_OPTION_TREE_LAST_COL_EMPTY)
	{
		GtkCellRenderer *renderer = gtk_cell_renderer_text_new ();
		GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes
			("", renderer, NULL);
		gtk_tree_view_append_column (treeview, column);
	}
	
	//now expand everything
	gtk_tree_view_expand_all (treeview);
	
	//add control buttons, in reverse order (packed at end)
	GtkWidget *bbox = gtk_hbox_new (FALSE, E2_PADDING);
	gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 0);
	E2_Sextet *q = (E2_Sextet *) e2_utils_sextet_new ();
	g_object_set_data_full (G_OBJECT (treeview), "e2-config-buttons", q, g_free);

	if (set->ex.tree.flags & E2_OPTION_TREE_ADD_DEL)
	{
		if (! (set->ex.tree.flags & E2_OPTION_TREE_LIST))
			q->f = e2_button_add_end (bbox, FALSE, 0, _("Add c_hild"), GTK_STOCK_INDENT,
				_("Add a child row to the currently selected one"),
				_e2_option_tree_add_child_cb, set);
		q->e = e2_button_add_end (bbox, FALSE, 0, _("Add a_fter"), GTK_STOCK_ADD,
			_("Add a row after the currently selected row in the tree"),
			_e2_option_tree_add_below_cb, set);	
		q->d = e2_button_add_end (bbox, FALSE, 0, _("_Remove"), GTK_STOCK_REMOVE,
			_("Remove the selected row"),  _e2_option_tree_del_cb, set);
	}
	if (set->ex.tree.flags & E2_OPTION_TREE_UP_DOWN)
	{
		q->c = e2_button_add_end (bbox, FALSE, 0, _("_Down"), GTK_STOCK_GO_DOWN,
			_("Move the selected row one place down"),
			_e2_option_tree_move_down_cb, set);
		q->b = e2_button_add_end (bbox, FALSE, 0, _("_Up"), GTK_STOCK_GO_UP,
			_("Move the selected row one place up"),
			_e2_option_tree_move_up_cb, set);
	}
	q->a = e2_button_add_end (bbox, FALSE, 0, _("_Help"), GTK_STOCK_HELP,
		_("Get help on this option"),  _e2_option_tree_help_cb, set);

	//add context-menu
	E2_TreeContextMenuFlags context_flags = E2_TREE_CONTEXT_DEFAULT;
	if (!((gboolean) set->ex.tree.flags & E2_OPTION_TREE_LIST))
		context_flags |= E2_TREE_CONTEXT_EXP_COL;
	GtkWidget *context_menu = gtk_menu_new ();
	//E2_Sextet *s = (E2_Sextet *)
	_e2_option_tree_menu_items_add (treeview, context_menu, context_flags, set);

		//connect callbacks
	//no longer needed as it seems
//	g_signal_connect (treeview, "drag-drop", G_CALLBACK (drop_cb), set);
//	g_signal_connect (treeview, "drag-end", G_CALLBACK (drag_end_cb), set);

	//arrange for data backup if/when the page is opened
	g_signal_connect (treeview, "realize",
		G_CALLBACK (_e2_option_tree_page_opened_cb), set);
	g_signal_connect (treeview, "popup_menu",
		G_CALLBACK (_e2_option_tree_popup_menu_cb), context_menu);
	g_signal_connect_after (treeview, "key_press_event",
//		G_CALLBACK (_e2_option_tree_key_press_cb), NULL);
		G_CALLBACK (e2_confdlg_key_press_cb), NULL);
	//this one handles left-button clicks
	g_signal_connect (treeview, "button-press-event",
		G_CALLBACK (_e2_option_tree_button_press_cb), set);
	//this one handles right-button clicks
	g_signal_connect (treeview, "button-press-event",
		G_CALLBACK (_e2_option_tree_button_press_cb2), context_menu);
	
	//goto first row & column
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (set->ex.tree.model), &iter))
	{
		GtkTreePath *path = gtk_tree_model_get_path (GTK_TREE_MODEL
			(set->ex.tree.model), &iter);
		gtk_tree_view_set_cursor (treeview, path,
			gtk_tree_view_get_column (treeview, 0), FALSE);
		gtk_tree_path_free (path);
	}
}
/**
 * @brief register new tree option
 *
 * register new tree option.
 *
 * @param name name of the option, generally a constant string but sometimes runtime-created
 * @param group group the option belongs to, used in config dialog, can be a r-t string FREEME or a _()
 * @param desc textual description of the option used in config dialog, a r-t _() string
 * @param depends name of another option this one depends on
 * @param selection_check_func function to check whether row is selectable
 * @param draggable_check_func function to check whether selected row is draggable
 * @param flags flags set when registering a tree option
*
 * @return  E2_OptionSet of the new option
 **/
E2_OptionSet *e2_option_tree_register (gchar *name, gchar *group, gchar *desc,
	gchar *depends, gpointer selection_check_func, gpointer draggable_check_func,
	E2_OptionTreeTypeFlags flags, E2_OptionFlags flags2)
{
	E2_OptionSet *set;
	if ((set = e2_option_get_simple (name)) != NULL)
		return set;
	set = g_malloc (sizeof (E2_OptionSet));
	//set up E2_OptionSet
	set->name = name;
	set->group = group;
	set->type = E2_OPTION_TYPE_TREE;
	set->desc = desc;
	set->tip = desc;
	set->depends = depends;
	set->ival = -1;
	set->flags = flags2 | E2_OPTION_FLAG_TREE;
	set->sval = NULL;
	set->ex.tree.def = NULL;
	set->ex.tree.columns_num = 0;
	set->ex.tree.columns = NULL;
	set->ex.tree.flags = flags;	//flags specified by the registering function
	set->ex.tree.model = gtk_tree_store_new (1, G_TYPE_INT);
	set->ex.tree.found_in_config_file = FALSE;
//	set->ex.tree.unknown = NULL;
	set->ex.tree.selection_check_func = selection_check_func;
	set->ex.tree.draggable_check_func = draggable_check_func;
	e2_option_register (set);
	return set;
}
#if 0
/**
 * @brief register unknown tree option
 *
 * read @a buf until the end of the tree option and register it
 * with a GList helper as unknown option.
 *
 * @param name string with the name of the option
 * @param i    current line counter
 * @param buf  NULL-terminated array of strings in config file format
 *
 * @return     new line counter
 **/
gint e2_option_tree_register_unknown (gchar *name, gint i, gchar *buf[])
{
	printd (DEBUG, "e2_option_tree_register_unknown (name:%s,i:%d,buf:)", name, i);
	gint save = i;
	gchar *line;
	GList *list = NULL;
	while ((line = buf[i++]) != NULL)
	{
		list = g_list_append (list, g_strdup (line));
		if (g_str_equal (line, ">"))
		{
			gchar **strv = e2_list_to_strv (list);
			g_list_free (list);
			e2_option_unknown_register2 (name, strv);
			return (i - 1);
		}
	}
	//failure, this doesn't look like a tree option
	return save;
}
#endif //0
//E2_OptionTreeColumn *
void e2_option_tree_add_column (E2_OptionSet *set, gchar *name,
	E2_OptionTreeType type, gint ivalue, gchar *svalue, void *extra,
	void *extra_data, E2_OptionTreeColFlags flags, gpointer visible_check_func,
	gpointer visible_check_func_data)
{
	E2_OptionTreeColumn *col = g_malloc (sizeof (E2_OptionTreeColumn));
	col->name = name;
	col->type = type;
	col->ival = ivalue;
	col->sval = svalue;
	col->extra = extra;
	col->extra_data = extra_data;
	col->flags = flags;
	col->visible_check_func = visible_check_func;
	col->visible_check_func_data = visible_check_func_data;
	set->ex.tree.columns = g_list_append (set->ex.tree.columns, col);
	set->ex.tree.columns_num++;
//	return col;
}

void e2_option_tree_sync_columns (E2_OptionSet *set)
{
	GType t[set->ex.tree.columns_num + 1];
	GList *column;
	gint i = 0;
	for (column = set->ex.tree.columns; column != NULL; column = g_list_next (column))
	{
		E2_OptionTreeColumn *tcol = column->data;
		switch (tcol->type)
		{
			case E2_OPTION_TREE_TYPE_BOOL:
				t[i] = G_TYPE_INT;
				break;
			case E2_OPTION_TREE_TYPE_STR:
			case E2_OPTION_TREE_TYPE_ICON:
			case E2_OPTION_TREE_TYPE_SEL:
			case E2_OPTION_TREE_TYPE_KEY:
				t[i] = G_TYPE_STRING;
				break;
			default:
				printd (ERROR, "internal error, list options may only be bool,"
					" string or int");
				break;
		}
		i++;
	} 
	//column for visible check
	t[set->ex.tree.columns_num] = G_TYPE_BOOLEAN;
	gtk_tree_store_set_column_types (set->ex.tree.model, set->ex.tree.columns_num + 1, t); 
/*	//check if there are values for this option that have been read from the
	//config file before
	if (set->ex.tree.unknown != NULL)
	{
		e2_option_read_from_string (set->ex.tree.unknown);
		g_strfreev (set->ex.tree.unknown);
		set->ex.tree.unknown = NULL;
	}
*/
}

void e2_option_tree_revert (gchar *option_name)
{
	E2_OptionSet *set = e2_option_tree_get (option_name);
	_e2_option_tree_revert (set, FALSE);
}

void _e2_option_tree_revert (E2_OptionSet *set, gboolean empty_check)
{
	if (!(set->ex.tree.found_in_config_file && empty_check) &&
		(set->ex.tree.def != NULL))
	{
		g_object_unref (set->ex.tree.model);
		set->ex.tree.model = gtk_tree_store_new (1, G_TYPE_INT);
		e2_option_tree_sync_columns (set);
		//option needs default configuration which is stored in ex.tree.def
		//of the OptionSet
		e2_option_read_from_string (set->ex.tree.def);
		//make sure the set is recognised as clean
		set->ex.tree.flags &= ~E2_OPTION_TREE_SET_EDITED;
	}
}
/**
 * @brief create a string representing a tree row
 *
 * This is a helper fn for backup and config file writing
 * Depth is represented by leading tabs
 * Columns are separated by '|'
 * Any other embedded '|' is escaped with '\'
 *
 * @param set  set data structure
 * @param iter treeiter used for scanning the treemodel
 * @param level tree path-depth of the processed line
 *
 * @return  the constructed string
 **/
gchar *e2_option_tree_row_write_to_string (E2_OptionSet *set, 
	GtkTreeIter *iter, gint level)
{
	GString *treerow = g_string_sized_new (128);
	gint j;
	for (j = 0; j < level; j++)
		treerow = g_string_append (treerow, "\t");
	GList *column;
	j = 0;
	for (column = set->ex.tree.columns; column != NULL; column = g_list_next (column))
	{
		E2_OptionTreeColumn *opt = column->data;
		switch (opt->type)
		{
			case E2_OPTION_TREE_TYPE_BOOL:
			{
				gboolean int_data;
				gtk_tree_model_get (set->ex.tree.model, iter,  j, &int_data, -1);
				if (int_data)
					treerow = g_string_append (treerow, "true");
				else
					treerow = g_string_append (treerow, "false");
			}
			break;
			case E2_OPTION_TREE_TYPE_STR:
			case E2_OPTION_TREE_TYPE_ICON:
			case E2_OPTION_TREE_TYPE_SEL:
			case E2_OPTION_TREE_TYPE_KEY:
			{
				gchar *str_data;
				gtk_tree_model_get (set->ex.tree.model, iter,  j, &str_data, -1);

				if (str_data != NULL)
				{
					if ((j == 0) && ((str_data[0] == '<') || (str_data[0] == '>')))
						treerow = g_string_append (treerow, "\\");
					if ((str_data[0] != '\0') && ((strchr (str_data, '|')) != NULL))	//if always ascii |, don't need g_utf8_strchr()
					{
						gchar **split = g_strsplit (str_data, "|", -1);
						g_free (str_data);
						str_data = g_strjoinv ("\\|", split);
						g_strfreev (split);
					}
//					if (str_data != NULL)
//					{
						treerow = g_string_append (treerow, str_data);
						g_free (str_data);
//					}
				}
			}
			break;
			default:
			break;
		}
		treerow = g_string_append (treerow, "|");
		j++;
	}
	//get rid of superfluous trailing '|'
	treerow = g_string_truncate (treerow, treerow->len -1);

	gchar *result = g_string_free (treerow, FALSE);
	return result;
}
/**
 * @brief create a string array of option-tree row data for a specified optionset
 *
 * This is a helper fn for backup
 * It is recursive, to deal with descendants
 *
 * @param lines ptr-array where the strings are recorded
 * @param set the set data structure
 * @param iter treeiter used for scanning the treemodel
 * @param level the current path depth in the tree
 *
 * @return 
 **/
static void e2_option_tree_write_to_storage (GPtrArray *lines, E2_OptionSet *set,
	GtkTreeIter *iter, gint level)
{
	do
	{
		gchar *rowstring = e2_option_tree_row_write_to_string (set, iter, level);
		g_ptr_array_add (lines, (gpointer) rowstring);

		GtkTreeIter iter2;
		if (gtk_tree_model_iter_children (set->ex.tree.model, &iter2, iter))
			e2_option_tree_write_to_storage (lines, set, &iter2, level + 1);

	} while (gtk_tree_model_iter_next (set->ex.tree.model, iter));
}
/**
 * @brief create backup data for a specified treeoption set
 *
 * Data are stored as strings, one for each tree row, in the same format
 * as used for the config file
 * Each row is referred to in a NULL-terminate gpointer array
 * This fn does nothing if there is already backup data for the set
 *
 * @param set the set data structure
 *
 * @return 
 **/
void e2_option_tree_backup (E2_OptionSet *set)
{
	if (set->ex.tree.def == NULL)
	{	//backup data does not exist now, so ok to create it
		GPtrArray *lines = g_ptr_array_sized_new (32);
		g_ptr_array_add (lines, (gpointer) g_strdup_printf ("<%s", set->name));
		GtkTreeIter iter;
		if ((gtk_tree_model_get_iter_first (set->ex.tree.model, &iter)) == TRUE)
			e2_option_tree_write_to_storage (lines, set, &iter, 0);
		g_ptr_array_add (lines, (gpointer) g_strdup (">"));
//		set->ex.tree.def_num = (gint) lines->len;
		g_ptr_array_add (lines, NULL);
		set->ex.tree.def = (gchar **) lines->pdata;
		g_ptr_array_free (lines, FALSE);
	}
}
/**
 * @brief optionally revert, then clean, tree option backup data
 *
 * @param set the set data structure
 * @param revert TRUE to restore the backup data to the tree, FALSE to cleanup only
 *
 * @return 
 **/
void e2_option_tree_unbackup (E2_OptionSet *set, gboolean revert)
{
	if (set->ex.tree.def != NULL)
	{	//backup data exists
		if (revert)
			_e2_option_tree_revert (set, FALSE);
		//cleanup
//		gint j = 0;
		gchar **cur = set->ex.tree.def;
		while (*cur != NULL)
		{
			g_free (*cur);
			cur++;
//			j++;
		}
//		set->ex.tree.def_num = j;
		g_free (set->ex.tree.def);
		set->ex.tree.def = NULL;
	}
}
/**
 * @brief iterate over all option trees, reverting any with backup data
 *
 * @return 
 **/
void e2_option_tree_restore_all (void)
{
	gint i;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
		if (set->type == E2_OPTION_TYPE_TREE)
			e2_option_tree_unbackup (set, TRUE);
	}
}
/**
 * @brief prepare default config tree contents for installation
 *
 * This creates a pointer array, referring to tree-row strings, and stores a
 * pointer, so that it can be processed after the config file (if any) is loaded
 *
 * @param set the set data structure
 * @param first start of a null-terminated series of arguments, 
 *  each a string which is to be converted to a tree row
 *
 * @return 
 **/
void _e2_option_tree_prepare_default (E2_OptionSet *set, gchar *first, ...)
{
	if (first == NULL) return;
//	printd (DEBUG, "preparing default option tree for %s>", first);
	va_list args;
	va_start (args, first);
	gchar *cur = first;
	GPtrArray *lines = g_ptr_array_sized_new (32);
	while (cur)
	{
//		cur = g_strdup (cur);
		//store ptr to constant string (do not free, later)
		g_ptr_array_add (lines, (gpointer) cur);
		cur = va_arg (args, gchar *);
	}
	va_end (args);
//	set->ex.tree.def_num = (gint) lines->len; UNUSED
	g_ptr_array_add (lines, NULL);  //null signals end, to parser
	set->ex.tree.def = (gchar **) lines->pdata;
	g_ptr_array_free (lines, FALSE);
}
/**
 * @brief install default config trees where needed
 *
 * after loading the config file (if any) at session start, check if any
 * tree option needs a default config, (stored in arrays) 
 * and install it if so
 * then free the default tree 'branches'
 *
 * @return 
 **/
void e2_option_trees_confirm (void)
{
	gint i;
//	gchar **cur;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
		if (set->type == E2_OPTION_TYPE_TREE)
		{
			if ((!set->ex.tree.found_in_config_file) &&
				(set->ex.tree.def != NULL))
			//option needs default configuration which is
			//stored in strings, each referred to in null-terminated
			//ptr array at ex.tree.def of the OptionSet
			{
//				printd (DEBUG, "installing default option tree %s>", *set->ex.tree.def);
				e2_option_read_from_string (set->ex.tree.def);
			}
			//in any event, free the stored default tree strings
			//if (set->ex.tree.found_in_config_file) &&
			 if (set->ex.tree.def != NULL)
			{
//				printd (DEBUG, "freeing init strings for default option tree %s>", *(set->ex.tree.def));
/* now uses constant strings, do not free
				cur = set->ex.tree.def;
				while (*cur != NULL)
				{
					g_free (*cur);
					cur++;
				} */
				g_free (set->ex.tree.def);
				set->ex.tree.def = NULL;
			}
		}
	}
}

E2_OptionSet *e2_option_tree_get (gchar *option)
{
	E2_OptionSet *set;
	set = (E2_OptionSet *) g_hash_table_lookup (options_hash, option);
	if (set == NULL)
	{
		printd (WARN, "trying to access option '%s' which doesn't exist", option);
		return NULL;
	}
	if (set->type != E2_OPTION_TYPE_TREE)
	{
		printd (WARN, "trying to access tree option '%s' which isn't a tree", set->name);
		return NULL;
	}
	return set;
}

void e2_option_tree_add (gchar *option, GtkTreeIter *iter, gint n_options,
	void *options[])
{
	E2_OptionSet *set = e2_option_tree_get (option);
	if (set != NULL)
		_e2_option_tree_add (set, iter, NULL, FALSE, FALSE, n_options, options);
}

void _e2_option_tree_add (E2_OptionSet *set, GtkTreeIter *iter, GtkTreeIter *parent,
	gboolean sibling, gboolean before, gint n_options, void *options[])
{
	if (sibling == TRUE)
	{
		if (before == TRUE)
			gtk_tree_store_insert_before (set->ex.tree.model, iter, NULL, parent);
		else
			gtk_tree_store_insert_after (set->ex.tree.model, iter, NULL, parent);
	}
	else
	{
		if (before == TRUE)
			gtk_tree_store_insert_before (set->ex.tree.model, iter, parent, NULL);
		else
			gtk_tree_store_insert_after (set->ex.tree.model, iter, parent, NULL);
	}

	gtk_tree_store_set (set->ex.tree.model, iter, set->ex.tree.columns_num, TRUE, -1);

	gint i;
	for (i = 0; i < n_options; i++)
	{
		if (options[i] != NULL)
			gtk_tree_store_set (set->ex.tree.model, iter, i, options[i], -1);
	}
	e2_option_tree_flag_change (set);
}

void e2_option_tree_add_default (gchar *option, GtkTreeIter *iter, GtkTreeIter *parent, gboolean sibling)
{
	E2_OptionSet *set = e2_option_tree_get (option);
	if (set != NULL)
	_e2_option_tree_add_default (set, iter, parent, sibling);
}

void _e2_option_tree_add_default (E2_OptionSet *set, GtkTreeIter *iter, GtkTreeIter *parent, gboolean sibling)
{
	void *arg[set->ex.tree.columns_num];
	GList *column;
	gint i = 0;
	for (column = set->ex.tree.columns; column != NULL; column = g_list_next (column))
	{
		E2_OptionTreeColumn *opt = column->data;
		switch (opt->type)
		{
			case E2_OPTION_TREE_TYPE_BOOL:
				arg[i] = GINT_TO_POINTER (opt->ival);
				break;
			case E2_OPTION_TREE_TYPE_SEL:
			case E2_OPTION_TREE_TYPE_STR:
			case E2_OPTION_TREE_TYPE_ICON:
			case E2_OPTION_TREE_TYPE_KEY:
				arg[i] = opt->sval;
				break;
			default:
				break;
		}
		i++;
	}
	
	_e2_option_tree_add (set, iter, parent, sibling, FALSE, set->ex.tree.columns_num, arg);

	if ((parent != NULL) && (sibling == FALSE))
		gtk_tree_view_expand_all (GTK_TREE_VIEW (set->widget));
	gtk_tree_view_set_cursor  (GTK_TREE_VIEW (set->widget),
		gtk_tree_model_get_path (GTK_TREE_MODEL (set->ex.tree.model), iter),
		gtk_tree_view_get_column (GTK_TREE_VIEW (set->widget), 0), FALSE);

	if (set->widget)
	{
		gtk_widget_grab_focus (set->widget);
	}
}

void e2_option_tree_del (gchar *option, GtkTreeIter *iter)
{
	E2_OptionSet *set = e2_option_tree_get (option);
	if (set != NULL)
		_e2_option_tree_del (set, iter);
}

void _e2_option_tree_del (E2_OptionSet *set, GtkTreeIter *iter)
{
	gtk_tree_store_remove (set->ex.tree.model, iter);
	e2_option_tree_flag_change (set);
}

void _e2_option_tree_move (E2_OptionSet *set, GtkTreeIter *iter, GtkTreePath *path_dest, gboolean sibling, gboolean before)
{
	//find out real number of columns, we also want to move the meta data
	gint n = gtk_tree_model_get_n_columns (set->ex.tree.model);
	gpointer values[n];
	gint i;
	//save old row contents
	for (i = 0; i < n; i++)
	{
		gtk_tree_model_get (set->ex.tree.model, iter, i, values + i, -1);
	}
	GtkTreePath *path_old = gtk_tree_model_get_path (set->ex.tree.model, iter);
	GtkTreeRowReference *ref_old = gtk_tree_row_reference_new (set->ex.tree.model, path_old);
	GtkTreeIter iter2;
	if ((gtk_tree_model_get_iter (set->ex.tree.model, &iter2, path_dest)) == TRUE)
	{
		GtkTreeIter iter_new;
		_e2_option_tree_add (set, &iter_new, &iter2, sibling, before, n, values);

		gtk_tree_path_free (path_old);
		path_old = gtk_tree_row_reference_get_path (ref_old);
		if ((gtk_tree_model_get_iter (set->ex.tree.model, &iter2, path_old)) == TRUE)
		{
			if (gtk_tree_model_iter_has_child (set->ex.tree.model, &iter2))
			{
				gboolean expand = FALSE;
				if (gtk_tree_view_row_expanded (GTK_TREE_VIEW (set->widget), path_old))
				{
					expand = TRUE;
					gtk_tree_view_collapse_row (GTK_TREE_VIEW (set->widget), path_old);
				}
				
				GtkTreePath *path_new = gtk_tree_model_get_path (set->ex.tree.model, &iter_new);
				GtkTreeIter iter_child;
				while (gtk_tree_model_iter_children (set->ex.tree.model, &iter_child, &iter2))
					_e2_option_tree_move (set, &iter_child, path_new, FALSE, TRUE);

				if (expand)
					gtk_tree_view_expand_row (GTK_TREE_VIEW (set->widget), path_new, FALSE);
				gtk_tree_path_free (path_new);
			}
			_e2_option_tree_del (set, &iter2);
		}

		gtk_tree_view_set_cursor  (GTK_TREE_VIEW (set->widget),
			path_dest, gtk_tree_view_get_column (GTK_TREE_VIEW (set->widget), 0), FALSE);
	}
	gtk_tree_path_free (path_old);
	gtk_tree_row_reference_free (ref_old);
}

void _e2_option_tree_set (E2_OptionSet *set, GtkTreeIter *iter, gint col, void *data)
{
	gtk_tree_store_set (GTK_TREE_STORE (set->ex.tree.model), iter, col, data, -1);
	e2_option_tree_flag_change (set);
}

void *e2_option_tree_get_void_simple (E2_OptionTreeColumn *col, gchar *option)
{
	switch (col->type)
	{
		case E2_OPTION_TREE_TYPE_BOOL:
			if (g_str_equal (option, "true")) return (void *) TRUE;
			else return (void *) FALSE;
			break;
		case E2_OPTION_TREE_TYPE_STR:
		case E2_OPTION_TREE_TYPE_ICON:
		case E2_OPTION_TREE_TYPE_SEL:
		case E2_OPTION_TREE_TYPE_KEY:
			return (void *) option;
		default:
			return NULL;
			break;
	}
	return NULL;
}

void e2_option_tree_write_to_file (E2_FILE *f, E2_OptionSet *set,
	GtkTreeIter *iter, gint level)
{
	do
	{
		gchar *rowstring = e2_option_tree_row_write_to_string (set, iter, level);
		gchar *writestring = g_strconcat (rowstring, "\n", NULL);
		e2_fs_file_put (writestring, f);
		g_free (rowstring);
		g_free (writestring);

		GtkTreeIter iter2;
		if (gtk_tree_model_iter_children (set->ex.tree.model, &iter2, iter))
			e2_option_tree_write_to_file (f, set, &iter2, level + 1);

	} while (gtk_tree_model_iter_next (set->ex.tree.model, iter));
}
