/*
 * 
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 * Portions copyright (C) 2004 Florian Zaehringer <flo.zaehringer@web.de>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/* Florian Zaehringer is the source of inspiration, and some of the code, here */
 
/**
 * @file src/config/e2_option_tree_context_menu.c
 * @brief context menu functions for config-dialog tree options
 * This file is #included in e2_option_tree.c
 * (i.e. this won't rebuild unless that is rebuilt)
 */

//prevent building of separate object file
#ifdef INCLUDED_IN_PARENT

  /*******************/
 /***** private ******/
/*******************/

/**
 * @brief set popup menu position
 *
 * This function is supplied when calling gtk_menu_popup(), to position
 * the displayed menu.
 * set @a push_in to TRUE for menu completely inside the screen,
 * FALSE for menu clamped to screen size
 *
 * @param menu the GtkMenu to be positioned
 * @param x	place to store gint representing the menu left
 * @param y  place to store gint representing the menu top
 * @param push_in place to store pushin flag
 * @param treeview where the menu is to be shown
 *
 * @return
 */
static void _e2_option_tree_menu_set_position (GtkMenu *menu,
	gint *x, gint *y, gboolean *push_in, GtkWidget *treeview)
{
	gint left, top;
	gtk_window_get_position (GTK_WINDOW (config_dialog),
		&left, &top);
	GtkAllocation alloc = treeview->allocation;
	*x = left + alloc.x + alloc.width;
	*y = top + alloc.y + alloc.height/2;
	*push_in = TRUE;
}
/**
 * @brief create treeview signature
 *
 * This gets the set name of the current view
 * The string is used as a signature to confirm the correct treeview,
 * to enable pasting
 *
 * @param id_string ptr to string where the result is to be stored
 * @param length ptr to no. of columns in model
 * @param treeview widget where the menu was initiated
 *
 * @return
 */
static void _e2_option_tree_menu_build_signature
	(gchar **id_string, gint *length, GtkWidget *treeview)
{
	GtkTreeModel *model = gtk_tree_view_get_model 
		(GTK_TREE_VIEW (treeview));
	E2_OptionSet *set = g_object_get_data (G_OBJECT (model), "e2-option-set");
	*id_string = g_strdup (set->name);
	*length = gtk_tree_model_get_n_columns (model);
}
/**
 * @brief tree expand-all callback
 *
 * @param widget the menu item widget which activated the callback
 * @param treeview the treeview to be collapsed
 *
 * @return TRUE (to prevent further handlers)
 */
static gboolean _e2_option_tree_expand_all_cb (GtkWidget *widget, GtkTreeView *treeview)
{
	gtk_widget_set_sensitive (widget, FALSE);
	//get the paired item
	gpointer m = g_object_get_data (G_OBJECT (widget), "e2-menu-button");
	gtk_widget_set_sensitive (GTK_WIDGET (m), TRUE);

	return e2_tree_expand_all_cb (widget, treeview);
}
/**
 * @brief tree collapse-all callback
 *
 * @param widget the menu item widget which activated the callback
 * @param treeview the treeview to be collapsed
 *
 * @return TRUE (to prevent further handlers)
 */
static gboolean _e2_option_tree_collapse_all_cb (GtkWidget *widget, GtkTreeView *treeview)
{
	gtk_widget_set_sensitive (widget, FALSE);
	//get the paired button
	gpointer m = g_object_get_data (G_OBJECT (widget), "e2-menu-button");
	gtk_widget_set_sensitive (GTK_WIDGET (m), TRUE);

	return e2_tree_collapse_all_cb (widget, treeview);
}
/**
 * @brief add model row, and all its descendants, to buffer list
 *
 * A depth-first copy process converts each row to an array,
 * then appends that array to a specified glist.
 * It is recursive, to handle descendants
 *
 * @param model the treemodel that is being processed
 * @param iter _pointer_ to tree iter to be processed
 * @param columncount no. of columns in the treemodel that is being processed
 * @param rowscopy **glist that holds the result
 *
 * @return
 */
static void _e2_option_tree_menu_copy_treerow (GtkTreeModel *model, 
	GtkTreeIter *iter, gint columncount, GList **rowscopy)
{
	//add path of copied node to the start of the buffer entry, to support proper pasting
	GtkTreePath *path = gtk_tree_model_get_path (model, iter);
	gchar *pathstring = gtk_tree_path_to_string (path);
	gtk_tree_path_free (path);

	GValueArray *row_array = g_value_array_new (columncount + 1);
	GValue value = {0, };
	GValue *valptr = g_value_init (&value, G_TYPE_STRING);
	g_value_set_static_string (valptr, pathstring);  //no copy created, don't free it here
	g_value_array_append (row_array, valptr);
	g_value_unset (valptr);
	//add store column values to the buffer entry
	gint i;
	for (i = 0; i < columncount ; i++)
	{
		GValue value = {0, };
		gtk_tree_model_get_value (model, iter, i, &value);
		g_value_array_append (row_array, &value);
		g_value_unset (&value);
	}
	//update the buffer list
	*rowscopy = g_list_append (*rowscopy, row_array);
	//repeat for all children
	gint children;
	if ((children = gtk_tree_model_iter_n_children (model, iter)) > 0)
	{
		GtkTreeIter iter2;
		gint n;
		for (n=0; n<children; n++)
		{
			gtk_tree_model_iter_nth_child (model, &iter2, iter, n);
			_e2_option_tree_menu_copy_treerow (model, &iter2, 
				columncount, rowscopy);
		}
	}
}
/**
 * @brief context menu copy action
 *
 * This is called in response to the copy and cut menu items
 * It works on all selected rows, and all their descendants
 * (though robust relation-checks among various selected rows
 * have NOT been implemented, as multiple selection is not
 * currently allowed)
 * A signature string for the treeview and a list of arrays,
 * each with a path string and a whole row contents, are added 
 * to a hash table.
 * The table may contain entries for more than 1 treeview
 *
 * @param menu_item the selected menu widget
 * @param treeview widget where the action is to occur
 *
 * @return TRUE (to prevent further handlers) if any treeview item was selected
 */
static gboolean _e2_option_tree_menu_copy_cb 
	(GtkWidget *menu_item, GtkWidget *treeview)
{
	GtkTreeSelection *sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
	if (gtk_tree_selection_count_selected_rows (sel) > 0)
	{
		gchar *id_string;
		gint columncount;
		_e2_option_tree_menu_build_signature (&id_string, &columncount, treeview);

		GtkTreeModel *model;
		GList *base = gtk_tree_selection_get_selected_rows (sel, &model);
		GList *selpaths;
		GtkTreeIter iter;
		
		GList *rowscopied = NULL;
		//recursive copy of all selected items and their descendants
		for (selpaths = base; selpaths != NULL; selpaths = selpaths->next)
		{
			if (gtk_tree_model_get_iter (model, &iter, (GtkTreePath *) selpaths->data))
				_e2_option_tree_menu_copy_treerow (model, &iter, columncount, &rowscopied);
		}

		g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
		g_list_free (base);

		printd (DEBUG, "adding to buffer: key %s and %d rows", id_string,
			g_list_length (rowscopied));

		g_hash_table_insert (tree_view_buffer_hash, id_string, rowscopied);
		//id_string is freed by g_hash_table_insert()

//		return TRUE;
	}
	return TRUE;  //FALSE;
}
/**
 * @brief context menu paste action
 *
 * This is called in response to the paste menu item
 * An id string for the treeview is first constructed, to ensure that the 
 * paste is appropriate
 * Pasted row(s) are placed after the selected row in the
 * destination treeview, as child(ren) if the path depth is appropriate
 *
 * @param menu_item the selected menu widget
 * @param set data struct for the set being processed
 *
 * @return TRUE (to prevent further handlers) if there is a selected item in the treeview
 */
static gboolean _e2_option_tree_menu_paste_cb
	(GtkWidget *menu_item, E2_OptionSet *set)
{
	GtkWidget *treeview = g_object_get_data (G_OBJECT (menu_item), "tree-view");

	GtkTreeSelection *sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
	if (gtk_tree_selection_count_selected_rows (sel) > 0)
	{
		//get data for the current page/view
		gchar *id_string;
		gint columncount;
		_e2_option_tree_menu_build_signature (&id_string, &columncount, treeview);
		//get the copied data for the current page
		GList *rowscopied = g_hash_table_lookup (tree_view_buffer_hash, id_string);
		//get path data for the 1st buffered item, the root of the copied branch
		GValueArray *value_array = rowscopied->data;
		GValue *value = g_value_array_get_nth (value_array, 0);
		gchar *pathstring = g_value_dup_string (value);
		GtkTreePath *path = gtk_tree_path_new_from_string (pathstring);
		gint copyrootdepth = gtk_tree_path_get_depth (path);
		gtk_tree_path_free (path);
		//store for references (hence iters) needed to preserve parent/child relations for pasted rows
		GHashTable *ref_hash = NULL;
		GtkTreeRowReference *ref;

		GtkTreeModel *model;
		GList *selpaths = gtk_tree_selection_get_selected_rows (sel, &model);
		GList *tmp;
		GtkTreeIter local1;
		GtkTreeIter local2;
		GtkTreeIter *iter = &local1;
		GtkTreeIter *parent = &local2;

		//paste all buffered items, after _all_ valid selected items !!
		for (tmp = selpaths; tmp!=NULL; tmp=tmp->next)
		{
			if (gtk_tree_model_get_iter (model, iter, (GtkTreePath *) tmp->data))
			{
				gint thisdepth = gtk_tree_path_get_depth (tmp->data);
				if (thisdepth == copyrootdepth)
				{	//the selected row is a valid sibling of the buffered row(s)
					//paste the root of the copied branch
					printd (DEBUG, "pasting branch root, at depth %d", thisdepth);
					if (gtk_tree_model_iter_parent (model, parent, iter))
					{	//pasted row (iter) will be _pre_pended to parent's children
						gtk_tree_store_insert_after (GTK_TREE_STORE (model), iter, parent, NULL);
					}
					else //parent is the root node
					{ //pasted row (iter) will be inserted after selected row
						gtk_tree_store_insert_after (GTK_TREE_STORE (model), parent, NULL, iter);
						iter = parent;
					}
				}
				else if (thisdepth == copyrootdepth - 1)
				//the selected item is a valid parent of the buffered row(s)
				{	//pasted row (iter) will be _pre_pended to parent's children
					printd (DEBUG, "pasting branch root, at depth %d", thisdepth);
					gtk_tree_store_insert_after (GTK_TREE_STORE (model), parent, iter, NULL);
					iter = parent;
				}
				else
				//see if we can find another suitable paste-place
					continue;

				//populate root
				gint i;
				for (i = 1; i <= columncount; i++)
				{
					value = g_value_array_get_nth (value_array, i);
					gtk_tree_store_set_value (GTK_TREE_STORE (model), iter, i-1, value);
				}

				//remember how to find root iter, for preserving branch relativities
				//(key is used determine parent/child relations between pasted rows)
				ref_hash = g_hash_table_new_full (g_str_hash, g_str_equal,
					g_free, (GDestroyNotify) gtk_tree_row_reference_free);
				path = gtk_tree_model_get_path (model, iter);
				ref = gtk_tree_row_reference_new (model, path);
				g_hash_table_insert (ref_hash, pathstring, ref);
				gtk_tree_path_free (path);

				GList *tmp2 = rowscopied;
				//now the rest of the branch, relative to the root
				while ((tmp2=tmp2->next) != NULL)
				{
					value_array = tmp2->data;
					//get path string
					value = g_value_array_get_nth (value_array, 0);
					pathstring = g_value_dup_string (value);
					//get former parent path string, by truncating the last segment
					gchar *s = g_strrstr (pathstring, ":");
					if (s != NULL)
						*s = '\0';
					//get parent iter
					ref = g_hash_table_lookup (ref_hash, pathstring);
					path = gtk_tree_row_reference_get_path (ref);
					gtk_tree_model_get_iter (model, parent, path);
					//append row (iter) to parent's children
					gtk_tree_store_insert_before (GTK_TREE_STORE (model), iter, parent, NULL);
					gtk_tree_path_free (path);
					g_free (pathstring);
					//remember new iter in case it's a parent
					path = gtk_tree_model_get_path (model, iter);
					ref = gtk_tree_row_reference_new (model, path);
					g_hash_table_insert (ref_hash, g_value_dup_string (value),
						ref);
					gtk_tree_path_free (path);
					//populate row
					gint i;
					for (i = 1; i <= columncount; i++)
					{
						value = g_value_array_get_nth (value_array, i);
						gtk_tree_store_set_value (GTK_TREE_STORE (model), iter, i-1, value);
					}
				}
			}
		}

		e2_option_tree_flag_change (set);

		//no change to (single row) selection, as all new rows are after

		g_free (id_string);
		if (ref_hash != NULL)
			g_hash_table_destroy (ref_hash);
		g_list_foreach (selpaths, (GFunc) gtk_tree_path_free, NULL);
		g_list_free (selpaths);
//		return TRUE;
	}
	return TRUE;  //FALSE;
}
/**
 * @brief add treerowreferences for row and its descendants to list
 *
 * A depth-first search scan creates references for the specified iter and
 * all its descendants, appending each ref. to a specified glist.
 * It is recursive, to handle descendants
 *
 * @param model the treemodel that is being processed
 * @param iter _pointer_ to tree iter to be processed
 * @param rowrefs **glist that holds the result
 *
 * @return
 */
static void _e2_option_tree_menu_reference_treerow
	(GtkTreeModel *model, GtkTreeIter *iter, GList **rowrefs)
{
	GtkTreePath *path = gtk_tree_model_get_path (model, iter);
	GtkTreeRowReference *ref;
	if ((ref = gtk_tree_row_reference_new (model, path)) != NULL)
		*rowrefs = g_list_append (*rowrefs, ref);
	gtk_tree_path_free (path);

	gint children;
	if ((children = gtk_tree_model_iter_n_children (model, iter)) > 0)
	{
		GtkTreeIter iter2;
		gint n;
		for (n=0; n<children; n++)
		{
			gtk_tree_model_iter_nth_child (model, &iter2, iter, n);
			_e2_option_tree_menu_reference_treerow (model, &iter2, rowrefs);
		}
	}
}
/**
 * @brief delete treestore row(s)
 *
 * This is called after row(s) have been copied, as part of the response to the cut menu item
 *
 * @param treeview where the action is to occur
 *
 * @return
 */
static void _e2_option_tree_menu_del (GtkTreeView *treeview, E2_OptionSet *set)
{
	GtkTreeSelection *sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
	if (gtk_tree_selection_count_selected_rows (sel) > 0)
	{
		GList *rowrefs = NULL;

		GtkTreeModel *model;
		GList *selpaths = gtk_tree_selection_get_selected_rows (sel, &model);
		GList *tmp;
		//get a place to goto after the deletion
		GtkTreePath *newpath = gtk_tree_path_copy (selpaths->data);
		if (!gtk_tree_path_prev (newpath))
			if (!gtk_tree_path_up (newpath))
				newpath = gtk_tree_path_new_first ();
		GtkTreePath *path;
		GtkTreeIter iter;

		//create treerowrefs for all buffered rows
		for (tmp = selpaths; tmp!=NULL; tmp=tmp->next)
		{
			if (gtk_tree_model_get_iter (model, &iter, (GtkTreePath *) tmp->data))
				_e2_option_tree_menu_reference_treerow (model, &iter, &rowrefs);
		}
		
		for (tmp = rowrefs; tmp!=NULL; tmp=tmp->next)
		{
			if ((path = gtk_tree_row_reference_get_path
				((GtkTreeRowReference *) tmp->data)) != NULL)
			{
				//brute force, unselect them all
				gtk_tree_selection_unselect_path (sel, tmp->data);
				if (gtk_tree_model_get_iter (model, &iter, path))
					gtk_tree_store_remove (GTK_TREE_STORE (model), &iter);
			}
		}

		if ( gtk_tree_model_get_iter (model, &iter, newpath)
			|| gtk_tree_model_get_iter_first (model, &iter) )
		{
			newpath = gtk_tree_model_get_path (model, &iter);
			gtk_tree_view_set_cursor (treeview, newpath,
				gtk_tree_view_get_column (treeview, 0),
					FALSE);
		}

		e2_option_tree_flag_change (set);

		g_list_foreach (rowrefs, (GFunc) gtk_tree_row_reference_free, NULL);
		g_list_free (rowrefs);
		g_list_foreach (selpaths, (GFunc) gtk_tree_path_free, NULL);
		g_list_free (selpaths);
		gtk_tree_path_free (newpath);
	}
}
/**
 * @brief context menu cut action
 *
 * This is called in response to the cut menu item
 * It involves a row(s) copy action, followed by row(s) delete
 *
 * @param menu_item the selected menu widget
 * @param set data struct for the set being processed
 *
 * @return TRUE (to prevent further handlers)
 */
static gboolean _e2_option_tree_menu_cut_cb
	(GtkWidget *menu_item, E2_OptionSet *set)
{
	 GtkWidget *treeview = g_object_get_data (G_OBJECT (menu_item), "tree-view");
	_e2_option_tree_menu_copy_cb (menu_item, treeview);
	_e2_option_tree_menu_del (GTK_TREE_VIEW (treeview), set);
	return TRUE;
}
/**
 * @brief add treeview context menu items
 *
 * @param treeview where the action is to occur
 * @param menu the menu widget to which the items are to be added
 * @param flags indicator of menu items to be created
 * @param set pointer to data struct for the option that is being edited
 *
 * @return
 */
static void _e2_option_tree_menu_items_add (GtkTreeView *treeview, 
	GtkWidget *menu, E2_TreeContextMenuFlags flags, E2_OptionSet *set)
{
	E2_Nontet *n = e2_utils_nontet_new ();
	g_object_set_data_full (G_OBJECT (menu), "e2-config-menu", n, g_free);
	//meu items which modify store content need set for callback data,
	//other items 
	if ((flags & E2_TREE_CONTEXT_CP_PST) || (flags & E2_TREE_CONTEXT_DEFAULT))
	{
		n->a = e2_menu_add (menu, _("Copy"), GTK_STOCK_COPY,
			_("Copy selected row and any descendant(s)"),
			_e2_option_tree_menu_copy_cb, treeview);

		n->b = e2_menu_add (menu, _("Paste"), GTK_STOCK_PASTE,
			_("Paste previously copied or cut row(s) after current row"),
			_e2_option_tree_menu_paste_cb, set);
		g_object_set_data (G_OBJECT (n->b), "tree-view", treeview);	//tree edit fn in this file needs this data

		n->c = e2_menu_add (menu, _("Cut"), GTK_STOCK_CUT, 
			_("Cut selected row and any descendant(s)"),
			_e2_option_tree_menu_cut_cb, set);
		g_object_set_data (G_OBJECT (n->c), "tree-view", treeview);	//tree edit fn in this file needs this data

		e2_menu_add_separator (menu);
	}
	if (flags & E2_TREE_CONTEXT_EXP_COL)
	{
		n->d = e2_menu_add (menu, _("_Expand"), GTK_STOCK_ZOOM_IN, 
				_("Expand all rows on this page"), _e2_option_tree_expand_all_cb, treeview);
		gtk_widget_set_sensitive (GTK_WIDGET(n->d), FALSE);	//widget starts with rows expanded
		n->e = e2_menu_add (menu, _("C_ollapse"), GTK_STOCK_ZOOM_OUT,
				_("Collapse all rows on this page"), _e2_option_tree_collapse_all_cb, treeview);
		//set pointers to the paired buttons
		g_object_set_data_full (G_OBJECT (n->d), "e2-menu-button", n->e, NULL);
		g_object_set_data_full (G_OBJECT (n->e), "e2-menu-button", n->d, NULL);

		e2_menu_add_separator (menu); //more items added in the calling function
	}
	n->f = e2_menu_add (menu, _("_Add"), GTK_STOCK_ADD, 
		_("Add a row after the current one"), _e2_option_tree_add_below_cb, set);

	if (flags & E2_TREE_CONTEXT_EXP_COL)
		n->g = e2_menu_add (menu, _("Add _child"), GTK_STOCK_INDENT, 
			_("Add a child to the selected row"), _e2_option_tree_add_child_cb, set);

	e2_menu_add_separator (menu);
	n->h = e2_menu_add (menu, _("_Up"), GTK_STOCK_GO_UP, 
		_("Move selected row up"), _e2_option_tree_move_up_cb, set);

	n->i = e2_menu_add (menu, "_Down", GTK_STOCK_GO_DOWN, 
		_("Move selected row down"), _e2_option_tree_move_down_cb, set);

}

  /*******************/
 /***** public ******/
/*******************/

/**
 * @brief (de)sensitize tree-option context-menu items
 *
 * This changes the sensitivity of the paste menu item, depending on 
 * whether the current treeview's set name is a key in the hash buffer
 *
 * @param menu the context-menu widget
 * @param treeview widget where the menu was initiated
 *
 * @return
 */
void e2_option_tree_menu_set_sensitive
	(GtkWidget *menu, GtkWidget *treeview)
{
	E2_Nontet *m = g_object_get_data (G_OBJECT (menu), "e2-config-menu");
	if (m == NULL)
		return;
	//first deal with the paste item
	//get current view's signature
	GtkTreeModel *model = gtk_tree_view_get_model 
		(GTK_TREE_VIEW (treeview));
	E2_OptionSet *set = g_object_get_data (G_OBJECT (model), "e2-option-set");

	gboolean state = (g_hash_table_lookup (tree_view_buffer_hash, set->name) != NULL);
	//FIXME also check for valid path depth of currently-selected item

	if (m->b != NULL)	//paste item
		gtk_widget_set_sensitive (m->b, state);

//	now adjust other menu items to conform with the page buttons
	E2_Sextet *s2 = g_object_get_data (G_OBJECT (treeview), "e2-config-buttons");
	if (s2 == NULL)
		return;
	if (s2->b != NULL)	//up
	{
		state = GTK_WIDGET_SENSITIVE(s2->b);
		if (m->h != NULL)	//up
			gtk_widget_set_sensitive (m->h, state);
	}
	if (s2->c != NULL)	//down
	{
		state = GTK_WIDGET_SENSITIVE(s2->c);
		if (m->i != NULL)	//down
			gtk_widget_set_sensitive (m->i, state);
	}
	if (s2->d != NULL)	//remove
	{
		state = GTK_WIDGET_SENSITIVE(s2->d);
		if (m->a != NULL)	//copy
			gtk_widget_set_sensitive (m->a, state);
		if (m->c != NULL)	//cut
			gtk_widget_set_sensitive (m->c, state);
	}
	if (s2->e != NULL)	//add
	{
		state = GTK_WIDGET_SENSITIVE(s2->e);
		if (m->f != NULL)	//add
			gtk_widget_set_sensitive (m->f, state);
		}
	if (s2->f != NULL)	//add child
	{
		state = GTK_WIDGET_SENSITIVE(s2->f);
		if (m->g != NULL)	//add child
			gtk_widget_set_sensitive (m->g, state);
	}
}
/**
 * @brief free memory used by tree-option context menu hash table entry
 *
 * Each entry in the hash buffer is a glist of value arrays, one array for each
 * row copied to the buffer
 *
 * @param list  glist to be freed
 *
 * @return
 */
void e2_option_tree_menu_hash_clean (GList *list)
{
	g_list_foreach (list, (GFunc) g_value_array_free, NULL);
	g_list_free (list);
}

#endif //def INCLUDED_IN_PARENT
