/* $Id: e2_config_dialog.c 1095 2005-08-28 07:17:41Z tpgww $
 *
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 * Copyright (C) 1999 Michael Clark.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "emelfm2.h"
#include <string.h>
#include "e2_dialog.h"
#include "e2_config_dialog.h"
#include "e2_option_tree.h"
#include "e2_filelist.h"
#include "e2_task.h"
#include "e2_plugins.h"
#include "e2_filetype.h"

//common vars we need in several functions
//static
GtkWidget *config_dialog = NULL;	//this is accessed by the list-cell-renderer popup
static GtkWidget *treeview;
static GtkTreeRowReference *page_last_ref = NULL;
static gchar *page_selected;
static gint page_last = 0;	//the page number of the notebook page last opened

// =========== functions supporting single-page config dialogs ===============

/**
 * @brief dummy 'apply' function used when none is supplied by caller
 **/
static void _e2_cd1_no_application (void)
{}
/**
 * @brief clean up at the end of a one-page config dialog
 *
 * This relates to tree-options only
 *
 * @param rt pointer to data struct for the dialog
 * @param flushtype TRUE to restore backed-up data to the tree, FALSE to cleanup only
 *
 * @return 
 **/
static void _e2_cd1_cancel (E2_SpecificConfDialogRuntime *rt, gboolean flushtype)
{
	//restore or abandon the default tree data
	e2_option_tree_unbackup (rt->set, flushtype);
	//cleanup any copy/cut buffer data
	g_hash_table_destroy (tree_view_buffer_hash);
	gtk_widget_destroy (rt->dialog);
	g_free (rt);
	config_dialog = NULL;
	gtk_widget_grab_focus (curr_view->treeview);
}

  /*********************/
 /***** callbacks *****/
/*********************/

/**
 * @brief callback for dialog "delete-event" signal
 *
 * @param widget UNUSED pointer to the widget that was 'deleted'
 * @param rt pointer to data struct for the dialog
 *
 * @return 
 **/
static void _e2_cd1_close_cb (GtkWidget *widget, E2_SpecificConfDialogRuntime *rt)
{
	_e2_cd1_cancel (rt, TRUE);
}
/**
 * @brief Close the dialog when esc key is pressed
 *
 * This is the callback for @a widget keypress signal
 *
 * @param widget UNUSED pointer to the widget that received the keypress
 * @param event event data struct
 * @param rt pointer to data struct for the dialog
 *
 * @return TRUE if esc key pressed, otherwise FALSE
 **/
static gboolean _e2_cd1_key_press_cb (GtkWidget *widget, GdkEventKey *event, 
	E2_SpecificConfDialogRuntime *rt)
{
	if (event->keyval == GDK_Escape)
	{  //restore tree data and clean up
		_e2_cd1_cancel (rt, TRUE);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief handle button click in a one-page config dialog
 *
 * This is the callback for @a button response signal
 * Note: applies only to tree options
 *
 * @param button the widget that was clicked
 * @param response the response assigned to @a button
 * @param rt pointer to data struct for the dialog
 *
 * @return 
 **/
static void _e2_cd1_response_cb (GtkWidget *button,
	gint response, E2_SpecificConfDialogRuntime *rt)
{
	switch (response)
	{
		case GTK_RESPONSE_OK:
			//perform any post-dialog data processing, via caller-specified 'apply' action
			(*rt->apply_function) ();
			//clean up without restoring tree data
			_e2_cd1_cancel (rt, FALSE);
			e2_option_disable_config_checks ();
			//update stored config data
			e2_option_file_write ();
			//update stamp, to prevent auto-refresh
			e2_fs_touch_config_dir ();
			e2_option_enable_config_checks ();
			break;
		case GTK_RESPONSE_APPLY:
			(*rt->apply_function) ();
/*			//clear old backup data
			e2_option_tree_unbackup (rt->set, FALSE);
			//renew backup data, in case we cancel later
			e2_option_tree_backup (rt->set); */
			break;
		default:
			_e2_cd1_cancel (rt, TRUE);
			break;
	}
}
/**
 * @brief create single-page config dialog for a tree-type option set
 *
 * This creates and optionally shows a dialog similar to a page in the overall config
 * dialog. It applies only to tree-type option sets.
 * Clicking 'apply' or 'ok' performs a caller-specified fn (if any) and updates the config file
 *
 * @param set_name string, non-translated 'internal' name of the tree optionset to be worked on
 * @param function ptr to caller's <void fn(void)> 'apply changes' function, or NULL if nothing to be done
 * @param showit boolean TRUE to show the dialog immediately, FALSE if not
 *
 * @return E2_SpecificConfDialogRuntime created, with dialog data
 **/
E2_SpecificConfDialogRuntime *e2_config_dialog_single (gchar *set_name,
	void *function, gboolean showit)
{
	//receptacle for cut/copied row(s) data generated from context menu
	tree_view_buffer_hash = g_hash_table_new_full (g_str_hash, g_str_equal, 
		g_free, (GDestroyNotify) e2_option_tree_menu_hash_clean);
	E2_SpecificConfDialogRuntime *rt = g_malloc (sizeof (E2_SpecificConfDialogRuntime));

	rt->set = e2_option_get_simple (set_name);
	rt->apply_function = (function != NULL) ? function : _e2_cd1_no_application;
  //set config_dialog too, so that list-cell-renderer can access the dialog
	config_dialog =
	rt->dialog = e2_dialog_create (NULL, NULL, _("configuration"),
		_e2_cd1_response_cb, rt);
	gtk_widget_set_size_request (rt->dialog, 600, 400);
	GtkWidget *vbox = GTK_DIALOG (rt->dialog)->vbox;
	gtk_container_set_border_width (GTK_CONTAINER (vbox), E2_PADDING);
	GtkWidget *frame = e2_widget_add_frame (vbox, FALSE, 0, NULL, FALSE);
	gchar *label_text2 = g_strconcat ("<b>", rt->set->group, "</b>", NULL);
	GtkWidget *label = e2_widget_add_label (NULL, label_text2, 0.5, FALSE, 0);
	g_free (label_text2);
	gtk_misc_set_padding (GTK_MISC (label), E2_PADDING_XSMALL, E2_PADDING_XSMALL);
	gtk_container_add (GTK_CONTAINER (frame), label);

	e2_option_connect (rt->dialog, FALSE);

	//add tree config data
	//(this also arranges backup of tree data)
	e2_option_tree_add_widget (rt->dialog, vbox, rt->set);

	g_signal_connect (G_OBJECT (rt->dialog), "delete_event",
		G_CALLBACK (_e2_cd1_close_cb), rt);
	g_signal_connect_after (G_OBJECT (rt->dialog), "key_press_event",
		G_CALLBACK (_e2_cd1_key_press_cb), rt);
	//prepare dialog with buttons
	E2_BUTTON_CANCEL.showflags |= E2_BTN_DEFAULT;
	E2_BUTTON_OK.showflags &= !E2_BTN_DEFAULT;
	e2_dialog_show (rt->dialog, app.main_window, 0,
		&E2_BUTTON_CANCEL, &E2_BUTTON_APPLY, &E2_BUTTON_OK, NULL);
	if (showit)
		gtk_widget_show_all (rt->dialog);

	return rt;
}

// ================== end of single-page config things ==================

  /*********************/
 /***** callbacks *****/
/*********************/

/**
 * @brief callback for show dialog 
 *
 * @param dialog UNUSED the dialog widget being shown
 * @param sw scrolled window ...
 *
 * @return 
 **/
static void _e2_confdlg_show_cb (GtkWidget *dialog, GtkWidget *sw)
{
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
}
/**
 * @brief callback for show dialog
 *
 * @param dialog
 * @param notebook
 *
 * @return 
 **/
static void _e2_confdlg_show_cb2 (GtkWidget *dialog, GtkNotebook *notebook)
{
	gtk_notebook_set_current_page (notebook, page_last);
	gtk_widget_grab_focus (treeview);
}
/**
 * @brief callback on changes on the dependencies between option widgets
 *
 * @param button
 * @param data
 *
 * @return 
 **/
static gboolean _e2_confdlg_depends_cb (GtkWidget *button,
	GtkWidget *data)
{
	gtk_widget_set_sensitive (data,
		!GTK_TOGGLE_BUTTON (button)->active);
	return TRUE;
}
/**
 * @brief
 *
 * @param button
 * @param data
 *
 * @return 
 **/
static gboolean _e2_confdlg_depends_cb_ (GtkWidget *button,
	GtkWidget *data)
{
	gtk_widget_set_sensitive (data,
		GTK_TOGGLE_BUTTON (button)->active);
	return TRUE;
}
/**
 * @brief
 *
 * @return 
 **/
static void _e2_confdlg_clean (void)
{
	if (config_dialog != NULL)
	{
		printd (DEBUG, "config dialog widget destroy");
		cfg.window_width = config_dialog->allocation.width;
		cfg.window_height = config_dialog->allocation.height;
		gtk_widget_destroy (config_dialog);
		config_dialog = NULL;
	}
	//cleanup any copy/cut buffer data
	g_hash_table_destroy (tree_view_buffer_hash);
}
/**
 * @brief
 *
 * @param button
 * @param apply
 *
 * @return 
 **/
static gboolean _e2_confdlg_ok_cb (GtkWidget *button, gboolean apply)
{
	printd (DEBUG, "config dialog callback: %s", (apply) ? "apply" : "ok");

	//in case monitoring option changes, before it does so ..
	e2_option_disable_config_checks ();
	if (cfg.auto_refresh_id != 0)
		g_source_remove (cfg.auto_refresh_id);

	gboolean bchoice;
	gint ichoice;
	gchar *cchoice, *ccurrent;

	E2_OptionFlags buildflags = 0;
	gint i;
	//update all options that were included in the dialog
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
		if (set->widget != NULL)	//this is the signal for not included in dialog
		{
			switch (set->type)
			{
				case E2_OPTION_TYPE_BOOL:
					bchoice = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (set->widget));
					if (_e2_option_bool_get (set) != bchoice)
					{
						_e2_option_bool_set (set, bchoice);
						buildflags |= set->flags;
					}
					break;
				case E2_OPTION_TYPE_INT:
					ichoice = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (set->widget));
					if (_e2_option_int_get (set) != ichoice)
					{
						_e2_option_int_set (set, ichoice);
						buildflags |= set->flags;
					}
					break;
				case E2_OPTION_TYPE_FONT:
				case E2_OPTION_TYPE_STR:
				case E2_OPTION_TYPE_COLOR:
					cchoice = gtk_editable_get_chars (GTK_EDITABLE (set->widget), 0, -1);
					ccurrent = _e2_option_str_get (set);
					if (! g_str_equal (cchoice, ccurrent))
					{
						if (set->type == E2_OPTION_TYPE_COLOR)
							_e2_option_color_set_str (set, cchoice);
						else
							_e2_option_str_set (set, cchoice);
						buildflags |= set->flags;
					}
					g_free (cchoice);
					break;
				case E2_OPTION_TYPE_SEL:
					ichoice = gtk_combo_box_get_active (GTK_COMBO_BOX (set->widget));
					if (_e2_option_sel_get (set) != ichoice)
					{
						_e2_option_sel_set (set, ichoice);
						buildflags |= set->flags;
					}
					break;
				case E2_OPTION_TYPE_TREE:
					if (set->ex.tree.flags & E2_OPTION_TREE_SET_EDITED)
					{
						buildflags |= set->flags;
						set->ex.tree.flags &= ~E2_OPTION_TREE_SET_EDITED;
					}
					//abandon backed-up tree-option data
					e2_option_tree_unbackup (set, FALSE);
					//if the config dialog is not finished, get ready for next ok/apply
					//FIXME this not needed if backup triggered by dirty tree
					if (apply)
						e2_option_tree_backup (set);
					break;
				default:
					break;
			}
		}
	}

	//if this is an 'ok' button response, close the dialog etc
	if (!apply)
		_e2_confdlg_clean ();

	//rebuild only what is needed ...
	//the 'special' (non-window-specific) rebuilds are handled in the
	//same order as for a session-start
	if (buildflags & E2_OPTION_FLAG_BUILDKEYS)
		e2_keybinding_clean ();

	if (buildflags & E2_OPTION_FLAG_BUILDPLUGS)
		e2_plugins_apply_new ();
	
	if (buildflags & E2_OPTION_FLAG_BUILDSAMEBARS)
	{	//user wants toolbar buttons conformed
		//FIXME distinguish between style & size
		gint index1 = e2_option_sel_get ("allbars-style");
		gint index2 = e2_option_sel_get ("allbars-isize");
		E2_ToolbarData **thisbar;
		for (thisbar = app.bars; *thisbar != NULL; thisbar++)
		{
			_e2_option_sel_set ((*thisbar)->rt->style, index1);
			_e2_option_sel_set ((*thisbar)->rt->isize, index2);
		}
	}
	if (buildflags & E2_OPTION_FLAG_BUILDALL)
		e2_window_recreate (&app.window);	//also registers all key-bindings
	else if (buildflags & E2_OPTION_FLAG_BUILDPANES)
	{
		//rebuild filepane treeviews, & toolbars (in case they are destroyed with the panes)
		e2_window_recreate (&app.window);
/*		//FIXME more intelligent rebuild for this e.g no need to do output pane
		if (buildflags & E2_OPTION_FLAG_BUILDKEYS)
		//create all bindings, then register _("general"
		//(must be done before file-pane contents are created)
			e2_keybinding_register (_C(17), app.main_window);
		e2_pane_recreate (&app.pane1);
		e2_pane_recreate (&app.pane2);
//		e2_toolbar_create (&app.toolbar);	//this might be inside a destroyed pane
//		e2_toolbar_create (&app.commandbar);
		e2_pane_flag_active ();

		WAIT_FOR_EVENTS; //make sure pane parameters are set, before connecting cb's
		e2_window_pane_cb_setup (&app.window);

		e2_pane_change_dir (&app.pane1, app.pane1.path);
		e2_pane_change_dir (&app.pane2, app.pane2.path);
*/
	}
	else
	{
		if (buildflags & E2_OPTION_FLAG_BUILDBARS)
			e2_toolbar_recreate_all ();

		if (buildflags & E2_OPTION_FLAG_BUILDKEYS)
//CHECKME ok to re-register command-line key bindings, if bars re-created ?
			e2_keybinding_register_all ();

		if (buildflags & E2_OPTION_FLAG_BUILDLISTS)
		{
			e2_fileview_set_font ();	//FIXME do this only if needed
			e2_filelist_refresh_list (curr_view);
			e2_filelist_refresh_list (other_view);
		}
	}

	//aliases
	if ((buildflags & E2_OPTION_FLAG_BUILDALIAS)
		&& e2_option_bool_get ("command-use-aliases"))
			e2_alias_sync (&app.aliases);

	//filetypes
	if (buildflags & E2_OPTION_FLAG_BUILDFILES)
		e2_filetype_apply_allnew ();

	if (!apply)	//if this is an 'ok' button response
	{
		//update the stored config data
		e2_option_file_write ();
		//update stamp, to prevent auto-refresh
		e2_fs_touch_config_dir ();
		//may need to show the revised config file data
		//FIXME this may cause crash
//		if (!(buildflags & (E2_OPTION_FLAG_BUILDALL | E2_OPTION_FLAG_BUILDPANES)))
//			e2_filelist_check_dirty (NULL);
	}

	//after config file update, set config monitoring according to current option
	e2_option_enable_config_checks ();
	//set filelist monitoring according to current option
	if (e2_option_bool_get ("auto-refresh"))
		cfg.auto_refresh_id = g_timeout_add 
			(E2_FILESCHECK_INTERVAL, (GSourceFunc) e2_filelist_check_dirty, NULL);

	return TRUE;
}
/**
 * @brief
 *
 * @param button
 *
 * @return TRUE always
 **/
static gboolean _e2_confdlg_cancel_cb (GtkWidget *button)
{
	if (config_dialog != NULL)
	{
		// revert all option trees
		e2_option_tree_restore_all ();

		_e2_confdlg_clean ();
		gtk_widget_grab_focus (curr_view->treeview);
	}
	return TRUE;
}
/**
 * @brief
 *
 * @param button
 * @param response
 * @param rt
 *
 * @return 
 **/
static void _e2_confdlg_response_cb (GtkWidget *button,
	gint response, E2_ConfigDialogRuntime *rt)
{
	switch (response)
	{
		case GTK_RESPONSE_APPLY:
			_e2_confdlg_ok_cb (button, TRUE);
			break;
		case GTK_RESPONSE_OK:
			_e2_confdlg_ok_cb (button, FALSE);
			g_free (rt);
			break;
		case E2_RESPONSE_USER1:	//revert to default options
			e2_task_configure_default (NULL);
		//delete current content ??
		//	_e2_config_dialog_content_create (rt);
			//FIXME status message
			_e2_confdlg_cancel_cb (button);
			g_free (rt);
			return e2_config_dialog_create (NULL);
			break;
		case E2_RESPONSE_USER2:	//toggle between advanced & basic
			//same response for both buttons - only 1 is displayed
			e2_option_bool_toggle ("advanced-config");
			_e2_confdlg_clean ();
			g_free (rt);
			e2_config_dialog_create (NULL);
			break;
		default:
			_e2_confdlg_cancel_cb (button);
			g_free (rt);
			break;
	}
	return;
}
/**
 * @brief
 *
 * @param widget
 * @param event
 * @param data
 *
 * @return 
 **/
gboolean e2_confdlg_key_press_cb (GtkWidget *widget,
	GdkEventKey *event, gpointer data)
{
//	if (event->keyval == GDK_Return)  return activates any focused button
//		return e2_confdlg_ok_cb (NULL, FALSE);
	if (event->keyval == GDK_Escape)
		return _e2_confdlg_cancel_cb (NULL);
	return FALSE;
}
/**
 * @brief
 *
 * @param treeview
 * @param event
 * @param data
 *
 * @return 
 **/
static gboolean _e2_confdlg_clicked_cb (GtkWidget *treeview,
	GdkEventButton *event, gpointer data)
{
	if (event->button == 3)
	{
		GtkWidget *menu = gtk_menu_new ();
		e2_menu_add (menu, _("_Expand"), GTK_STOCK_ZOOM_IN, 
			_("Expand all rows"), e2_tree_expand_all_cb, treeview);
		e2_menu_add (menu, _("C_ollapse"), GTK_STOCK_ZOOM_OUT, 
			_("Collapse all rows"), e2_tree_collapse_all_cb, treeview);
		e2_menu_popup (menu, event->button, event->time);
		return TRUE;
	}
	return FALSE;
}

  /**************************/
 /***** font callbacks *****/
/**************************/

/**
 * @brief
 *
 * @param button
 * @param entry
 *
 * @return 
 **/
static void _e2_confdlg_font_select_ok_cb (GtkWidget *button, GtkWidget *entry)
{
	//FIXME: very bad
	GtkWidget *dialog = button->parent->parent->parent;
	const gchar *title = gtk_window_get_title (&(GTK_DIALOG (dialog)->window));
	gchar *font_str = gtk_font_selection_dialog_get_font_name
		(GTK_FONT_SELECTION_DIALOG (dialog));
	gtk_entry_set_text (GTK_ENTRY (entry), font_str);
	g_free (font_str);
	g_object_set_data (G_OBJECT (config_dialog), title, NULL);
}
/**
 * @brief
 *
 * @param button
 * @param dialog
 *
 * @return 
 **/
static void _e2_confdlg_font_select_cancel_cb (GtkWidget *button, GtkWidget *dialog)
{
	const gchar *title = gtk_window_get_title (&(GTK_DIALOG (dialog)->window));
	g_object_set_data (G_OBJECT (config_dialog), title, NULL);
}
/**
 * @brief
 *
 * @param button
 * @param set
 *
 * @return 
 **/
static void _e2_confdlg_font_select_cb (GtkWidget *button, E2_OptionSet *set)
{
	gchar *title = g_strdup_printf (_("Choose font: %s"), set->desc);
	GtkFontSelectionDialog *dialog = g_object_get_data (G_OBJECT (config_dialog), title);
	if (dialog != NULL)
		gtk_window_present (GTK_WINDOW (dialog));
	else
	{
		dialog = GTK_FONT_SELECTION_DIALOG (gtk_font_selection_dialog_new (title));
		gtk_widget_destroy (dialog->apply_button);
		gtk_font_selection_dialog_set_font_name (dialog,
			gtk_entry_get_text (GTK_ENTRY (set->widget)));
		g_signal_connect (G_OBJECT (dialog->ok_button), "clicked",
			G_CALLBACK (_e2_confdlg_font_select_ok_cb), set->widget);
		g_signal_connect (G_OBJECT (dialog->cancel_button), "clicked",
			G_CALLBACK (_e2_confdlg_font_select_cancel_cb), dialog);
		g_signal_connect (G_OBJECT (dialog), "delete_event",
			G_CALLBACK (_e2_confdlg_font_select_cancel_cb), NULL);
		g_object_set_data_full (G_OBJECT (config_dialog), title, dialog, (void *) gtk_widget_destroy);
		e2_dialog_show (GTK_WIDGET (dialog), config_dialog, 0, NULL);
	}
	g_free (title);
}
/**
 * @brief
 *
 * @param entry
 * @param label
 *
 * @return 
 **/
static void _e2_confdlg_font_entry_cb (GtkEditable *entry, GtkWidget *label)
{
	gchar *font_str = gtk_editable_get_chars (entry, 0, -1);
	if ((font_str != NULL) && (*font_str != '\0') && (strlen (font_str) > 2))
	{
		gchar *label_text = g_strdup_printf ("\t<small><i>%s</i></small> <span font_desc=\"%s\">%s</span>\n", 
			_("example:"), font_str, _("abcd efgh ABCD EFGH"));
		gtk_label_set_markup (GTK_LABEL (label), label_text);
		g_free (label_text);
	}
	g_free (font_str);
}

  /**************************/
 /***** color callbacks *****/
/**************************/

/**
 * @brief
 *
 * @param button
 * @param entry
 *
 * @return 
 **/
static void _e2_confdlg_color_select_ok_cb (GtkWidget *button, GtkWidget *entry)
{
	GtkWidget *dialog = button->parent->parent->parent;
	const gchar *title = gtk_window_get_title (&(GTK_DIALOG (dialog)->window));
	GdkColor color;
	gtk_color_selection_get_current_color (GTK_COLOR_SELECTION 
		(GTK_COLOR_SELECTION_DIALOG (dialog)->colorsel), &color);
	gchar *color_str = e2_utils_color2str (&color);
	gtk_entry_set_text (GTK_ENTRY (entry), color_str);
	g_free (color_str);
	g_object_set_data (G_OBJECT (config_dialog), title, NULL);
}
/**
 * @brief
 *
 * @param button
 * @param dialog
 *
 * @return 
 **/
static void _e2_confdlg_color_select_cancel_cb (GtkWidget *button, GtkWidget *dialog)
{
	const gchar *title = gtk_window_get_title (&(GTK_DIALOG (dialog)->window));
	g_object_set_data (G_OBJECT (config_dialog), title, NULL);
}
/**
 * @brief
 *
 * @param button
 * @param set
 *
 * @return 
 **/
static void _e2_confdlg_color_select_cb (GtkWidget *button, E2_OptionSet *set)
{
	gchar *title = g_strdup_printf (_("Choose color: %s"), set->desc);
	GtkColorSelectionDialog *dialog = g_object_get_data (G_OBJECT (config_dialog), title);
	if (dialog != NULL)
		gtk_window_present (GTK_WINDOW (dialog));
	else
	{
		dialog = GTK_COLOR_SELECTION_DIALOG (gtk_color_selection_dialog_new (title));
		gtk_widget_destroy (GTK_COLOR_SELECTION_DIALOG (dialog)->help_button);

		GdkColor color;
		gchar *color_str = gtk_editable_get_chars (GTK_EDITABLE (set->widget), 0, -1);
		gdk_color_parse (color_str, &color);
		gtk_color_selection_set_current_color (GTK_COLOR_SELECTION (dialog->colorsel), &color);
		g_free (color_str);
		g_signal_connect (G_OBJECT (dialog->ok_button), "clicked",
			G_CALLBACK (_e2_confdlg_color_select_ok_cb), set->widget);
		g_signal_connect (G_OBJECT (dialog->cancel_button), "clicked",
			G_CALLBACK (_e2_confdlg_color_select_cancel_cb), dialog);
		g_signal_connect (G_OBJECT (dialog),
			"delete_event", G_CALLBACK (_e2_confdlg_color_select_cancel_cb), dialog);
		g_object_set_data_full (G_OBJECT (config_dialog), title, dialog,
			(void *) gtk_widget_destroy);
		gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);
		e2_dialog_show (GTK_WIDGET (dialog), config_dialog, 0, NULL);
	}
	g_free (title);
}
/**
 * @brief
 *
 * @param entry
 * @param label
 *
 * @return 
 **/
static void _e2_confdlg_color_entry_cb (GtkEditable *entry, GtkWidget *label)
{
	gchar *color_str = gtk_editable_get_chars (entry, 0, -1);
	GdkColor color;
	if ((gdk_color_parse (color_str, &color)) == TRUE)
	{
		gchar *label_text = g_strdup_printf ("\t"
		"<small><i>%s </i></small><span foreground=\"%s\" background=\"%s\">%s</span>"
		" <span foreground=\"%s\">%s</span><small>  |  <i>%s</i></small> %s\n", 
		_("example:"), color_str, color_str, _("abCD"), color_str, _("abCD"), _("current color:"), color_str );
		gtk_label_set_markup (GTK_LABEL (label), label_text);
		g_free (label_text);
	}
	g_free (color_str);
}
/**
 * @brief
 *
 * @param selection
 * @param notebook
 *
 * @return 
 **/
static gboolean _e2_confdlg_category_selected_cb 
	(GtkTreeSelection *selection, GtkNotebook *notebook)
{
	printd (DEBUG, "callback: category selected");
	GtkTreeIter iter;
	GtkTreeModel *model;
    gint page_num;

	if (gtk_tree_selection_get_selected (selection, &model, &iter))
		gtk_tree_model_get (model, &iter, 1, &page_num, -1);

	if (page_num >= 0)
	{
		page_last = page_num;
		gtk_notebook_set_current_page (notebook, page_num);
		gtk_widget_grab_focus (treeview);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief
 *
 * @param parent_iter
 * @param rt
 * @param label_text
 *
 * @return 
 **/
static GtkWidget *_e2_confdlg_add_page (GtkTreeIter *parent_iter, 
	E2_ConfigDialogRuntime *rt, gchar *label_text)
{
	//add page to the treeview model
	GtkTreeIter child_iter;
	gtk_tree_store_append (rt->store, &child_iter, parent_iter);
	gint page_num = gtk_notebook_get_n_pages (rt->notebook);
	gtk_tree_store_set (rt->store, &child_iter, 0, label_text,
		1, page_num, -1);

	//is this one the one nominated to open at startup?
	if (page_selected != NULL
		&& g_str_equal (page_selected, label_text))
			page_last = page_num;

	//remember where the selected row will be
	if (page_last == page_num)
	{
		GtkTreePath *path = gtk_tree_model_get_path
			(GTK_TREE_MODEL (rt->store), &child_iter);
		gtk_tree_row_reference_free (page_last_ref); //ok to free a NULL ref
		page_last_ref = gtk_tree_row_reference_new
			(GTK_TREE_MODEL (rt->store), path);
		gtk_tree_path_free (path);
	}

	//page vbox containing the headline and config widgets
	GtkWidget *vbox = gtk_vbox_new (FALSE, E2_PADDING);
	g_object_set_data_full (G_OBJECT (vbox), "iter", gtk_tree_iter_copy (&child_iter),
		(void *) gtk_tree_iter_free);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), E2_PADDING);

	//page headline
	GtkWidget *frame = e2_widget_add_frame (vbox, FALSE, 0, NULL, FALSE);
	gchar *label_text2 = g_strconcat ("<b>", label_text, "</b>", NULL);
	GtkWidget *label = e2_widget_add_label (NULL, label_text2, 0.5, FALSE, 0);
	g_free (label_text2);
	gtk_misc_set_padding (GTK_MISC (label), E2_PADDING_XSMALL, E2_PADDING_XSMALL);
	gtk_container_add (GTK_CONTAINER (frame), label);
	gtk_notebook_append_page (rt->notebook, vbox, NULL);

	return vbox;
}
/**
 * @brief
 *
 * @param group
 * @param iter
 * @param rt
 *
 * @return 
 **/
static GtkWidget *_e2_confdlg_get_page (gchar *group, 
	GtkTreeIter *iter, E2_ConfigDialogRuntime *rt)
{
	GtkWidget *box;
	if ((box = (GtkWidget *) g_hash_table_lookup (rt->opthash, group)) == NULL)
	{
		group = g_strdup (group);
		gchar *frame = strrchr (group, ':');	//always ascii :, don't need g_utf8_strrchr()
		if (frame != NULL)
		{
			*frame = '\0';
			GtkWidget *frame_box;
			if ((frame_box = (GtkWidget *) g_hash_table_lookup (rt->opthash, group)) == NULL)
				frame_box = _e2_confdlg_get_page (group, iter, rt);
			gchar *frame_title = e2_utils_str_stretch (frame + 1);
			GtkWidget *frame_widget = gtk_frame_new (frame_title);
			g_free (frame_title);
			GtkWidget *vbox = gtk_vbox_new (FALSE, 0);
			gtk_container_add (GTK_CONTAINER (frame_widget), vbox);
			gtk_container_set_border_width (GTK_CONTAINER (vbox), E2_PADDING);
			gtk_box_pack_start (GTK_BOX (frame_box), frame_widget, FALSE, FALSE, 0);
			box = vbox;
			*frame = ':';
		}
		else
		{
			gchar *child = strrchr (group, '.');	//always ascii '.', don't need g_utf8_strrchr()
			if (child != NULL)
			{
				*child = '\0';
				box = _e2_confdlg_get_page (group, iter, rt);
				GtkTreeIter *iter2;
				iter2 = g_object_get_data (G_OBJECT (box), "iter");
				if (iter2)
					box = _e2_confdlg_add_page (iter2, rt, child + 1);
				*child = '.';
			} else 
				box = _e2_confdlg_add_page (iter, rt, group);
		}
		g_hash_table_insert (rt->opthash, group, box);
	}
	return box;
}
/**
 * @brief create and show configuration dialog
 *
 * @param page name of page to show when dialog is shown
 *
 * @return 
 **/
void e2_config_dialog_create (gchar *page)
{
	printd (DEBUG, "create config dialog (%s)", page);
	//check if there is already a config dialog opened
	if (config_dialog != NULL)
	{
		gtk_window_present (GTK_WINDOW (config_dialog));
		return;
	}
	//note which page we want to open at - with NULL for none
	page_selected = (page != NULL && *page == '\0') ? NULL : page;

	//receptacle for cut/copied row(s) data generated from context menu
	tree_view_buffer_hash = g_hash_table_new_full (g_str_hash, g_str_equal, 
		g_free, (GDestroyNotify) e2_option_tree_menu_hash_clean);
	E2_ConfigDialogRuntime *rt = g_malloc (sizeof (E2_ConfigDialogRuntime));

	config_dialog = e2_dialog_create (NULL, NULL, _("configuration"), _e2_confdlg_response_cb, rt);

	e2_option_connect (config_dialog, FALSE);

	//the main hpane - contains the categories treeview and the frame
	GtkWidget *hpane = gtk_hpaned_new ();
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (config_dialog)->vbox), hpane);
	
	//scrolled window for the category treeview
	GtkWidget *sw = e2_widget_get_sw (GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_widget_set_size_request (sw, 200, 0);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw), GTK_SHADOW_IN);
	gtk_paned_pack1 (GTK_PANED (hpane), sw, FALSE, TRUE);

	//categories treestore has single node, 
	//2 columns = (displayed) label and (hidden) corresponding enumerator
	rt->store = gtk_tree_store_new (2, G_TYPE_STRING, G_TYPE_INT);

	//categories tree view
	treeview = gtk_tree_view_new_with_model (GTK_TREE_MODEL (rt->store));
	g_object_unref (rt->store);
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (treeview), TRUE);
	GtkTreeSelection *selection = gtk_tree_view_get_selection (GTK_TREE_VIEW
		(treeview));
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);

	GtkCellRenderer *renderer = gtk_cell_renderer_text_new ();
	g_object_set (renderer, "xalign", 0.0, NULL);

	gint col_offset = gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW
		(treeview), -1, _("Categories"), renderer, "text", 0, NULL);

	GtkTreeViewColumn *column = gtk_tree_view_get_column (GTK_TREE_VIEW
		(treeview), col_offset - 1);
	gtk_tree_view_column_set_clickable (GTK_TREE_VIEW_COLUMN (column), FALSE);
	gtk_container_add (GTK_CONTAINER (sw), treeview);

	//the notebook
	rt->notebook = GTK_NOTEBOOK (gtk_notebook_new ());
	gtk_notebook_set_show_tabs (rt->notebook, FALSE);
	gtk_notebook_set_show_border (rt->notebook, FALSE);
	sw = e2_widget_get_sw (GTK_POLICY_NEVER, GTK_POLICY_NEVER);
	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (sw),
		GTK_WIDGET (rt->notebook));
	gtk_viewport_set_shadow_type (GTK_VIEWPORT (GTK_BIN (sw)->child),
		GTK_SHADOW_NONE);
	gtk_paned_pack2 (GTK_PANED (hpane), sw, TRUE, TRUE);

	//iterate through the options, setting up category tree and page contents
//	gboolean basic = ! e2_option_bool_get ("advanced-config");
	E2_OptionFlags includemask = (e2_option_bool_get ("advanced-config")) ? 
	E2_OPTION_FLAG_BASIC | E2_OPTION_FLAG_ADVANCED :
	E2_OPTION_FLAG_BASIC | E2_OPTION_FLAG_BASICONLY ;
	
	rt->opthash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
	gint i;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
//		if ((basic && (set->flags & E2_OPTION_FLAG_ADVANCED))
//			|| set->flags & E2_OPTION_FLAG_HIDDEN)
		if (!(set->flags & includemask))
		{
			set->widget = NULL;	//make sure that this is ignored when updating after 'ok' or 'apply'
			continue;
		}
		//make sure there's a size group for the page
		gchar *size_group_key = g_strconcat (set->group, "_sizegroup", NULL);	//no translation
		GtkSizeGroup *size_group = (GtkSizeGroup *) g_hash_table_lookup (rt->opthash, size_group_key);
		if (GTK_IS_SIZE_GROUP (size_group))
			g_free (size_group_key);
		else
		{
			size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
			g_hash_table_insert (rt->opthash, size_group_key, size_group);
		}

		GtkWidget *box = _e2_confdlg_get_page (set->group, NULL, rt);

		GtkWidget *hbox, *label;
		gchar *label_text;
		switch (set->type)
		{
			case E2_OPTION_TYPE_TREE:
				e2_option_tree_add_widget (config_dialog, box, set);
				break;
			case E2_OPTION_TYPE_BOOL:
				e2_option_bool_add_dialog_widget (config_dialog, box, size_group, set);
				break;
			case E2_OPTION_TYPE_STR:
				hbox = gtk_hbox_new (FALSE, E2_PADDING);
				gtk_box_pack_start (GTK_BOX (box), hbox, TRUE, TRUE, E2_PADDING);
				label = e2_widget_add_label (hbox, set->desc, 0.0, FALSE, 0);
				gtk_size_group_add_widget (size_group, label);
				set->widget = e2_widget_add_entry (hbox, set->sval, TRUE, FALSE);
				e2_widget_set_tooltip (NULL, set->widget, set->tip);
				break;
			case E2_OPTION_TYPE_INT:
				e2_option_int_add_widget (config_dialog, box, size_group, set);
				break;
			case E2_OPTION_TYPE_SEL:
				e2_option_sel_add_widget (config_dialog, box, size_group, set);
				break;
			case E2_OPTION_TYPE_FONT:
				//hbox for option label, font entry and font dialog button
				hbox = gtk_hbox_new (FALSE, E2_PADDING);
				gtk_box_pack_start (GTK_BOX (box), hbox, TRUE, TRUE, E2_PADDING);
				//option label
				label_text = g_strconcat (set->desc, ": ", NULL);
				label = e2_widget_add_label (hbox, label_text, 0.0, FALSE, 0);
				g_free (label_text);
				gtk_size_group_add_widget (size_group, label);
				//font entry for font name
				set->widget = e2_widget_add_entry (hbox, NULL, TRUE, FALSE);
				e2_widget_set_tooltip (NULL, set->widget, set->tip);
				//button for font select dialog
				e2_button_add (hbox, FALSE, 0, _("change"), GTK_STOCK_SELECT_FONT,
					_("Click to open a font select dialog"), _e2_confdlg_font_select_cb, set);
				//label for font example
				label = e2_widget_add_label (box, "", 0.0, FALSE, 0);
				gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
				//react on changes to the font name
				g_signal_connect (GTK_EDITABLE (set->widget), "changed",
					G_CALLBACK (_e2_confdlg_font_entry_cb), label);
				//set entry content _after_ signal connect so the example label
				//gets some contents - it's empty at the beginning
				gtk_entry_set_text (GTK_ENTRY (set->widget), set->sval);
				break;
			case E2_OPTION_TYPE_COLOR:
				//hbox for option label, color entry and color dialog button
				hbox = gtk_hbox_new (FALSE, E2_PADDING);
				gtk_box_pack_start (GTK_BOX (box), hbox, TRUE, TRUE, E2_PADDING);
				//option label
				label_text = g_strconcat (set->desc, ": ", NULL);
				label = e2_widget_add_label (hbox, label_text, 0.0, FALSE, 0);
				g_free (label_text);
				gtk_size_group_add_widget (size_group, label);
				//color entry for color names/color hex values
				set->widget = e2_widget_add_entry (hbox, NULL, TRUE, FALSE);
				e2_widget_set_tooltip (NULL, set->widget, set->tip);
				//button for color select dialog
				e2_button_add (hbox, FALSE, 0, _("change"), GTK_STOCK_SELECT_COLOR,
					_("click to open a color selection dialog"), _e2_confdlg_color_select_cb, set);
				//label for color example
				label = e2_widget_add_label (box, "", 0.0, FALSE, 0);
				gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
				//react on changes to the color name/color hex value
				g_signal_connect (GTK_EDITABLE (set->widget), "changed",
					G_CALLBACK (_e2_confdlg_color_entry_cb), label);
				//set entry after signal connect so the example label gets
				//some contents, it's empty at the beginning
				gtk_entry_set_text (GTK_ENTRY (set->widget), set->sval);
				break;
			default:
				break;
		}
	}
	//a set's dependency (if any) may be 'forward' in the array, in which
	//case the relevant widget not known when the set is processed, so
	//need to setup dependencies after all options are setup i.e. iterate again ...
	//FIXME in pass 1, log the items with dependencies, then just do those
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
//		if (basic && (set->flags & E2_OPTION_FLAG_ADVANCED))
		if (!(set->flags & includemask))
			continue;
		if (set->depends != NULL)
		{
			if (set->depends[0] == '!')
			{	//negated dependency
				E2_OptionSet *dep = e2_option_get (set->depends + 1);
				if ((dep != NULL) && (dep->type == E2_OPTION_TYPE_BOOL))
				{
					if (dep->widget != NULL)	//dep is in this dialog
						g_signal_connect (G_OBJECT (dep->widget), "toggled",
							G_CALLBACK (_e2_confdlg_depends_cb), set->widget);
					gtk_widget_set_sensitive (set->widget,
						! _e2_option_bool_get (dep));	//if dep is not in the dialog, it will remain in its current state
				}
			}
			else	//not negated
			{
				E2_OptionSet *dep = e2_option_get (set->depends);
				if ((dep != NULL) && (dep->type == E2_OPTION_TYPE_BOOL))
				{
					if (dep->widget != NULL)	//dep is in this dialog
						g_signal_connect (G_OBJECT (dep->widget), "toggled",
							G_CALLBACK (_e2_confdlg_depends_cb_), set->widget);
					gtk_widget_set_sensitive (set->widget,
						_e2_option_bool_get (dep));
				}
				else	//always insensitive if dep not boolean
					if (dep->type != E2_OPTION_TYPE_BOOL)
				{
					gtk_widget_set_sensitive (set->widget, FALSE);
				}
				//ignore instruction if dep set  doesn't exist
			}
		}
	}
	//clean up
	g_hash_table_destroy (rt->opthash);

	//expand the category tree view
	gtk_tree_view_expand_all (GTK_TREE_VIEW (treeview));

	//get the path to the startup row
	GtkTreePath *path = gtk_tree_row_reference_get_path (page_last_ref);
	if (path != NULL)
	{
//	no initial selection, just set cursor position
//	note - selection cb not connected yet
//		gtk_tree_selection_select_path (selection, path);
		//make sure the selected row is in the window
		if (page_last > 0)
		{
			GtkTreeViewColumn *col = gtk_tree_view_get_column
				(GTK_TREE_VIEW (treeview), 0);
			gtk_tree_view_set_cursor (GTK_TREE_VIEW (treeview), path, col, FALSE);
			gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (treeview), path, col, TRUE, 0.5, 0.5);
		}
	}

//	gtk_window_set_default_size -- does not work ... 
	gtk_window_resize (GTK_WINDOW(config_dialog), cfg.window_width,
		cfg.window_height);	//revert to last size if not too small (initially -1,-1 = default)

	g_signal_connect (G_OBJECT (config_dialog), "delete_event",
		G_CALLBACK (_e2_confdlg_cancel_cb), NULL);
	g_signal_connect_after (G_OBJECT (config_dialog), "key_press_event",
		G_CALLBACK (e2_confdlg_key_press_cb), NULL);
	g_signal_connect (G_OBJECT (treeview), "key_press_event",
		G_CALLBACK (e2_confdlg_key_press_cb), NULL);
	g_signal_connect (G_OBJECT (treeview), "button_press_event",
		G_CALLBACK (_e2_confdlg_clicked_cb), NULL);
	g_signal_connect (G_OBJECT (selection), "changed",
		G_CALLBACK (_e2_confdlg_category_selected_cb), rt->notebook);
	//make the notebook open at the correct page
	//(can't get it to work, if set sooner than in the 'show' cb ...)
	g_signal_connect (G_OBJECT (config_dialog), "show",
		G_CALLBACK (_e2_confdlg_show_cb2), rt->notebook);
	g_signal_connect (G_OBJECT (config_dialog), "show",
		G_CALLBACK (_e2_confdlg_show_cb), sw);

	e2_dialog_add_undefined_button_custom (config_dialog, FALSE, 
		E2_RESPONSE_USER1, _("_Default"), GTK_STOCK_CLEAR, _("Revert all options to their default settings"),
		NULL, NULL);
	if (e2_option_bool_get ("advanced-config"))
		e2_dialog_add_undefined_button_custom (config_dialog, FALSE, 
			E2_RESPONSE_USER2, _("_Basic"), GTK_STOCK_REDO, _("Display only the basic configuration options"),
			NULL, NULL);
	else
		e2_dialog_add_undefined_button_custom (config_dialog, FALSE, 
			E2_RESPONSE_USER2, _("Ad_vanced"), GTK_STOCK_REDO, _("Display all configuration options"),
			NULL, NULL);

	E2_BUTTON_CANCEL.showflags |= E2_BTN_DEFAULT;
	E2_BUTTON_OK.showflags &= !E2_BTN_DEFAULT;
	e2_dialog_show (config_dialog, app.main_window, 0,
		&E2_BUTTON_CANCEL, &E2_BUTTON_APPLY, &E2_BUTTON_OK, NULL);
}
