/* $Id: e2_filelist.c 1128 2005-09-14 02:15:03Z tpgww $
 *
 * Portions copyright (C) 2004-2005 tooar <tooar@gmx.net>.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/e2_filelist.c 
 * @brief directory content liststore functions
 *
 * This file contains functions related to creation, filling and emptying a
 * liststore of directory content data, and fns for the associated model.
 * Also filtering and sorting of store content, and other related utilties.
 **/

#include "e2_filelist.h"
#include <string.h>
#include <sys/time.h>
#include <langinfo.h>
#include <dirent.h>
#include <pwd.h>
#include <time.h>
#include <grp.h>
#include "e2_option.h"
#include "e2_task.h"
#include "e2_dialog.h"

typedef struct _DirHistoryEntry
{
	gchar path[PATH_MAX];
	gint row;
	gboolean selected;  //TRUE if row was selected
} DirHistoryEntry;

#ifdef E2_SELTXT_RECOLOR
extern GdkColor selectedtext;
#endif

//disable-refresh counter
gint refresh_ref;
//static gboolean case_sensitive; //local copy of option value, for faster access
struct stat statfast;	//statbuf used to speed sorting

static gint _e2_filelist_name_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_filelist_size_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_filelist_perm_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_filelist_user_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_filelist_group_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_filelist_mdate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_filelist_adate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_filelist_cdate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);

static void _e2_filelist_store_empty (FileView *view);
static void _e2_filelist_store_fill (FileView *view);
static void _e2_filelist_sortNselect (FileView *view);

E2_Column e2_all_columns[MAX_COLUMNS] = {
	{N_("Filename"), 120, _e2_filelist_name_sort},
	{N_("Size"), 80, 	_e2_filelist_size_sort},
	{N_("Permissions"), 80, _e2_filelist_perm_sort},
	{N_("Owner"), 60, _e2_filelist_user_sort},
	{N_("Group"), 60, _e2_filelist_group_sort},
	{N_("Modified"), 120, _e2_filelist_mdate_sort},
	{N_("Accessed"), 120, _e2_filelist_adate_sort},
	{N_("Changed"), 120, _e2_filelist_cdate_sort} };

/**
 * @brief filename sort-order comparison function
 *
 * This compares 'sort-key' row data, which is utf-8 compatible
 * If the case-insensitive sorting option was in force when the data 
 * was loaded into the model, the keys are 'case-folded'
 * ie this is not so dynamic as we might want
 * Directories are placed before other things.
 * NB utf8 collate keys for ascii text are case insensitive !!
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_filelist_name_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	gchar *keya; // = data;
	gchar *keyb; // = data + E2_UTFNAME_MAX;  //this is the size set in e2_fileview_column_header_clicked_cb() etc
	FileInfo *infoa, *infob;
	gint result;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, NAMEKEY, &keya, FINFO, &infoa, -1);
	gtk_tree_model_get (model, b, NAMEKEY, &keyb, FINFO, &infob, -1);

	if (g_str_equal (infoa->filename, ".."))
		result = (order) ? -1 : 1;
	else
		if (g_str_equal (infob->filename, ".."))
			result = (order) ? 1 : -1;
	else
		if (e2_fs_is_dir_fast (infoa, &statfast))
		{
			if (e2_fs_is_dir_fast (infob, &statfast))
				result = strcmp (keya, keyb);
			else
		  		result = (order) ? -1 : 1;
		}
	else
		if (e2_fs_is_dir_fast (infob, &statfast))
			result = (order) ? 1 : -1;
	else
		result = strcmp (keya, keyb);

	g_free (keya);
	g_free (keyb);

	return result;
}
/**
 * @brief filesize sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_filelist_size_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_size - infob->statbuf.st_size);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;
	
	return (infoa->statbuf.st_size - infob->statbuf.st_size);
}
/**
 * @brief file permission sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_filelist_perm_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;
	
	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);

	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_mode - infob->statbuf.st_mode);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;
	
	return (infoa->statbuf.st_mode - infob->statbuf.st_mode);
}
/**
 * @brief file owner sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_filelist_user_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_uid - infob->statbuf.st_uid);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_uid - infob->statbuf.st_uid);
}
/**
 * @brief file group sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_filelist_group_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_gid - infob->statbuf.st_gid);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_gid - infob->statbuf.st_gid);
}
/**
 * @brief file acess date sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_filelist_adate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_atime - infob->statbuf.st_atime);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_atime - infob->statbuf.st_atime);
}
/**
 * @brief file modification date sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_filelist_mdate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_mtime - infob->statbuf.st_mtime);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_mtime - infob->statbuf.st_mtime);
}

/**
 * @brief file status-chage date sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_filelist_cdate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_ctime - infob->statbuf.st_ctime);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_ctime - infob->statbuf.st_ctime);
}

// some filter menu callbacks (others are in filter-dialog files)

void e2_filelist_filter_toggle (FileView *view)
{
	gboolean filter_on = view->filename_filter.active 
		|| view->size_filter.active
		|| view->date_filter.active;
	gchar *name = (view == &app.pane1_view) ? _A(9) : _A(10);
	name = g_strconcat (name, ".", _A(22), NULL);
	gboolean filter_button = e2_toolbar_toggle_button_get_state (name);
	//filter_button = TRUE is 'normal' un-filtered state
	if (filter_on && filter_button)
		e2_toolbar_toggle_button_set_state (name, FALSE);
	else if (!filter_on && !filter_button)
		e2_toolbar_toggle_button_set_state (name, TRUE);
	g_free (name);
}

void e2_filelist_filter_dirs_cb (GtkWidget *widget, FileView *view)
{  //toggle cached flag to match the menu check button
//	view->filter_directories = gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (widget));
	view->filter_directories = !view->filter_directories;
	e2_filelist_refresh_list (view);
}

void e2_filelist_remove_filters_cb (GtkWidget *widget, FileView *view)
{
/* menu items now re-created for each popup
	//turn off the check menu items
	gulong handler = g_signal_handler_find (view->check_name, G_SIGNAL_MATCH_FUNC, 
		0, 0, NULL, e2_name_filter_dialog_create_cb, NULL);
	g_signal_handler_block (view->check_name, handler);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view->check_name), FALSE);
	g_signal_handler_unblock (view->check_name, handler);

	handler = g_signal_handler_find (view->check_size, G_SIGNAL_MATCH_FUNC, 
		0, 0, NULL, e2_size_filter_dialog_create_cb, NULL);
	g_signal_handler_block (view->check_size, handler);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view->check_size), FALSE);
	g_signal_handler_unblock (view->check_size, handler);

	handler = g_signal_handler_find (view->check_date, G_SIGNAL_MATCH_FUNC, 
		0, 0, NULL, e2_date_filter_dialog_create_cb, NULL);
	g_signal_handler_block (view->check_date, handler);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view->check_date), FALSE);
	g_signal_handler_unblock (view->check_date, handler);

 	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view->check_dirs), FALSE);
*/
	// turn off local flags
	view->filename_filter.active = FALSE;
	view->size_filter.active = FALSE;
	view->date_filter.active = FALSE;
	view->filter_directories = FALSE;
	//show the results
	e2_filelist_filter_toggle (view);
	e2_filelist_refresh_list (view);
}
/**
 * @brief initialise file filter flags to show everything
 *
 * @param view data structure for view being processed
 *
 * @return
 */
void e2_filelist_initialize_filters (FileView *view)
{
	view->filter_directories = FALSE;

	view->filename_filter.active = FALSE;
	g_strlcpy (view->filename_filter.pattern, "*",
		      sizeof(view->filename_filter.pattern));
	view->filename_filter.case_sensitive = TRUE;

	view->size_filter.active = FALSE;
	view->size_filter.size = 0;
	view->size_filter.op = GT;

	view->date_filter.active = FALSE;
	view->date_filter.time = time(NULL);
	view->date_filter.time_type = ATIME;
	view->date_filter.op = GT;
}
/**
 * @brief decide whether an item should be included in the filelist, on the basis of its name
 *
 * Item name is compared with current filters
 * Item name string is localised
 * though this fn assumes ascii = FIXME
 *
 * @param view data structure for view being processed
 * @param info data structure for item being checked
 *
 * @return TRUE if item 'passes' (i.e. to be included in file list)
 */
static gboolean _e2_filelist_match_filename_filter (FileView *view, FileInfo *info)
{
	gchar *s, *pattern, *free_this, *dup, *utf;
	gchar save;
	gboolean result;

	if (e2_fs_is_dir (info) && !view->filter_directories)
		 return TRUE;

	free_this = pattern = g_strdup (view->filename_filter.pattern);
	if (!view->filename_filter.case_sensitive)
	{
		pattern = g_utf8_strdown (pattern, -1);
		g_free (free_this);
		free_this = pattern;
	}
	F_DECLARE;
	utf = F_FILENAME_FROM_LOCALE (info->filename);
	//maybe several patterns, separated by commas
	while ((s = strchr (pattern, ',')) != NULL)	//if always ascii ',', don't need g_utf8_strchr()
	{	//check each pattern that is followed by a ','
		save = *s;
		*s = '\0';
		if (!view->filename_filter.case_sensitive)
		{
		  dup = g_utf8_strdown (utf, -1);
		  if (g_pattern_match_simple (pattern, dup))
		  {
		    g_free (free_this);
		    g_free (dup);
			F_FREE (utf);
		    return !view->filename_filter.invert_mask;
		  }
		  g_free(dup); 
		}
		else if (g_pattern_match_simple (pattern, utf))
		{
		  g_free(free_this);
		  F_FREE (utf);
		  return !view->filename_filter.invert_mask;
		}
		*s = save;	//in case the matched char is bigger than 1 byte
		NCHR(s);
		pattern = s;
	}
	//check the last (or only) pattern
	if (!view->filename_filter.case_sensitive)
	{
		dup = g_utf8_strdown (utf, -1);
		result = g_pattern_match_simple (pattern, dup);
		g_free (dup);
	}
	else
		result = g_pattern_match_simple (pattern, utf);

	g_free (free_this);
	F_FREE (utf);
	return (view->filename_filter.invert_mask ? !result : result);
}
/**
 * @brief decide whether an item should be included in the filelist, on the basis of its size
 *
 * Item statbuf.st_size is compared with curent filters
 *
 * @param view data structure for view being processed
 * @param info data structure for item being checked
 *
 * @return TRUE if item does NOT need to be filtered out
 */
static gboolean _e2_filelist_match_size_filter (FileView *view, FileInfo *info)
{
	if (e2_fs_is_dir (info) && !view->filter_directories)
		 return TRUE;

	switch (view->size_filter.op)
	{
		case GT:
		  return (info->statbuf.st_size > view->size_filter.size);
		  break;
		case LT:
		  return (info->statbuf.st_size < view->size_filter.size);
		  break;
		case EQ:
		  return (info->statbuf.st_size == view->size_filter.size);
		  break;
		default:
		  return TRUE;
	}
	// this path can never happen
	return TRUE;
}
/**
 * @brief decide whether an item should be included in the filelist, on the basis of its access time
 *
 * Item statbuf.st_atime is compared with curent filters
 *
 * @param view data structure for view being processed
 * @param info data structure for item being checked
 *
 * @return TRUE if item does NOT need to be filtered out
 */
static gboolean _e2_filelist_match_date_filter (FileView *view, FileInfo *info)
{
	if (e2_fs_is_dir (info) && !view->filter_directories)
		 return TRUE;

	switch (view->date_filter.time_type)
	{
		case ATIME:
		  if (view->date_filter.op == GT)
		    return (difftime(view->date_filter.time, info->statbuf.st_atime) < 0);
		  else
		    return (difftime(view->date_filter.time, info->statbuf.st_atime) > 0);
		  break;
		case MTIME:
		  if (view->date_filter.op == GT)
		    return (difftime(view->date_filter.time, info->statbuf.st_mtime) < 0);
		  else
		    return (difftime(view->date_filter.time, info->statbuf.st_mtime) > 0);
		  break;
		case CTIME:
		  if (view->date_filter.op == GT)
		    return (difftime(view->date_filter.time, info->statbuf.st_ctime) < 0);
		  else
		    return (difftime(view->date_filter.time, info->statbuf.st_ctime) > 0);
		  break;
		default:
		  return TRUE;
	}
	// this path can't happen
	return TRUE;
}
/**
 * @brief reload list store for specified view and re-select items 
 *
 * This records the names of currently-selected items in the view,
 * reloads the directory contents (which clears the old selection)
 * finds the rows in the new store that match the selected names,
 * and selects them.
 * NB this no longer switches views
 *
 * @param view data structure for view being processed
 *
 * @return
 */
void e2_filelist_refresh_list (FileView *view)
{
	g_return_if_fail (view != NULL);
// if threads are in force  ...
//	if (!view->completed)
//		return;		//the view is still being processed from last time

//	gdk_threads_enter ();	//protect thread glitches
#ifdef E2_INCLIST
	//FIXME
#else
	//log the names of selected items in this view
	view->selected_names = e2_fileview_log_selected_names (view);
	//clear and refill the store, and re-select items where possible
	_e2_filelist_store_fill (view);
#endif
//	gdk_threads_leave ();
}
/* These functions must encapsulate any instructions that access the FileInfo
 * structs produced when a selection is interrogated.
 * This is because a refresh in the middle of such an operation would free
 * the FileInfo structs as a side-effect.
 */
/**
 * @brief  disable refreshing of config data and pane filelists
 *
 * Increase ref count, if == 1, stop the timer process which calls the refresh fns
 * Why is config data check bundled here ?
 *
 * @return
 */
void e2_filelist_disable_refresh (void)
{
	refresh_ref++;
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, refresh ref now = %d", refresh_ref);
#endif
	if (refresh_ref == 1)
	{
		if (e2_option_bool_get ("auto-refresh"))
			g_source_remove (cfg.auto_refresh_id);
		if (e2_option_bool_get ("auto-refresh-config"))
			g_source_remove (cfg.check_config_id);
		e2_window_disable_status_update ();
	}
}
/**
 * @brief  cause polling of config data and pane filelists at specified intervals
 *
 * Polling (re)starts after the interval 
 * reference count is back to 0
 *
 * @param interval  milliseconds between polls
 *
 * @return
 */
/*void e2_filelist_set_refresh (guint interval)
{
	e2_filelist_disable_refresh ();
	e2_filelist_enable_refresh_custom (interval);
}

void e2_filelist_set_refresh_default (void)
{
	e2_filelist_disable_refresh ();
	e2_filelist_enable_refresh_custom (E2_FILESCHECK_INTERVAL);
} */
/**
 * @brief  re-enable polling of config data and pane filelists
 *
 * Polling (re)starts after the default 1-second interval, if the 
 * reference count is back to 0
 *
 * @return
 */
void e2_filelist_enable_refresh (void)
{
	e2_filelist_enable_refresh_custom (E2_FILESCHECK_INTERVAL);
}
/**
 * @brief  re-enable polling of config data and pane filelists, with a specified interval
 *
 * Decrease ref count, if == 0, restart the timers which call the refresh fns
 *  Config timer is bundled here because the fn refreshes the panes
 *
 * @param interval milliseconds between polls
 *
 * @return
 */
void e2_filelist_enable_refresh_custom (guint interval)
{
	refresh_ref--;
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, refresh ref now = %d", refresh_ref);
#endif
	if (refresh_ref < 0)
	{
		fprintf(stderr, _("Something is going wrong with the auto-refresh. Exit quickly\n"));
		e2_window_output_show (NULL, NULL);
		status_message(
		_("Something is going wrong with the auto-refresh.\nPlease send a bug report to author.\n"));
	}

	if (refresh_ref == 0)
	{
		if (e2_option_bool_get ("auto-refresh"))
			cfg.auto_refresh_id = g_timeout_add 
				(interval, (GSourceFunc) e2_filelist_check_dirty, NULL);

		if (e2_option_bool_get ("auto-refresh-config"))
			cfg.check_config_id = g_timeout_add
				(E2_CONFIGCHECK_INTERVAL, (GSourceFunc) e2_option_check_config_files, NULL);

		e2_window_enable_status_update (-1);
	}
}
/**
 * @brief refresh filepane contents (and with FAM, config too) if the corresponding changed is detected
 *
 * Among other uses, this is the timeout function for filelists auto refresh
 * When FAM/gamin in use, it also checks for and flags (not processes)
 * config file updates, as that data comes from the same data stream.
 *
 * @param userdata data specified when the timer was initialised, always NULL, or when called directly, maybe non-NULL to trigger a pre-check delay
 *
 * @return TRUE so the timer keeps working
 */
gboolean e2_filelist_check_dirty (gpointer userdata)
{
	static gboolean p1deferred = FALSE;
	static gboolean p2deferred = FALSE;
#ifdef E2_FAM
	static gboolean cfgdeferred = FALSE;
#endif

	gboolean p1dirty, p2dirty;
#ifdef E2_FAM
	gboolean localcfgdirty;	//use a local copy so that the extern value doesn't get cleared before it's noticed
	extern gboolean cfgdirty;
#endif

#ifdef E2_FAM
	if (userdata != NULL 
		 && cfg.monitor_type != E2_MONITOR_DEFAULT)
			g_usleep (E2_FAMWAIT);	//wait for things to be noticed
#endif

	e2_fs_FAM_check_dirty (&p1dirty, &p2dirty
#ifdef E2_FAM
	, &localcfgdirty
#endif
	);

	//CHECKME process active pane first, more likely to be attending to that ?
	if (p1dirty || p1deferred)
	{
		if (!app.pane1_view.completed)
			p1deferred = TRUE;
		else
		{
			p1deferred = FALSE;
			e2_filelist_refresh_list (&app.pane1_view);
			printd (DEBUG, "pane 1 refreshed");
		}
	}

	if (p2dirty || p2deferred)
	{
		if (!app.pane2_view.completed)
			p2deferred = TRUE;
		else
		{
			p2deferred = FALSE;
			e2_filelist_refresh_list (&app.pane2_view);
			printd (DEBUG, "pane 2 refreshed");
		}
	}
#ifdef E2_FAM
	if (localcfgdirty || cfgdeferred)
	{
		if (!cfg.completed)
			cfgdeferred = TRUE;
		else
		{
			cfgdeferred = FALSE;
			if (localcfgdirty)
				cfgdirty = TRUE;	//set, but don't clear, the main flag
		}
	}
#endif
	return TRUE;
}
/**
 * @brief find matching path entry in history list
 *
 *
 *
 * @param list Glist of past path strings
 * @param path path string to be matched
 *
 * @return list entry whose data matches path, or else NULL
 */
static GList *_e2_filelist_history_find (GList *list, gchar *path)
{
	GList *tmp;
	DirHistoryEntry *entry;

	for (tmp = list; tmp != NULL; tmp = tmp->next)
	{
		entry = tmp->data;
		if (g_str_equal (entry->path, path))
		  return tmp;
	}
	return NULL;
}
/**
 * @brief get new directory contents for a view
 *
 * @param view rt data structure for the view to be changed
 * @param path utf string with new path
 *
 * @return
 */
void e2_filelist_change_dir (FileView *view, gchar *path)
{
	printd (DEBUG, "change dir: %s", path);
	DirHistoryEntry *entry;
	//update row history for current dir
	GList *outdir = _e2_filelist_history_find (view->dir_history, view->dir);
	if (outdir != NULL)	//can never be NULL ??
	{
		entry = outdir->data;
/*	this remembers the state of the current row
		GtkTreePath *path = gtk_tree_path_new_from_indices
			(view->row, -1);
		if (path != NULL)
		{
			GtkTreeSelection *sel = gtk_tree_view_get_selection
				(GTK_TREE_VIEW (view->treeview));
			entry->selected = gtk_tree_selection_path_is_selected
				(sel, path);
			gtk_tree_path_free (path);
		}
		else 
			entry->selected = FALSE
		entry->row = view->row
*/
		/*even though the sort-order and/or dir contents might
		  change before we re-visit here,
		  if there is a selection now, remember its start,
		  or else remember the row currently focused */
		GtkTreeSelection *sel = gtk_tree_view_get_selection
			(GTK_TREE_VIEW (view->treeview));
		gint selcount = gtk_tree_selection_count_selected_rows
			(sel);
		if (selcount > 0)
		{
			GtkTreeModel *model = GTK_TREE_MODEL (view->store);
			GList *selpaths = gtk_tree_selection_get_selected_rows
				(sel, &model);
			//get the 1st selected item's row
			entry->row = *gtk_tree_path_get_indices (selpaths->data);
			entry->selected = TRUE;
			g_list_foreach (selpaths, (GFunc) gtk_tree_path_free, NULL);
			g_list_free (selpaths);
		}
		else
		{	//just remember the current row
			entry->row = view->row;
			entry->selected = FALSE;
		}
	}
	//now update the local path store
	g_strlcpy (view->dir, path, sizeof (view->dir));

	//re-populate the view's data store
	view->selected_names = NULL;
	_e2_filelist_store_fill (view);
	
	//update history list
	GList *indir = _e2_filelist_history_find (view->dir_history, view->dir);
	gboolean oldflag;
	if (indir != NULL)
	{  //new dir is already in history list
		oldflag = TRUE;
		entry = indir->data;
//		gboolean swt = (view != curr_view);
//		if (swt)  WHY BOTHER ??
//			e2_pane_activate_other (); //make this pane active if not so already

		//goto, maybe select, the row we were at last time
		//(tuff luck if there's been content change or sort-order change)
		e2_fileview_focus_row (view, entry->row, entry->selected, FALSE, TRUE, FALSE);

//		if (swt)  WHY BOTHER
//			e2_pane_activate_other ();  //come back again
	}
	else
	{  //need a new entry for the history list
		oldflag = FALSE;
		entry = g_new0 (DirHistoryEntry, 1);
		//just set the path, for now - other data set when dir is departed
		g_strlcpy (entry->path, view->dir, sizeof(entry->path));
	}

	if (outdir == NULL)  //no entry for the dir we are departing (at seeion start)
	{
		//initialise the list
//		if (oldflag)
//			view->dir_history = g_list_delete_link (view->dir_history, indir);
		view->dir_history = g_list_append (view->dir_history, entry);
//		printd (DEBUG, "NO OLD DIR, dir history item %s appended", entry->path);
	}
	else
	{
//		DirHistoryEntry *dbg1 = outdir->data;
//		printd (DEBUG, "history entry for departing dir %s", dbg1->path);
		//move/put this one after prior dir, if not there already
		GList *tail = outdir->next;
		DirHistoryEntry *taildata;
		if (tail == NULL)
		{
			if (oldflag)
				view->dir_history = g_list_delete_link (view->dir_history, indir);
			view->dir_history = g_list_append (view->dir_history, entry);
//			printd (DEBUG, "dir history item %s appended as new tail", entry->path);
		}
		else
		{
			taildata = tail->data;
			if (oldflag && g_str_equal (taildata->path, entry->path))
			{	//already in place, just update the data
				//(could be a duplicate of outdir?)
				taildata->row = entry->row;
				taildata->selected = entry->selected;
//				printd (DEBUG, "existing dir history item %s updated", entry->path);
			}
			else
			{
				//kill the old tail
				gint k = g_list_position (view->dir_history, tail);
				while (tail != NULL)
				{
					taildata = tail->data;
					if (!g_str_equal (taildata->path, entry->path))
						view->dir_history = g_list_delete_link (view->dir_history, tail);
					else
						k++;	//pass this one

					tail = g_list_nth (view->dir_history, (guint) k);
				}
				//move the new tail into place
				if (oldflag)
					view->dir_history = g_list_delete_link (view->dir_history, indir);
				view->dir_history = g_list_append (view->dir_history, entry);
//				printd (DEBUG, "dir history item %s appended as replacement tail", entry->path);
			}
		}
	}
	
/*	GList *debug = view->dir_history;
	for ( ; debug != NULL ; debug=debug->next)
	{
		DirHistoryEntry *dirdata = debug->data;
		printd (DEBUG, "dir history item %s", dirdata->path);
	}
*/
	gint maxhistory = e2_option_int_get ("dir-line-history-max");
	if (g_list_length (view->dir_history) > maxhistory)
	{
//		indir = g_list_nth (view->dir_history, maxhistory);
		indir = g_list_first (view->dir_history);
		view->dir_history = g_list_delete_link (view->dir_history, indir);
	}
}
/**
 * @brief re-populate list store with unsorted data for a specified directory
 *
 * Only called from e2_filelist_store_fill ()
 * Directory whose content we want is named (not utf8) in view->dir
 * Current store content is cleared after successful preliminary things, but
 * before new data is added, of course
 * If there's a problem accessing the new directory, will try to cd to 
 * somewhere else that is accessible (to avoid crashing)
 * Active filter(s) are applied, but no sorting.
 * If relevant, the text-color for selected items is updated here, rather 
 * than in the selection function (too many un-necessary repeats, there)
 *
 * @param view data structure for the view to which the file list belongs
 *
 * @return
 */
static void _e2_filelist_store_refill (FileView *view)
{
//	printd (DEBUG, "start store refill");
	//FIXME adapt this 
	DIR *dp;
	struct dirent entry_store;
	struct dirent *entry;
//#ifndef E2_FAM
	struct stat statbuf;
//#endif
	FileInfo info;

	g_return_if_fail (view != NULL);

	//make sure the specified dir is still accessible
	//(note - if this is part of a cd, not a refresh, a valid dir will already have been found)
	E2_PaneRuntime *rt = (view == curr_view ) ? app.pane_current : app.pane_other ;
	gchar *utf = g_strdup (rt->path);	//copy, so that rt->path is not clobbered
	if (e2_fs_get_valid_path (&utf, TRUE))
		g_free (utf);
	else
	{	//if desired path not accessible, go to somewhere that is ..
//		printd (DEBUG, "Cannot access %s, going to %s instead", view->dir, utf);
		gchar *msg = g_strdup_printf ("Cannot access %s, going to %s instead", rt->path, utf);
		e2_output_print_error (msg);
		e2_pane_change_dir (rt, utf);	//CHECKME this will re-enter this fn ?
		g_free (utf);
		g_free (msg);
		return;
	}

	if (! e2_fs_chdir (view->dir))
		return;	//can't open the dir

	F_DECLARE;
	gchar *loc_dir = F_FILENAME_TO_LOCALE (view->dir);

//#ifndef E2_FAM
	e2_fs_stat (loc_dir, &statbuf); // stat before we opendir
//#endif
	if ((dp = opendir (loc_dir)) == NULL)
	{
		printd (WARN, _("Unable to open directory: %s\n"), view->dir);
		F_FREE (loc_dir);
		return;
	}
//#ifndef E2_FAM
	//these are not really necessary when FAM is working,
	//but any time saving by testing for that is not material ..
	view->dir_mtime = statbuf.st_mtime;
	view->dir_ctime = statbuf.st_ctime;
//#endif
	//get format for dates display
	gchar *strf_string;
	switch	(e2_option_int_get ("date-string"))
	{
		case 1:
			strf_string = "%d/%m/%y %H:%M";	//stadard
			break;
		case 2:
			strf_string = "%m/%d/%y %H:%M";	//american
			break;
		case 3:
/*			{
				strf_string = nl_langinfo (D_T_FMT);
				if (strf_string != NULL && *strf_string != '\0')
					break;
			} */
			strf_string = "%x %X";	//localised
			break;
		default:
			strf_string = "%b %d %H:%M";
			break;
	}
	//get format for size display
	gchar *comma;
	switch	(e2_option_int_get ("size-string"))	
	{
		case 1:
		{
			comma = nl_langinfo (THOUSEP);
			if (comma == NULL || *comma == '\0')
				comma = ",";
			break;
		}
		default:
			comma = NULL;	//signal to use the condensed version
			break;
	}
	
	gboolean caseignore = ! e2_option_bool_get ("namesort-case-sensitive");

	GtkStyle *style = gtk_rc_get_style (view->treeview);
	GdkColor *default_color = &style->text[GTK_STATE_NORMAL];
//	GdkColor *background = &style->base[GTK_STATE_NORMAL];
#ifdef E2_SELTXT_RECOLOR
	// set the current foreground color for selected items
	//used in selection function, but don't want to do this each selection
	//CHECKME get the right one - ~bg[*] is no good nor is ~base[*] 
	//~text is likely to be a light color
	selectedtext.red = ~style->text[GTK_STATE_NORMAL].red;
	selectedtext.green = ~style->text[GTK_STATE_NORMAL].green;
	selectedtext.blue = ~style->text[GTK_STATE_NORMAL].blue;
#endif

	//zap the current store contents, and selection
	_e2_filelist_store_empty (view);

//	printd (DEBUG, "start dir load for %s", view->dir);

	FileInfo *this_info;
	struct tm *tm_ptr;
	struct passwd *pwd_buf;
	struct group *grp_buf;
	gchar filename_buf[sizeof(this_info->filename)]; //[NAME_MAX];
	gchar size_buf[20];	//enough for 999 Gb
	gchar modified_buf[25];
	gchar accessed_buf[25];
	gchar changed_buf[25];
	gchar perm_buf[11];
	gchar uid_buf[20];
	gchar gid_buf[20];
	gchar *buf[NAMEKEY+1];	//pointers to NAMEKEY strings

/* seems like setting up a GValue array is not worth it ...
#if defined (E2_CURRENTGTK_VERSION) && GTK_MINOR_VERSION >= 6
#ifdef E2_SELTXT_RECOLOR
#define STORECOLCOUNT MODEL_COLUMNS
#else
//BACKCOLOR info not set in store when it's filled, so not needed
//in setup arrays unless SELCOLOR is in effect
#define STORECOLCOUNT MODEL_COLUMNS-1
#endif

static gint colnums[STORECOLCOUNT] = {
		FILENAME, SIZE, PERM, OWNER, GROUP,
		MODIFIED, ACCESSED, CHANGED, NAMEKEY,
		FINFO, FORECOLOR
#ifdef E2_SELTXT_RECOLOR
	, BACKCOLOR
	, SELCOLOR
#endif
		};

	//FIXME do this init just once
	GValue *colvalues[STORECOLCOUNT];
	gint i;
	for (i=FILENAME; i <= NAMEKEY ; i++)
	{
		g_value_init (colvalues[i], G_TYPE_STRING);
		g_value_set_static_string (colvalues[i], buf[i]);
	}
	for (; i < STORECOLCOUNT; i++);
		g_value_init (colvalues[i], G_TYPE_???);

#endif
*/
	gchar *freeme;
	//look these up once, for speed
	GdkColor *link_color = e2_option_color_get ("color-ft-link");
	GdkColor *dir_color = e2_option_color_get ("color-ft-dir");
	GdkColor *dev_color = e2_option_color_get ("color-ft-dev");
	GdkColor *sock_color = e2_option_color_get ("color-ft-socket");
	GdkColor *exec_color = e2_option_color_get ("color-ft-exec");
	gboolean test1 = !e2_option_bool_get ("show-updir-entry")
			|| g_str_equal (view->dir, G_DIR_SEPARATOR_S);  //at root directory now?

/*	//DEBUG
	struct timeval result, strt, nd;
	result.tv_sec = 0;
	result.tv_usec = 0;
*/
	GtkTreeIter iter;
	//iterate through the directory contents
//	while ((entry = readdir(dp)) != NULL)
	while (!readdir_r (dp, &entry_store, &entry) && entry != NULL)
	{
		if (g_str_equal (entry->d_name, "."))
		  continue;
		if (test1 && g_str_equal (entry->d_name, ".."))
//			&& (!e2_option_bool_get ("show-updir-entry")
//			|| g_str_equal (view->dir, G_DIR_SEPARATOR_S)))  //at root directory now?
		  continue;
		if (((entry->d_name[0] == '.') && (!g_str_equal(entry->d_name, "..")))
			&& !view->show_hidden )
		  continue;

		//get data into local structure (mimic strlcpy)
/*		length = strlen (entry->d_name);
		if (length +1 <= sizeof (info.filename))
			strcpy (info.filename, entry->d_name);
		else
		{	//this truncation means it will never be able to be stat'd anymore
			strncpy (info.filename, entry->d_name. sizeof(info.filename)-1);
			*(info.filename + sizeof(info.filename) -1) = '\0';
		} */
		g_strlcpy (info.filename, entry->d_name, sizeof (info.filename));
		e2_fs_lstat (entry->d_name, &info.statbuf);
		if (view->filename_filter.active && !_e2_filelist_match_filename_filter(view, &info))
		    continue;
		if (view->size_filter.active && !_e2_filelist_match_size_filter(view, &info))
		    continue;
		if (view->date_filter.active && !_e2_filelist_match_date_filter(view, &info))
		    continue;

		//this one is to keep
		this_info = g_memdup (&info, sizeof(FileInfo));
		//localised string, no glib copying
		//name is already truncated to buffer size, if need be
		strcpy (filename_buf, info.filename);

		//directories get a trailing separator, if there's space
		if (e2_fs_is_dir (this_info))
		{
			gint len = strlen (filename_buf);
			if (len < sizeof(filename_buf) -1)
			{
				*(filename_buf+len) = G_DIR_SEPARATOR;
				*(filename_buf+len+1) = '\0';
			}
		}
		//convert to utf if appropriate
		buf[FILENAME] = D_FILENAME_FROM_LOCALE (filename_buf);

		if (comma == NULL)
		{	//use condensed size format
			if (info.statbuf.st_size < 10240) // less than 10k
			{
			  g_snprintf(size_buf, sizeof(size_buf), _("%d b"),
				(gint) info.statbuf.st_size);
			}
			else if (info.statbuf.st_size < 1048576) // less than a meg
			{
			  g_snprintf(size_buf, sizeof(size_buf), _("%.2f kb"),
				((gfloat) info.statbuf.st_size / 1024.0));
			}
			else  // more than a meg
			{
			  g_snprintf(size_buf, sizeof(size_buf), _("%.2f Mb"),
				((gfloat) info.statbuf.st_size / 1048576.0));
			}
		}
		else
		{	//use actual size, with commas
			g_snprintf(size_buf, sizeof(size_buf), "%d",
				(gint) info.statbuf.st_size);

			gint len = strlen (size_buf);
			gint ths = len-1;  //0-based index
			gint i;
			while (ths > 2 && len < sizeof(size_buf))
			{
				for (i = len-1; i > ths-3; i--)
					size_buf[i+1] = size_buf[i];

				size_buf[i+1] = *comma;
				size_buf[++len] = '\0';
				ths = i;
			}
		}
		buf[SIZE] = g_strdup (size_buf);	//content is ascii, no need to convert to utf

		e2_fs_get_perm_string (perm_buf, sizeof(perm_buf), info.statbuf.st_mode);
//		buf[PERM] = e2_utf8_from_locale (perm_buf);
		buf[PERM] = g_strdup (perm_buf); //content is ascii, no need to convert to utf

		if ((pwd_buf = getpwuid (info.statbuf.st_uid)) == NULL)
		{
		  g_snprintf (uid_buf, sizeof(uid_buf), "%d",
		     (gint) info.statbuf.st_uid);
		  buf[OWNER] = g_strdup (uid_buf); //content is ascii number, no need to convert to utf
		}
		else
		{
		  buf[OWNER] = e2_utf8_from_locale (pwd_buf->pw_name);
		}

		if ((grp_buf = getgrgid (info.statbuf.st_gid)) == NULL)
		{
		  g_snprintf(gid_buf, sizeof(gid_buf), "%d",
		       (gint) info.statbuf.st_gid);
		  buf[GROUP] = g_strdup (gid_buf); //content is ascii number, no need to convert to utf
		}
		else
		{
		  buf[GROUP] = e2_utf8_from_locale (grp_buf->gr_name);
		}
		tm_ptr = localtime (&(info.statbuf.st_mtime));
		strftime (modified_buf, sizeof(modified_buf), strf_string, tm_ptr);
		buf[MODIFIED] = e2_utf8_from_locale (modified_buf);

		tm_ptr = localtime (&(info.statbuf.st_atime));
		strftime (accessed_buf, sizeof(accessed_buf), strf_string, tm_ptr);
		buf[ACCESSED] = e2_utf8_from_locale (accessed_buf);

		tm_ptr = localtime(&(info.statbuf.st_ctime));
		strftime(changed_buf, sizeof(changed_buf), strf_string, tm_ptr);
		buf[CHANGED] = e2_utf8_from_locale (changed_buf);

		if (caseignore)
		{
			freeme = g_utf8_casefold (buf[FILENAME], -1);
#if  !defined (E2_CURRENTGTK_VERSION) || GTK_MINOR_VERSION < 8
			buf[NAMEKEY] = g_utf8_collate_key (freeme, -1);
#else
			buf[NAMEKEY] = g_utf8_collate_key_for_filename (freeme, -1);
#endif
			g_free (freeme);
		}
		else
#if  !defined (E2_CURRENTGTK_VERSION) || GTK_MINOR_VERSION < 8
			buf[NAMEKEY] = g_utf8_collate_key (buf[FILENAME], -1);
#else
			buf[NAMEKEY] = g_utf8_collate_key_for_filename (buf[FILENAME], -1);
#endif
		GdkColor *foreground;
		switch (this_info->statbuf.st_mode & S_IFMT)
		{
		  case S_IFLNK:
		    foreground = link_color;
		    break;
		  case S_IFDIR:
		    foreground = dir_color;
		    break;
		  case S_IFCHR:
		  case S_IFBLK:
		    foreground = dev_color;
		    break;
		  case S_IFSOCK:
		    foreground = sock_color;
		    break;
		  case S_IFREG:
			  //show as executable if _anyone_ has that permission
		    if ((S_IXUSR & this_info->statbuf.st_mode)
		  		|| (S_IXGRP & this_info->statbuf.st_mode)
		  		|| (S_IXOTH & this_info->statbuf.st_mode))
			  foreground = exec_color;
			else
			  foreground = default_color;
		    break;
		  default:
			foreground = default_color;
		    break;
		}
#if !defined (E2_CURRENTGTK_VERSION) || GTK_MINOR_VERSION < 6
		//append empty row to the list store
		gtk_list_store_append (view->store, &iter);

//		gettimeofday (&strt, NULL); //FOR DEBUGGING, check time this load takes
		
		//populate it
		gtk_list_store_set (view->store, &iter,
#else
		gtk_list_store_insert_with_values (view->store, &iter, -1,
#endif
			FILENAME, buf[FILENAME],
			SIZE, buf[SIZE],
			PERM, buf[PERM],
			OWNER, buf[OWNER],
			GROUP, buf[GROUP],
			MODIFIED, buf[MODIFIED],
			ACCESSED, buf[ACCESSED],
			CHANGED, buf[CHANGED],
			NAMEKEY, buf[NAMEKEY],
			FINFO, this_info,
			FORECOLOR, foreground,
//			BACKCOLOR, background,	don't set this till needed
#ifdef E2_SELTXT_RECOLOR
			SELCOLOR, foreground,  //for reverting when item unselected
#endif
//			VISIBLE, TRUE,
			-1);
//		gettimeofday (&nd, NULL);

		//cleanup
		gint i;
		for (i = 0; i <= NAMEKEY; i++)
			g_free (buf[i]);

/*		if (nd.tv_usec < strt.tv_usec)
		{
		    int nsec = (strt.tv_usec - nd.tv_usec) / 1000000 + 1;
    		strt.tv_usec -= 1000000 * nsec;
    		strt.tv_sec += nsec;
		}
		if (nd.tv_usec - strt.tv_usec > 1000000)
		{
			int nsec = (strt.tv_usec - nd.tv_usec) / 1000000;
			strt.tv_usec += 1000000 * nsec;
			strt.tv_sec -= nsec;
		}
		result.tv_sec = result.tv_sec + nd.tv_sec - strt.tv_sec;
		result.tv_usec = result.tv_usec + nd.tv_usec - strt.tv_usec;
		if (result.tv_usec > 1000000)
		{
			int nsec = result.tv_usec / 1000000 + 1;
    		result.tv_usec -= 1000000 * nsec;
    		result.tv_sec += nsec;
		}
*/		
	}
	closedir(dp);

//	printd (DEBUG, "finish store refill");
//	printd (DEBUG, "populating rows took %f seconds", result.tv_sec + result.tv_usec / 1000000.0 );
	F_FREE (loc_dir);
}
/**
 * @brief create list store structure
 *
 * Creates basic structure of list store, for panes filelists, and stores
 * pointer to the store in the view rt data.
 * No rows or data are added.
 * This should only be needed at session start ?
 * In any event, it destroys any existing list-store belonging to the view (superfluous ?)
 *
 * @param view for which the list store is to be created
 *
 * @return
 */
void e2_filelist_make_store (FileView *view)
{
//	if (GTK_LIST_STORE (view->store) != NULL)
//		gtk_widget_destroy (GTK_LIST_STORE(view->store));   how to destroy old version ?
	view->store = gtk_list_store_new (MODEL_COLUMNS,
	G_TYPE_STRING,  //FILENAME
	G_TYPE_STRING,  //SIZE
	G_TYPE_STRING,  //PERM
	G_TYPE_STRING,  //OWNER
	G_TYPE_STRING,  //GROUP
	G_TYPE_STRING,  //MODIFIED
	G_TYPE_STRING,  //ACCESSED
	G_TYPE_STRING,  //CHANGED
	// the rest are not displayed
	G_TYPE_STRING,  //NAMEKEY for i18n name sorts
	G_TYPE_POINTER,  //FINFO pr to FileInfo for the item
	GDK_TYPE_COLOR,  //FORECOLOR line colour
	GDK_TYPE_COLOR  //BACKCOLOR line colour
#ifdef E2_SELTXT_RECOLOR
	, GDK_TYPE_COLOR  //SELCOLOR  foreground of selected lines
#endif
//	G_TYPE_BOOLEAN	//VISIBLE whether filtered or not
	);

// FIXME = make filtermodel and sorting work together ?
// adding sortmodel doesn't help ...
/*	view->sortmodel = gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL (view->store));
	view->filtermodel = gtk_tree_model_filter_new
		(view->sortmodel, NULL);		//NULL = main root node ?
	gtk_tree_model_filter_set_visible_column
		(GTK_TREE_MODEL_FILTER (view->filtermodel), VISIBLE);
*/
}
/**
 * @brief  de-populate list store for current directory
 *
 * Treeview is detached from the model, before data is cleared (for speed)
 *
 * @param view runtime data for pane containing the view
 *
 * @return
 **/
static void _e2_filelist_store_empty (FileView *view)
{
//	printd (DEBUG, "start clearing list store");
	//make sure the model stays alive after the treeview unrefs it
	g_object_ref (view->store);
//	g_object_ref (view->filtermodel);  //and this too
//	g_object_ref (view->sortmodel);  //and this too
	//for speed, detach model from view
	//(re-attached after store is (re)filled, elsewhere)
	gtk_tree_view_set_model (GTK_TREE_VIEW (view->treeview), NULL);

	GtkTreeModel *model = GTK_TREE_MODEL (view->store);
	FileInfo *info;
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter_first (model, &iter))
	{	//it's not empty already
		//clear file info data structs referred to in the store
		do
		{
			gtk_tree_model_get (model, &iter, FINFO, &info, -1);
			g_free (info);
		} while (gtk_tree_model_iter_next (model, &iter));
	//CHECKME no need to separately clear gdk color structs ?

		//clear the store itself
		gtk_list_store_clear (view->store);
		//clear out any old selection
		GtkTreeSelection *sel = gtk_tree_view_get_selection 
			(GTK_TREE_VIEW (view->treeview));
//		if (sel != NULL)
			gtk_tree_selection_unselect_all (sel);
	}
	//don't want old tagged items, either
//	g_list_free(view->tagged);
//	view->tagged = NULL;
}
/**
 * @brief  populate list store with items from a specified directory
 *
 *  The exisiting store contents are removed, first.
 *
 * @param view runtime data for the view to be displayed
 *
 * @return
 **/
static void _e2_filelist_store_fill (FileView *view)
{
	//prevent re-entrant refreshes
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, store fill");
#endif
	e2_filelist_disable_refresh ();
	e2_widget_set_cursor (GDK_WATCH);

	//prevent repetitive counts in case of long refreshes
//	if (view == curr_view)
//		e2_window_disable_status_update ();

//	e2_window_show_status_message (_("<b>Processing ...</b>"));

	//turn off model sorting before any rows are added
	GtkTreeModel *mdl = GTK_TREE_MODEL (view->store);
	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (mdl);
	//remember the current sorting arrangements
	view->sort_order = GTK_SORT_ASCENDING;
	gtk_tree_sortable_get_sort_column_id (sortable, &view->sort_column, &view->sort_order);
	if (view->sort_column < 0
		//don't want the GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID (-2) column that is startup default
		|| view->sort_column > NAMEKEY
		//in case gtk does something undocumented ...
		)
		view->sort_column = FILENAME;
	//now we actually turn it off
	//GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID (-2) appeared in gtk 2.6
	gtk_tree_sortable_set_sort_column_id (sortable,
		GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, GTK_SORT_ASCENDING);

	//if we can, empty current, and get the new directory data into the liststore
	_e2_filelist_store_refill (view);

	view->completed = FALSE;  //set blocked signal

	_e2_filelist_sortNselect (view);

//	gdk_threads_enter ();
//	gtk_main_iteration_do (FALSE);
//	gdk_threads_leave ();

	//fork to finish the sort and (re)selection process
//	g_thread_create ((void *)_e2_filelist_sortNselect, view, FALSE, G_ERR_UNKNOWN);

#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, store fill");
#endif
	e2_widget_set_cursor (GDK_LEFT_PTR);
	e2_filelist_enable_refresh ();
}

/*gboolean e2_filelist_store_fill (FileView *view, gboolean newload)
{
	//get the directory data into the liststore
	_e2_filelist_load_dir (view);
	//and sort it using the current sort column
	GtkTreeModel *mdl = GTK_TREE_MODEL (view->store);
	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (mdl);
	gint sort_column_id = FILENAME;
	GtkSortType sort_order = GTK_SORT_ASCENDING;
	//workaround, sort column is -2, at session start !!
	gtk_tree_sortable_get_sort_column_id (sortable, &sort_column_id, &sort_order);
	if (sort_column_id < 0)
		sort_column_id = FILENAME;
	//perform the sort
	gtk_tree_sortable_set_sort_column_id (sortable, sort_column_id, sort_order);
	//(re)attach model to view
	gtk_tree_view_set_model (GTK_TREE_VIEW (view->treeview), mdl);
	// destroy model automatically with view
	g_object_unref (view->store);
//	g_object_unref (view->filtermodel);  //and this too
//	g_object_unref (view->sortmodel);  //and this too

	if (newload)
	{
		//select 1st row of view, if any
		GtkTreeIter iter;
		if (gtk_tree_model_get_iter_first (mdl, &iter))
		{
			GtkTreeSelection *sel = gtk_tree_view_get_selection
				(GTK_TREE_VIEW (view->treeview));
			gtk_tree_selection_select_iter (sel, &iter);
			view->row = 0;
		}
		else
			//unselect the dir line entry
			gtk_widget_grab_focus (view->treeview);
	}
} */
/**
 * @brief  sort filelist and (re) select item(s)
 *
 * (this is separate, to support threading if desired)
 *  Current store contents are removed, first.
 * Treeview is (re) attached to the model, after data is loaded (for speed)
 * First time, the filename column is sorted ascending, thereafter the current
 * arrangment is retained.
 * The 1st row of the view is selected if view->selected_names is NULL
 *
 * @param view runtime data for the view to be displayed
 *
 * @return
 **/
static void _e2_filelist_sortNselect (FileView *view)
{
//    g_mutex_lock (app.filelist_mutex);

	GtkTreeModel *mdl = GTK_TREE_MODEL (view->store);
/*	//check if we expect the sort to be protracted
	gint total_rows = gtk_tree_model_iter_n_children (mdl, NULL);
	gint threshold = e2_option_int_get ("sort-before-show-count");
	gboolean fast = total_rows <= threshold;
	if (!fast)
		//FIXME prevent duplicated messages
		e2_window_show_status_message (_("<b>Processing ...</b>"));  */
	//get the sort direction
	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (mdl);
/*	gint sort_column_id;  //this will be set to a non-relevant value
	GtkSortType sort_order = GTK_SORT_ASCENDING;

//	gdk_threads_enter ();

	gtk_tree_sortable_get_sort_column_id (sortable, &sort_column_id, &sort_order);
*/
//	g_mutex_unlock (app.filelist_mutex);

//	printd (DEBUG, "sorting list for %s", view->dir);
	//sort the view using the current sort column & direction
	gtk_tree_sortable_set_sort_column_id (sortable, view->sort_column, view->sort_order);
//	printd (DEBUG, "finished sorting list for %s", view->dir);

	//(re)attach model to view
	gtk_tree_view_set_model (GTK_TREE_VIEW (view->treeview), mdl);
	// destroy model automatically with view
	g_object_unref (view->store);
//	g_object_unref (view->filtermodel);  //and this too
//	g_object_unref (view->sortmodel);  //and this too

//	g_mutex_unlock (app.filelist_mutex);

/*	if (view->selected_names == NULL)
	{	//this is an initial load of the directory,
		//select 1st row of view, if any
		GtkTreeIter iter;
		if (gtk_tree_model_get_iter_first (mdl, &iter))
		{
			GtkTreeSelection *sel = gtk_tree_view_get_selection
				(GTK_TREE_VIEW (view->treeview));
			gtk_tree_selection_select_iter (sel, &iter);
			view->row = 0;
		}
		else
			//nothing in the list, just unselect the dir line entry
			gtk_widget_grab_focus (view->treeview);
	}
	else //this is a refresh, re-select the old names where possible
*/
#ifdef E2_INCLIST
	//FIXME
#else
	if (e2_fileview_reselect_names (view, view->selected_names))
	{	//there was a re-selection, so also restore old scroll positions
		GtkAdjustment *adj;
		adj = gtk_tree_view_get_vadjustment (GTK_TREE_VIEW (view->treeview));
		//the vadjustment value will be unchanged, but the sw 
		//doesn't know about it unless this signal is issued
		gtk_adjustment_value_changed (adj);
		adj = gtk_tree_view_get_hadjustment (GTK_TREE_VIEW (view->treeview));
		gtk_adjustment_value_changed (adj);
	}
	else
	//nothing in the list, or nothing selected, just unselect the dir line entry
		gtk_widget_grab_focus (view->treeview);
#endif //def E2_INCLIST

//	printd (DEBUG, "finished new selection for %s", view->dir);
	//signal to the world ...
	view->completed = TRUE;

	if (view == curr_view)
	{
		e2_window_update_status_bar (NULL);
//		e2_window_enable_status_update (-1);
	}
	else
	{
		//shift focus back to where it belongs
		gtk_widget_grab_focus (app.pane_current->focus_widget);
		e2_hook_list_run (&app.hook_pane_focus_changed, app.pane_current);
	}

	//if both views are completed, re-start refreshing
//DEBUG this is for threaded processiing - otherwise, it stuffs up the refresh ref count
//	FileView *othr = (view == app.pane1.view) ? app.pane2.view : app.pane1.view;
//	if (othr->completed)
//	{
//		refresh_ref = 1;	//probably was an extra ++ for the other file pane
//		e2_filelist_enable_refresh ();
//	}

//	if (!fast)
//		e2_window_remove_status_message ();
//	gdk_flush();
//	gdk_threads_leave ();
}
