/* $Id: e2_fileview.c 1114 2005-09-12 07:39:55Z tpgww $
 *
 * Portions copyright (C) 2004-2005 tooar <tooar@gmx.net>.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/* TODO 26/10/04
confirn whether column-sort hack (see e2_fileview_column_header_buttonrel_cb) is a general solution
(minor) check - mutually-exclusive updir started and drag started ?
(minor) get proper cursor size for dnd
(trivial) set default background colour once, not each treeview ??
(trivial) un-select dirline text for empty _startup_ dirs
*/
 
/**
 * @file src/e2_fileview.c 
 * @brief directory content view functions
 *
 * This file contains functions related to creation of treeviews to display
 * directory content, plus related selections and drag'ndrop.
 **/

#include "emelfm2.h"
//#include <unistd.h>
#include <string.h>
#include "e2_filelist.h"
#include "e2_context_menu.h"
#include "e2_option.h"
#include "e2_dialog.h"
#include "e2_task.h"

//mouse pointer height, used for getting the path at the _top_ of the pointer
gint pointer_height;
//default pointer height, used when we can't find the runtime value
//NOTE: setting the height too small eg 16 causes problems for
//un-lighting a row when the mouse moves slowly
#define POINTER_HEIGHT 26

gint col_width_store[2][MAX_COLUMNS];
//this holds, in fixed column order (used in model & config, cache data),
// the displayed column order, a row for each pane
gint stored_col_order[2][MAX_COLUMNS];
//this holds column data from the above arrays, for cache reading/writing
GList *cols_data;
//this holds, in displayed column order
// the fixed column order, a row for each pane
gint displayed_col_order[2][MAX_COLUMNS];

// Colors
GdkColor LIST_COLOR;
#ifdef E2_SELTXT_RECOLOR
GdkColor selectedtext;
#endif

// Stuff for drag and drop
enum { TARGET_URL, TARGET_STRING };

static GtkTargetEntry target_table[] = {
	{ "text/uri-list", 0, TARGET_URL },
	{ "text/plain",    0, TARGET_STRING }
};
static guint n_targets = sizeof(target_table) / sizeof(target_table[0]);
static gboolean btn2_released;	//flag used to decide id middle-btn DnD menu is wanted

//local copy of flag
static gboolean button2updir;
//number of last mouse button press, for distinguishing drags
static gboolean pane1_all = FALSE, pane2_all=FALSE;
//these are related to column dragging
static gint header_button;
static gboolean block;

static gboolean _e2_fileview_drag_leave_cb (GtkWidget *treeview, 
	GdkDragContext *context, guint time, FileView *view);
static void _e2_fileview_set_row_background (FileView *view,
	GtkTreeIter *iter, GdkColor *color);
static void _e2_fileview_clear_row_background (FileView *view,
	GtkTreeIter *iter);

// Callbacks

/**
 * @brief column-header button-press detection
 *
 * This fn records which button was clicked, so that when a "release"
 * callback occurs (which may be when a drag starts, not an actual
 * release) we can check whether it is still pressed
 * This is a workaround for gtk's (2.4 at least) behaviour
 * It relies on detecting clicks on the undocumented column->button widget
 *
 * @param header the col button header widget which was clicked
 * @param rt data structure for the pane to which the file list belongs unused
 *
 * @return FALSE, so related signals are not blocked
 */
static gboolean _e2_fileview_column_header_buttonpress_cb
	(GtkWidget *header, GdkEventButton *event, E2_PaneRuntime *rt)
{
	printd (DEBUG, "callback: column header button press");
	header_button = event->button;
	return FALSE;
}
/**
 * @brief column-header button-release detection
 *
 * This checks whether a release event is real. In gtk 2.4 
 * at least, such an event may be when a column-drag starts,
 * not an actual release.
 * It relies on detecting clicks on the undocumented 
 * column->button widget
 *
 * @param header the col button widget which was clicked
 * @param rt data structure for the pane to which the file list belongs unused
 *
 * @return TRUE, if this is a 'bogus' release, else FALSE
 */
static gboolean _e2_fileview_column_header_buttonrel_cb
	(GtkWidget *header, GdkEventButton *event, E2_PaneRuntime *rt)
{
	printd (DEBUG, "callback: column header button release");

	if (rt->view != curr_view)
		e2_pane_activate_other ();

	//check for whether button is really still pressed
	//gtk 2.4 emits a bogus extra release signal when column-drag occurs
	gboolean pressed = FALSE;
	if (!block && event->send_event)
	{
		pressed = TRUE;
		block = TRUE;
	}
	else if (block)
		block = FALSE;  //on the second pass, just unblock
//	else
//		pressed = FALSE;

	if (!pressed)
		header_button = -1;  //signal to "clicked" callback not to do anything
	return pressed;	//stop further action on this button if not released yet
}
/**
 * @brief column-click callback
 *
 * This fn updates the view sorting, by applying an ascending sort
 * to a column that is not sorted, or by toggling the sort direction 
 * of a column that is already sorted.
 * Sort indicators are updated.
 * This is called when a column header is clicked, before any
 * associated "column-change" signal.
 *
 * @param col the view column which was clicked
 * @param rt data structure for the pane to which the file list belongs
 *
 * @return
 */
static void _e2_fileview_column_header_clicked_cb 
	(GtkTreeViewColumn *col, E2_PaneRuntime *rt)
{
	if (header_button != -1)
		return;		//the button is still down, this is drag action
	
	printd (DEBUG, "callback: column header click");

	//get which column was cliicked
	GList *cols = gtk_tree_view_get_columns (GTK_TREE_VIEW (rt->view->treeview));
	gint clicked_col = g_list_index (cols, (gpointer) col);
	g_list_free (cols);

	//get colum-order array for the pane
	gint *order_array = (rt->view == app.pane1.view) ? displayed_col_order[0] : displayed_col_order[1];
	//get clicked-model-column number
	gint i = order_array[clicked_col];
/*	if (i == 0)
	{ //filename column  DO WE WANT TO MAKE THIS DYNAMIC BUT SLOWER ??
		//get a local copy of the case-sensitive flag
		E2_OptionSet *set = e2_option_get_simple ("namesort-case-sensitive");
		case_sensitive = _e2_option_bool_get (set);
	} */
	//get current-sorted-column details
	gint m;
	GtkSortType old_order;
	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (rt->view->store);
	gtk_tree_sortable_get_sort_column_id (sortable, &m, &old_order);
	
	if (m == FILENAME)
	//revert any sort-function change by the extension-sort plugin
		gtk_tree_sortable_set_sort_func (sortable, FILENAME, 
		e2_all_columns[FILENAME].sort_func, &rt->view->sort_order, NULL);

	GtkSortType sort_order;
	if (m == i)
	{	//column is already sorted, toggle it
		rt->view->sort_order = sort_order = (old_order == GTK_SORT_ASCENDING) ?
			GTK_SORT_DESCENDING : GTK_SORT_ASCENDING;
		//set flag for row-comparison fn
//		order = (sort_order == GTK_SORT_ASCENDING);
	}
	else
	{  //this is a new sort column
		//replace sort indicator
		gtk_widget_hide (rt->view->sort_arrows[m]);
		gtk_widget_show (rt->view->sort_arrows[i]);
		//1st click makes it ascending
		rt->view->sort_order = sort_order = GTK_SORT_ASCENDING;
//		order = TRUE;
	}
	//set appropriate arrow type
	gtk_arrow_set (GTK_ARROW (rt->view->sort_arrows[i]),
		(sort_order == GTK_SORT_ASCENDING) ? GTK_ARROW_DOWN : GTK_ARROW_UP,
			GTK_SHADOW_IN);
/* no use, gtk always returns gchar data to heap space
	//setup stack variable space for speedier name sorts
	if (i == FILENAME)
	{
		gchar namespace [2][E2_UTFNAME_MAX];
		gtk_tree_sortable_set_sort_func (sortable, FILENAME,
			e2_all_columns[FILENAME].sort_func,
            (gpointer *) namespace, NULL);
	} */

	//do the sort
	gtk_tree_sortable_set_sort_column_id (sortable, i, sort_order);

	gtk_widget_grab_focus (rt->view->treeview);
	return;
}
/**
 * @brief column-change callback
 *
 * This fn refreshes the columns-order array for the pane.
 * It is called when a column is dropped to a different position
 * and when the window is closed
 * (also, multiple times when the window is re-created - CHECKME why)
 *
 * @param treeview the widget where the action occurred
 * @param rt data structure for the pane to which the file list belongs
 *
 * @return
 */
static void _e2_fileview_col_change_cb (GtkTreeView *treeview,
	E2_PaneRuntime *rt)
{
	printd (DEBUG, "callback: column change");
	e2_fileview_set_col_order (rt);
}
/**
 * @brief mouse button double-click callback
 *
 * This "executes" the clicked row item
 * view->row will have been set by the button-release callback
 *
 * @param treeview the widget where the button was pressed
 * @param path model path to the clicked row
 * @param col clicked view column unused
 * @param view rt data for the view to be worked on
 *
 * @return
**/
static void _e2_fileview_row_activated_cb (
		GtkTreeView        *treeview,
		GtkTreePath        *path,
		GtkTreeViewColumn  *col,
		FileView              *view)
{
	printd (DEBUG, "callback: double-click");
	GtkTreeIter iter;
	GtkTreeModel *model = gtk_tree_view_get_model (treeview);
    if (gtk_tree_model_get_iter (model, &iter, path))
	{
//		view->row = *gtk_tree_path_get_indices (path); //keep a record of which row it is
		FileInfo *info;
		gtk_tree_model_get (model, &iter, FINFO, &info, -1);
		e2_task_backend_open (info);
//revert the bogus view swap, from e2_task_open_backend ()
//		e2_fileview_switch_views_simple ();
	}
}
/**
 * @brief  mouse button release callback
 *
 * This checks if the click was a left click. If so, and it's not the
 * end of a drag, or an 'un-select' click, the clicked row is logged.
 * Detection of the end of a drag relies on an undocumented 
 * check whether the release happened in the same pane as
 * was clicked.
 *
 * @param treeview where the button was _pressed_
 * @param event
 * @param view rt data for the view where the button was _pressed_
 * 
 * @return TRUE (stop other handlers) if updir initiated, else FALSE (allow other handlers)
**/
static gboolean _e2_fileview_mouse_release_cb (GObject *treeview, 
	GdkEventButton *event, FileView *view)
{
	printd (DEBUG, "button release");

/* support better dragging, but blocks double-clicks
	if (event->button == 1)
	{
		//correctly distinguish drag releases, and normal clicks
		//this is an undocumented property of releases in the same pane as clicked??
		if (event->send_event == 0) 
		{	//not a drag release
			GtkTreePath *path;
			if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (view->treeview),
					event->x, event->y, &path, NULL, NULL, NULL))
			{
				GtkTreeSelection *sel = gtk_tree_view_get_selection (GTK_TREE_VIEW(treeview));
				guint modifiers = gtk_accelerator_get_default_mod_mask ();

				if ((event->state & modifiers) == GDK_CONTROL_MASK)
				{	//mimic ctrl-click = toggle selection of clicked item
					if (gtk_tree_selection_path_is_selected (sel, path))
						gtk_tree_selection_unselect_path (sel, path);
					else
						gtk_tree_selection_select_path (sel, path);
					gtk_tree_path_free (path);
					return TRUE;
				}
				//FIXME mimic shift-clicks, & remove that from press cb
				if ((event->state & modifiers) != GDK_SHIFT_MASK)
				{	//mimic nortmal-click = select clicked item
					gtk_tree_selection_unselect_all (sel);
					gtk_tree_selection_select_path (sel, path);
					gtk_tree_path_free (path);
					return TRUE;
				}
			}
			gtk_tree_path_free (path);
		}
	}
//	GtkTreePath *path;
	else */
		if (event->button == 2)
	{
//		printd (DEBUG, "callback: mouse button-2 release");
		//correctly distinguish drag releases, and normal clicks
		//this is an undocumented property of releases in the same pane as clicked??
		if (event->send_event == 0) 
		{	//not a drag release
			if (button2updir)
			{
				printd (DEBUG, "up");
				e2_pane_change_dir (NULL, "..");
				return TRUE;
			}
			//otherwise, maybe log the row
			guint modifiers = gtk_accelerator_get_default_mod_mask ();
			if ((event->state & modifiers) != GDK_CONTROL_MASK)
			{	//not an "unselect" click
				GtkTreePath *path;
				if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (view->treeview),
					event->x, event->y, &path, NULL, NULL, NULL))
				{
					// row was NOT selected when the mouse was clicked
					GtkTreeSelection *sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (view->treeview));
					gtk_tree_selection_select_path (sel, path);
					//record the row as the last-selected one
					view->row = *gtk_tree_path_get_indices (path);
				}
				gtk_tree_path_free (path);
			}
		}
		else
			//set signal for DnD handler
			btn2_released = TRUE;
	}
	else
		btn2_released = FALSE;

	return FALSE;
}
/**
 * @brief mouse button press callback
 *
 * This make the clicked view active, if it wasn't already.
 * A button-1 press when there is nothing else selected logs
 * the clicked model row, if any, in case it's a drag start. 
 * (For anything other than a drag, the row is logged whem the btn is
 * released.) The gtk handler will select the row.
 * A button-2 press does an updir, if that option is in force, or else
 * logs the clicked model row, if any.
 * A button-3 press sets up the relevant context menu, and may
 * select the line or unselect everything, depending on options in force
 *
 * @param treeview the widget where the button was pressed
 * @param event gdk event data
 * @param view rt data for the view to be worked on
 *
 * @return TRUE (stop other handlers) for btn 2 (updir) or btn 3 press, else FALSE (allow other handlers)
**/
static gboolean _e2_fileview_mouse_press_cb (GtkWidget *treeview, 
	GdkEventButton *event, FileView *view)
{
//	printd (DEBUG, "callback: mouse button press");
	//make the clicked pane active, if it wasn't already
	if (view != curr_view)  // && !panechange_started)
		e2_pane_activate_other ();

	GtkTreePath *path;
	GtkTreeSelection *sel;
	//filter out the special cases for btns 1 or 2
/*	if (event->button == 1)
	{
//		printd (DEBUG, "callback: mouse button-1 press");
		sel = gtk_tree_view_get_selection (GTK_TREE_VIEW(treeview));
		if (gtk_tree_selection_count_selected_rows (sel) > 0)
			return FALSE;
	}
	else
	can't do this here, if middle-button dragging is in force
		if (event->button == 2)
	{
//		printd (DEBUG, "callback: mouse button-2 press");
		if (button2updir)
		{
			printd (DEBUG, "up");
			e2_pane_change_dir (NULL, "..");
			return TRUE;
		}
	}
*/
/* this supports better dragging, but blocks double-clicks ;-(
if (event->button == 1)
	{
		if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
			event->x, event->y, &path, NULL, NULL, NULL))
		{
			guint modifiers = gtk_accelerator_get_default_mod_mask ();
			if ((event->state & modifiers) == GDK_CONTROL_MASK)
			{
				gtk_tree_path_free (path);
				return TRUE;	//wait until release, to see what to do with the selected row
			}
			sel = gtk_tree_view_get_selection (GTK_TREE_VIEW(treeview));
			if (gtk_tree_selection_path_is_selected (sel, path))
			{
				//we can't easily mimic the 'shift-click' action when button is released
				if ((event->state & modifiers) != GDK_SHIFT_MASK)
				{
					gtk_tree_path_free (path);
					return TRUE;	//wait until release, to see what to do with the selected row
				}
			}
			else
			//log the clicked row
			view->row = *gtk_tree_path_get_indices (path);
		}
		gtk_tree_path_free (path);
//		return FALSE; this is the default
	}
*/
	if (event->button == 1)
	{
		if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
			event->x, event->y, &path, NULL, NULL, NULL))
		{
			//log the clicked row
			view->row = *gtk_tree_path_get_indices (path);
		}
		gtk_tree_path_free (path);
	}
//	else if (event->button == 2)
//		return TRUE;	//wait until release, to see whether to select the row

	else if (event->button == 3)
	{
//		printd (DEBUG, "callback: mouse button-3 press");
		if (e2_option_bool_get ("windows-right-click")
//			&& (view->tagged == NULL)
		)
		{
			//if windows right click is enabled, we want to also select the item,
			//if not already done, and clear any other selected item(s) unless this
			//one is selected already
			sel = gtk_tree_view_get_selection (GTK_TREE_VIEW(treeview));
			if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
				event->x, event->y, &path, NULL, NULL, NULL)	)
			{
				if (!gtk_tree_selection_path_is_selected (sel, path))
					gtk_tree_selection_unselect_all (sel);
				gtk_tree_selection_select_path (sel, path);
				view->row = *gtk_tree_path_get_indices (path);
			}
			else if (! e2_option_bool_get ("windows-right-click-extra"))
				gtk_tree_selection_unselect_all (sel);

			gtk_tree_path_free (path);
		}
		//context menu
		gint type = 0;
		guint modifiers = gtk_accelerator_get_default_mod_mask ();
		if ((event->state & modifiers) == GDK_SHIFT_MASK) type = 1;
		else if ((event->state & modifiers) == GDK_CONTROL_MASK) type = 2;
		e2_context_menu_show (event->button, event->time, type);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief set popup menu position
 *
 * This function is supplied when calling gtk_menu_popup(), to position
 * the displayed menu.
 * set @a push_in to TRUE for menu completely inside the screen,
 * FALSE for menu clamped to screen size
 *
 * @param menu the GtkMenu to be positioned
 * @param x	place to store gint representing the menu left
 * @param y  place to store gint representing the menu top
 * @param push_in place to store pushin flag
 * @param treeview in focus when the menu key was pressed
 *
 * @return
 */
void e2_fileview_set_menu_position (GtkMenu *menu,
	gint *x, gint *y, gboolean *push_in, GtkWidget *treeview)
{
	gint left, top;
	gtk_window_get_position (GTK_WINDOW (app.main_window),
		&left, &top);
	GtkAllocation alloc = treeview->allocation;
	*x = left + alloc.x + alloc.width/2;
	*y = top + alloc.y +alloc.height/2 - 30;
	*push_in = TRUE;
}
/**
 * @brief menu button callback
 *
 * Sets up the relevant context menu
 *
 * @param treeview the widget where the button was pressed
 * @param view rt data for the view to be worked on
 *
 * @return TRUE (stop other handlers) for button 2 or 3 press, else FALSE (allow other handlers)
**/
static gboolean _e2_fileview_popup_cb (GtkWidget *treeview, 
	FileView *view)
{
	gint menu_type = 0;	//default to 'normal' context menu
/*	GdkModifierType state;
	if (gtk_get_current_event_state (&state))
	{
		if (state & GDK_SHIFT_MASK) menu_type = 1;
		else
			if (state & GDK_CONTROL_MASK) menu_type = 2;
	}
*/	gint event_time = gtk_get_current_event_time ();
	//button 0 signals this is not a mouse-button click
	e2_context_menu_show (0, event_time, menu_type);
	return TRUE;
}
/** @brief  UNUSED find filename begining with specified char, in current view
 *
 *
 * @param ch character at the start of the name
 * 
 * @return the row number where the matching name was found, or -1 if not found
**/
/*static gint e2_fileview_find_filename_begining_with (gchar ch)
{
	gint i;
	//number of rows in the model
	gint n = gtk_tree_model_iter_n_children
		(GTK_TREE_MODEL (curr_view->store), NULL);
	GtkTreeIter iter;
	GtkTreeModel *model = GTK_TREE_MODEL (curr_view->store);
	gchar *name;
	// search from after the current row to the end of the store
	for (i = curr_view->row+1; i < n; i++)
	{
		if (! gtk_tree_model_iter_nth_child
			(model, &iter, NULL, i))
				continue;  //the specified row doesn't exist any more
		gtk_tree_model_get (model, &iter, FILENAME, &name, -1);
		
		if (name[0] == ch)
		{
			g_free (name);
			return i;
		}
		else
			g_free (name);
	}
	// search from the first row to the current row
	for (i = 0; i < curr_view->row; i++)
	{
		if (! gtk_tree_model_iter_nth_child
			(model, &iter, NULL, i))
				continue;  //the specified row doesn't exist any more
		gtk_tree_model_get (model, &iter, FILENAME, &name, -1);
		
		if (name[0] == ch)
		{
			g_free (name);
			return i;
		}
		else
			g_free (name);
	}
	return -1;
}
*/
/**
 * @brief
 *
 *
 * @param widget the widget where the button was released
 * @param event 
 * @param view rt data for ther view to be worked on
 * 
 * @return
**/
static gboolean _e2_fileview_key_press_cb (GtkWidget *widget, 
	GdkEventKey *event, FileView *view)
{
	/*
	GList *tmp;

	// get rid of all the modifiers that we don't care about
	event->state &= ~(GDK_LOCK_MASK | GDK_MOD2_MASK | GDK_MOD3_MASK |
		GDK_MOD4_MASK | GDK_MOD5_MASK);
	for (tmp = cfg.key_bindings; tmp != NULL; tmp = tmp->next)
	{
		KeyBinding *kb = tmp->data;

		//key bindings are case insensitive
		if ((event->state == kb->state) &&
			(gdk_keyval_to_lower (event->keyval) == kb->keyval))
		{
			e2_command_run (kb->action, E2_COMMAND_RANGE_DEFAULT);
			return TRUE;
		}
	} */
	guint modifiers = gtk_accelerator_get_default_mod_mask ();
	modifiers &= event->state;

	if (g_ascii_isalpha (event->keyval) && (event->keyval < 0x100) 
		&& (modifiers == 0 || modifiers == GDK_SHIFT_MASK))
//	if ((event->keyval < 0x100) && (event->state <= 1))
	{	// the key is a letter and the only possible modifier is Shift
		if (e2_option_bool_get ("use-vi-keys"))
		{
			switch (event->keyval)
			{
				case 'j':
				{
					gint n = view->row;
					gint m = gtk_tree_model_iter_n_children
						(GTK_TREE_MODEL (view->store), NULL);
					if (n < m-1)
						e2_fileview_focus_row (view, n+1, TRUE, TRUE, TRUE, TRUE);
					return TRUE;
				}
				case 'k':
				{
					gint n = view->row;
					if (n > 0)
						e2_fileview_focus_row (view, n-1, TRUE, TRUE, TRUE, TRUE);
					return TRUE;
				}
				case 'l':
					e2_task_open (NULL, NULL, NULL);
					gtk_widget_grab_focus (curr_view->treeview);
					return TRUE;
				case 'h':
					e2_pane_change_dir (NULL, "..");
					return TRUE;
				case 'g':
					e2_fileview_focus_row (view, 0, TRUE, TRUE, TRUE, TRUE);
					return TRUE;
				case 'G':
					{
						gint n = gtk_tree_model_iter_n_children
							(GTK_TREE_MODEL (view->store), NULL);
						e2_fileview_focus_row (view, n-1, TRUE, TRUE, TRUE, TRUE);
					}
					return TRUE;
				default:
				break;
			}
		}
/*  no need for this: the treeview handles it itself, BUT it fails to set the focus row !!
		else  //not vi keys
		{
			//select row beginning with letter entered, if we can
			gint row = e2_fileview_find_filename_begining_with (event->keyval);
			if (row > -1)
				e2_fileview_focus_row (view, row, TRUE, TRUE, TRUE, TRUE);
			return TRUE;
		} */
	}
	return FALSE;
}

/*static gboolean __e2_fileview_column_button_cb (GtkWidget *widget, gint col, FileView *view)
{  superseded fn
	printd (DEBUG, "callback: column button");
	GtkSortType direction;

	// reverse the sorting direction if the list is already sorted by this col
	if (GTK_CLIST(view->clist)->sort_column == col)
		direction = (GTK_CLIST(view->clist)->sort_type == GTK_SORT_ASCENDING
		 ? GTK_SORT_DESCENDING : GTK_SORT_ASCENDING);
	else
		direction = GTK_CLIST(view->clist)->sort_type;
		
	sort_list(view, all_columns[col].sort_func, direction, col);
	return TRUE;
}

static gboolean column_resize_cb (GtkCList *clist,
	gint column, gint width, gint *width_array)
{
	width_array[column] = width;
	return TRUE;
}  */

// D&D stuff

gchar *uri_list;
gchar *dest_dir;
guint drop_time;
/**
 * @brief get drag op data from menu item and initiate the drag op task
 *
 * @param menuitem the clicked menu item widget
 * @param context drag data context struct
 *
 * @return  TRUE, always
**/
static gboolean _e2_fileview_drag_op_selected_cb (GtkWidget *menuitem, GdkDragContext *context)
{
 	//get code for backend function (copy etc) to be performed on
	//the selected items, or -1 if the user cancelled
	GdkDragAction chosen = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (menuitem), "op-func"));
	gtk_widget_destroy (app.drag_op_menu);
	if (chosen != -1)
	{	//the user did not cancel
		gpointer func;
		switch (chosen)
		{
//copy & wait for signal from dest seems not to work
			case GDK_ACTION_MOVE:
				func = e2_task_backend_move;
				break;
			case GDK_ACTION_LINK:
				func = e2_task_backend_link;
				break;
			default:
				func = e2_task_backend_copy;
				break;
		}
		gboolean result = e2_task_drop (func, uri_list, dest_dir);
//		gboolean delete = (chosen & GDK_ACTION_MOVE) && result;
		context->action = chosen;	//CHECKME not needed if cancelled ?
//		gtk_drag_finish (context, result, delete, drop_time);	//CHECKME need timestamp from the "drag_data_drop" signal ?
		gtk_drag_finish (context, result, FALSE, drop_time);
	}
	g_free (dest_dir); //this was allocated in drag_data_received_cb(), before menu was created
	g_free (uri_list);	//this was dup'd in drag_data_received_cb(), before menu was created
	return TRUE;
}
/**
 * @brief execute drop
 *
 * This is the callback for "drag_data_received" signals, emitted
 * on @a treeview when dragged data has been received
 * It initiates action for the dropped items
 * dest_dir used here is a newly-allocated string, must be freed
 *
 * @param treeview widget that received the signal
 * @param context the dragcontext
 * @param x where the drop happened, mouse-cursor-left
 * @param y where the drop happened, mouse-cursor-BOTTOM
 * @param data pointer to selected items data struct
 * @param drag_info the info that has been registered with the target in the GtkTargetList
 * @param time the timestamp at which the data was received
 * @param view pointer to FileView data struct for the pane shown in @a treeview
 * 
 * @return 
**/
static void _e2_fileview_drag_data_received_cb (
		GtkWidget *treeview,
		GdkDragContext *context,
		gint x,
		gint y,
		GtkSelectionData *data,
		guint drag_info,
		guint time,
		FileView *view)
{
	printd (DEBUG, "callback: drag data received");
	FileInfo *info;
//	gchar *dest_dir; use global instead

	//create relevant drop-directory path string
	gboolean pane_dir = TRUE;
	GtkTreePath *path;
	if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
		x, y-pointer_height, &path, NULL, NULL, NULL))
	{
		GtkTreeIter iter;
		gchar *filename;
		GtkTreeModel *model = GTK_TREE_MODEL (view->store);
		if (gtk_tree_model_get_iter (model, &iter, path))
		{			
			gtk_tree_model_get (model, &iter, FILENAME, &filename, FINFO, &info, -1);
			//NICE to add check for executable, here, and do a different drop menu
			//check if dropping onto a dir
			if (e2_fs_is_dir (info))
			{
				//now that we're finishing, un-light the row
				_e2_fileview_clear_row_background (view, &iter);
				//append target dir's name to the pane dir path
				//view->dir, filename both utf, both have trailing /
				dest_dir = g_strconcat (view->dir, filename, NULL);
				pane_dir = FALSE;
			}
			g_free (filename);
		}
		gtk_tree_path_free (path);
	}
	//default is to just drop to the pane dir
	if (pane_dir)
	{
		dest_dir = g_strdup (view->dir);	//utf, trailing /
	}

	//get the sender's or user's choice on what to do with the 
	//dragged items, and do it
	gboolean menu;
	if (btn2_released)
		menu = TRUE;
	else
	{
		GdkDisplay *display = gdk_display_manager_get_default_display (gdk_display_manager_get());
		GdkModifierType mask;
		gdk_display_get_pointer (display, NULL, NULL, NULL, &mask);
		guint modifiers = gtk_accelerator_get_default_mod_mask ();
		menu = ((mask & modifiers) == GDK_MOD1_MASK); //alt key pressed
	}
	//make data (if any) available to the handler
	uri_list = (gchar *)data->data;
	if (*uri_list == '\0')
	{
		g_free (dest_dir);
		return;
	}

	GdkDragAction action = context->actions;
	if (menu && (action & GDK_ACTION_ASK))
	{
		//if not processing immediately, prevent data from getting lost
		uri_list = g_strdup (uri_list);
		drop_time = time;

		if (app.drag_op_menu != NULL)
			gtk_widget_destroy (app.drag_op_menu);

		app.drag_op_menu = gtk_menu_new();

		GtkWidget *item;

		if (action & GDK_ACTION_COPY)
		{
			item = e2_menu_add (app.drag_op_menu, _("_Copy"), NULL, NULL,
				_e2_fileview_drag_op_selected_cb, context);
			g_object_set_data (G_OBJECT (item), "op-func", GINT_TO_POINTER (GDK_ACTION_COPY));
		}

		if (action & GDK_ACTION_MOVE)
		{
			item = e2_menu_add (app.drag_op_menu, _("_Move"), NULL, NULL,
				_e2_fileview_drag_op_selected_cb, context);
			g_object_set_data (G_OBJECT (item), "op-func", GINT_TO_POINTER (GDK_ACTION_MOVE));
		}

		if (action & GDK_ACTION_LINK)
		{
			item = e2_menu_add (app.drag_op_menu, _("_Link"), NULL, NULL,
				_e2_fileview_drag_op_selected_cb, context);
			g_object_set_data (G_OBJECT (item), "op-func", GINT_TO_POINTER (GDK_ACTION_LINK));
		}

		item = e2_menu_add (app.drag_op_menu, _("C_ancel"), NULL, NULL,
			_e2_fileview_drag_op_selected_cb, context);
		g_object_set_data (G_OBJECT (item), "op-func", GINT_TO_POINTER (-1)); //this signals 'cancel' to the handler

		gtk_menu_popup (GTK_MENU (app.drag_op_menu), NULL, NULL, NULL, NULL, 0, (guint32) time);
	}
	else	//don't ask
	{
		gboolean result;
		action = context->suggested_action;

		if (action & GDK_ACTION_COPY)
			result = e2_task_drop (e2_task_backend_copy, uri_list, dest_dir);
		else if (action & GDK_ACTION_MOVE)
			//copying and waiting for signal from dest seems not to work
			result = e2_task_drop (e2_task_backend_move, uri_list, dest_dir);
		else if (action & GDK_ACTION_LINK)
			result = e2_task_drop (e2_task_backend_link, uri_list, dest_dir);
//		gboolean delete = (action & GDK_ACTION_MOVE) && result;
		context->action = action;
		g_free (dest_dir);
//		gtk_drag_finish (context, result, delete, time);
		gtk_drag_finish (context, result, FALSE, time);
	}
}
/**
 * @brief "drag_begin" signal callback
 *
 * set dnd pointer height
 * set relevant drag icon
 *
 * @param treeview widget where drag started unused
 * @param dc dragcontext
 * @param view rt data for view where drag started unused
 * 
 * @return
**/
static void _e2_fileview_drag_begin_cb (GtkWidget *treeview,
	GdkDragContext *dc, FileView *view)
{
	printd (DEBUG, "drag begin");
	GtkSettings *s = gtk_settings_get_default ();
	gint pointer_width;
	if (!gtk_icon_size_lookup_for_settings (s, GTK_ICON_SIZE_LARGE_TOOLBAR, //GTK_ICON_SIZE_DND is too big !
		&pointer_width, &pointer_height))
			pointer_height = POINTER_HEIGHT;	//can't find useful size, use the default
	GtkTreeSelection *sel = gtk_tree_view_get_selection (GTK_TREE_VIEW(treeview));	
	gint count = gtk_tree_selection_count_selected_rows (sel);
	if (count > 1)
		gtk_drag_set_icon_stock (dc, GTK_STOCK_DND_MULTIPLE, 0, 1);	
	else //no point in checking for > 0  as default will be used, anyway
		gtk_drag_set_icon_stock (dc, GTK_STOCK_DND, 0, 1);	
	return;
}
/**
 * @brief UNUSED "drag-data-delete" signal callback
 *
 * This is called when a drag of type GDK_ACTION_MOVE 
 * was successfully completed.
 *
 * @param treeview widget where drag started unused
 * @param dc dragcontext
 * @param view rt data for view where drag started
 * 
 * @return
**/
/*static void _e2_fileview_drag_delete_cb (GtkWidget *treeview,
	GdkDragContext *dc, FileView *view)
{
	printd (DEBUG, "drag data delete");
	e2_task_delete (NULL, NULL, NULL);	//delete active-pane selected items
	return;
} */
/**
 * @brief callback for "drag-data-get" signal
 *
 * This is the emitted on the drag source
 * when the drop site requests the data which is dragged.
 * It builds a URI list of the selected filenames to set as the drag data
 * Any directory item will not have a trailing separator, as the names
 * are freshly converted from the relevant FileInfo field
 *
 * @param widget the treeview which received the signal unused
 * @param context the drag context unused
 * @param data the GtkSelectionData to be filled with the dragged data
 * @param info_arg the info that has been registered with the target in the GtkTargetList.
 * @param time the timestamp at which the data was requested
 * @param view rt data structure for the view 
 * 
 * @return 
**/
static void _e2_fileview_drag_data_get_cb (
		GtkWidget            *widget,
		GdkDragContext       *context,
		GtkSelectionData     *data,
		guint                info_arg,
		guint               time,
		FileView        *view)
{
	printd (DEBUG, "callback: drag data get");

#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, drag date get");
#endif
	e2_filelist_disable_refresh();
	GList *base, *tmp;
	base = tmp = e2_fileview_get_selected_local (view);
	if (base != NULL)
	{	//there is actually a selection to process ...
		F_DECLARE;
		FileInfo *info;
		gchar *utf;
		GString *uri_list = g_string_sized_new (PATH_MAX);
	
	/* this is the emerging standard for uir list entries ?
	<?xml version="1.0" encoding="UTF-8"?>
   <resource-lists xmlns="urn:ietf:params:xml:ns:resource-lists" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\r\n
	 <list>\r\n
	   <entry uri="sip:bill@example.com" />\r\n
	   <entry uri="sip:joe@example.org" />\r\n
	   <entry uri="sip:ted@example.net" />\r\n
	 </list>\r\n
   </resource-lists>\r\n
for now, at least, in accord with old w3c standard (which is currently under review)
/foo/bar => file:///foo/bar
OR
/foo/bar => file://localhost/foo/bar
NOTE some apps omit the leading "//" from the first example, resulting in file:/foo/bar
*/
		for (; tmp != NULL; tmp = tmp->next)
		{
			info = tmp->data;
			utf = F_FILENAME_FROM_LOCALE (info->filename);
			g_string_append_printf (uri_list, "file://%s%s\r\n", view->dir, utf);
			F_FREE (utf);
		}
	
		gtk_selection_data_set (data, data->target, 8, (guchar *)uri_list->str, uri_list->len);
		g_string_free (uri_list, TRUE);
		g_list_free (base);
	}
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, drag date get");
#endif
	e2_filelist_enable_refresh ();
}
/**
 * @brief highlight drop zone
 *
 * This is the callback for "drag_motion" signals
 * It is called when the user moves the cursor over the drop-widget
 * during a drag
 * It handles the highlighting of drop-target rows
 *
 * @param treeview widget ptr
 * @param context drag context
 * @param x the left coordinate of the current cursor position
 * @param y the BOTTOM coordinate of the current cursor position
 * @param time the timestamp of the motion event
 * @param view rt data for the view where the drag [started?]
 * 
 * @return  TRUE, ie the whole widget is a valid drop zone
**/
static gboolean _e2_fileview_drag_motion_cb (GtkWidget *treeview,
	GdkDragContext *context, gint x, gint y, guint time, FileView *view)
{
	printd (DEBUG, "callback: drag motion");
	GtkTreeIter iter;
	GtkTreePath *path;
	GtkTreeModel *model = GTK_TREE_MODEL (view->store);
	//select the path of the item under the cursor TOP ie y-pointer_height
	if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
		x, y-pointer_height, &path, NULL, NULL, NULL))
	{
		//list store path depth is 1, so row no. is 1st index
		gint row = *gtk_tree_path_get_indices (path);
		if (row != view->drop_row)
		{	//mouse has moved, possibly to another row
			if (gtk_tree_model_iter_nth_child (model, &iter, NULL, view->drop_row))
			{
//				gtk_list_store_set (view->store, &iter, BACKCOLOR, &LIST_COLOR, -1);
				if (view->lit)
					_e2_fileview_clear_row_background (view, &iter);
				if (gtk_tree_model_get_iter (model, &iter, path))
				{
					//only highlight dirs (either pane)
					FileInfo *info;
					gtk_tree_model_get (model, &iter, FINFO, &info, -1);
					if (e2_fs_is_dir (info))
					{
//						gtk_list_store_set (view->store, &iter, BACKCOLOR, 
//							e2_option_color_get ("color-drag-highlight"), -1);
						_e2_fileview_set_row_background (view, &iter, e2_option_color_get ("color-drag-highlight"));
					}
					view->drop_row = row;
				}
			}
		}
	}
	else //there is no row under the mouse
		if (view->lit
			&& gtk_tree_model_iter_nth_child (model, &iter, NULL, view->drop_row))
			//restore the color of the last row that was lit
//			gtk_list_store_set (view->store, &iter, BACKCOLOR, &LIST_COLOR, -1);
			_e2_fileview_clear_row_background (view, &iter);

	gtk_tree_path_free (path);
	return TRUE;  
}
/**
 * @brief set background color of the drop target row back to the default value
 *
 * This is the callback for "drag-leave" signals
 *
 * @param treeview widget where the drop occurred unused
 * @param context  dragcontext unused
 * @param time unused
 * @param view rt data for the view where the drop ocurred
 * 
 * @return  TRUE if the conversion completed successfully, else FALSE
**/
static gboolean _e2_fileview_drag_leave_cb (GtkWidget *treeview, 
	GdkDragContext *context, guint time, FileView *view)
{
	GtkTreeIter iter;
	if (! gtk_tree_model_iter_nth_child
		(GTK_TREE_MODEL (view->store), &iter, NULL, view->drop_row))
			return FALSE;  //the specified row doesn't exist
//	gtk_list_store_set (view->store, &iter, BACKCOLOR, &LIST_COLOR, -1);
	if (view->lit)
		_e2_fileview_clear_row_background (view, &iter);

	return TRUE;
}

// end of DND stuff

/**
 * @brief set background color of treeview line
 *
 *
 * @param view data struct for the view to process
 * @param iter pointer to iter for the row to be changed
 * @param color pointer to struct with data for the new color
 *
 * @return
**/
static void _e2_fileview_set_row_background (FileView *view,
	GtkTreeIter *iter, GdkColor *color)
{
	gtk_list_store_set (GTK_LIST_STORE (view->store), iter, BACKCOLOR, color, -1);

	GList* columns, *renderers, *base1, *base2;
	columns = base1 = gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	for (; columns != NULL ; columns = columns->next)
	{
		GtkTreeViewColumn *column = columns->data;
		{
			renderers = base2 = gtk_tree_view_column_get_cell_renderers (column);
			for (; renderers != NULL ; renderers = renderers->next)
				gtk_tree_view_column_add_attribute (column,
					(GtkCellRenderer *) renderers->data,
					"cell-background-gdk", BACKCOLOR);
			g_list_free (base2);
		}
	}
	g_list_free (base1);
	view->lit = TRUE;
}
/**
 * @brief revert background color of treeview line
 *
 *
 * @param view data struct for the view to process
 * @param iter pointer to iter for the row to be changed
 *
 * @return
**/
static void _e2_fileview_clear_row_background (FileView *view, GtkTreeIter *iter)
{
	//make sure the row does not show if another row is colored
	gtk_list_store_set (GTK_LIST_STORE (view->store), iter, BACKCOLOR, NULL, -1);
	//unfortunately, no way to remove a single attribute, so set
	//all attributes the way they were when the treeview was created
	GList* columns, *renderers, *base1, *base2;
	columns = base1 = gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	for (; columns != NULL ; columns = columns->next)
	{
		GtkTreeViewColumn *column = columns->data;
		//the column number is 'textified' as the title
		const gchar *title = gtk_tree_view_column_get_title (column);
		gint m = title[0] - '0';
		renderers = base2 = gtk_tree_view_column_get_cell_renderers (column);
		for (; renderers != NULL ; renderers = renderers->next)
			gtk_tree_view_column_set_attributes (column,
					(GtkCellRenderer *) renderers->data,
					"text", m,
					"foreground-gdk", FORECOLOR,
					NULL);
			g_list_free (base2);
	}
	g_list_free (base1);
	view->lit = FALSE;
}

#ifdef E2_SELTXT_RECOLOR
/**
 * @brief treeview selection function
 *
 * This toggles the foreground text color for the selected item
 * It is called prior to the selection-state change
 *
 * @param selection the selection object (unused)
 * @param model the tree model
 * @param path path to item whose state is to change
 * @param path_currently_selected TRUE if the row is about to be UNselected (unused)
 * @param view  rt data for the view,  assigned when function was initiated (unused)
 *
 * @return TRUE to allow selection state to change
**/
static gboolean _e2_fileview_selection_func (GtkTreeSelection *selection,
	GtkTreeModel *model, GtkTreePath *path,
	gboolean path_currently_selected, FileView *view)
{
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter (model, &iter, path))
	{
		GdkColor *forenow, *selnow;  //, *temp;
		gtk_tree_model_get (model, &iter, FORECOLOR, &forenow, SELCOLOR, &selnow, -1);
		if (path_currently_selected)
		//revert the stored color
			forenow = selnow;
		else
		//impose the selected text color			
			forenow = &selectedtext;

		gtk_list_store_set (GTK_LIST_STORE (model), &iter, FORECOLOR, forenow, -1);
	}
	return TRUE;
}
#endif

// Public functions

/**
 * @brief display (NOT select) a numbered row, optionally show its middle column center-pane
 *
 *
 * @param view rt data for ther view to be worked on
 * @param row number of the row that is to be focused
 * @param clear_selection TRUE if whole selection for the view is to be un-selected first
 * @param center TRUE if cursor focus should be moved to the center column of the view
 * @param grab_focus TRUE if the treeview is to be given the focus
* 
 * @return
**/
void e2_fileview_focus_row (FileView *view,
		      gint row,
		      gboolean select_row,
		      gboolean clear_selection,
		      gboolean center,
		      gboolean grab_focus)
{
	if (view != NULL)
	{
		GtkTreeIter iter;
		if (! gtk_tree_model_iter_nth_child
			(GTK_TREE_MODEL (view->store), &iter, NULL, row) )
			return;  //the specified row doesn't exist
		view->row = row;	//remember the focus row
		GtkTreePath* path = gtk_tree_model_get_path
			(GTK_TREE_MODEL (view->store), &iter);
		GtkTreeView *tvw = GTK_TREE_VIEW (view->treeview);	
		gfloat h_align = (center) ? 0.5 : 0.0;
		gtk_tree_view_scroll_to_cell (tvw, path, NULL, TRUE, 0.5, h_align);
		GtkTreeSelection *sel = gtk_tree_view_get_selection (tvw);
		if (clear_selection)
			gtk_tree_selection_unselect_all (sel);
		if (select_row)
			gtk_tree_selection_select_path (sel, path);
		if (grab_focus)
			gtk_widget_grab_focus(view->treeview);
		gtk_tree_path_free (path);
	}
}
/**
 * @brief UNUSED select 1st row, in specified view, where the filename matches a specified name
 *
 * model content update is suspended while this is performed
 *
 * @param view rt data for ther view to be worked on
 * @param filename string with name of file to be used for selection
 * 
 * @return
**/
/*void e2_fileview_select_row_by_filename (FileView *view, gchar *filename)
{
	gint i;
	//number of rows in the model
	gint n = gtk_tree_model_iter_n_children
		(GTK_TREE_MODEL (view->store), NULL);
	GtkTreeIter iter;
	GtkTreeModel *model = GTK_TREE_MODEL (view->store);
	gchar *thisname;
	e2_filelist_disable_refresh();
	// search from after the current row to the end of the store
	for (i = 0; i < n; i++)
	{
		if (! gtk_tree_model_iter_nth_child
			(model, &iter, NULL, i))
				continue;  //the specified row doesn't exist any more
		gtk_tree_model_get (model, &iter, FILENAME, &thisname, -1);

		if (g_str_equal (thisname, filename))
		{
			e2_fileview_focus_row (view, i, TRUE, FALSE, TRUE, TRUE);
			g_free (thisname);
			break;
		}
		else
			g_free (thisname);
	}
	e2_filelist_enable_refresh();
} */
/**
 * @brief get file info for 1st tagged or selected row
 *
 *
 * @param view rt data for ther view to be worked on
 * 
 * @return file info structure for the 1st valid selected row, or if none, then NULL
**/
FileInfo *e2_fileview_get_selected_first_local (FileView *view)
{
	FileInfo *info = NULL;
	GtkTreeView *tvw = GTK_TREE_VIEW (view->treeview);	
	GtkTreeSelection *sel = gtk_tree_view_get_selection (tvw);
	GtkTreeModel *model;
	GList *base = gtk_tree_selection_get_selected_rows (sel, &model);
	GList *rowpaths;
	GtkTreeIter iter;
	for (rowpaths=base; rowpaths!=NULL; rowpaths=rowpaths->next)
	{
		GtkTreePath *first_sel = (GtkTreePath *) rowpaths->data;
		if (gtk_tree_model_get_iter (model, &iter, first_sel))
		{
			gtk_tree_model_get (model, &iter, FINFO, &info, -1);
			break;
		}
	}
	g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (base);

	return info;
}

gint list_str_data_compare_func (gconstpointer a, gconstpointer b)
{
	return (strcmp ((gchar *)a, (gchar *)b));
}
/**
 * @brief get list of selected-item names in list store for specified view
 *
 * This logs the names in current sort-order, for speedier re-selection later
 * It is used as part of the view refresh or recreate processes
 * (even with FAM, needed for refresh action, (un)hide, (un)filter,
 *  window re-create)
 *
 * @param view data structure for view being processed
 *
 * @return sorted list of names
 */
GList *e2_fileview_log_selected_names (FileView *view)
{
	GtkTreeIter iter;
	GtkTreeSelection *refresh_sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (view->treeview));
	GtkTreeModel *model;
	GList *base = gtk_tree_selection_get_selected_rows (refresh_sel, &model);
	GList *selected_paths;
	GList *selected_names = NULL;
	gchar *itemname;

	for (selected_paths=base; selected_paths!=NULL; selected_paths=selected_paths->next)
	{
		if (gtk_tree_model_get_iter (model, &iter, selected_paths->data))
  		{
			gtk_tree_model_get (model, &iter, FILENAME, &itemname, -1);
			selected_names = g_list_append (selected_names, itemname);
		}
	}
	g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (base);

	selected_names = g_list_sort (selected_names,
		list_str_data_compare_func);

	return selected_names;
}
/**
 * @brief re-select items in list store for specified view  REPLACED BY INLINE CODE
 *
 *
 * @param view data structure for view being processed
 *
 * @return TRUE to stop the list scan when a match is found, else FALSE
 */
/*static GtkTreeSelection *refresh_sel;

gboolean e2_filelist_find_selected (GtkTreeModel *model,
	GtkTreePath *path, GtkTreeIter *iter, gchar *match_name)
{
	gchar *thisname;
	gtk_tree_model_get (model, iter, FILENAME, &thisname, -1);
	gboolean result = g_str_equal (thisname, match_name);
	if (result)
		//reselect this path
		gtk_tree_selection_select_path (refresh_sel, path);
	g_free (thisname);
	return result;
} */
/*		gint next_scan = 0;
		gint current_row = 0;
//		gchar namestore[E2_UTFNAME_MAX];  gtk won't use this !!
		gchar *itemname;
		gchar *this_name;  // = namestore;  // put this on the stack to save time !!
		for (; selected_names != NULL; selected_names=selected_names->next)
		{
			itemname = (gchar *) selected_names->data;
			do
			{
				gtk_tree_model_get (model, &iter, FILENAME, &this_name, -1);
				if (g_str_equal (this_name, itemname))
				{	//select the row
					gtk_tree_selection_select_iter (refresh_sel, &iter);
					next_scan = (gtk_tree_model_iter_next (model, &iter)) ?
					++current_row : -1;
					g_free (this_name);
	printd (DEBUG, "resel msg 2");
					break;
				}
				g_free (this_name); //not a local variable, now
				current_row++;
			} while (gtk_tree_model_iter_next (model, &iter));
			if (next_scan == -1)
				//found match at end, no more searching
				break;
			else if (next_scan < current_row)
				// reached end without finding a match, go back and restart after last match, for next name
				gtk_tree_model_iter_nth_child (model, &iter, NULL, next_scan);	
			//otherwise, the seach for the next name begins again at the next row,
	printd (DEBUG, "resel msg 3");
		}  */
//REPLACED by code above	gtk_tree_model_foreach (model, (GFunc) e2_filelist_find_selected, (gchar *) selected_names->data);

static gint list_item_data_compare_func (gconstpointer a, gconstpointer b)
{
	gint result =
	strcmp (((E2_ViewItem *)a)->name, ((E2_ViewItem *)b)->name);
	return result;
}

/**
 * @brief re-select items in list store for specified view
 *
 * This iterates forward through the store, presuming that the loaded order
 * is the same as the logged names
 * If there is nothing to re-select, the first row is selected as a default
 *
 * @param view data structure for view being processed
 * @param list sorted list of names of selected items
 *
 * @return TRUE if reselection not needed, or completed successfully, FALSE if store is empty
 */
gboolean e2_fileview_reselect_names (FileView *view, GList *selected_names)
{
	if (selected_names==NULL)
		return TRUE;
	GList *base=selected_names;
	GList *new_items = NULL;
	GList *base2=NULL;
	GList *next_start = NULL;
	GtkTreeIter iter;
	GtkTreeModel *model = GTK_TREE_MODEL (view->store);
	GtkTreeSelection *refresh_sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (view->treeview));
	if (gtk_tree_model_get_iter_first (model, &iter))
	{
		//grab a name-sorted list of new names and corresponding paths
		do
		{
			E2_ViewItem *itemdata = g_malloc (sizeof(E2_ViewItem));
			gtk_tree_model_get (model, &iter, FILENAME, &(itemdata->name), -1);
			itemdata->path = gtk_tree_model_get_path (model, &iter);
			new_items = g_list_append (new_items, itemdata);
		} while (gtk_tree_model_iter_next (model, &iter));

		new_items = base2 = next_start = g_list_sort (new_items,
			list_item_data_compare_func);

		//now we have sorted old names and new names,
		//walk through old ones
		for (; selected_names != NULL; selected_names=selected_names->next)
		{
			for (new_items=next_start; new_items != NULL;
					new_items=new_items->next)
			{
				gchar *oldname=(gchar *)selected_names->data;
				E2_ViewItem *newitem = (E2_ViewItem *)new_items->data;
				gchar *newname=newitem->name;
				gint compare = strcmp (newname, oldname);
				if (compare < 0)
					continue;	//keep looking
				else if (compare == 0)
				{
					GtkTreePath *path = newitem->path;
					gtk_tree_selection_select_path (refresh_sel, path);
					next_start=new_items->next;  //resume scanning from the same place
					break;
				}
				else
				{	//not in the list anymore
					next_start=new_items;  //keep scanning from the same place
					break;
				}
			}
		}
	}
	//cleanup
	g_list_foreach (base, (GFunc) g_free, NULL);
	g_list_free (base);
	
	if (base2 != NULL)
	{
		for (new_items=base2; new_items != NULL;
					new_items=new_items->next)
		{
			E2_ViewItem *newitem = (E2_ViewItem *)new_items->data;
			g_free (newitem->name);
			gtk_tree_path_free (newitem->path);
			g_free (newitem);
		}
		g_list_free (base2);
	}
	if (!gtk_tree_model_get_iter_first (model, &iter))
	{
		gtk_widget_grab_focus (view->treeview);
		return FALSE;
	}
/* don't select anything else
	if (gtk_tree_selection_count_selected_rows (refresh_sel) == 0)
	{
		gtk_tree_selection_select_iter (refresh_sel, &iter);
		view->row = 0;
	} */
	return TRUE;
}

/**
 * @brief get list of selected rows' info items
 *
 * This is usable only when the data is used before any chance
 * of filelist refresh, which would invalidate the listed data pointers
 *
 * @param view rt data for the view to be worked on
 * 
 * @return glist with the selected rows' info items, excluding ".."
**/
GList *e2_fileview_get_selected_local (FileView *view)
{
//	g_signal_emit_by_name (G_OBJECT (view->treeview), "end-selection");

//	if (view->tagged != NULL)
//		return g_list_copy (view->tagged);

	GList *selectedrow_data = NULL;
	FileInfo *info;
	GtkTreePath *path;
	GtkTreeIter iter;
	GtkTreeView *tvw = GTK_TREE_VIEW (view->treeview);	
	GtkTreeSelection *sel = gtk_tree_view_get_selection (tvw);
	GtkTreeModel *model;
	GList *base = gtk_tree_selection_get_selected_rows (sel, &model);
	GList *rowpaths;

	for (rowpaths=base; rowpaths!=NULL; rowpaths=rowpaths->next)
	{
		path = (GtkTreePath *) rowpaths->data;
		if (gtk_tree_model_get_iter (model, &iter, path))
		{
			gtk_tree_model_get (model, &iter, FINFO, &info, -1);
			if (!g_str_equal (info->filename, ".."))	//CHECKME ok, even with localised string?
				selectedrow_data = g_list_append (selectedrow_data, info);
		}
	}
	g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (base);
	return selectedrow_data;
}
/**
 * @brief translate array into another form sorted in index order
 *
 * This is needed to convert the 'fixed' column order used in data models
 * and config/cache data, to variable order in accord with the displayed view columns
 * and vice versa
 *
 * @param src_array pointer to array of ints that is to be translated
 * @param dest_array pointer to array of ints that stores the translated order
 * @param size the number of elements to be translated
 *
 * @return
 */
void e2_fileview_translate_cols_array (gint *src_array, gint *dest_array,
 gint size)
{
	gint i;
	for (i=0; i< size; i++)
		dest_array[src_array[i]] = i;
}
/**
 * @brief exchange active and inactive panes
 *
 * This function is called from e2_pane_activate_other (), i.e.
 * whenever we need to exchange curr_view and other_view.
 * It toggles pointers and changes colours
 * The latter is done by interrogating the selections in both file views,
 * and so this fn should be blocked if either view is still being
 * populated at the time this is called ?? CHECKME
 *
 * @return
 */
void e2_fileview_switch_views (void)
{
/*	if (curr_view == app.pane1.view)
		printd (DEBUG, "switch views - from 1 to 2");
	else if (curr_view == app.pane2.view)
		printd (DEBUG, "switch views - from 2 to 1");
	else
		printd (DEBUG, "switch views - unknown");
*/
	FileView *temp;
//	GList *tmp;

	if (! e2_fs_chdir (other_view->dir))
		return;
	
/* THIS DOES NOT WORK
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *sel;
	GList *paths, *base;

	E2_OptionSet *set = e2_option_get ("active-pane-signal");
	if (_e2_option_int_get (set) == 2)	//want to change sensitivity
	{	//so change the colours before-hand
#ifdef E2_REFRESH_DEBUG
		printd (DEBUG, "disable refresh, switch views");
#endif
		e2_filelist_disable_refresh ();

		GdkColor *background = e2_option_color_get ("color-inactive-selected");

//	model = gtk_tree_view_get_model (GTK_TREE_VIEW (curr_view->treeview));
		sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (curr_view->treeview));
		base = gtk_tree_selection_get_selected_rows (sel, &model);
		for (paths = base; paths != NULL; paths = paths->next)
		{
			gtk_tree_model_get_iter (model, &iter, paths->data);
			gtk_list_store_set (curr_view->store, &iter, BACKCOLOR, &LIST_COLOR, -1);
		}
		g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
		g_list_free (base);

//	_e2_fileview_clear_row_background (curr_view, &iter);

//	model = gtk_tree_view_get_model (GTK_TREE_VIEW (other_view->treeview));
		sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (other_view->treeview));
		base = gtk_tree_selection_get_selected_rows (sel, &model);
		for (paths = base; paths != NULL; paths = paths->next)
		{
			gtk_tree_model_get_iter (model, &iter, paths->data);
			gtk_list_store_set (other_view->store, &iter, BACKCOLOR, background, -1);
//			_e2_fileview_set_background (GTK_TREE_VIEW (other_view->treeview),
//				&iter, background);
		}
		g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
		g_list_free (base);

#ifdef E2_REFRESH_DEBUG
		printd (DEBUG, "enable refresh, switch views");
#endif
		e2_filelist_enable_refresh ();
	}
*/
	temp = other_view;
	other_view = curr_view;
	curr_view = temp;

	e2_pane_flag_active ();	//change the status-indicator for the panes
}
/**
 * @brief UNUSED swap view pointers 
 *
 * Exchange curr_view and other_view, 
 * without doing anything else
 *
 * @return 
 **/
/*void e2_fileview_switch_views_simple (void)
{
	FileView *temp = other_view;
	other_view = curr_view;
	curr_view = temp;
}*/
/**
 * @brief helper to set file-list font
 * @param view pointer to data struct for the view to be changed
 * @param fontstr name of font to apply to the filelist associated with @a view
 *
 * @return
**/
static void _e2_fileview_set_font (FileView *view, gchar *fontstr)	
{
	GList* columns, *renderers, *base1, *base2;
	columns = base1 = gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	for (; columns != NULL ; columns = columns->next)
	{
		GtkTreeViewColumn *column = columns->data;
		renderers = base2 = gtk_tree_view_column_get_cell_renderers (column);
		for (; renderers != NULL ; renderers = renderers->next)
			g_object_set (renderers->data, "font", fontstr, NULL);	
		g_list_free (base2);
	}
	g_list_free (base1);
}
/**
 * @brief set file-lists font
 * This is needed to change font after a config dialog that does
 * not rebuild the whole window
 * @return
**/
void e2_fileview_set_font (void)
{
	gchar *fontstr = e2_option_str_get ("list-font");
	_e2_fileview_set_font (curr_view, fontstr);
	_e2_fileview_set_font (other_view, fontstr);
}
/**
 * @brief update array of displayed columns for a view
 *
 * this fn interates over pane columns glist, grabbing each title
 * like "4", converting it to corresponding integer, and storing in array
 *
 * @param rt data structure for the pane to which the file list belongs
 *
 * @return
 */
void e2_fileview_set_col_order (E2_PaneRuntime *rt)
{
	gboolean first_pane = (strchr (rt->name, '1') != NULL);	//always ascii
	gint *order_array = (first_pane) ? displayed_col_order[0] : displayed_col_order[1];
	GList *cols = gtk_tree_view_get_columns (GTK_TREE_VIEW (rt->view->treeview));
	GList *base = cols;
	G_CONST_RETURN gchar *title;
	gint colnum = 0;
	for ( ; cols != NULL; cols = cols->next )
	{
		title = gtk_tree_view_column_get_title (cols->data);
		order_array[colnum++] = atoi (title);
	}
	g_list_free (base);
}
/**
 * @brief update columns data, that is not automatically updated, for caching
 *
 * Columns order and widths are udated and stored in a list
 * List format is: pane1 {order, width, ...}, pane2 {order, width, ...}
 *
 * @return
**/
void e2_fileview_update_col_cachedata (void)
{
	//get current column widths
	GtkTreeViewColumn *col;
	G_CONST_RETURN gchar *title;
	gint i;
	FileView *view;
	view = app.pane1.view;
	GList *columns = 	gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	GList *base = columns;
	for (; columns != NULL; columns=columns->next)
	{
		col = columns->data;
		title = gtk_tree_view_column_get_title (col);
		//title is "0" ... "7", for cols 0 ... 7
		i = atoi (title);
		col_width_store [0][i] = gtk_tree_view_column_get_width (col);
	}
	g_list_free (base);

	view = app.pane2.view;
	columns = 	gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	base = columns;
	for (; columns != NULL; columns=columns->next)
	{
		col = columns->data;
		title = gtk_tree_view_column_get_title (col);
		i = atoi (title);
		col_width_store [1][i] = gtk_tree_view_column_get_width (col);
	}
	g_list_free (base);

	//get columns displayed order
	e2_fileview_translate_cols_array (displayed_col_order[0], stored_col_order[0], MAX_COLUMNS);
	e2_fileview_translate_cols_array (displayed_col_order[1], stored_col_order[1], MAX_COLUMNS);

	//setup for list-cache writing
	//start fresh list
	e2_list_free_with_data (&cols_data);
	gchar *buffer;
	gint j;
	for (j = 0; j < 2; j++)
	{ for (i = 0; i < MAX_COLUMNS; i++) {
		//data need to be strings, for the cache writer
		buffer = g_strdup_printf ("%d", stored_col_order[j][i]);
		cols_data = g_list_append (cols_data, buffer);
		buffer = g_strdup_printf ("%d", col_width_store[j][i]);
		cols_data = g_list_append (cols_data, buffer);
	}}
}
/**
 * @brief UNUSED find which view column shows filenames
 *
 * Columns may be in any order
 * Scans column titles, looking for "0"
 *
 * @param view data structure for the view to which the file list belongs
 *
 * @return integer column number, 0 to MAX_COLUMNS-1
 */
/* gint e2_fileview_find_name_col (FileView *view)
{
	GList *cols = gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	GList *base = cols;
	G_CONST_RETURN gchar *title;
	gint result = 0;
	for ( ; cols != NULL; cols = cols->next )
	{
		title = gtk_tree_view_column_get_title (cols->data);
		if (g_str_equal (title, "0"))  break;
		result++;
	}
	g_list_free (base);
//	return (result < MAX_COLUMNS) ? result : -1;
	return result;
} */
/**
 * @brief  create file view for the specified pane
 *
 * This creates the on-screen elements of a file view, for a specified pane
 * Should be run only once for each pane
 * Column order is determined by array displayed_col_order[].
 *  Columns are named "0" to "7" for later searching.
 * Sets default background color
 * The associated treemodel must exist when this is called
 *
 * @param rt runtime data for pane containing the view
 *
 * @return the treeview widget
 **/
GtkWidget *e2_fileview_create_view (E2_PaneRuntime *rt)
{
	gint i, m;
	gchar *option_name;
	E2_OptionSet *set;
	gboolean show_this;

	gboolean first_pane = (strchr (rt->name, '1') != NULL);	//always ascii
	gint *width_array = (first_pane) ? col_width_store[0] : col_width_store[1];
	gint *order_array = (first_pane) ? displayed_col_order[0] : displayed_col_order[1];

	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeSelection *sel;
	gchar *fontstr = e2_option_str_get ("list-font");
	static gchar id[] = {'0', '\0'};  //for column header tiles, used to check which column

	GtkTreeModel *mdl = GTK_TREE_MODEL (rt->view->store);
	GtkWidget *tvw = gtk_tree_view_new_with_model (mdl);
	g_object_unref (rt->view->store);  //don't want the extra ref count from the view attachment
	//save it for others to use
	rt->view->treeview = tvw;

	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (mdl);
//	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (rt->view->sortmodel);
	//setup for non-sorted dir loads
	gtk_tree_sortable_set_default_sort_func (sortable, NULL, NULL, NULL);

	//set general treeview properties
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (tvw), TRUE);
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (tvw),	//useless with managed backcolors 
		e2_option_bool_get ("panes-hinted"));
	sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (tvw));
	gtk_tree_selection_set_mode (sel, GTK_SELECTION_MULTIPLE);
#ifdef E2_SELTXT_RECOLOR
	if (e2_option_bool_get ("selectedtext-colorchange"))
		gtk_tree_selection_set_select_function (sel,
			(void *) _e2_fileview_selection_func, rt->view, NULL);
#endif

	//set columns' properties
	for (i = 0; i < MAX_COLUMNS; i++)
	{
		m = order_array[i];  //get the model/'external' index
		gtk_tree_sortable_set_sort_func (sortable, m,
			e2_all_columns[m].sort_func, &rt->view->sort_order, NULL);

		//get the column visibility
		option_name = g_strdup_printf ("%s-show-column%d", rt->name, m);
		// e2_option_bool_get  defaults to FALSE
		set = e2_option_get (option_name);
		show_this = (set != NULL) ? _e2_option_bool_get (set) : TRUE;
		g_free (option_name);

		//create header widget, to support styling and avoid gtk bug with arrows
		GtkWidget *headerbox = gtk_hbox_new (FALSE, 1);
		GtkWidget *label = gtk_label_new (gettext(e2_all_columns[m].title)); //same title for each pane
		//expand = TRUE prevents col width < title width + sort indicator (if any)
		//spacing 0 when label expand = FALSE (or ese it crashes when col is too narrow to show all)
		gtk_box_pack_start (GTK_BOX(headerbox), label, FALSE, TRUE, 0);
		gtk_widget_show (label);
		if (m == FILENAME)
			//remember the Filename column label in case we need to use it for active-pane flagging
			rt->view->name_label = GTK_LABEL (label);
		rt->view->sort_arrows[m] = gtk_arrow_new (GTK_ARROW_DOWN, GTK_SHADOW_IN);
		//this doesn't show the indicator at the end of the header ! (and different exp, fill don't help)
		gtk_box_pack_end (GTK_BOX(headerbox), rt->view->sort_arrows[m], FALSE, FALSE, 0);
		gtk_widget_show(headerbox);

		renderer = gtk_cell_renderer_text_new ();
		g_object_set (renderer, 
			"yalign", 0.0,
			"ypad", 0, //supposedly default, but this makes the rows closer
			"font", fontstr,	//this property not stored in model, as it doesn't change on the fly ...
			NULL);
//		gtk_cell_renderer_text_set_fixed_height_from_font
//			(GTK_CELL_RENDERER_TEXT (renderer), 1);
		//right-align the size column only
		if (m == SIZE)
			g_object_set (renderer, 
			"xalign", 1.0,
			"xpad", 0,
			NULL);
		
  		id[0] = m+'0';
		column = gtk_tree_view_column_new_with_attributes (
			id, //this is for seaching & sorting, displayed title is in widget label
			renderer,
			"text", m,
			"foreground-gdk", FORECOLOR,
//			"cell-background-gdk", BACKCOLOR,
			NULL);

		gint thiswidth = width_array[m];
		//doesn't like 0 widths set by gtk when a col is hidden
		if (thiswidth == 0)
			thiswidth = e2_all_columns[m].size;

		g_object_set (column,
			"sizing", GTK_TREE_VIEW_COLUMN_FIXED,
			"resizable", TRUE,
			"fixed-width", thiswidth,   //width_array[m],
			"visible", show_this,
			"clickable", TRUE,
			"reorderable", TRUE,
			"widget", headerbox,
			NULL);

		gtk_tree_view_append_column (GTK_TREE_VIEW (tvw), column);

		g_signal_connect (column, "clicked",
		 G_CALLBACK (_e2_fileview_column_header_clicked_cb), rt);
		 
		//NB undocumented ->button
		//workaround for detecting header clicks to distinguish from drags
		g_signal_connect (column->button, "button_press_event",
		 G_CALLBACK (_e2_fileview_column_header_buttonpress_cb), rt);
 		g_signal_connect (column->button, "button_release_event",
		 G_CALLBACK (_e2_fileview_column_header_buttonrel_cb), rt); 

	}
	//setup inital sort parameters, ascending name col
//	order = TRUE;
	rt->view->sort_order = GTK_SORT_ASCENDING;
//	m = e2_fileview_find_name_col (rt->view);
//	gtk_widget_show (rt->view->sort_arrows[m]);
	gtk_widget_show (rt->view->sort_arrows[0]);

/*  background color now done on-the-fly
	//FIXME do this once (not for both panes)
	//set the background color for resetting from the DRAG_HILIGHT color
	GtkStyle *style = gtk_rc_get_style (tvw);
	LIST_COLOR.red   = style->base[GTK_STATE_NORMAL].red;
	LIST_COLOR.green = style->base[GTK_STATE_NORMAL].green;
	LIST_COLOR.blue  = style->base[GTK_STATE_NORMAL].blue;
*/
	//do this once, too
	button2updir = e2_option_bool_get ("button2-updir");

	g_signal_connect (tvw, "columns-changed",
	 G_CALLBACK(_e2_fileview_col_change_cb), rt);
	g_signal_connect (tvw, "row-activated", 
	 G_CALLBACK (_e2_fileview_row_activated_cb), rt->view);

	//DnD connections
//	gtk_drag_source_set (tvw, GDK_BUTTON1_MASK, target_table, n_targets,
//		GDK_ACTION_COPY); //can't use 2 of this fn, it seems
	gtk_drag_source_set (tvw, GDK_BUTTON1_MASK | GDK_BUTTON2_MASK, target_table, n_targets,
		GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_LINK | GDK_ACTION_ASK);
	gtk_drag_dest_set (tvw, GTK_DEST_DEFAULT_MOTION
		| GTK_DEST_DEFAULT_DROP, target_table, n_targets,
//		, target_table, n_targets,
		GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_LINK | GDK_ACTION_ASK);

	g_signal_connect (G_OBJECT (tvw), "drag_begin",
		G_CALLBACK (_e2_fileview_drag_begin_cb), rt->view);
	g_signal_connect (G_OBJECT (tvw), "drag_data_get",
		G_CALLBACK (_e2_fileview_drag_data_get_cb), rt->view);
	g_signal_connect (G_OBJECT (tvw), "drag_motion",
		G_CALLBACK (_e2_fileview_drag_motion_cb), rt->view);
	g_signal_connect (G_OBJECT (tvw), "drag_leave",
		G_CALLBACK (_e2_fileview_drag_leave_cb), rt->view);
//needed if GTK_DEST_DEFAULT_DROP not set, above
//	g_signal_connect (G_OBJECT (tvw), "drag_drop",
//		G_CALLBACK (_e2_fileview_drag_drop_cb), rt->view);
	g_signal_connect (G_OBJECT (tvw), "drag_data_received",
		 G_CALLBACK(_e2_fileview_drag_data_received_cb), rt->view);
//	g_signal_connect (G_OBJECT (tvw), "drag-data-delete",
//		G_CALLBACK (_e2_fileview_drag_delete_cb), rt->view);

	g_signal_connect (G_OBJECT (tvw), "button_press_event",
		G_CALLBACK (_e2_fileview_mouse_press_cb), rt->view);
	g_signal_connect (G_OBJECT (tvw), "button_release_event",
		G_CALLBACK (_e2_fileview_mouse_release_cb), rt->view);
	g_signal_connect (G_OBJECT (tvw), "popup-menu",
		G_CALLBACK (_e2_fileview_popup_cb), rt->view);
	g_signal_connect (G_OBJECT (tvw), "key_press_event",
		G_CALLBACK (_e2_fileview_key_press_cb), rt->view);

	gtk_widget_show (tvw);
	return tvw;
}
/**
 * @brief  create scrolled window for the specified pane
 *
 * Should be run only once for each pane
 *
 * @param rt runtime data for pane containing the view
 *
 * @return the scrolled window widget
 **/
GtkWidget *e2_fileview_create_sw (E2_PaneRuntime *rt)
{
	//create liststore structure
	e2_filelist_make_store (rt->view);
	//setup column-order array
	gboolean first_pane = (strstr(rt->name, "1") != NULL);
	gint array_row = (first_pane) ? 0 : 1;
	e2_fileview_translate_cols_array (stored_col_order[array_row],
		displayed_col_order[array_row], MAX_COLUMNS);
//	gint i;
//	GList *order = (first_pane) ? pane1_order : pane2_order;
//	for (i=0; i<MAX_COLUMNS; i++, order=order->next)
//		displayed_col_order[array_row][i] = GPOINTER_TO_INT (order->data);
	//create file list view for the pane
	GtkWidget *tvw = e2_fileview_create_view (rt);
	//make it all scrollable
	GtkWidget *sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW(sw),
		GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_placement(GTK_SCROLLED_WINDOW(sw),
		e2_option_int_get ("scrollbar-position"));

	gtk_container_add (GTK_CONTAINER (sw), tvw);
	gtk_widget_show (sw);

	//support for workaround of gtk 'feature' = session startup with vadjustment > 0
//	rt->vadj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (sw));

	return sw; 
}

/**
 * @brief  toggle selection of all items in a pane
 *
 * This function selects or unselects all** items in a pane, depending on
 * the current state of the corresponding flag, pane1_all or pane2_all
 *  ** except '..' entries, (if displayed)
 * The pane will be activated, if it wasn't already.
 * The function handles clicks of the select-all toggle buttons 
 *
 * @param blah unused widget ptr (in case this is a callback)
 * @param view fileview for pane that is to be (un)selected
 *
 * @return 
 **/
void e2_fileview_select_all (GtkWidget *blah, FileView *view)
{
	if (curr_view != view)
		e2_pane_activate_other ();
	GtkTreeSelection *selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (curr_view->treeview));
	gboolean *allnow = (view == app.pane1.view) ? &pane1_all : &pane2_all ;
	if (*allnow)
	{
		gtk_tree_selection_unselect_all (selection);
		*allnow = FALSE;
	}
	else //select
	{
		gtk_tree_selection_select_all (selection);
		if (e2_option_bool_get ("show-updir-entry")
			&& !g_str_equal (curr_view->dir, G_DIR_SEPARATOR_S))  //not at root dir now
		{
//			gtk_clist_unselect_row (GTK_CLIST (curr_view->clist), 0, 0);
//			GtkTreeIter iter = <iter for .. row>;  //not necessarily child 0 ?
//			gtk_tree_selection_unselect_iter (selection, &iter);
		}
		*allnow = TRUE;
	}		
}
/**
 * @brief cleanup content of selected file info struct
 *
 * foreach function for e2_fileview_clean_selected()
 *
 * @param seldata ptr to the struct to be processed
 * @param user_data user data passed to g_ptr_array_foreach() UNUSED
 * 
 * @return
**/
void e2_fileview_selected1_clean (E2_SelectedItemInfo *seldata,
	gpointer user_data)
{
	g_free (seldata->filename);
#ifdef E2_ROWREFS
	gtk_tree_row_reference_free (seldata->ref);
#else
	gtk_tree_path_free (seldata->path);
#endif
}
/**
 * @brief cleanup pointer array of selected file info structs
 *
 *
 * @param selected ptr to the array to be freed
 * 
 * @return
**/
void e2_fileview_clean_selected (GPtrArray *selected)
{
	g_ptr_array_foreach (selected, (GFunc) e2_fileview_selected1_clean, NULL);
	g_ptr_array_free (selected, TRUE);
}
/**
 * @brief create pointer array of selected file info structs
 *
 * This fn creates a pointer array, with pointers to structs containing
 * selected items' names and other info.
 * If found, ".." is filtered out.
 * If found, any trailing / is removed.
 * Item name strings are utf-8
 * Data is duplicated, so the selection is refresh-proof.
 * Treerow refs support incremental store changes as the selection is processed
 * The returned array needs to be freed with its data.
 *
 * @param view is data structure for the view to be interrogated
 * 
 * @return pointer to GPtrArray structure, or NULL if nothing selected
**/
GPtrArray *e2_fileview_get_selected (FileView *view)
{
//	if (view == curr_view)
//		g_signal_emit_by_name (G_OBJECT(view->treeview), "end-selection");
//	F_DECLARE;
	GtkTreeSelection *sel = gtk_tree_view_get_selection 
		(GTK_TREE_VIEW (view->treeview));
	gint count = gtk_tree_selection_count_selected_rows (sel);
	if (count > 0)
	{
		GPtrArray *array = g_ptr_array_sized_new (count);
		GtkTreeModel *model;
		GList *base = gtk_tree_selection_get_selected_rows (sel, &model);
		GList *paths;
		GtkTreeIter iter;
		gchar *filename, *localfilename;
//		FileInfo *info;
		E2_SelectedItemInfo *fileinfo;
		for (paths = base; paths != NULL; paths = paths->next)
		{
			if (!gtk_tree_model_get_iter (model, &iter, paths->data))
				continue;  //the row has been altered since it was selected
//			gtk_tree_model_get (model, &iter, FILENAME, &filename, FINFO, &info, -1);
			gtk_tree_model_get (model, &iter, FILENAME, &filename, -1);
			if (!g_str_equal (filename, ".."))
			{
				fileinfo = g_malloc (sizeof (E2_SelectedItemInfo));
				//always replicate filename
				localfilename = g_strdup (filename);

				if (g_str_has_suffix (localfilename, G_DIR_SEPARATOR_S))
				{	//strip the trailer (can't have it on links, at least)
					gchar *s = localfilename + strlen (localfilename) - 1;
					*s = '\0';
				}
				fileinfo->filename = localfilename;  //this is utf8
//				fileinfo->statbuf = info->statbuf; NO LONGER USED
#ifdef E2_ROWREFS
				//get a reference in case the model changes during the operation on the selection
				fileinfo->ref = gtk_tree_row_reference_new (model, paths->data);
#else
				//this is only relevant if store data not changed on the fly
				fileinfo->path = gtk_tree_path_copy (paths->data);
#endif
				g_ptr_array_add (array, fileinfo);
			}
// always free it now			else
			g_free (filename);
		}
		g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
		g_list_free (base);
		return array;
	}
	return NULL;
}
/**
 * @brief get itemname for a specified view and row
 *
 *
 * @param view is data structure for the view to be interrogated
 * @param row number of the row to be interrogated
 * 
 * @return pointer to duplicated string, or NULL if there's nothing found
**/
gchar *e2_fileview_get_row_name (FileView *view, gint row)
{
	GtkTreeIter iter;
	GtkTreeModel *model = GTK_TREE_MODEL (view->store);
	if (! gtk_tree_model_iter_nth_child (model, &iter, NULL, row))
		return NULL;
	gchar *name;
	gtk_tree_model_get (model, &iter, FILENAME, &name, -1);
	return name;
}
