/* $Id: e2_toolbar.c 1101 2005-08-28 10:55:48Z tpgww $
 *
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/e2_toolbar.c 
 * @brief toolbar functions
 *
 * This file contains the functions for creation, re-creation and
 * manipuation of toolbars and their contents, and bar context menu.
 * Except that initialisation of panebars is largely handled 
 * in e2_pane.c
 **/

#include "emelfm2.h"
#include <string.h>
#include "e2_toolbar.h"
#include "e2_dialog.h"
#include "e2_filelist.h"
#include "e2_complete.h"

//static void resize_cb (GtkWidget *handlebox, GtkAllocation *alloc, E2_ToolbarRuntime *rt);
static GtkWidget *_e2_toolbar_add_button (E2_ToolbarRuntime *rt, gchar *name,
	gchar *icon, gchar *tip, gpointer data);
static void _e2_toolbar_add_menu_items (GtkWidget *menu, 
	GtkTreeModel *model, GtkTreeIter *iter, gint level, E2_ToolbarRuntime *rt);
//static void e2_toolbar_add_items (GtkTreeModel *model, GtkTreeIter *iter, gint level,
//	E2_ToolbarRuntime *rt);

  /*******************/
 /***** actions *****/
/******************/
/**
 * @brief toggle whether specified toolbar exists or not
 *
 * This is the action for 'toggle <toolbar>'
 * 
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
/*static void _e2_toolbar_toggle (E2_ToolbarRuntime *rt)
{
	if (_e2_option_bool_toggle (rt->show))
		e2_toolbar_create (rt);
	else
		e2_toolbar_destroy (rt);
}*/

  /***************************************/
 /***** bar context-menu callbacks *****/
/***************************************/

/**
 * @brief edit the contents of a toolbar, via a config dialog
 *
 * This is performed when the toolbar context-menu item 'edit bar' is selected
 *
 * @param menuitem UNUSED the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime for the bar
 *
 * @return
 **/
static void _e2_toolbar_edit_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	//FIXME just recreate the edited bar (need data for the 'apply' function
	e2_config_dialog_single (rt->name, e2_toolbar_recreate_all, TRUE);
}
/**
 * @brief hide specified toolbar
 *
 * This is performed when the toolbar context-menu item 'show bar' is selected
 *
 * @param menuitem UNUSED the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime for the bar
 *
 * @return
 **/
static void _e2_toolbar_hide_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked)  //bar's menu is not now being created
	{
		_e2_option_bool_toggle (rt->show);  //now FALSE
		e2_toolbar_recreate (rt);
	}
}
/**
 * @brief toggle display of tooltips for a specified bar
 *
 * This is performed when the toolbar context-menu item 'show tips' is selected
 *
 * @param menuitem UNUSED the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime for the bar
 *
 * @return
 **/
static void _e2_toolbar_toggle_tt_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		gtk_toolbar_set_tooltips (GTK_TOOLBAR (rt->toolbar),
			_e2_option_bool_toggle (rt->tooltips));
	}
}
/**
 * @brief change a toolbar's space-handling mechanism
 *
 * This is performed when any of the options on the space handling
 * sub-menu of the bar's context-menu is selected
 *
 * @param menuitem the clicked sub-menu widget
 * @param rt ptr to E2_ToolbarRuntime for the bar
 *
 * @return
 **/
static void _e2_toolbar_update_space_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		GSList *group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (menuitem));
		gint index = 2 - g_slist_index (group, menuitem);
		if (_e2_option_int_get (rt->space) != index)
		{
			_e2_option_sel_set (rt->space, index);
			e2_toolbar_recreate (rt);
		}
	}
}
/**
 * @brief toggle toolbar between horizontal and vertical layout
 *
 * This is performed when the 'horizontal' option on the 'type'
 * sub-menu of the bar's context-menu is selected
 *
 * @param menuitem UNUSED the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
static void _e2_toolbar_toggle_hori_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		_e2_option_bool_toggle (rt->hori);
		e2_toolbar_recreate (rt);
	}
}
/**
 * @brief change type of toolbar
 *
 * type = main toolbar, big panebar etc
 * This is performed when any of the 'types' on the 'type'
 * sub-menu of the bar's context-menu is selected
 *
 * @param menuitem the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
static void _e2_toolbar_update_type_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		GSList *group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (menuitem));
		gint index = 3 - g_slist_index (group, menuitem);
		if (_e2_option_int_get (rt->type) != index)
		{
			_e2_option_sel_set (rt->type, index);
			e2_toolbar_recreate (rt);
		}
	}
}
/**
 * @brief  move bar 1 place right or down in its container
 *
 * This is performed when the corresponding choice is made
 * from the 'type' sub-menu of the bar's context-menu
 *
 * @param menuitem UNUSED the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
static void _e2_toolbar_increase_priority_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		gint i = _e2_option_int_get (rt->priority);
		if (_e2_option_int_set (rt->priority, i + 1) != i)
			e2_toolbar_recreate (rt);
	}
}
/**
 * @brief move bar 1 place left or up in its container
 *
 * This is performed when the corresponding choice is made
 * from the 'type' sub-menu of the bar's context-menu
 * It causes the bar to be moved left or up in its container
 *
 * @param menuitem UNUSED the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
static void _e2_toolbar_decrease_priority_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		gint i = _e2_option_int_get (rt->priority);
		if (_e2_option_int_set (rt->priority, i - 1) != i)
			e2_toolbar_recreate (rt);
	}
}
/**
* @brief move toolbar to the top or left of its container
 *
 * This is performed when the corresponding choice is made
 * from the 'type' sub-menu of the bar's context-menu
 *
 * @param menuitem UNUSED the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
static void _e2_toolbar_reset_priority_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked && (_e2_option_int_get (rt->priority) != 0))
	{
		_e2_option_int_set (rt->priority, 0);
		e2_toolbar_recreate (rt);
	}
}
/**
 * @brief change whether bar buttons are shown with relief
 *
 * This is performed when the corresponding choice is made
 * from the bar's context-menu
 *
 * @param menuitem UNUSED the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
static void _e2_toolbar_toggle_relief_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		_e2_option_bool_toggle (rt->relief);
		e2_toolbar_recreate (rt);
	}
}
/**
 * @brief change whether bar buttons are shown all with the same width
 *
 * This is performed when the corresponding choice is made
 * from the bar's context-menu
 *
 * @param menuitem UNUSED the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
static void _e2_toolbar_toggle_same_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		_e2_option_bool_toggle (rt->same);
		e2_toolbar_recreate (rt);
	}
}
/**
 * @brief change the way bar buttons are presented
 *
 * This is performed when the corresponding choice is made
 * from the bar's context-menu
 * options include: with/without icon, with/without label, 
 * label below/beside icon
 *
 * @param menuitem the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
static void _e2_toolbar_update_style_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		GSList *group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (menuitem));
		gint index = 4 - g_slist_index (group, menuitem);
		if (_e2_option_int_get (rt->style) != index)
		{
			_e2_option_sel_set (rt->style, index);
			e2_toolbar_recreate (rt);
		}
	}
}
/**
 * @brief change the size of bar button icons
 *
 * This is performed when the corresponding choice is made
 * from the icon size sub-menu of the bar's context-menu
 *
 * @param menuitem the clicked menu widget
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
static void _e2_toolbar_update_isize_cb (GtkWidget *menuitem, E2_ToolbarRuntime *rt)
{
	if (!rt->blocked) //bar's menu is not now being created
	{
		GSList *group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (menuitem));
		gint index = 6 - g_slist_index (group, menuitem);
		if (_e2_option_int_get (rt->isize) != index)
		{
			_e2_option_sel_set (rt->isize, index);
			e2_toolbar_recreate (rt);
		}
	}
}
  /*******************/
 /***** menus ******/
/*******************/

/**
 * @brief create toolbar context menu
 *
 * This is the callback for a button click on a toolbar. Right clicks cause
 * a context menu to be created
 *
 * @param widget the clicked bar container
 * @param event Gdk event data struct
 * @param rt ptr to E2_ToolbarRuntime for the bar
 *
 * @return TRUE (ie handled) for a right-click, else FALSE
 **/
static gboolean _e2_toolbar_click_cb (GtkWidget *widget, 
	GdkEventButton *event, E2_ToolbarRuntime *rt)
{
	if ((event->button == 3) && (event->type == GDK_BUTTON_PRESS))
	{
		//block local callbacks while the menu is being constructed
		rt->blocked = TRUE;
		GtkWidget *menu, *submenu;
		menu = gtk_menu_new ();
		e2_menu_add_check (menu, _("Show _bar"),
			_e2_option_bool_get (rt->show), _e2_toolbar_hide_cb, rt);
		e2_menu_add_check (menu, _("Show _tooltips"),
			_e2_option_bool_get (rt->tooltips), _e2_toolbar_toggle_tt_cb, rt);
		submenu = e2_menu_add_submenu (menu, _("Space _handling"));
		GSList *group = NULL;
		gint i = _e2_option_int_get (rt->space);
		e2_menu_add_radio (submenu, &group, _("_none"),
			(i == 0), _e2_toolbar_update_space_cb, rt);
		e2_menu_add_radio (submenu, &group, _("_scrollbars"),
			(i == 1), _e2_toolbar_update_space_cb, rt);
		e2_menu_add_radio (submenu, &group, _("_rest button"),
			(i == 2), _e2_toolbar_update_space_cb, rt);
		submenu = e2_menu_add_submenu (menu, _("_Placement"));
		e2_menu_add_check (submenu, _("_horizontal"),
			_e2_option_bool_get (rt->hori), _e2_toolbar_toggle_hori_cb, rt);
		GtkWidget *submenu2 = e2_menu_add_submenu (submenu, _("_Container"));
		group = NULL;
		i = _e2_option_int_get (rt->type);
		e2_menu_add_radio (submenu2, &group, _("_main window"),
			(i == 0), _e2_toolbar_update_type_cb, rt);
		e2_menu_add_radio (submenu2, &group, _("_both panes"),
			(i == 1), _e2_toolbar_update_type_cb, rt);
		e2_menu_add_radio (submenu2, &group, _("file-pane _1"),
			(i == 2), _e2_toolbar_update_type_cb, rt);
		e2_menu_add_radio (submenu2, &group, _("flle-pane _2"),
			(i == 3), _e2_toolbar_update_type_cb, rt);
		if (_e2_option_bool_get (rt->hori))
		{
			e2_menu_add (submenu, _("_Up"), 
				GTK_STOCK_GO_UP, NULL, _e2_toolbar_decrease_priority_cb, rt);
			e2_menu_add (submenu, _("_Down"),
				GTK_STOCK_GO_DOWN, NULL, _e2_toolbar_increase_priority_cb, rt);
		} else
		{
			e2_menu_add (submenu, _("_Left"),
				GTK_STOCK_GO_BACK, NULL, _e2_toolbar_decrease_priority_cb, rt);
			e2_menu_add (submenu, _("_Right"),
				GTK_STOCK_GO_FORWARD, NULL, _e2_toolbar_increase_priority_cb, rt);
		}
		e2_menu_add (submenu, _("_Reset order"), NULL, NULL,
			_e2_toolbar_reset_priority_cb, rt);
		e2_menu_add_check (menu, _("Buttons _relief"),
			_e2_option_bool_get (rt->relief), _e2_toolbar_toggle_relief_cb, rt);
		e2_menu_add_check (menu, _("Buttons _same size"),
			_e2_option_bool_get (rt->same), _e2_toolbar_toggle_same_cb, rt);
		submenu = e2_menu_add_submenu (menu, _("_Button style"));
		group = NULL;
		i = _e2_option_int_get (rt->style);
		//conform these labels with those in the config setup, below
		e2_menu_add_radio (submenu, &group, _("_theme"),
			(i == 0), _e2_toolbar_update_style_cb, rt);
		e2_menu_add_radio (submenu, &group, _("icon _only"),
			(i == 1), _e2_toolbar_update_style_cb, rt);
		e2_menu_add_radio (submenu, &group, _("_label only"),
			(i == 2), _e2_toolbar_update_style_cb, rt);
		e2_menu_add_radio (submenu, &group, _("icon _above label"),
			(i == 3), _e2_toolbar_update_style_cb, rt);
		e2_menu_add_radio (submenu, &group, _("icon _beside label"),
			(i == 4), _e2_toolbar_update_style_cb, rt);
		submenu = e2_menu_add_submenu (menu, _("_Icon size"));
		group = NULL;
		i = _e2_option_int_get (rt->isize);
		//conform these labels with those in the config setup, below
		e2_menu_add_radio (submenu, &group, _("_theme"),
			(i == 0), _e2_toolbar_update_isize_cb, rt);
		e2_menu_add_radio (submenu, &group, _("_menu"),
			(i == 1), _e2_toolbar_update_isize_cb, rt);
		e2_menu_add_radio (submenu, &group, _("toolbar _small"),
			(i == 2), _e2_toolbar_update_isize_cb, rt);
		e2_menu_add_radio (submenu, &group, _("toolbar _large"),
			(i == 3), _e2_toolbar_update_isize_cb, rt);
		e2_menu_add_radio (submenu, &group, _("_button"),
			(i == 4), _e2_toolbar_update_isize_cb, rt);
		e2_menu_add_radio (submenu, &group, _("drag '_n' drop"),
			(i == 5), _e2_toolbar_update_isize_cb, rt);
		e2_menu_add_radio (submenu, &group, _("_dialog"),
			(i == 6), _e2_toolbar_update_isize_cb, rt);
//		e2_menu_add_separator (menu);
		e2_menu_add (menu, _("Bar i_tems"), GTK_STOCK_PREFERENCES,
			NULL, _e2_toolbar_edit_cb, rt);

		e2_menu_popup (menu, event->button, event->time);
		rt->blocked = FALSE;
		return TRUE;
	}
	return FALSE;
}

/**
 * @brief destroy rest-menu after a choice has been made
 *
 * @param menu the rest-menu widget
 * @param rt pointer to data struct for the toolbar that has the rest button
 *
 * @return
 **/
static void _e2_toolbar_destroyrest_menu_cb (GtkWidget *menu, E2_ToolbarRuntime *rt)
{
	gtk_container_foreach (GTK_CONTAINER (menu), (GtkCallback) gtk_widget_destroy, NULL);
	gtk_widget_destroy (menu);
	gtk_button_released (GTK_BUTTON (rt->button_rest));	//pop the button back up
}
/**
 * @brief create a rest-menu for a bar
 *
 * Menu content is determined by data "path" of bar item
 * rt->menu_starter
 *
 * @param rt ptr to E2_ToolbarRuntime for the toolbar
 *
 * @return the menu widget
 **/
static GtkWidget *_e2_toolbar_createrest_menu (E2_ToolbarRuntime *rt)
{
	GtkTreePath *path = g_object_get_data (G_OBJECT (rt->menu_starter), "path");
	if (path != NULL)
	{
		GtkTreeIter iter;
		if (gtk_tree_model_get_iter (rt->set->ex.tree.model, &iter, path))
		{
			GtkWidget *menu = gtk_menu_new ();
			_e2_toolbar_add_menu_items (menu,
				rt->set->ex.tree.model, &iter, 1, rt);
			g_signal_connect (G_OBJECT (menu), "selection-done",
				G_CALLBACK (_e2_toolbar_destroyrest_menu_cb), rt);
			return menu;
		}
	}
	return NULL;
}
/**
 * @brief pop up the button pressed to initiate a toolbar submenu
 *
 * This is the callback for 'selection done' signal for the respective menu
 * and in the bookmarks setup, and a general submenu setup ...
 *
 * @param menu UNUSED the menu that is now being closed
 * @param button the widget clicked to initiate the menu
 *
 * @return
 **/
static void _e2_toolbar_menu_finished_cb (GtkWidget *menu, GtkWidget *button)
{
	//for permanent menus, do not destroy, but only applies to bookmarks now ??
	//LEAKS ??
//	gtk_container_foreach (GTK_CONTAINER (menu), (GtkCallback) gtk_widget_destroy, NULL);
	gtk_button_released (GTK_BUTTON (button));
}
/**
 * @brief calculate left and top coordinates for @a menu to pop up for @a button
 *
 * @param button the activated widget on the toolbar
 * @param menu the GtkMenu to be positioned
 * @param menu_position pointer to allocation struct, with window offset and in which to store the results
 * @param rt data struct for the toolbar containing @a button
 *
 * @return
 */
void e2_toobar_menu_position_set (GtkWidget *button, GtkWidget *menu, 
	GtkAllocation *menu_position, E2_ToolbarRuntime *rt)
{
	GtkRequisition menu_size;
	gtk_widget_size_request (menu, &menu_size);

	gint left = menu_position->x;
	gint top = menu_position->y;

	if (_e2_option_bool_get (rt->hori))	//horizontal bar
	{	//place below or above button, left-aligned
		menu_position->x = left;
		menu_position->y = top + rt->toolbar_container_box->allocation.height;
		gint screen_height = gdk_screen_height ();
		if (menu_position->y + menu_size.height > screen_height)
			menu_position->y = top - menu_size.height;
	}
	else	//vertical bar
	{	//place right or left of button, top-aligned
		menu_position->x = left + rt->toolbar_container_box->allocation.width;
		gint screen_width = gdk_screen_width ();
		if (menu_position->x + menu_size.width > screen_width)
			menu_position->x = left - menu_size.width;
		menu_position->y = top;
	}
}
/**
 * @brief set popup menu position
 *
 * This function is supplied when calling gtk_menu_popup(), to position
 * the displayed menu.
 * set @a push_in to TRUE for menu completely inside the screen,
 * FALSE for menu clamped to screen size
 *
 * @param menu the GtkMenu to be positioned
 * @param x	place to store gint representing the menu left
 * @param y  place to store gint representing the menu top
 * @param push_in place to store pushin flag
 * @param menu_position struct with data for where the menu is to be shown
 *
 * @return
 */
void e2_toolbar_menu_position (GtkMenu *menu, gint *x, gint *y,
	gboolean *push_in, GtkAllocation *menu_position)
{
	*x = menu_position->x;
	*y = menu_position->y;
	*push_in = TRUE;
}
/**
 * @brief pop up a menu, if it's a 'rest' menu, create it first
 *
 * If it's a rest-menu, @a item will have data "menu" = NULL,
 * otherwise "menu" points to an existing menu widget
 *
 * @param button widget clicked to initiate the menu
 * @param x coordinate of left of window relative to left of screen
 * @param y coordinate of top of window relative to screen (WAS + button height)
 * @param mouse_button no. of mouse button which was pressed
 * @param time time at which the activation event occurred
 * @param rt ptr to E2_ToolbarRuntime for the bar
 *
 * @return FALSE, always (though that seems to be irrelevant)
 **/
static gboolean _e2_toolbar_menu_popup (GtkWidget *button,
	gint x, gint y, gint mouse_button, guint32 time, E2_ToolbarRuntime *rt)
//	gint mouse_button, guint32 time, E2_ToolbarRuntime *rt)
{
	GtkWidget *menu = g_object_get_data (G_OBJECT (button), "menu");
	if (menu == NULL)
	{
		if (rt != NULL)
			menu = _e2_toolbar_createrest_menu (rt);
		else //can never happen ??
			return FALSE;
	}

	GtkAllocation menu_position;
	menu_position.x = x;
	menu_position.y = y;
	e2_toobar_menu_position_set (button, menu, &menu_position, rt);

	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, 
		(GtkMenuPositionFunc) e2_toolbar_menu_position, &menu_position, mouse_button, time);

	return FALSE;
}
/**
 * @brief create and pop-up menu at a position related to the button pos
 *
 * This is the callback for 'mnemonic activate' signal for rest-menu button, 
 * and in the bookmarks setup, and a general submenu setup ...
 *
 * @param button activated widget
 * @param arg UNUSED T/F
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return FALSE always
 **/
static gboolean _e2_toolbar_menu_popup_mnemonic_cb 
	(GtkWidget *button, gboolean arg, E2_ToolbarRuntime *rt)
{
	gint x, y;
	e2_utils_get_abs_pos (button, &x, &y);
//	y += button->allocation.height;
	_e2_toolbar_menu_popup (button, x, y, 0, 0, rt);
//	_e2_toolbar_menu_popup (button, 0, 0, rt);
	return FALSE;
}
/**
 * @brief create and pop-up the menu at a position related to the mouse pos
 * this is the callback for a button release on menu button, 
 * rest, bookmarks button, and a general submenu setup ...
 *
 * @param button clicked button widget
 * @param event Gdk event data struct
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return TRUE for btn 1 event, else FALSE
 **/
static gboolean _e2_toolbar_menu_popup_cb (GtkWidget *button, 
	GdkEventButton *event, E2_ToolbarRuntime *rt)
{
	if (event->button == 1)
	{
/*		gint x = event->x_root - event->x;
		gint y = event->y_root - event->y + button->requisition.height;
		return _e2_toolbar_menu_popup (button, x, y, event->button, event->time, rt); */
		//check mouse has not moved from the clicked button
		if ( event->x <= button->allocation.width 
			&& event->y <= button->allocation.height )
		{
			gint x = event->x_root - event->x;
			gint y = event->y_root - event->y;
			return _e2_toolbar_menu_popup (button, x, y, 1, event->time, rt);
		}
	}
	return FALSE;
}
/**
 * @brief UNUSED resize all command lines in a toolbar
 *
 * This is called when a bar is resized, so that commandlines
 * Commandlines are recognised by a data field "command_line"
 *  Seems to be not needed, for GTK26 at least
 * CHECKME = this is needed only if max width of command line is implemented
 *
 * @param child toolbar-container child widget
 * @param rt ptr to E2_ToolbarRuntime for the bar
 *
 * @return
 **/
/*static void _e2_toolbar_resize_combo (GtkWidget *child, E2_ToolbarRuntime *rt)
{
	GtkWidget *command_line = g_object_get_data (G_OBJECT (child), "command_line");
	if (command_line != NULL)
	{
		gint max_width = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (command_line), "cl_max_width"));
		if (max_width == -1)  return;
		GtkRequisition req_cl, req_hb;
		gtk_widget_size_request (GTK_WIDGET (rt->toolbar), &req_hb);
		if (req_hb.width == rt->alloc->width)
			return;
		printd (DEBUG, "resizing command line for %s", rt->option);

		gtk_widget_size_request (command_line, &req_cl);
		//req_hb.width _should_ be bigger than req_cl.width
		gint new_width = rt->alloc->width - req_hb.width + req_cl.width;
		//check upper limit
		if (new_width > max_width)
			new_width = max_width;
		if (new_width > rt->alloc->width)  //bug, hb width < cl width !!
			new_width = rt->alloc->width;
		//check lower limit
		if (new_width < 1)
			new_width = 1;
		gint min_width = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (command_line), "cl_min_width"));
		if ((min_width != -1) && (new_width < min_width))
			new_width = min_width;

		if (req_cl.width != new_width)
			gtk_widget_set_size_request (GTK_WIDGET (child), new_width, -1);
	} 
}
*/
/**
 * @brief show 'rest' button at end of toolbar
 *
 * if not done before, create a 'rest' button, add it to end of 
 * bar container, and show the button
 * if button already created, just show it
 * the button visibility is signalled by rt->is_too_small
 *  
 * @param rt ptr to data struct for toolbar that has been resized (=rt->toolbar)
 *
 * @return
 **/
static void _e2_toolbar_add_rest_button (E2_ToolbarRuntime *rt)
{
	if (rt->is_too_small)
	{	//button is already displayed
		printd (DEBUG, "rest button already displayed");
		return;
	}
	if (rt->button_rest != NULL)
	{	//button is already created
		gtk_widget_show (rt->button_rest);
		rt->is_too_small = TRUE;
		printd (DEBUG, "re-displayed existing rest button");
		return;
	}
	//create button with style according to options in force
	gint i = _e2_option_int_get (rt->style);
	if (i == 0)
	{ //convert default bar style to corresponding local enumerator
		GtkSettings* defs = gtk_settings_get_default ();
        g_object_get (G_OBJECT (defs), "gtk-toolbar-style", &i, NULL);
		i++;	//adjust for 0=default
	}
	gchar *img;
	if (i == 2)
		img = NULL;
	else
		img = (_e2_option_bool_get (rt->hori)) ? GTK_STOCK_GOTO_LAST : GTK_STOCK_GOTO_BOTTOM;
	rt->button_rest = _e2_button_get (i == 1 ? "" : _("_Rest"),
		img, rt->icon_size, _("Show a menu of hidden items"), NULL, NULL,
			(i == 3 ? E2_BUTTON_ICON_ABOVE_LABEL : 0) |
			(i == 1 ? E2_BUTTON_SHOW_MISSING_ICON : 0));
	gtk_button_set_relief (GTK_BUTTON (rt->button_rest),
		_e2_option_bool_get (rt->relief) ? GTK_RELIEF_NORMAL : GTK_RELIEF_NONE);
	g_signal_connect (rt->button_rest, "button-press-event",
		G_CALLBACK (_e2_toolbar_menu_popup_cb), rt);
	g_signal_connect (rt->button_rest, "mnemonic-activate",
		G_CALLBACK (_e2_toolbar_menu_popup_mnemonic_cb), rt);
	gtk_box_pack_end (GTK_BOX (rt->toolbar_container_box), rt->button_rest, FALSE, FALSE, 0);
	gtk_widget_show_all (rt->button_rest);
	rt->is_too_small = TRUE;
	printd (DEBUG, "rest button created and added to %s",rt->name);
}
//#ifdef GTK26
/**
 * @brief after a toolbar item is mapped, adjust the rest-menu start item or hide the rest button
 *
 * This callback is connected to toolitems only when the toolbar space-handling mode is "rest-menu"
 *
 * @param widget ptr to toolitem that has just been mapped
 * @param rt data struct for the toolbar
 *
 * @return
 **/
static void _e2_toolbar_mapitem_cb (GtkWidget *widget, E2_ToolbarRuntime *rt)
{
//	if (_e2_option_int_get (rt->space) == 2)
//	{	//space handling by rest menu
		gint index = g_list_index (rt->bar_items_list, widget);
		printd (DEBUG, "%s item %d mapped", rt->name, index);
		gint count = g_list_length (rt->bar_items_list);
		gboolean allvisible = (index == count - 1);

		if (index < count -1)
		{  //find the next unhidden item (if any) on the bar
			GList *iterator = g_list_nth (rt->bar_items_list, ++index);
			while ( ! GTK_WIDGET_VISIBLE (GTK_BIN (iterator->data)->child)
				&& (iterator = iterator->next) != NULL )
			{}
			if (iterator != NULL)
				rt->menu_starter = GTK_BIN (iterator->data)->child;
			else
				allvisible = TRUE;	 //make sure we pass the following test
		}
		if (allvisible)
		{	//last visible item on the bar is now mapped
			if (rt->button_rest != NULL)
			{
				rt->is_too_small = FALSE;
				gtk_widget_hide (rt->button_rest);
			}
		}
//	}
}
/**
 * @brief after a toolbar item is unmapped, adjust the space-handing arrangements accordingly
 *
 * This callback is connected to toolitems only when the toolbar space-handling mode is
 * scrollbar or rest-menu
 * As appropriate, the rest-menu start item will be altered, the rest button displayed
 * If scrollbar in force, and this is the "last" visible item on the bar, then a size request
 * is issued, locking the minimum bar size and displaying the scrollbar (and this cb will
 * never be called again)
 *
 * @param widget ptr to toolitem that has just been mapped
 * @param rt data struct for the toolbar
 *
 * @return
 **/
static void _e2_toolbar_unmapitem_cb (GtkWidget *widget, E2_ToolbarRuntime *rt)
{
	//ignore unmapping toggles
	if ( ! GTK_WIDGET_VISIBLE (GTK_BIN (widget)->child))
		return;

	gint index = g_list_index (rt->bar_items_list, widget);
//	if (index == -1)
//		return;
	printd (DEBUG, "%s item %d UNMAPPED", rt->name, index);

	gint handlertype = _e2_option_int_get (rt->space);
	switch (handlertype)
	{
		case 1:
		{	//space handling by scroll bar
			gint count = g_list_length (rt->bar_items_list);
			if (index == count -1)
			{
				gboolean horiz = _e2_option_bool_get (rt->hori);
				GtkAllocation alloc = GTK_WIDGET (rt->toolbar)->allocation;
				gint sizenow = (horiz) ? alloc.width : alloc.height;
//				printd (DEBUG, "allocated bar size = %d", sizenow);
				//get back the size at which the first button was hidden, 
				// to activate the scrollbar
				if (horiz)
					//+1 ensures last visible item is displayed
					gtk_widget_set_size_request (GTK_WIDGET (rt->toolbar) , sizenow + 1, -1);
				else
					gtk_widget_set_size_request (GTK_WIDGET (rt->toolbar), -1, sizenow + 1);
			}
			break;
		}
		case 2:
		{	//space handling by rest menu
			rt->menu_starter = GTK_BIN (widget)->child;  //flag the initial item in the rest-menu
			_e2_toolbar_add_rest_button (rt);  //create/(re)display rest menu btn
			break;
		}
		default:
			break;
	}
	return;
}
/**
 * @brief after a toolbar is torn-off, retain its size so that its contents are still displayed
 *
 * @param handlebox the object which received the signal = rt->toolbar_container
 * @param widget the child widget of the handlebox = rt->toolbar_container_box
 * @param rt data struct for the toolbar
 *
 * @return
 **/
static void _e2_toolbar_detach_cb (GtkHandleBox *handlebox,
	GtkWidget *widget, E2_ToolbarRuntime *rt)
{
	gint barsize;
	GtkOrientation bar_direction = gtk_toolbar_get_orientation (rt->toolbar);
	if (bar_direction == GTK_ORIENTATION_HORIZONTAL)
	{
		barsize = widget->allocation.x + widget->allocation.width;
		gtk_widget_set_size_request (widget, barsize, -1);
	}
	else
	{
		barsize = widget->allocation.y + widget->allocation.height;
		gtk_widget_set_size_request (widget, -1, barsize);
	}
}
/**
 * @brief after a torn-off toolbar is re-connected, allow its size to be adjusted as normal
 *
 * @param handlebox UNUSED the object which received the signal = rt->toolbar_container
 * @param widget the child widget of the handlebox = rt->toolbar_container_box
 * @param data UNUSED pointer to data specified when callback was connected
 *
 * @return
 **/
static void _e2_toolbar_attach_cb (GtkHandleBox *handlebox,
	GtkWidget *widget, gpointer data)
{
	gtk_widget_set_size_request (widget, -1, -1);
}
/**
 * @brief connect all toolbars' attach/detach signals
 *
 * This is called at session-start, when the main window is displayed,
 * and after window re-builds
 * Toolbars are in handle boxes, which can be de/attached, and
 * if so, the size needs to be manipulated to preserve the contents display
 *
 * @return
 **/
void e2_toolbar_initialise_tearoffs (void)
{
	E2_ToolbarData **thisbar;
	for (thisbar = app.bars; *thisbar != NULL; thisbar++)
	{
		E2_ToolbarRuntime *trt = (*thisbar)->rt;
		g_signal_connect (G_OBJECT (trt->toolbar_container), "child-detached",
			G_CALLBACK (_e2_toolbar_detach_cb), trt);
		g_signal_connect (G_OBJECT (trt->toolbar_container), "child-attached",
			G_CALLBACK (_e2_toolbar_attach_cb), NULL);
	}
}
/**
 * @brief initialise all one or all (4) toolbars' space-handing arrangements
 *
 * This is called at session-start, when the main window is displayed,
 * and after window re-builds
 * Since this is after the bars and their contents are mapped, the 
 * map, unmap, resize signals for bar buttons can be connected to their
 * callbacks without irrelevant calls (hence delays).
 * As appropriate, the rest-menu start item will be set, and the rest button displayed
 * @param rt pointer to data struct for bar to be processed, or NULL for all bars
 *
 * @return
 **/
void e2_toolbar_initialise_space_handler (E2_ToolbarRuntime *rt)
{
	gint i, count;
	gint last_item_index;
	gint barsize, fullsize=100;	//nominal size for warning prevention, should always be replaced
	
	if (rt != NULL)
		count = 1;
	else
	{
		count = 0;
		E2_ToolbarData **thisbar;
		for (thisbar = app.bars; *thisbar != NULL; thisbar++)
			count++;
	}

	E2_ToolbarRuntime *barsrt [count];

	if (rt != NULL)
		barsrt [0] = rt;
	else
	{
		for (i = 0; i < count; i++)
			barsrt [i] = app.bars[i]->rt;
	}
	
	for (i=0; i<count; i++)
	{
		if (!_e2_option_bool_get (barsrt[i]->show))
			continue;	//don't do anything if the bar is hidden
		gint htype = _e2_option_int_get (barsrt[i]->space);
		if ( htype != 0)
		//don't need to do anything if there is no space handler for the bar
		{
			GList *iterator;
			//find the currently-allocated end of the bar
			GtkToolbar *tbar = barsrt[i]->toolbar;
			GtkOrientation bar_direction = gtk_toolbar_get_orientation (tbar);
			GtkWidget *wid = GTK_WIDGET (tbar);
			barsize = (bar_direction == GTK_ORIENTATION_HORIZONTAL) ?
				wid->allocation.x + wid->allocation.width :
				wid->allocation.y + wid->allocation.height ;
			//find the last visible item on the bar
			gboolean visflg = FALSE;
			gint hidden_count;
			for (hidden_count = 0, last_item_index = gtk_toolbar_get_n_items (tbar) - 1;
				last_item_index > -1; 
				last_item_index--, 
				//hidden items are allocated a size of 1, so this allows compensation for all such items
				hidden_count++
				)
			{
				wid = GTK_WIDGET (gtk_toolbar_get_nth_item (tbar, last_item_index));
				//x == -1 is a hack test for bar hidden items, for horiz. or vert. bar
				// GTK_WIDGET_VISIBLE (wid) always succeeds here - probably some child is the hidden thing
				if ( wid->allocation.x > -1 )
				{
					//find the notional (no-space-constraint) end of that last item
					fullsize = (bar_direction == GTK_ORIENTATION_HORIZONTAL) ?
						wid->allocation.x + wid->allocation.width + hidden_count:
						wid->allocation.y + wid->allocation.height + hidden_count;
					visflg = TRUE;
					break;
				}
			}
			if (!visflg)
				fullsize = 1;	//nominal size when no item is visible on bar

			//checking barsize against fullsize can cause last btn to be hidden without any rest btn
			//FIXME = find a way to show that btn (show_all, size_request N/A)
			if (barsize < fullsize+1)
			{ //there is a restriction, initial handler setup is needed
				if (htype == 2)
				{	//space handling by rest menu
					//find the first (if any) unmapped + visible button on the bar
					iterator = g_list_first (barsrt [i]->bar_items_list);
					while (iterator != NULL)
					{
						if ( ! GTK_WIDGET_MAPPED (iterator->data)
							&& GTK_WIDGET_VISIBLE (GTK_BIN (iterator->data)->child) )
						{	//we found an unmapped, visible one
							barsrt [i]->menu_starter = GTK_BIN (iterator->data)->child;
							//create/display rest menu btn
							//this will also decrease the bar size accordingly, and therefore hide other button(s) !!
							_e2_toolbar_add_rest_button (barsrt [i]);
							break;
						}
						iterator = iterator->next;
					}
//					if ( (iterator == NULL)
//						// this is a kludge for the 1-pixel offset workaround, above
//						&& (barsize < fullsize) )
//						printd (DEBUG, "no index for menu starter !!");
				}
				else  //not rest-menu
					if (htype == 1)
				{  //make sure scrollbar will be visible, if bar starts too small
					if (_e2_option_bool_get (barsrt [i]->hori))
						//+1 to ensure the last item is actually shown
						gtk_widget_set_size_request (GTK_WIDGET (tbar) , fullsize + 1, -1);
					else
						gtk_widget_set_size_request (GTK_WIDGET (tbar), -1, fullsize + 1);
				}
			}
			//space handler needs to know when things change
			iterator = g_list_first (barsrt[i]->bar_items_list);
			while (iterator != NULL)
			{
				if (htype == 2)  //rest-menu applies
					g_signal_connect (GTK_WIDGET (iterator->data), "map", G_CALLBACK (_e2_toolbar_mapitem_cb), barsrt[i]);
				//scrollbar or rest-menu applies
				g_signal_connect (GTK_WIDGET (iterator->data), "unmap", G_CALLBACK (_e2_toolbar_unmapitem_cb), barsrt[i]);
				iterator = iterator->next;
			}
		}
	}
}
//#endif  //def GTK26

  /*****************/
 /***** utils *****/
/*****************/

/**
 * @brief UNUSED pack toolbar into a container
 *
 * @param box container widget into which the bar will be packed
 * @param rt ptr to E2_ToolbarRuntime
 *
 * @return
 **/
/*static void _e2_toolbar_pack_and_prio (GtkWidget *box, E2_ToolbarRuntime *rt)
{
	gtk_box_pack_start (GTK_BOX (box), rt->toolbar_container, FALSE, FALSE, 0);
	gtk_box_reorder_child (GTK_BOX (box), rt->toolbar_container, _e2_option_int_get (rt->priority));
} */
/**
 * @brief get option data from a row in the relevant treestore
 *
 * @param rt ptr to E2_ToolbarRuntime
 * @param model model for the treestore
 * @param iter pointer to iter to use for interrogating the model
 * @param name pointer to location to store the retrieved name
 * @param icon pointer to location to store the retrieved icon identifier string
 * @param tip pointer to location to store the retrieved tooltip string
 * @param type pointer to location to store the retrieved type string
 * @param arg pointer to location to store the retrieved argument(s) string
 *
 * @return
 **/
static void _e2_toolbar_get_option_row (E2_ToolbarRuntime *rt, 
	GtkTreeModel *model, GtkTreeIter *iter, gchar **name,
	gchar **icon, gchar **tip, gchar **type, gchar **arg)
{
//	gchar *_name, *_icon, *_tip, *_type, *_arg;
	gchar *_tip;
	if (type == NULL)
		gtk_tree_model_get (model, iter, 0, name, 1, icon, 2, &_tip,
			3, arg, -1);
	else
	{
		gtk_tree_model_get (model, iter, 0, name, 1, icon, 2, &_tip,
			3, type, 4, arg, -1);
//		*type = _type;
	}
//	*name = e2_utils_replace_vars (_name);
//	*icon = _icon;
	*tip = e2_utils_replace_vars (_tip);
//	*arg = e2_utils_replace_vars (_arg); DON'T EXPAND UNTIL USED
	if ((*tip != NULL) && (**tip == '\0'))
	{
		g_free (*tip);
		*tip = NULL;
	}
//	g_free (_name);
	g_free (_tip);
//	g_free (_arg);
}
/**
 * @brief add button to (the end of) a toolbar
 *
 * Button content and style are determined by option in force, including 
 * whether buttons are same-size or not
 * For gtk2.4 and later, the button is added to a toolitem, which is added to the bar
 *
 * @param rt toolbar data struct
 * @param name translated string used for button label
 * @param icon path/name string of button icon (may be name of a stock icon)
 * @param tip translated string for button tooltip
 * @param data data to be provided to the button callback
 *
 * @return the button widget
 **/
static GtkWidget *_e2_toolbar_add_button (E2_ToolbarRuntime *rt,
	gchar *name, gchar *icon, gchar *tip, gpointer data)
{
	gint i = _e2_option_int_get (rt->style);
	if (i == 0)
	{ //convert default bar style to corresponding local enumerator
		GtkSettings* defs = gtk_settings_get_default ();
        g_object_get (G_OBJECT (defs), "gtk-toolbar-style", &i, NULL);
		i++;	//adjust for 0=default
	}
	GtkWidget *button = _e2_button_get (i == 1 ? "" : name,
		i == 2 ? NULL : icon, rt->icon_size, NULL, e2_action_run, data,
		(i == 3 ? E2_BUTTON_ICON_ABOVE_LABEL : 0) |
		(i == 1 ? E2_BUTTON_SHOW_MISSING_ICON : 0));
//	WHAT IS THIS FOR ?
//	if (data != NULL)
//		g_object_set_data_full (G_OBJECT (button), "free-callback-data",
//		data, (GDestroyNotify) e2_action_free_runtime);
	gtk_button_set_relief (GTK_BUTTON (button), _e2_option_bool_get (rt->relief) ?
		GTK_RELIEF_NORMAL : GTK_RELIEF_NONE);

	GtkToolItem *tool = gtk_tool_item_new ();
	//this line is based on default status being FALSE, 
	//which seems to be correct, but is contrary to the API doc
	if (_e2_option_bool_get (rt->same))  // if buttons are to be same size
		gtk_tool_item_set_homogeneous (tool, TRUE);
	gtk_container_add (GTK_CONTAINER (tool), button);
	gtk_tool_item_set_tooltip (tool, app.tooltips, tip, NULL);	
	gtk_toolbar_insert (rt->toolbar, tool, -1);
	gtk_widget_show_all (GTK_WIDGET (tool));
//#ifdef GTK26
	//add to bar-items cache
	rt->bar_items_list = g_list_append (rt->bar_items_list, tool);
//#endif  //def GTK26
	return button;
}

  /**********************************/
 /***** toggle-button functions *****/
/**********************************/
/**
 * @brief swap the state of toggle button associated with @a action
 *
 * For a specified action, this hides the current toggle button,
 * and shows the alternate button for its new state
 * The button rt data is accessed via the action->data2
 *
 * @param action_name string with (translated) name of the action initiated by the button
 *
 * @return boolean T/F, the state of the toggle button after this toggle
 **/
 gboolean e2_toolbar_button_toggle (gchar *action_name)
{
	E2_Action *action = e2_action_get (action_name);
	E2_ToggleActionEx *ex = action->data2;
	ex->current_state = ! ex->current_state;
	gboolean state_new = ex->current_state;
	if (state_new)
	{
		gtk_widget_hide (ex->false_button);
		gtk_widget_show (ex->true_button);
	}
	else //false now
	{
		gtk_widget_hide (ex->true_button);
		gtk_widget_show (ex->false_button);
	}
	return state_new;
}
/**
 * @brief set state of toggle button associated with @a action
 *
 * For a specified action, this shows the appropriate toggle button,
 * The button rt data is accessed via the action->data2
 * Buttons must have been created before this is called
 *
 * @param action_name string with (translated) name of the action initiated by the button
 * @param state value to be set, T or F
 *
 * @return
 **/
void e2_toolbar_toggle_button_set_state (gchar *action_name, gboolean state)
{
	E2_Action *action = e2_action_get (action_name);
	E2_ToggleActionEx *ex = action->data2;
	if (ex == NULL) return;
	if (state == ex->current_state) return;
	ex->current_state = state;
	if (state)
	{
		gtk_widget_hide (ex->false_button);
		gtk_widget_show (ex->true_button);
	}
	else //false now
	{
		gtk_widget_hide (ex->true_button);
		gtk_widget_show (ex->false_button);
	}
}
/**
 * @brief get current state of toggle button associated with @a action
 *
 * The button rt data is accessed via the action->data2
 * Buttons must have been created before this is called
 *
 * @param action_name string with (translated) name of the action initiated by the button
 *
 * @return current state of the button, T or F, of F if there is no data
 **/
gboolean e2_toolbar_toggle_button_get_state (gchar *action_name)
{
	E2_Action *action = e2_action_get (action_name);
	E2_ToggleActionEx *ex = action->data2;
	if (ex == NULL) return FALSE;
	return (ex->current_state);
}
/**
 * @brief clear toggle data flag
 *
 * @param data
 * @param rt ptr to E2_ToolbarRuntime for a bar, or NULL for all bars
 *
 * @return
 **/
static void _e2_toolbar_toggle_button_clear_state
	(E2_ToggleActionEx *data, E2_ToolbarRuntime *rt)
{
	if (rt == NULL || rt == data->bar_rt)
		data->state = E2_TOGGLE_DESTROYED;
}
/**
 * @brief clear all toggle button default flags for a bar, or for all bars
 *
 * To ensure that each toggle button (pair) is recreated, 
 * clear each toggle-button's 'default_set' flag
 *
 * @param rt ptr to rt data struct for a bar, or NULL for all bars
 *
 * @return
 **/
void e2_toolbar_toggle_buttons_set_destroyed (E2_ToolbarRuntime *rt)
{
	g_ptr_array_foreach (toggles_array,
		(GFunc) _e2_toolbar_toggle_button_clear_state, rt);
}

  /***************************/
 /***** toolbar creation *****/
/***************************/

/**
 * @brief add bookmarks to toolbar specified by @a rt
 *
 * This is the handler for bookmark.<list> pseudo-action in a
 * toolbar. It adds buttons to the bar, each with a menu of
 * 'children' if they exist.
 * The bookmark 'context menu' is disabled for each bookmark
 * button (but not for any children)
 *
 * @param set data struct for the bookmarks option
 * @param rt ptr to E2_ToolbarRuntime
 * @param iter pointer to iter to be used for interrogating the set model
 * @param action
 *
 * @return
 **/
static void _e2_toolbar_add_bookmarks (E2_OptionSet *set,
	E2_ToolbarRuntime *rt, GtkTreeIter *iter, E2_Action *action)
{
	do
	{
		gchar *name, *icon, *tip, *path;
		_e2_toolbar_get_option_row (rt, set->ex.tree.model,
			iter, &name, &icon, &tip, NULL, &path);
		GtkWidget *button = _e2_toolbar_add_button (rt, name, icon, tip,
			e2_action_pack_runtime (action, path, g_free));	//WHAT is this fn & args for ??
		g_signal_connect_data (G_OBJECT (button), "button-press-event",
			G_CALLBACK (e2_bookmark_click_cb),
			gtk_tree_model_get_path (set->ex.tree.model, iter),
			(GClosureNotify) gtk_tree_path_free, 0);
		g_object_set_data (G_OBJECT (button), "bookmark-path", path);
		//prevent context menu for this button
		//(can't use FALSE, as that returns NULL when queried)
		g_object_set_data (G_OBJECT (button), "with-context", GINT_TO_POINTER (1));

		if (gtk_tree_model_iter_has_child (set->ex.tree.model, iter))
		{
			GtkWidget *menu = gtk_menu_new ();
			GtkTreeIter iter2;
			gtk_tree_model_iter_children (set->ex.tree.model, &iter2, iter);
			e2_menu_add_bookmarks_submenu (action, set, menu, &iter2);	//create menu of child marks
			g_object_set_data_full (G_OBJECT (button), "menu", menu,
				(GDestroyNotify) gtk_widget_destroy);
			g_signal_connect (button, "button-press-event", 
				G_CALLBACK (_e2_toolbar_menu_popup_cb), rt);
			g_signal_connect (button, "mnemonic-activate", 
				G_CALLBACK (_e2_toolbar_menu_popup_mnemonic_cb), rt);
			g_signal_connect (G_OBJECT (menu), "selection-done",
				G_CALLBACK (_e2_toolbar_menu_finished_cb), button);
		}
		g_free (name);
		g_free (icon);
		g_free (tip);
	} while ((gtk_tree_model_iter_next (set->ex.tree.model, iter)) == TRUE);
}
/**
 * @brief
 *
 * @param menu widget to which the items will be added
 * @param model pointer to model for treeestore with the menu's contents
 * @param iter pointer to iter to be used for interrogating @a model
 * @param level treestore depth
 * @param rt ptr to toolbar data struct
 *
 * @return
 **/
static void _e2_toolbar_add_menu_items (GtkWidget *menu, 
	GtkTreeModel *model, GtkTreeIter *iter, gint level, E2_ToolbarRuntime *rt)
{
	do
	{
		gchar *name, *icon, *tip, *type, *arg;
		_e2_toolbar_get_option_row (rt, model, iter, &name, &icon, &tip, &type, &arg);

		E2_Action *action = e2_action_get (type);
		if (action != NULL) 
		{
			switch (action->type)
			{
				case E2_ACTION_TYPE_SUBMENU:
				{
					GtkWidget *item = e2_menu_add (menu, name,
						icon, tip, NULL, NULL);
					GtkTreeIter iter2;
					if (gtk_tree_model_iter_children (model, &iter2, iter))
					{
						GtkWidget *menu2 = gtk_menu_new ();
						_e2_toolbar_add_menu_items (menu2, model, &iter2, level + 1, rt);
						gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu2);
					}
					break;
				}
				case E2_ACTION_TYPE_BOOKMARKS:
				{
					E2_OptionSet *set = e2_option_get ("bookmarks");
					GtkTreeIter iter;
					if ((arg != NULL) && (*arg != '\0'))
					{
						if (e2_tree_find_iter_from_str (set->ex.tree.model, 0, arg, &iter, TRUE))
						{
							GtkTreeIter iter2;
							if (gtk_tree_model_iter_children (set->ex.tree.model, &iter2, &iter))
							{
								e2_menu_add_bookmarks_submenu (action, set, menu, &iter2);
								rt->has_bookmarks = TRUE;
							}
						}
					}
					else if (gtk_tree_model_get_iter_first (set->ex.tree.model, &iter))
					{
						e2_menu_add_bookmarks_submenu (action, set, menu, &iter);
						rt->has_bookmarks = TRUE;
					}
				}
					break;
				case E2_ACTION_TYPE_PLUGINS:
					e2_menu_create_plugins_menu (menu);
					break; 
				case E2_ACTION_TYPE_SEPARATOR:
					e2_menu_add_separator (menu);
					break;
/* filters converted to a toggle, with customised menu created on-the-fly
				case E2_ACTION_TYPE_FILTERS:
					{	// create filters sub menu for current pane
						FileView *view = (rt == &app.pane1.toolbar) ? 
							&app.pane1_view : &app.pane2_view;
						view->check_name = e2_menu_add_check (menu, _("_Name filter"),
							FALSE, e2_name_filter_dialog_create_cb, view);
						view->check_size = e2_menu_add_check (menu, _("_Size filter"),
							FALSE, e2_size_filter_dialog_create_cb, view);
						view->check_date = e2_menu_add_check (menu, _("_Date filter"),
							FALSE, e2_date_filter_dialog_create_cb, view);
						e2_menu_add_separator (menu);
						view->check_dirs = e2_menu_add_check (menu, _("_Directories too"),
							FALSE, e2_filelist_filter_dirs_cb, view);
						e2_menu_add_separator (menu);
						e2_menu_add (menu, _("_Remove all filters"), NULL, NULL, 
							e2_filelist_remove_filters_cb, view);
					}
					break; */
				case E2_ACTION_TYPE_TOGGLE:
					{
						gboolean this_state = g_str_equal (arg, "true");
						E2_ToggleActionEx *ex = action->data2;
						//if the current state matches this tree item, display it
						if (this_state == ex->current_state)
						{
							E2_ActionRuntime *actionrt = e2_action_pack_runtime (action, arg, NULL);	
//							GtkWidget *item = 
							e2_menu_add (menu, name, icon, tip,
								e2_action_run, actionrt);
//	WHAT IS THIS FOR ?
//							g_object_set_data_full (G_OBJECT (item), "free-callback-data", actionrt,
//								(GDestroyNotify) e2_action_free_runtime);
						}
					}
					break;
				case E2_ACTION_TYPE_ITEM:
					{
						E2_ActionRuntime *actionrt = e2_action_pack_runtime (action, arg, NULL);
//						GtkWidget *item = 
						e2_menu_add (menu, name, icon, tip,
							e2_action_run, actionrt);
//						g_object_set_data_full (G_OBJECT (item), "free-callback-data", actionrt,
//							(GDestroyNotify) e2_action_free_runtime);
					}
					break;
				default:
					break;
			}
		}
		//cleanup
		//CHECKME = the item  may need to be shown 
		g_free (name);
		g_free (icon);
		g_free (tip);
		g_free (type);
	} while ((gtk_tree_model_iter_next (model, iter)) == TRUE);
}
/**
 * @brief add items to the toolbar specified by @a rt
 *
 * @param model pointer to model for treeestore with the bar's contents
 * @param iter pointer to iter to be used for interrogating @a model
 * @param level treestore depth
 * @param rt ptr to toolbar data struct
 *
 * @return
 **/
static void _e2_toolbar_add_items (GtkTreeModel *model, 
	GtkTreeIter *iter, gint level, E2_ToolbarRuntime *rt)
{
	do
	{
		gchar *name, *icon, *tip, *type, *arg;
		_e2_toolbar_get_option_row (rt, model, iter, &name, &icon, &tip, &type, &arg);

		E2_Action *action = e2_action_get (type);
		if (action != NULL) 
		{
			GtkWidget *button = NULL;
			switch (action->type)
			{
				case E2_ACTION_TYPE_COMMAND_LINE:
				{  //name is 'public' version, now not stored in cl_rt->name
					//CHECKME = is <has_arg> a sufficient differentiator for cmd lines vs dir lines ?
					E2_CommandLineRuntime *cl_rt = e2_command_line_create (name,
						action->data, action->func, action->has_arg ?
						E2_COMMAND_LINE_ORIGINAL : E2_COMMAND_LINE_DIR_LINE, action->data2);
					GtkWidget *command_line = GTK_WIDGET (cl_rt->cl);
					button = command_line;

					GtkToolItem *tool = gtk_tool_item_new ();
 					gtk_container_add (GTK_CONTAINER (tool), command_line);
					gtk_toolbar_insert (GTK_TOOLBAR (rt->toolbar), tool, -1);
					gtk_widget_show_all (GTK_WIDGET (tool));
					g_object_set_data_full (G_OBJECT (rt->toolbar), cl_rt->name,
						cl_rt, (GDestroyNotify) e2_command_line_destroy);
					//get the width of the box
					gint min, max;
					if (strlen (arg) > 0)
					{
						gchar **split = g_strsplit (arg, ",", 2);
						gchar *end = NULL;
						gint i;
						if (split[0])
						{
							i = (gint) g_ascii_strtoull (split[0], &end, 10);
							if (end != split[0])
								min = i;
							else
								min = 100;	//fallback value
						}
						else
							min = 100;

/*						if (_e2_option_bool_get (rt->hori))
//							gtk_widget_set_size_request (command_line, min, -1);
							gtk_widget_set_size_request (GTK_WIDGET (tool), min, -1);
						else
						//with horiz min = -1, GTK makes vertical bars too wide
							gtk_widget_set_size_request (GTK_WIDGET (tool), -1, min);
						40 or so gives a reasonable width, but GTK combo-box
						doesn't do drop-across, so when the line is that narrow,
						you can't read drop-down items
							gtk_widget_set_size_request (command_line, 50, min);
							gtk_widget_set_size_request (GTK_WIDGET (tool), 50, min);
*/						gtk_widget_set_size_request (GTK_WIDGET (tool), min, -1);
						
						if (split[1])
						{
							if ((g_str_equal (split[1], _A(89))) || (g_str_equal (split[1], "*")))	//_("expand"
							{
								max = -1;
								gtk_tool_item_set_expand (tool, TRUE);
							}
							else
							{
								i = (gint) strtoull (split[1], &end, 10);	
								if (end != split[0])
									max = i;
								else
									max = 200;	//fallback value
								g_object_set_data (G_OBJECT (tool), "command_line", command_line);
								//min is also needed when max applies
//FIXME distinguish horiz & vert bars	if (_e2_option_bool_get (rt->hori))
								g_object_set_data (G_OBJECT (command_line), "cl_min_width", GINT_TO_POINTER(min));
								g_object_set_data (G_OBJECT (command_line), "cl_max_width", GINT_TO_POINTER(max));
							}
						}
						g_strfreev (split);
					}
//					gtk_widget_set_size_request (command_line, max, -1);  no point if max = -1, bug if max != -1
//					gtk_widget_set_size_request (command_line, min, -1);

					rt->has_command_line = TRUE;
				}
					break;
				case E2_ACTION_TYPE_SUBMENU:
				{
					button = _e2_toolbar_add_button (rt, name, icon, tip, NULL);
					gtk_widget_show_all (button);

					GtkTreeIter iter2;
					if (gtk_tree_model_iter_children (model, &iter2, iter))
					{
						GtkWidget *menu = gtk_menu_new ();
						_e2_toolbar_add_menu_items (menu, model, &iter2, level + 1, rt);

						g_object_set_data_full (G_OBJECT (button), "menu", menu,
							(GDestroyNotify) gtk_widget_destroy);
						g_signal_connect (G_OBJECT (menu), "selection-done",
							G_CALLBACK (_e2_toolbar_menu_finished_cb), button);
//						g_signal_connect (button, "button-press-event",
						g_signal_connect (button, "button-release-event",
							G_CALLBACK (_e2_toolbar_menu_popup_cb), rt);
						g_signal_connect (button, "mnemonic-activate",
							G_CALLBACK (_e2_toolbar_menu_popup_mnemonic_cb), rt);
					}
				}
					break;
				case E2_ACTION_TYPE_SEPARATOR:
				{
					GtkWidget *sep, *align;
					if (_e2_option_bool_get (rt->hori))
					{
						sep = gtk_vseparator_new ();
						align = gtk_alignment_new (0.5, 0.5, 1.0, 0.6);
					} else
					{
						sep = gtk_hseparator_new ();
						align = gtk_alignment_new (0.5, 0.5, 0.6, 1.0);
					}
					gtk_container_add (GTK_CONTAINER (align), sep);
					gtk_container_set_border_width (GTK_CONTAINER (align), E2_PADDING_XSMALL);
					GtkToolItem *tool = gtk_tool_item_new();
					gtk_container_add (GTK_CONTAINER (tool), align);
					gtk_toolbar_insert (GTK_TOOLBAR (rt->toolbar), tool, -1);
					gtk_widget_show_all (GTK_WIDGET (tool));
				}
					break;
				case E2_ACTION_TYPE_BOOKMARKS:
					{
						E2_OptionSet *set = e2_option_get ("bookmarks");
						GtkTreeIter iter;
						if ((arg != NULL) && (*arg != '\0'))
						{
							if (e2_tree_find_iter_from_str (set->ex.tree.model, 0, arg, &iter, TRUE))
							{
								GtkTreeIter iter2;
								if (gtk_tree_model_iter_children (set->ex.tree.model, &iter2, &iter))
								{
									rt->has_bookmarks = TRUE;
									_e2_toolbar_add_bookmarks (set, rt, &iter2, action);
								}
							}
						}
						else if (gtk_tree_model_get_iter_first (set->ex.tree.model, &iter))
						{
							rt->has_bookmarks = TRUE;
							_e2_toolbar_add_bookmarks (set, rt, &iter, action);
						}
					}
					break;
				case E2_ACTION_TYPE_TOGGLE:
					{
						rt->has_toggle = TRUE;	//remember that this bar has toggle(s)
						button = _e2_toolbar_add_button (rt, name, icon, tip,
							e2_action_pack_runtime (action, arg, g_free));	//WHAT is this fn & args for ??
							//NULL);
						gboolean thisarg = g_str_equal (arg, "true");	//no translation
						//setup extended toggle data
						E2_ToggleActionEx *ex = action->data2;
						if (thisarg)
							ex->true_button = button;
						else
							ex->false_button = button;
						ex->bar_rt = rt;	//remember the bar, for specific cleanups
						//decide whether to re-hide the button
						switch (ex->state)
						{
							case E2_TOGGLE_NONE:
								//when first created, 1st button of any pair sets the default
								ex->current_state = thisarg;
								ex->state = E2_TOGGLE_INIT;
								break;
							case E2_TOGGLE_INIT:
								//when first created, 2nd button of any pair is hidden
								gtk_widget_hide (button);
								break;
							case E2_TOGGLE_DESTROYED:
								//when destroyed, keep the former status
//								ex->state = E2_TOGGLE_DONE;
//							case E2_TOGGLE_DONE:
								if (thisarg != ex->current_state)
									gtk_widget_hide (button);
							default:
								break;
						}
					}
					break;
				case E2_ACTION_TYPE_ITEM:
					button = _e2_toolbar_add_button (rt, name, icon, tip,
						e2_action_pack_runtime (action, arg, g_free)); //WHAT is this fn & args for ??
					gtk_widget_show_all (button);
					break;
				default:
					break;
			}
			if (button != NULL)
				g_object_set_data_full (G_OBJECT (button), "path",
					gtk_tree_model_get_path (model, iter),
					(GDestroyNotify) gtk_tree_path_free);
			//cleanup
			if ((action->type != E2_ACTION_TYPE_ITEM) &&
				(action->type != E2_ACTION_TYPE_TOGGLE))
				g_free (arg);
		} 
		//cleanup
		g_free (name);
		g_free (type);
		g_free (icon);
		g_free (tip);
	} while ((gtk_tree_model_iter_next (model, iter)) == TRUE);
/*	//DEBUG
	GtkRequisition req3;
	gtk_widget_size_request (GTK_WIDGET (rt->toolbar), &req3);
	gint count = gtk_toolbar_get_n_items (rt->toolbar);
	GtkAllocation *alloc;
	alloc = &(GTK_WIDGET (rt->toolbar))->allocation; */
}
/**
 * @brief create toolbar associated with @a rt
 *
 * @param rt ptr to toolbar data struct
 *
 * @return
 **/
void e2_toolbar_create (E2_ToolbarRuntime *rt)
{
	if (!_e2_option_bool_get (rt->show)) return;
	
	rt->blocked = FALSE;
	rt->hidden = FALSE;
	rt->is_too_small = FALSE;
	rt->button_rest = NULL;
	rt->toolbar = GTK_TOOLBAR (gtk_toolbar_new ());
	//setup toolbar
	gtk_toolbar_set_tooltips (rt->toolbar, _e2_option_bool_get (rt->tooltips));
	//strictly, we are not supposed to set style/size - the theme should do it
	gtk_toolbar_set_style (rt->toolbar, _e2_option_int_get (rt->style) - 1);	//-1 corrects for our "default" = 0
	//don't want to change the option size if it's 'default', so keep a working copy
	//(can't use rt->toolbar->icon_size, it reverts to default whenever the bar changes size)
	//e2 enumerator has same order as gtk enumerator
	rt->icon_size = _e2_option_int_get (rt->isize);
	if (rt->icon_size == 0)	//default
	{
/* gtk 2.6.8 spits warning about gtk-toolbar-icon-size property
	not existing, tho' API doco says it does !
		GtkSettings* defs = gtk_settings_get_default ();
        g_object_get (G_OBJECT (defs), "gtk-toolbar-icon-size", &(rt->icon_size), NULL);
		if (rt->icon_size == 0) */
			rt->icon_size = GTK_ICON_SIZE_LARGE_TOOLBAR;
	}

	//handlebox as toolbar container
	rt->toolbar_container = gtk_handle_box_new ();
	gtk_handle_box_set_shadow_type (GTK_HANDLE_BOX (rt->toolbar_container), GTK_SHADOW_NONE);
//don't want these signals before window is displayed, they are redundant until then
//SO DO THIS IN TOOLBAR SPACE-HANDLER INITIALIZER
//	gulong handler = g_signal_connect (G_OBJECT (rt->toolbar), "size-allocate", G_CALLBACK (_e2_toolbar_resize_cb), rt);
//	if (!GTK_WIDGET_VISIBLE (app.main_window))
//		g_signal_handler_block (rt->toolbar, handler);

	//setup toolbar orientation
	if (_e2_option_bool_get (rt->hori))
	{
		rt->toolbar_container_box = gtk_hbox_new (FALSE, 0);
		gtk_toolbar_set_orientation (rt->toolbar, GTK_ORIENTATION_HORIZONTAL);
		gtk_handle_box_set_handle_position (GTK_HANDLE_BOX (rt->toolbar_container), GTK_POS_LEFT);
		gtk_handle_box_set_snap_edge (GTK_HANDLE_BOX (rt->toolbar_container), GTK_POS_TOP);
	} else
	{
		rt->toolbar_container_box = gtk_vbox_new (FALSE, 0);
		gtk_toolbar_set_orientation (rt->toolbar, GTK_ORIENTATION_VERTICAL);
		gtk_handle_box_set_handle_position (GTK_HANDLE_BOX (rt->toolbar_container), GTK_POS_TOP);
		gtk_handle_box_set_snap_edge (GTK_HANDLE_BOX (rt->toolbar_container), GTK_POS_LEFT);
	}
	gtk_container_add (GTK_CONTAINER (rt->toolbar_container),
		rt->toolbar_container_box);

	//scrollbars in force
	if (_e2_option_int_get (rt->space) == 1)
	{
		GtkWidget *sw;
		if (_e2_option_bool_get (rt->hori))
			sw = e2_widget_get_sw (GTK_POLICY_AUTOMATIC, GTK_POLICY_NEVER);
		else
			sw = e2_widget_get_sw (GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
		gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw), GTK_SHADOW_NONE);
		gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (sw), GTK_WIDGET (rt->toolbar));
		gtk_box_pack_start (GTK_BOX (rt->toolbar_container_box), sw, TRUE, TRUE, 0);
		gtk_widget_show (sw);
//#ifdef GTK26
//		rt->hide_threshold = 0;
//#endif
	} else
		gtk_box_pack_start (GTK_BOX (rt->toolbar_container_box), GTK_WIDGET (rt->toolbar), TRUE, TRUE, 0);

//#ifdef GTK26
	if (rt->bar_items_list != NULL)
	{
		g_list_free (rt->bar_items_list) ;
		rt->bar_items_list = NULL;
	}
//#endif

	//add toolbar to the main window
	gboolean hor_now =  _e2_option_bool_get (rt->hori);
	GtkWidget *box;
	gint type = _e2_option_int_get (rt->type);
	switch (type)
	{
		case 3:	//pane 2
			box = (hor_now) ? app.pane2.inner_box : app.pane2.outer_box;
			break;
		case 2:	//pane 1
			box = (hor_now) ? app.pane1.inner_box : app.pane1.outer_box;
			break;
		case 1:	//both panes
			{
			gboolean hor_cfg = e2_option_bool_get ("panes-horizontal");
			if (hor_now)
				box = (hor_cfg) ? app.window.panes_inner_box : app.window.panes_outer_box;
			else
				box = (hor_cfg) ? app.window.panes_outer_box : app.window.panes_inner_box;
			}
			break;
		default:	//main window
			box = (hor_now) ? app.vbox_main : app.hbox_main;
			break;
	}
//	_e2_toolbar_pack_and_prio (packer, rt);
	gtk_box_pack_start (GTK_BOX (box), rt->toolbar_container, FALSE, FALSE, 0);
	gtk_box_reorder_child (GTK_BOX (box), rt->toolbar_container, _e2_option_int_get (rt->priority));

	//bar items
	GtkTreeIter iter;
	if ((gtk_tree_model_get_iter_first (rt->set->ex.tree.model, &iter)) == TRUE)
		_e2_toolbar_add_items (rt->set->ex.tree.model, &iter, 0, rt);

	//hookup the context menu
	g_signal_connect (G_OBJECT (rt->toolbar_container), "button-press-event", 
		G_CALLBACK (_e2_toolbar_click_cb), rt);

/* this is redundant: non-homog. buttons are setup with these properties
	GList *children = rt->toolbar->children;
	while (children != NULL)
	{
		GtkToolbarChild *child = children->data;  //deprecated
		if ((GTK_IS_BUTTON (child->widget)) && (GTK_IS_LABEL (child->label)))
		{
			gtk_label_set_use_markup (GTK_LABEL (child->label), TRUE);
			gtk_label_set_use_underline (GTK_LABEL (child->label), TRUE);
		}
		children = g_list_next (children);
	}
*/

//	gtk_widget_show_all (rt->toolbar_container); //can't show all as it displays toggle buttons
	gtk_widget_show (GTK_WIDGET (rt->toolbar));
	gtk_widget_show (rt->toolbar_container);
// need to show bar before its space handler can be properly initialized
	gtk_widget_show (rt->toolbar_container_box);

	//FIXME: add minimum sizes to panes if the toolbar is a small panebar
/*	if (type > 1)
	{
		
	} */
}
/**
 * @brief destroy and re-create toolbar specified by @a rt
 *
 * @param rt ptr to toolbar data struct
 *
 * @return
 **/
void e2_toolbar_recreate (E2_ToolbarRuntime *rt)
{
	e2_toolbar_destroy (rt);
	e2_toolbar_create (rt);
//#ifdef GTK26
	e2_toolbar_initialise_space_handler (rt);
//#endif
	//FIXME set appropriate toggle buttons
}
/**
 * @brief destroy toolbar widget for toolbar specified by @a rt
 *
 * @param rt ptr to toolbar data struct
 *
 * @return
 **/
void e2_toolbar_destroy (E2_ToolbarRuntime *rt)
{
/*	if (rt->toolbar != NULL)
	{
		gtk_widget_destroy (GTK_WIDGET (rt->toolbar));
		rt->toolbar = NULL;
	}
	if (rt->toolbar_container_box != NULL)
	{
		gtk_widget_destroy (rt->toolbar_container_box);
		rt->toolbar_container_box = NULL;
	}
	if (rt->toolbar_container != NULL)
	{
		gtk_widget_destroy (rt->toolbar_container);
		rt->toolbar_container = NULL;
	}
*/
	//destroy the outer-most container widget of the bar
	gtk_widget_destroy (rt->toolbar_container);
	//ensure any and all toggle buttons are rebuilt with their current state
	if (rt->has_toggle)
		e2_toolbar_toggle_buttons_set_destroyed (rt);
}
/**
 * @brief re-create all toolbars
 *
 * @return
 **/
void e2_toolbar_recreate_all (void)
{
	E2_ToolbarData **thisbar;
	for (thisbar = app.bars; *thisbar != NULL; thisbar++)
		e2_toolbar_recreate ((*thisbar)->rt);
}
/**
 * @brief initialise various toolbar data items, and one action
 *
 * @param barnum enumerator which tells which bar to process
 *
 * @return
 **/
void e2_toolbar_initialise (E2_BarType barnum)
{
	gchar *barname = app.bars[barnum]->name;	//internal (untranslated) bar identifier, "taskbar" etc
	E2_ToolbarRuntime *rt = app.bars[barnum]->rt;

//	these prevent proper toolbar destruction as part of config refresh ...
//	rt->toolbar = NULL;
//	rt->toolbar_container = NULL;

	rt->has_bookmarks = FALSE;
	rt->has_command_line = FALSE;
	rt->is_too_small = FALSE;
	rt->button_rest = NULL;
	rt->name = barname;
	rt->public_name = app.bars[barnum]->public_name;
	//grab references to options registered at session-start
	rt->set = e2_option_get (barname);
	gchar *option_name = g_strconcat (barname, "-show", NULL);
	rt->show = e2_option_get (option_name);
	g_free (option_name);
	option_name = g_strconcat (barname, "-tooltips", NULL);
	rt->tooltips =  e2_option_get (option_name);
	g_free (option_name);
	option_name = g_strconcat (barname, "-space", NULL);
	rt->space =  e2_option_get (option_name);
	g_free (option_name);
	option_name = g_strconcat (barname, "-type", NULL);
	rt->type = e2_option_get (option_name);
	g_free (option_name);
	option_name = g_strconcat (barname, "-hori", NULL);
	rt->hori = e2_option_get (option_name);
	g_free (option_name);
	option_name = g_strconcat (barname, "-priority", NULL);
	rt->priority = e2_option_get (option_name);
	g_free (option_name);
	option_name = g_strconcat (barname, "-relief", NULL);
	rt->relief = e2_option_get (option_name);
	g_free (option_name);
	option_name = g_strconcat (barname, "-same", NULL);
	rt->same = e2_option_get (option_name);
	g_free (option_name);
	option_name = g_strconcat (barname, "-style", NULL);
	rt->style = e2_option_get (option_name);
	g_free (option_name);
	option_name = g_strconcat (barname, "-isize", NULL);
	rt->isize = e2_option_get (option_name);
	g_free (option_name);
}
/**
 * @brief setup data for all toolbars
 *
 * @return
 **/
void e2_toolbar_data_create (void)
{
	static gchar *barnames [E2_BAR_COUNT] =
	{
		"taskbar",
		"panebar1",
		"panebar2",
		"commandbar"
	};
	static gchar *transnames [E2_BAR_COUNT];
	transnames[E2_BAR_TASK] = _C(37);
	transnames[E2_BAR_PANE1] = _C(26);
	transnames[E2_BAR_PANE2] = _C(27);
	transnames[E2_BAR_COMMAND] = _C(6);
	static E2_ToolbarRuntime *barsrt [E2_BAR_COUNT] =
	{
		&app.toolbar,
		&app.pane1.toolbar,
		&app.pane2.toolbar,
		&app.commandbar
	};

	app.bars = g_malloc (sizeof (E2_ToolbarData*) * (E2_BAR_COUNT+1));	//space for bars and trailing null
	E2_ToolbarData **thisbar = app.bars;
	gint i;
	for (i=0; i<E2_BAR_COUNT; i++)
	{
		*thisbar = g_malloc (sizeof (E2_ToolbarData));
		(*thisbar)->type = (E2_BarType) i;
		(*thisbar)->name = barnames [i];
		(*thisbar)->public_name = transnames [i];
		(*thisbar)->rt = barsrt[i];
		thisbar++;
	}
	*thisbar = NULL;
}
/**
 * @brief register actions for toolbars
 *
 * @return
 **/
void e2_toolbar_actions_register (void)
{
//	E2_ToolbarData **thisbar;
//	for (thisbar = app.bars; *thisbar != NULL; thisbar++)
//	{
/*	this should be with the options fn, and in fact, it is already !
		gchar *option_name = g_strconcat ((*thisbar)->name, "-show", NULL);
		E2_OptionSet *set = e2_option_get (option_name);
		g_free (option_name);
*/
		//FIXME bar public names have spaces, not allowed in action names
		//anyway, why do we need this as an action ??
/*		gchar *action_name = g_strconcat(_A(2),".",_A(80),
			"_",(*thisbar)->public_name,NULL);  //_("configure.toggle_<barname>
		e2_action_register (action_name, E2_ACTION_TYPE_TOGGLE,
			_e2_toolbar_toggle, (*thisbar)->rt, FALSE, 0, set->name);
*/
//	}
}
/**
 * @brief set up options for toolbar @a name
 *
 * @param barnum code which identifies the bar
 *
 * @return
 **/
void e2_toolbar_options_register (E2_BarType barnum)
{
	gchar *name = app.bars[barnum]->name;	//internal (untranslated) bar identifier, "taskbar" etc
	gchar *public_name = app.bars[barnum]->public_name;

	gint i = (gint) barnum;
	//these must be in the same order as the E2_BarType enumerator
	gint data[4][10] = {  //defaults for ...
		{ TRUE, TRUE, 2, 2, FALSE, 1, FALSE, TRUE, 1, 3 },  //taskbar
		{ TRUE, TRUE, 2, 2, TRUE, 0, FALSE, FALSE, 1, 2 },  //panebar1
		{ TRUE, TRUE, 2, 3, TRUE, 0, FALSE, FALSE, 1, 2 },  //panebar2
		{ TRUE, TRUE, 2, 0, TRUE, 1, FALSE, FALSE, 1, 2 },  //commandbar
		};

	//show the bar, j=0
	gchar *group_name = g_strconcat (public_name, ".", _C(24), ":", _C(23), NULL); //_("<barname>.options:miscellaneous"
	gchar *option_name = g_strconcat (name, "-show", NULL);
	gchar *desc = g_strconcat (_("show the "), public_name, NULL);
	gchar *tip = g_strdup_printf (_("This determines whether the %s is displayed or hidden"), public_name);
	e2_option_bool_register (option_name, group_name, desc, tip, NULL, data[i][0],
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP 
		| E2_OPTION_FLAG_FREENAME | E2_OPTION_FLAG_FREEDESC 
		| E2_OPTION_FLAG_FREETIP | E2_OPTION_FLAG_BUILDBARS);
	//show tooltips, j=1
	option_name = g_strconcat (name, "-tooltips", NULL);
	desc = g_strdup_printf (_("show tooltips for %s buttons"), public_name);
	tip = g_strdup_printf(_("If deactivated, tooltips will not be displayed for %s buttons"), public_name);
	e2_option_bool_register (option_name, group_name, desc, tip, NULL, data[i][1],
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREENAME
		| E2_OPTION_FLAG_FREEDESC | E2_OPTION_FLAG_FREETIP);
	//restricted-space handling, j=2
	option_name = g_strconcat (name, "-space", NULL);
	tip = g_strdup_printf (_("This determines the method for accessing %s elements that are hidden due to lack of screen-space"), public_name);
	gchar *opt_toolbar_space[] =
		{_("none"), _("use scrollbars"), _("use rest button"), NULL};
	e2_option_sel_register (option_name, group_name, _("space handling"), tip, NULL, data[i][2], opt_toolbar_space,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREENAME
		| E2_OPTION_FLAG_FREETIP | E2_OPTION_FLAG_BUILDBARS);

	group_name = g_strconcat (public_name, ".", _C(24), ":", _C(32), NULL);  //_("<barname>.options:position"
	//bar type, j=3
	option_name = g_strconcat (name, "-type", NULL);
	desc = g_strconcat (public_name, _(" container"), NULL);
	tip = g_strdup_printf (_("This determines the 'box' into which the %s is placed"), public_name);
	gchar *opt_toolbar_type[] =
		{_("main window"), _("both panes"), _("file-pane 1"), _("file-pane 2"), NULL};
	e2_option_sel_register (option_name, group_name, desc, tip, NULL, data[i][3], opt_toolbar_type,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP 
		| E2_OPTION_FLAG_FREENAME  | E2_OPTION_FLAG_FREEDESC 
		| E2_OPTION_FLAG_FREETIP | E2_OPTION_FLAG_BUILDALL);
	//bar is horizontal, j=4
	option_name = g_strconcat (name, "-hori", NULL);
	desc =	g_strconcat (public_name, _(" horizontal"), NULL);
	tip = g_strdup_printf (_("This determines whether the %s is displayed horizontally or vertically"), public_name);
	e2_option_bool_register (option_name, group_name, desc, tip,	NULL, data[i][4],
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREENAME 
		| E2_OPTION_FLAG_FREEDESC | E2_OPTION_FLAG_FREETIP 
		| E2_OPTION_FLAG_BUILDBARS);
	//bar placement, j=5
	option_name = g_strconcat (name, "-priority", NULL);
	desc = g_strconcat (public_name, _(" priority"), NULL);
	tip = _("This determines the order of toolbars (0 = left or top)");
	e2_option_int_register (option_name, group_name, desc, tip, NULL, data[i][5], 0, 10,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREENAME 
		| E2_OPTION_FLAG_FREEDESC | E2_OPTION_FLAG_BUILDALL);

	group_name = g_strconcat (public_name, ".", _C(24), ":", _C(35),  NULL);  //_(".options:style"
	//buttons always have relief, j=6
	option_name = g_strconcat (name, "-relief", NULL);
	desc = g_strconcat (public_name, _(" buttons have relief"), NULL);
	e2_option_bool_register (option_name, group_name, desc, 
		_("Buttons with relief show their outline continually, not just when 'moused'"),
		NULL, data[i][6],
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP 
		| E2_OPTION_FLAG_FREENAME | E2_OPTION_FLAG_FREEDESC
		| E2_OPTION_FLAG_BUILDBARS);
	//buttons are same size, j=7
	option_name = g_strconcat (name, "-same", NULL);
	desc = g_strconcat (public_name, _(" buttons are the same size"), NULL);
	e2_option_bool_register (option_name, group_name, desc,
		_("Equal-width buttons look good on a vertical bar, bad on a horizontal bar"),
		NULL, data[i][7],
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREENAME
		| E2_OPTION_FLAG_FREEDESC | E2_OPTION_FLAG_BUILDBARS);
	//buttons style, j=8
	option_name = g_strconcat (name, "-style", NULL);
	gchar *opt_toolbar_style[] =
		{_("theme"),_("icon only"), _("label only"), _("icon above label"), _("icon beside label"), NULL};
	tip = g_strdup_printf (_("'%s' uses the Gtk default, '%s' leaves most space for other things"),
		opt_toolbar_style[0], opt_toolbar_style[1]);
	e2_option_sel_register (option_name, group_name, _("button style"), tip, NULL,
		data[i][8], opt_toolbar_style,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREENAME
		| E2_OPTION_FLAG_FREETIP | E2_OPTION_FLAG_BUILDBARS);
	//buttons icon size, j=9
	option_name = g_strconcat (name, "-isize", NULL);
	gchar *opt_toolbar_isize[] =
		{_("theme"),_("menu"), _("toolbar small"), _("toolbar large"), _("button"), _("dnd"), _("dialog"),NULL};
	tip = g_strdup_printf (_("'%s' uses the Gtk default, '%s' is smallest, '%s' is largest"), 
		opt_toolbar_isize[0], opt_toolbar_isize[1], opt_toolbar_isize[6]);
		e2_option_sel_register (option_name, group_name, _("icon size"), tip,
		NULL, data[i][9], opt_toolbar_isize,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREENAME 
		| E2_OPTION_FLAG_FREETIP | E2_OPTION_FLAG_BUILDBARS);
}
/**
 * @brief setup config options specific to the toolbar
 * In particular, this sets the default content of the bar
 * @return
 **/
void e2_toolbar_toolbaroptions_register (void)
{
	gchar *internal_name = app.bars[E2_BAR_TASK]->name;	//"taskbar", no translation
	gchar *public_name = app.bars[E2_BAR_TASK]->public_name;
	E2_OptionSet *set = e2_option_tree_register (internal_name, public_name, public_name,
		NULL, NULL, NULL, E2_OPTION_TREE_UP_DOWN | E2_OPTION_TREE_ADD_DEL,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_BUILDBARS);
	e2_option_tree_add_column (set, _("Label"), E2_OPTION_TREE_TYPE_STR, 0, "",
		NULL, NULL, 0, NULL, NULL);
	e2_option_tree_add_column (set, _("Icon"), E2_OPTION_TREE_TYPE_ICON, 0, "",
		NULL, NULL, 0, NULL, NULL);
	e2_option_tree_add_column (set, _("Tooltip"), E2_OPTION_TREE_TYPE_STR, 0, "",
		NULL, NULL, 0, NULL, NULL);
	e2_option_tree_add_column (set, _("Action"), E2_OPTION_TREE_TYPE_SEL, 0, _A(17),
		e2_actions_fill_store, GINT_TO_POINTER (E2_ACTION_EXCLUDE_TOOLBAR), 0, NULL, NULL);
	e2_option_tree_add_column (set, _("Argument"), E2_OPTION_TREE_TYPE_STR, 0, "",
		NULL, NULL, 0, NULL, NULL);
	e2_option_tree_sync_columns (set);

	_e2_option_tree_prepare_default (set,
		g_strconcat("<",internal_name, NULL),
		g_strconcat("|||",_A(0),".",_A(23),"|",_C(37),NULL),	//this arg needs to be same as in bookmarks.c
		g_strconcat("|||",_A(17),"|",NULL),
		g_strconcat(_("_Copy"),"|gtk-copy|",
			_("Copy item(s) selected in the active pane to the other one"),"|",_A(5),".",_A(33),"|",NULL),
		g_strconcat(_("_Move"),"|"ICON_DIR"move_"E2IP".png|",
			_("Move item(s) selected in the active pane to the other one"),"|",_A(5),".",_A(48),"|",NULL),
		g_strconcat(_("_Link"),"|"ICON_DIR"symlink_"E2IP".png|",
			_("Symlink item(s) selected in the active pane to the other one"),"|",_A(5),".",_A(68),"|",NULL),
		g_strconcat(_("Re_name.."),"|gtk-convert|",
			_("Rename item(s) selected in the active pane"),"|",_A(5),".",_A(60),"|",NULL),
		g_strconcat(_("_Trash"),"|"ICON_DIR"trash_"E2IP".png|",
			_("Move item(s) selected in the active pane to a trashbin"),"|",_A(5),".",_A(81),"|",NULL),
		g_strconcat(_("Ma_ke dir.."),"|gtk-open|",_("Create new directory(ies)"),"|",_A(3),".",_A(50),"|",NULL),
//		g_strconcat("|||",_A(17),"|",NULL),
//		g_strconcat("\\<span foreground='red'><b>",_("_Delete"),"</b></span>|gtk-delete|",
//			_("Delete item(s) selected in the active pane"),"|",_A(5),".",_A(35),"|",NULL),
//		g_strconcat("|||",_A(17),"|",NULL),
		g_strconcat("|||",_A(17),"|",NULL),
		g_strconcat(_("Re_fresh"),"|gtk-refresh|",_("Update pane contents"),"|",_A(11),".",_A(59),"|",NULL),
		g_strconcat(_("_Switch"),"|"ICON_DIR"switch_"E2IP".png|",
			_("Toggle the active pane"),"|",_A(11),".",_A(67),"|",NULL),
		g_strdup(">"),
		NULL);

	e2_toolbar_options_register (E2_BAR_TASK);  //setup options relevant to all toolbars
}
/**
 * @brief setup config options specific to the commandbar
 * In particular, this sets the default content of the bar
 * @return
 **/
void e2_toolbar_commandbaroptions_register (void)
{
	gchar *internal_name = app.bars[E2_BAR_COMMAND]->name;	// "commandbar", no translation
	gchar *public_name = app.bars[E2_BAR_COMMAND]->public_name;
	E2_OptionSet *set = e2_option_tree_register (internal_name, public_name, public_name,
		 NULL, NULL, NULL, E2_OPTION_TREE_UP_DOWN | E2_OPTION_TREE_ADD_DEL,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_BUILDBARS);
	e2_option_tree_add_column (set, _("Label"), E2_OPTION_TREE_TYPE_STR, 0, "",
		NULL, NULL, 0, NULL, NULL);
	e2_option_tree_add_column (set, _("Icon"), E2_OPTION_TREE_TYPE_ICON, 0, "",
		NULL, NULL, 0, NULL, NULL);
	e2_option_tree_add_column (set, _("Tooltip"), E2_OPTION_TREE_TYPE_STR, 0, "",
		NULL, NULL, 0, NULL, NULL);
	e2_option_tree_add_column (set, _("Action"), E2_OPTION_TREE_TYPE_SEL, 0, _A(17),
		e2_actions_fill_store, GINT_TO_POINTER (E2_ACTION_EXCLUDE_TOOLBAR), 0, NULL, NULL);
	e2_option_tree_add_column (set, _("Argument"), E2_OPTION_TREE_TYPE_STR, 0, "",
		NULL, NULL, 0, NULL, NULL);
	e2_option_tree_sync_columns (set);

	_e2_option_tree_prepare_default (set,
		g_strconcat("<",internal_name,NULL),
//		g_strconcat("|gtk-zoom-fit|",_("Toggle maximized output pane"),"|",_A(8),".",_A(28),"|","*,0",NULL),	//no translation
		g_strconcat(_("Full"),"|gtk-zoom-fit|",_("Maximize output pane"),"|",_A(8),".",_A(74),"|false",NULL),		//no translation
		g_strconcat(_("Shrink"),"|gtk-zoom-out|",_("Un-maximize output pane"),"|",_A(8),".",_A(74),"|true",NULL),		//no translation
//		g_strconcat("|gtk-go-up|",_("Toggle visible output pane"),"|",_A(8),".",_A(28),"|","*,1",NULL),  	//no translation
		g_strconcat(_("Hide"),"|"ICON_DIR"output_hide_"E2IP".png|",_("Hide output pane"),"|",_A(8),".",_A(72),"|false",NULL), 	//no translation
		g_strconcat(_("Show"),"|"ICON_DIR"output_show_"E2IP".png|",_("Show output pane"),"|",_A(8),".",_A(72),"|true",NULL), 	//no translation

		g_strconcat(_("Clear"),"|gtk-clear|",_("Clear output pane"),"|",_A(8),".",_A(29),"|",NULL),
		g_strconcat("|||",_A(1),".",_A(24),"|","100,",_A(89),NULL),  //1st arg is min size, integer, do not translate
		g_strconcat(_("cl"),"|"ICON_DIR"cl_clear_"E2IP".png|",_("Clear command line"),"|",_A(1),".",_A(29),"|",NULL),
		g_strconcat(_("ps"),"|"ICON_DIR"ps_"E2IP".png|",_("List child processes"),"|",_A(1),".",_A(47),"|",NULL),
		g_strconcat(_("du"),"|"ICON_DIR"plugin_du_"E2IP".png|",_("Calculate disk usage of selected item(s)"),"|",PLUGIN,".","du","|",NULL),  //conformed to new internal name
		g_strconcat(_("_Find.."),"|gtk-find|",_("Find item in active pane, by name"),"|",_A(5),".",_A(6),"|",NULL),
		g_strconcat(_("_X"),"|"ICON_DIR"terminal_"E2IP".png|",_("Open terminal at the active directory"),"|",_A(16),"|$[command-xterm] &",NULL),
		//this command works ok with xterm, but not with gnome-terminal
//		g_strconcat(_("su"),"|"ICON_DIR"su_"E2IP".png|",_("Run something as root"),"|",_A(16),"|$[command-xterm] -e 'su -c \"%{",_("Enter command:"),"}\";echo -n \"",_("Done, Press enter..."),"\";read'",NULL),
		g_strconcat(_("su.."),"|"ICON_DIR"su_"E2IP".png|",_("Run something as root"),"|",_A(16),"|xterm -e 'su -c \"%{",_("Enter command:"),"}\";echo -n \"",_("Done. Press enter "),"\";read'",NULL),
#ifdef E2_FS_MOUNTABLE
		g_strconcat(_("mts.."),"|"ICON_DIR"mounts_"E2IP".png|",_("Mount or unmount a device"),"|",_A(94),".",_A(23),"|",_C(6),NULL),
#endif
		g_strconcat("|||",_A(17),"|",NULL),
		g_strconcat(_("_Help"),"|gtk-help|",_("Get information about this program"),"|",_A(3),".",_A(26),"|",NULL),
		g_strconcat(_("_Settings.."),"|gtk-preferences|",_("View/change configuration settings for this program"),"|",_A(2),".",_A(32),"|",NULL),
		g_strconcat("|||",_A(17),"|",NULL),
		g_strconcat(_("_Quit"),"|gtk-quit|",_("Close this program"),"|",_A(1),".",_A(83),"|",NULL),
		g_strdup(">"),
		NULL);

	e2_toolbar_options_register (E2_BAR_COMMAND);  //setup options relevant to all toolbars
}
