/* $Id: e2_window.c 1036 2005-08-15 02:14:29Z tpgww $
 * 
 * Copyright (C) 2004-2005 tooar <tooar@gmx.net>
 * Portions copyright (C) 1999 Michael Clark.
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/e2_window.c 
 * @brief main window functions
 *
 * main window functions, including actions on panes
 *
 **/

#include "emelfm2.h"
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include "e2_window.h"
#include "e2_option.h"
#include "e2_action.h"
#include "e2_context_menu.h"
#include "e2_toolbar.h"
#include "e2_filelist.h"
#include "e2_task.h"

extern GList *cols_data;
extern gint col_width_store[2][MAX_COLUMNS];
extern gint stored_col_order[2][MAX_COLUMNS];

static gdouble _e2_window_get_pos (GtkWidget *paned);

static void _e2_window_set_icon (void);

  /*********************/
 /***** callbacks *****/
/*********************/
/**
 * @brief callback for signal emitted when file panes divider is moved
 *
 *
 * @param widget UNUSED, probably the widget clicked to initiate the action
 * @param alloc
 * @param rt pointer to window data struct
 *
 * @return TRUE, always
 **/
static gboolean _e2_window_filepanes_adjusted_cb (GtkWidget *widget, 
	GtkAllocation *alloc, E2_WindowRuntime *rt)
{
//	printd (DEBUG, "_e2_window_filepanes_adjusted_cb (widget:,alloc:%d-%d,rt:_)", alloc->width, alloc->height);
	rt->panes_paned_ratio = _e2_window_get_pos (rt->panes_paned);
//	printd (DEBUG, "1ratio: %f", rt->panes_paned_ratio);
	if (((rt->panes_paned_ratio == 0.0) && (app.pane_current == &app.pane1)) ||
		((rt->panes_paned_ratio == 1.0) && (app.pane_current == &app.pane2)))
		e2_pane_activate_other ();
	if ((rt->panes_paned_ratio != 0.0) && (rt->panes_paned_ratio != 1.0))
		rt->panes_paned_ratio_last = rt->panes_paned_ratio;
	return TRUE;
}
/**
 * @brief after a file pane is mapped or unmapped, ensure the correct toggle button is visible in the other pane's toolbar
 *
 * This fixes the toggle button when the panes divider is dragged
 *
 * @param widget ptr to scrolled window for the pane that has just been mapped
 * @param state TRUE for map signal, FALSE for unmap
 *
 * @return
 **/
void _e2_window_map_pane_cb (GtkWidget *widget, gpointer state)
{
	gchar *action_name = (widget == app.pane1.pane) ? _A(10) : _A(9) ; 
	action_name = g_strconcat (action_name,".",_A(74),NULL);  //pane 1 or pane 2 toggle-full
	if (e2_toolbar_toggle_button_get_state (action_name) == GPOINTER_TO_INT (state))
		e2_toolbar_button_toggle (action_name);
	g_free (action_name);
}
/**
 * @brief callback for signal emitted when output pane size is adjusted
 *
 *
 * @param widget UNUSED, probably the widget clicked to initiate the action
 * @param alloc
 * @param rt pointer to window data struct
 *
 * @return TRUE, always
 **/
static gboolean _e2_window_outputpane_adjusted_cb (GtkWidget *widget,
	GtkAllocation *alloc, E2_WindowRuntime *rt)
{
//	printd (DEBUG, "_e2_window_outputpane_adjusted_cb (widget:,alloc:%d-%d,rt:_)", alloc->width, alloc->height);
	rt->output_paned_ratio = _e2_window_get_pos (app.window.output_paned);
//	printd (DEBUG, "2ratio: %f", rt->output_paned_ratio);
	//remember visibility state
	app.output.visible = (rt->output_paned_ratio != 1.0);
	if ((rt->output_paned_ratio != 0.0) && (rt->output_paned_ratio != 1.0))
		rt->output_paned_ratio_last = rt->output_paned_ratio;
	//make sure the correct toggle btn is shown
	gchar *action_name = g_strconcat (_A(8),".",_A(74),NULL);  //output toggle-full
	if (rt->output_paned_ratio < 0.001)
	{
		if (!e2_toolbar_toggle_button_get_state (action_name))
			e2_toolbar_button_toggle (action_name);
	}
	else
		if (e2_toolbar_toggle_button_get_state (action_name))
			e2_toolbar_button_toggle (action_name);
	g_free (action_name);

	return TRUE;
}
/**
 * @brief after the output pane is mapped or unmapped, ensure the correct toggle button is visible in the commandbar
 *
 * This fixes the toggle button when the panes divider is dragged
 *
 * @param widget ptr to output pane scrolled window that has just been mapped
 * @param state TRUE when mapped, FALSE when unmapped
 *
 * @return
 **/
static void _e2_window_map_output_cb (GtkWidget *widget, gpointer state)
{
	//use toolbar runtime and its .option because that striing is not translated
	//make sure the toggle for the other pane is not 'split' form
	gchar *action_name = g_strconcat (_A(8),".",_A(72),NULL);
	if (e2_toolbar_toggle_button_get_state (action_name) == GPOINTER_TO_INT (state))
		e2_toolbar_button_toggle (action_name);
	g_free (action_name);
}
/**
 * @brief callback for signal emitted when the main window is first shown
 *
 * paned dividers' positions are 'crudely' set when those widgets are created,
 * but now all things are setup properly, the sizes are fine-tuned, here
 * Pane resize etc callbacks are then connected, so that we don't get unnecessary
 * callbacks during session setup
 * After all is stable, the toolbars space handling arrangements are initialized,
 * again, to avoid unnecessary callbacks
 *
 * @param window UNUSED, the window widget being shown
 * @param rt pointer to window data struct
 *
 * @return TRUE, always
 **/
static void _e2_window_show_cb (GtkWidget *window, E2_WindowRuntime *rt)
{
	printd (DEBUG, "show main window cb");
	gtk_paned_set_position (GTK_PANED (rt->panes_paned),
		GTK_PANED (rt->panes_paned)->max_position * rt->panes_paned_ratio);
	gtk_paned_set_position (GTK_PANED (app.window.output_paned),
		GTK_PANED (app.window.output_paned)->max_position * rt->output_paned_ratio);
	rt->panes_paned_pos = gtk_paned_get_position (GTK_PANED (rt->panes_paned));
	rt->output_paned_pos = gtk_paned_get_position (GTK_PANED (app.window.output_paned));

	g_signal_connect (G_OBJECT (app.pane1.outer_box), "size-allocate",
		G_CALLBACK (_e2_window_filepanes_adjusted_cb), rt);
	//ensure correct toggle button is displayed when a pane divider is dragged
	g_signal_connect (G_OBJECT (app.pane1.pane), "map",
			G_CALLBACK (_e2_window_map_pane_cb), GINT_TO_POINTER (TRUE));
	g_signal_connect (G_OBJECT (app.pane1.pane), "unmap",
			G_CALLBACK (_e2_window_map_pane_cb), GINT_TO_POINTER (FALSE));
	g_signal_connect (G_OBJECT (app.pane2.pane), "map",
			G_CALLBACK (_e2_window_map_pane_cb), GINT_TO_POINTER (TRUE));
	g_signal_connect (G_OBJECT (app.pane2.pane), "unmap",
			G_CALLBACK (_e2_window_map_pane_cb), GINT_TO_POINTER (FALSE));

	g_signal_connect (G_OBJECT (app.output.text), "size-allocate",
		G_CALLBACK (_e2_window_outputpane_adjusted_cb), rt);
	g_signal_connect (G_OBJECT (app.output.scroll), "map",
			G_CALLBACK (_e2_window_map_output_cb), GINT_TO_POINTER (TRUE));
	g_signal_connect (G_OBJECT (app.output.scroll), "unmap",
			G_CALLBACK (_e2_window_map_output_cb), GINT_TO_POINTER (FALSE));

	WAIT_FOR_EVENTS; //faster window completion if we hide buttons now,
					// if need be (ie dont wait till main loop)
//#ifdef GTK26
	//now that window is established, we are ready to initialise all toolbars' overflow handling
	e2_toolbar_initialise_space_handler (NULL);
	//setup for retaining toolbars' contents when torn-off
	e2_toolbar_initialise_tearoffs ();
//#endif
}

  /*****************/
 /***** utils *****/
/*****************/

/**
 * @brief decrement @a counter if @a iter points to a filelist updir entry ".."
 * This is a selection-foreach function.
 *
 * @param model treemodel for a filepane liststore
 * @param path UNUSED path to selected item being processed
 * @param iter pointer to iter to use for interrogating @a model
 * @param counter pointer to integer to be decremented
 *
 * @return 
 **/
static void _e2_window_updir_entry_check (GtkTreeModel *model,
	  GtkTreePath *path, GtkTreeIter *iter, gint *counter)
{
	gchar *filename;
	gtk_tree_model_get (model, iter, FILENAME, &filename, -1);
	if (g_str_equal (filename, ".."))
		(*counter)--;
	g_free (filename);
}
/**
 * @brief create paned widget to contain the file panes, and surrounding containers
 *
 * Depending on whether the file panes are horiz or vert, a vert or horiz
 * paned widget is packed inside two boxes
 * The 'separator' between the panes is set to the correct position
 * (this is before pane content is added, so that un-necessary content
 * mapping is minimized)
 * All created widgets are 'shown'
 *
 * @param rt ptr to window data struct
 *
 * @return 
 **/
static void _e2_window_create_pane_boxes (E2_WindowRuntime *rt)
{
	gint base;
	if (e2_option_bool_get ("panes-horizontal"))
	{
		rt->panes_horizontal = TRUE;
		rt->panes_outer_box = gtk_hbox_new (FALSE, 0);
		rt->panes_inner_box = gtk_vbox_new (FALSE, 0);
		rt->panes_paned = gtk_vpaned_new ();
		base = app.main_window->allocation.height - 28;  //28 = notional height of commandbar
	}
	else
	{
		rt->panes_horizontal = FALSE;
		rt->panes_outer_box = gtk_vbox_new (FALSE, 0);
		rt->panes_inner_box = gtk_hbox_new (FALSE, 0);
		rt->panes_paned = gtk_hpaned_new ();
		base = app.main_window->allocation.width - 5;  //5 = trial'n'error value to minimise separator jitter
	}
	if (base < 2)
		base = 2;
	//set interim panes sizes, to minimize un-necessary content mapping
	//real size set again, later
	gtk_paned_set_position (GTK_PANED (rt->panes_paned),
		base * rt->panes_paned_ratio);
	rt->panes_paned_pos = gtk_paned_get_position (GTK_PANED (rt->panes_paned));

	gtk_box_pack_start_defaults (GTK_BOX (rt->panes_outer_box), rt->panes_inner_box);
	gtk_box_pack_start_defaults (GTK_BOX (rt->panes_inner_box), rt->panes_paned);

	gtk_widget_show (rt->panes_outer_box);
	gtk_widget_show (rt->panes_inner_box);
	gtk_widget_show (rt->panes_paned);
}
/**
 * @brief get ratio value for position of paned widget divider
 *
 * @param paned the widget for which the ratio is to be calculated
 *
 * @return ratio: paned current position /  _paned->max_position
 **/
static gdouble _e2_window_get_pos (GtkWidget *paned)
{
	GtkPaned *_paned = GTK_PANED (paned);
	gint pos = gtk_paned_get_position (_paned);
	return ((gdouble) pos / _paned->max_position);
}
/**
 * @brief calculate position of the divider for @a paned
 *
 * This converts a string fraction or percentage (with "%" suffix)
 * to an integer, which corresponds to the specified fraction or
 * percentage of @a paned ->max_position
 *
 * @param str string with value to be converted 
 * @param paned the paned widget to which the value applies
 *
 * @return integer value of the 'position' of the paned divider, or -2 if invalid string is provided
 **/
static gint _e2_window_get_pos_from_str (gchar *str, GtkPaned *paned)
{
	if (str == NULL)
		return -2;

	gboolean percent = g_str_has_suffix (str, "%");

	gchar *end = NULL;
	gdouble num = g_ascii_strtod (str, &end);

	if (end == str)
		return -2;

	gint retval;
	if (percent)
		retval = paned->max_position * (num / 100.0);
	else
//		retval = num;
		retval =paned->max_position * num;

	return retval;
}

  /*******************/
 /***** actions *****/
/*******************/
/**
 * @brief toggle between horizontal and vertical file panes
 *
 * This is the method to implement the panes.toggle direction action
 *
 * @param widget UNUSED, probably the widget clicked to initiate the action
 *
 * @return 
 **/
static void _e2_window_toggle_panes_direction (GtkWidget *widget)
{
	e2_option_bool_toggle ("panes-horizontal");
	e2_window_recreate (&app.window);
}
/**
 * @brief change position of the separator between window file-panes or between file & output panes
 *
 * value(s) provided are each a fraction of the relevant window dimension,
 * or * which causes the stored prior value of the fraction to be used
 *
 * @param paned ptr to paned widget to be adjusted
 * @param pos pointer to store for paned position value
 * @param ratio_last pointer to store for the prior value of the 'paned ratio' (position / max. position)
 * @param arg string, possibly comma-separated into <  ,  >
 *
 * @return 
 **/
static void _e2_window_adjust_panes_ratio (GtkPaned *paned, gint *pos,
	gdouble *ratio_last, gchar *arg)
{
	printd (DEBUG, "_e2_window_adjust_panes_ratio (paned:,pos:%d,ratio_last:%f,arg:%s)",
		*pos, *ratio_last, arg);

	if (arg == NULL)
		return;

	gint cur;	// = -2;

	gchar *second = strchr (arg, ',');	//if always ascii ',', don't need g_utf8_strchr()
	if (second != NULL)
	{
		glong offset =  g_utf8_pointer_to_offset (arg, second);
		gchar *temp_str = e2_utf8_ndup (arg, offset);
		gint one = _e2_window_get_pos_from_str (temp_str, paned);
		g_free (temp_str);
		gint two = _e2_window_get_pos_from_str (second + 1, paned);  //assumes the ',' is a single byte long !
		if (two == -2)	//invalid value provided
			return;
		cur = gtk_paned_get_position (paned);
		if (cur == two)
		{
			if (one == -2)
			{
				if ((*pos == paned->max_position) || (*pos == paned->min_position))
					cur = paned->max_position * *ratio_last;
				else
					cur = *pos;
			}
			else
				cur = one;
		}
		else
		{
			if ((cur != 0) && (cur != paned->max_position))
				*pos = cur;
			cur = two;
		}
	} else //there is only one parameter provided
	{
		cur = _e2_window_get_pos_from_str (arg, paned);
		if (cur == -2)
		{	//invalid value provided
			if ((*pos == paned->max_position) || (*pos == paned->min_position))
				cur = paned->max_position * *ratio_last;
			//FIXME: what was that for? it creates a bug if the output pane is hidden
			//and adjust ratio is called with '*' as argument, because it then doesn't
			//adjust the output pane even though it should
//			else
//				return;
//				cur = *pos;
		} else
			*pos = cur;
	}
	gtk_paned_set_position (paned, cur);
}
/**
 * @brief implement adjust-pane-ratio action
 *
 * This is a wrapper for _e2_window_adjust_panes_ratio ()
 *
 * @param rt ptr to window data struct
 * @param arg string with new ratio to be set
 *
 * @return 
 **/
static void _e2_window_adjust_pane_ratio_action (E2_WindowRuntime *rt, gchar *arg)
{
	_e2_window_adjust_panes_ratio (GTK_PANED (rt->panes_paned), 
		&rt->panes_paned_pos, &rt->panes_paned_ratio_last, arg);
}
/**
 * @brief respond to pane 1 full-width toggle button
 *
 * The visible toggle button is swapped, and the panes-ajust fn
 * is called with either "1" or "*"
 *
 * @return 
 **/
static void _e2_window_toggle_full_pane1 (void)
{
	gchar *action_name = g_strconcat(_A(9),".",_A(74),NULL);
	gboolean next_state = e2_toolbar_button_toggle (action_name);
	gchar *arg = (next_state) ? "1" : "*";	//no translation
	E2_WindowRuntime *rt = &app.window;
	_e2_window_adjust_panes_ratio (GTK_PANED (rt->panes_paned), 
		&rt->panes_paned_pos, &rt->panes_paned_ratio_last, arg);
	g_free (action_name);
} 
/**
 * @brief respond to pane2 full-width toggle button
 *
 * The visible toggle button is swapped, and the panes-ajust fn
 * is called with either "0" or "*"
 *
 * @return 
 **/
static void _e2_window_toggle_full_pane2 (void)
{
	gchar *action_name = g_strconcat(_A(10),".",_A(74),NULL);
	gboolean next_state = e2_toolbar_button_toggle (action_name);
	gchar *arg = (next_state) ? "0" : "*" ;
	E2_WindowRuntime *rt = &app.window;
	_e2_window_adjust_panes_ratio (GTK_PANED (rt->panes_paned),
		&rt->panes_paned_pos, &rt->panes_paned_ratio_last, arg);
	g_free (action_name);
}
/**
 * @brief respond to output pane full-window toggle button
 *
 * The visible toggle button is swapped or changed, and the panes-ajust fn
 * is called with either "0" or "*"
 *
 * @return 
 **/
static void _e2_window_toggle_full_output (void)
{
	gchar *action_name = g_strconcat(_A(8),".",_A(74),NULL);
	gboolean next_state = e2_toolbar_button_toggle (action_name);
	gchar *arg = (next_state) ? "0" : "*";
	E2_WindowRuntime *rt = &app.window;
	_e2_window_adjust_panes_ratio (GTK_PANED (rt->output_paned), &rt->output_paned_pos,
		&rt->output_paned_ratio_last, arg);
	g_free (action_name);
	
	if (!next_state)
	{  //cancelling full-window reverts to "last" split ratio, never to hidden
		//make sure the other toggle btn is correct
		if (app.window.output_paned_ratio_last < 0.999)
		{
			action_name = g_strconcat(_A(8),".",_A(72),NULL);
			e2_toolbar_toggle_button_set_state (action_name, FALSE);
			g_free (action_name);
		}
	}
} 
/**
 * @brief respond to output pane visible toggle button
 *
 * The visible toggle button is swapped, and the panes-ajust fn
 * is called with either "1" or "*"
 *
 * @return 
 **/
static void _e2_window_toggle_visible_output (void)
{
	gchar *action_name = g_strconcat(_A(8),".",_A(72),NULL);
	gboolean next_state = e2_toolbar_button_toggle (action_name);
	gchar *arg = (next_state) ? "1" : "*";	//no translation
	E2_WindowRuntime *rt = &app.window;
	_e2_window_adjust_panes_ratio (GTK_PANED (rt->output_paned), &rt->output_paned_pos,
		&rt->output_paned_ratio_last, arg);
	g_free (action_name);
} 
/**
 * @brief show output pane, at its previous size
 *
 * This is a callback for command-line "focus-in-event",
 * and also used generally
 * Runs output adjust ratio action
 *
 * @param widget UNUSED
 * @param data UNUSED
 *
 * @return
 **/
gboolean e2_window_output_show (GtkWidget *widget, gpointer data)
{
//	if (GTK_WIDGET (app.output.text)->allocation.height == 1)
	if (!app.output.visible)
	{
		gchar *action_name = g_strconcat (_A(8),".",_A(28),NULL); //_("output.adjust ratio")
		e2_action_run_simple (action_name, "*");  //no arg translation
		g_free (action_name);
// done in callback
//		action_name = g_strconcat(_A(8),".",_A(72),NULL);
//		e2_toolbar_button_toggle (action_name);
//		g_free (action_name);
	}
	return FALSE;
}
/**
 * @brief hide output pane
 *
 * This is a callback for command-line "focus-out-event",
 * and also used generally
 * Runs output adjust ratio action
 *
 * @param widget UNUSED
 * @param data UNUSED
 *
 * @return
 **/
gboolean e2_window_output_hide (GtkWidget *widget, gpointer data)
{
//	if (GTK_WIDGET (app.output.text)->allocation.height > 1)
	if (app.output.visible)
	{
		gchar *action_name = g_strconcat (_A(8),".",_A(28),NULL); //_("output.adjust ratio")	
		e2_action_run_simple (action_name, ("1"));  //no translation
		g_free (action_name);
// done in callback
//		action_name = g_strconcat(_A(8),".",_A(72),NULL);
//		e2_toolbar_button_toggle (action_name);
//		g_free (action_name);
	}
	return FALSE;
}
/**
 * @brief implement adjust-output-ratio action
 *
 * This is a wrapper for _e2_window_toggle etc
 * The visible toggle button is swapped, and the panes-ajust fn
 * is called with "*" ",1" or ",0"
 *
 * @param rt ptr to window data struct
 * @param arg string with new ratio to be set
 *
 * @return 
 **/
static void _e2_window_adjust_output_ratio_action 
	(E2_WindowRuntime *rt, gchar *arg)
{
	gchar *newarg = NULL;
	//check for "1" or "100%" first
	if (strchr (arg, '1') != NULL	//always ascii, no translation
		&& rt->output_paned_ratio < 1.0)
			newarg = "1";	//no translation
	else if (strchr (arg, '0') != NULL
		&& rt->output_paned_ratio > 0)
			newarg = "0";	//no translation
	else if (strchr (arg, '*') != NULL)
			newarg = "*";	//no translation

	if (newarg != NULL)
		_e2_window_adjust_panes_ratio (GTK_PANED (rt->output_paned),
			&rt->output_paned_pos, &rt->output_paned_ratio_last, newarg);
}

  /******************/
 /***** public *****/
/******************/

/**
 * @brief set output pane size  UNUSED
 *
 * for internal use, not a toolbar toggle
 *
 * @param rt window rt data structure
 * @param arg string indicating the desired state of the pane e.g "*" or "0"
 *
 * @return 
 **/
/* void e2_window_adjust_output_pane_ratio (E2_WindowRuntime *rt, gchar *arg)
{
	_e2_window_adjust_panes_ratio (GTK_PANED (rt->output_paned), 
		&rt->output_paned_pos, &rt->output_paned_ratio_last, arg);
}  */

/**
 * @brief set default window icon list
 *
 * Set default window icon list from files in @a icon_dir, with names
 * starting with @a base, and with trailing 'xx.png' where xx are digits
 *
 * @param base filename base of the icon files
 *
 * @return 
 **/
static void _e2_window_set_icon (void)
{
	gchar *icon_dir = ICON_DIR;
	gchar *base = 	BINNAME"_";	//this is the 'prefix' for window icon filenames
	//this is ok not to vfs-ize, as it must work on the local system
	GError *error = NULL;
	GDir *dir = g_dir_open (icon_dir, 0, &error);
	if (error != NULL)
	{
		printd (WARN, "could not open window icon directory '%s' (%s)", icon_dir, error->message);
		g_error_free (error);
		return;
	}

	GList *list = NULL;
	const gchar *name;
	//name is like "emelfm2_24.png"
	gint len = strlen (base) + 6;	//assuming ascii, +6 allows for xx.png, filters out the 'alternates'
	while ((name = g_dir_read_name (dir)) != NULL)
	{
		if (g_str_has_prefix (name, base) && g_str_has_suffix (name, ".png") &&
			strlen (name) == len)
		{
			gchar *file = g_build_filename (icon_dir, name, NULL);
			GError *error = NULL;
			GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file (file, &error);
			if (error != NULL)
			{
				printd (WARN, "could not open image file '%s' (%s)", file, error->message);
				g_free (file);
				g_error_free (error);
				continue;
			}
			g_free (file);
			list = g_list_append (list, pixbuf);
		}
	}
	g_dir_close (dir);

	gtk_window_set_default_icon_list (list);
}

static gint last_selected_rows = -1;	//-1 ensures it always reports at session start
static gint last_total_rows;

/**
 * @brief update status line message about selected and total items counts
 *
 * This is called periodically by the status line timer function, when that is not suspended
 * Any ".." entry is filtered out from the count
 *
 * @param userdata UNUSED
 *
 * @return TRUE, always (so the timer is never cancelled)
 **/
gboolean e2_window_update_status_bar (gpointer userdata)
{
 	gchar status_text[128];
	gint selected_rows, total_rows;
//	gdk_threads_enter ();	//protect timer fn against thread glitches
	GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (curr_view->treeview));
	GtkTreeSelection *selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (curr_view->treeview));

	selected_rows = gtk_tree_selection_count_selected_rows (selection);
	total_rows = gtk_tree_model_iter_n_children (model, NULL);
	if ((selected_rows != last_selected_rows) || (total_rows != last_total_rows))
	{
		if (e2_option_bool_get ("show-updir-entry"))
		{	//make status-line counts ignore the updir entry
			gtk_tree_selection_selected_foreach (selection, 
				(GtkTreeSelectionForeachFunc) _e2_window_updir_entry_check, &selected_rows);
			total_rows--;
		}
		last_selected_rows = selected_rows;
		last_total_rows = total_rows;
		g_snprintf(status_text, sizeof(status_text), _("%d of %d items selected"),
			selected_rows, total_rows);
		gtk_label_set_text(GTK_LABEL(app.status_bar_label2), status_text);
	}
//	gdk_threads_leave ();
	return TRUE;  //never turn this off
}
/**
 * @brief turn on status bar 'selected files' message refreshing 
 *
 * @param interval no of milliseconds between updates, or -1 for default
 *
 * @return 
 **/
void e2_window_enable_status_update (gint interval)
{
	if (interval < 0)
		interval = E2_STATUSREPORT_INTERVAL;
	cfg.status_timer_id = g_timeout_add (interval, e2_window_update_status_bar, NULL);
}
/**
 * @brief turn off status bar 'selected files' message refreshing 
 *
 * @return 
 **/
void e2_window_disable_status_update (void)
{
	g_source_remove (cfg.status_timer_id);
}
/**
 * @brief display status bar custom message
 *
 * The specified string is added to the end of the hbox, but only displayed
 * if it is the first message in the box
 * This allows multiple messages to be queued
 * FIXME the messages may be cancelled in any order if threads are in force ...
 *
 * @param message string to be displayed, may have pango markup
 *
 * @return 
 **/
void e2_window_show_status_message (gchar *message)
{
	GtkWidget *label = gtk_label_new (NULL);
	gtk_label_set_markup (GTK_LABEL (label), message);
	gtk_box_pack_start (GTK_BOX (app.status_bar_box3), label, FALSE, FALSE, 0);
	GList *children = GTK_BOX (app.status_bar_box3)->children;
	if (g_list_length (children) == 1)
		gtk_widget_show_all (app.status_bar_box3);
}
/**
 * @brief cancel display of status bar custom message
 *
 * @return 
 **/
//FIXME the logic here is crap - needs to be thread-safe
void e2_window_remove_status_message (void)
{
	GList *children = GTK_BOX (app.status_bar_box3)->children;
	gint num = g_list_length (children);
	if (num > 1)
	{	//show the next label in the queue (stupid !!)
		GtkWidget *label = ((GtkBoxChild *) children->next->data)->widget;
		gtk_widget_show_all (label);
	}
	if (num ==1)
		gtk_widget_hide (app.status_bar_box3);
	if (num > 0)
	{
		GtkWidget *label = ((GtkBoxChild *) children->data)->widget;
		gtk_container_remove (GTK_CONTAINER (app.status_bar_box3), label);
	}
}
/**
 * @brief create main window
 * This is called only at session-start, from main () 
 *
 * @param rt ptr to window data struct
 *
 * @return 
 **/
void e2_window_create (E2_WindowRuntime *rt)
{
	_e2_window_set_icon ();

	//setup main window
	app.main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_role (GTK_WINDOW (app.main_window), "main");
	gtk_window_set_title (GTK_WINDOW (app.main_window), PROGNAME);
	gtk_window_set_wmclass (GTK_WINDOW (app.main_window), "main", BINNAME);
	gtk_widget_set_size_request (app.main_window, 0, 0);	//window as small as possible
	//default size 640 x 480
	e2_cache_int_register ("window-width", &app.main_window->allocation.width, 640);
	e2_cache_int_register ("window-height", &app.main_window->allocation.height, 480);
	gtk_window_set_default_size (GTK_WINDOW (app.main_window),
		app.main_window->allocation.width, app.main_window->allocation.height);
	gtk_window_set_resizable (GTK_WINDOW (app.main_window), TRUE);

	g_signal_connect (G_OBJECT (app.main_window), "show", 
		G_CALLBACK (_e2_window_show_cb), rt);
/*FIXME if the system needs shutdown-feedback. support 
	different processes for user- and system-initiated shutdowns?
	g_signal_connect (G_OBJECT (app.main_window),  ??, 
		G_CALLBACK (system_shutdown), NULL); */
	//trap this event so we can clean up gracefully
	g_signal_connect (G_OBJECT (app.main_window), "destroy",
		G_CALLBACK (e2_main_winclose_shutdown), NULL);

	app.hbox_main = gtk_hbox_new (FALSE, 0);
	app.vbox_main = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (app.main_window), app.hbox_main);
	gtk_box_pack_start_defaults (GTK_BOX (app.hbox_main), app.vbox_main);
	gtk_widget_show (app.hbox_main);
	gtk_widget_show (app.vbox_main);

	//file panes
	e2_cache_double_register ("file-pane-ratio-last", &rt->panes_paned_ratio_last, 0.55);
	e2_cache_double_register ("file-pane-ratio", &rt->panes_paned_ratio, 0.55);
	_e2_window_create_pane_boxes (rt);	//setup paned widget, with approximately 'correct' ratio
	
	//output pane
	e2_cache_double_register ("output-pane-ratio-last", &rt->output_paned_ratio_last, 0.85);
	e2_cache_double_register ("output-pane-ratio", &rt->output_paned_ratio, 0.85);
	GtkWidget *wid = e2_output_create (&app.output);
	//set visibility state
	app.output.visible = (rt->output_paned_ratio != 1.0);
	rt->output_paned = gtk_vpaned_new ();
	GTK_WIDGET_UNSET_FLAGS (rt->output_paned, GTK_CAN_FOCUS);
	gtk_paned_pack1 (GTK_PANED (rt->output_paned), rt->panes_outer_box, TRUE, TRUE);
	gtk_paned_pack2 (GTK_PANED (rt->output_paned), wid, TRUE, TRUE);

	gtk_paned_set_position (GTK_PANED (rt->output_paned),
		app.main_window->allocation.height * rt->output_paned_ratio);
	rt->output_paned_pos = gtk_paned_get_position (GTK_PANED (app.window.output_paned));

	gtk_widget_show (wid);
	gtk_widget_show (rt->output_paned);
	gtk_box_pack_start (GTK_BOX (app.vbox_main), rt->output_paned, TRUE, TRUE, 0);

	//do this registration here, rather that in the pane context,
	//as it covers both panes, and to avoid attempted re-registers
	//whenever the main window is re-created
	gint i; gint j;
	cols_data = NULL;
	E2_Cache *cache = e2_cache_list_register ("columns-data", &cols_data);
	//before writing cache, this will update list data
	cache->sync_func = e2_fileview_update_col_cachedata;
	if (cols_data == NULL)
	{  //there was no cache found, so setup defaults
		for (j = 0; j < 2; j++)
		{ for (i = 0; i < MAX_COLUMNS; i++) {
			stored_col_order[j][i] = i;
			col_width_store[j][i] = e2_all_columns[i].size;
		}}
	}
	else //column data found in cache
	{
		//get data from cached list (no data checks!!)
		//list format = pane1 { order, width, ...} pane2 {order, width, ... }
		GList *iterator = g_list_first (cols_data);
		for (j = 0; j < 2; j++)
		{ for (i = 0; i < MAX_COLUMNS; i++) {
			stored_col_order[j][i] = atoi (iterator->data);
			iterator = iterator->next;
			col_width_store[j][i] = atoi (iterator->data);
			iterator = iterator->next;
		}}
	}

	app.pane_current = &app.pane1;
	app.pane_other = &app.pane2;
	app.pane1.view = &app.pane1_view;
	app.pane2.view = &app.pane2_view;
	curr_view = &app.pane1_view;
	other_view = &app.pane2_view;

	e2_cache_int_register ("pane1-show-hidden", &app.pane1_view.show_hidden, FALSE);
	e2_cache_int_register ("pane2-show-hidden", &app.pane2_view.show_hidden, FALSE);

	//create all bindings, then register _("general"
	//(must be done before file-pane contents are created)
	e2_keybinding_register (_C(17), app.main_window);

	e2_pane_create (&app.pane1);
	e2_pane_create (&app.pane2);

	gtk_paned_pack1 (GTK_PANED (rt->panes_paned), app.pane1.outer_box, TRUE, TRUE);
	gtk_paned_pack2 (GTK_PANED (rt->panes_paned), app.pane2.outer_box, TRUE, TRUE);

	//show which pane (pane 1) is active
	e2_pane_flag_active ();
	//setup correct focus
	gtk_widget_grab_focus (curr_view->treeview);
	//other toolbars
	e2_toolbar_initialise (E2_BAR_TASK);
	e2_toolbar_create (&app.toolbar);
	e2_toolbar_initialise (E2_BAR_COMMAND);
	e2_toolbar_create (&app.commandbar);

	if (_e2_option_bool_get (app.commandbar.show)) //FIXME button could be on any bar
	{
		//show corrrect output pane toggle buttons according to cached pane ratio
		gchar *action_name;
		if (app.window.output_paned_ratio < 0.001)
		{
			action_name = g_strconcat(_A(8),".",_A(74),NULL);
			e2_toolbar_toggle_button_set_state (action_name, TRUE);
			g_free (action_name);
		}
		else if (app.window.output_paned_ratio > 0.999)
		{
			action_name = g_strconcat(_A(8),".",_A(72),NULL);
			e2_toolbar_toggle_button_set_state (action_name, TRUE);
			g_free (action_name);
		}
	}

	//status bar
	wid = gtk_table_new (1, 3, TRUE);
	gtk_box_pack_start (GTK_BOX (app.vbox_main), wid, FALSE, FALSE, 0);
	gtk_widget_show (wid);
	//find current user name (can't use environment string ...)
	struct passwd *pw_buf;
	gint myuid = getuid ();
	GString *who_where = g_string_new ("");
	while ((pw_buf = getpwent ()) != NULL)
		if ((gint) pw_buf->pw_uid == myuid)
			break;
	if (pw_buf != NULL)
		g_string_assign (who_where, pw_buf->pw_name);
	//find current machine name
	const gchar *env = g_getenv (_("HOSTNAME"));
	if (env != NULL)
		g_string_append_printf (who_where, _(" @ %s"), env);
	GtkWidget *wid2 = gtk_label_new (who_where->str);	
	g_string_free (who_where, TRUE);
//	gtk_misc_set_alignment (GTK_MISC (wid2), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE(wid), wid2, 0, 1, 0, 1,
		GTK_SHRINK, GTK_SHRINK,
		0, 0);
	gtk_widget_show (wid2);

	app.status_bar_label2 = gtk_label_new (" ");
//	gtk_misc_set_alignment (GTK_MISC (app.status_bar_label2), 0.5, 0.5);
//	gtk_box_pack_start (GTK_BOX (app.vbox_main), app.status_bar_label2, FALSE, TRUE, 0);
	gtk_table_attach (GTK_TABLE(wid), app.status_bar_label2, 1, 2, 0, 1,
		GTK_EXPAND, GTK_SHRINK,
		E2_PADDING_LARGE, 0);
	gtk_widget_show (app.status_bar_label2);
	//create hbox for any other status indicator eg for progress bars
	//not shown until needed
	app.status_bar_box3 = gtk_hbox_new (FALSE, E2_PADDING_SMALL);
	gtk_table_attach (GTK_TABLE(wid), app.status_bar_box3, 2, 3, 0, 1,
		GTK_EXPAND, GTK_SHRINK,
		E2_PADDING_SMALL, 0);
}
/**
 * @brief re-create main window
 *
 * This is used after config dialog, or detection of a config file that is updated 
 * (e.g. by another instance of e2), or a change of file-panes direction
 *
 * @param rt ptr to window data struct
 *
 * @return 
 **/
void e2_window_recreate (E2_WindowRuntime *rt)
{
	printd (DEBUG, "recreate main window");
	//make sure current cols data are used for the window recreate
	e2_fileview_update_col_cachedata ();
	//save current scroll positions
	GtkAdjustment *adj;
	gdouble curr_thumbx, other_thumbx, curr_thumby, other_thumby;
	gdouble curr_upperx, other_upperx, curr_uppery, other_uppery;
	// (no point in saving adj pointer(s) as they change when views rebuilt)
	adj = gtk_tree_view_get_hadjustment
		(GTK_TREE_VIEW (curr_view->treeview));
	curr_thumbx = gtk_adjustment_get_value (adj);
	curr_upperx = adj->upper;  //no fn for this
	adj = gtk_tree_view_get_hadjustment
		(GTK_TREE_VIEW (other_view->treeview));
	other_thumbx = gtk_adjustment_get_value (adj);
	other_upperx = adj->upper;  //no fn for this
	adj = gtk_tree_view_get_vadjustment
		(GTK_TREE_VIEW (curr_view->treeview));
	curr_thumby = gtk_adjustment_get_value (adj);
	curr_uppery = adj->upper;  //no fn for this
	adj = gtk_tree_view_get_vadjustment
		(GTK_TREE_VIEW (other_view->treeview));
	other_thumby = gtk_adjustment_get_value (adj);
	other_uppery = adj->upper;  //no fn for this
	//save current selections
#ifdef E2_INCLIST
	//FIXME
#else
	GtkTreeSelection *selection;
	GList *curr_selected_names = e2_fileview_log_selected_names (curr_view);
	GList *other_selected_names = e2_fileview_log_selected_names (other_view);
#endif
	//destroy the widgets
	//toolbar(s) may be outside rt->panes_outer_box, so destroy them independently
//	e2_toolbar_destroy (&app.pane1.toolbar);
//	e2_toolbar_destroy (&app.pane2.toolbar);
//	e2_toolbar_destroy (&app.toolbar);
//	e2_toolbar_destroy (&app.commandbar);
	gtk_widget_destroy (app.pane1.toolbar.toolbar_container);
	gtk_widget_destroy (app.pane2.toolbar.toolbar_container);
	gtk_widget_destroy (app.toolbar.toolbar_container);
	gtk_widget_destroy (app.commandbar.toolbar_container);
	gtk_widget_destroy (rt->panes_outer_box);
	//ensure all toolbar toggle buttons are recreated with their current state
	e2_toolbar_toggle_buttons_set_destroyed (NULL);

	e2_utils_update_gtk_settings ();	//why does this need to be redone, here ?

	e2_keybinding_register (_C(17), app.main_window);  //_("general"

	_e2_window_create_pane_boxes (rt);

	e2_pane_create_part (&app.pane1);
	e2_pane_create_part (&app.pane2);
	e2_toolbar_create (&app.toolbar);
	e2_toolbar_create (&app.commandbar);

	gtk_paned_pack1 (GTK_PANED (rt->panes_paned), app.pane1.outer_box, TRUE, TRUE);
	gtk_paned_pack2 (GTK_PANED (rt->panes_paned), app.pane2.outer_box, TRUE, TRUE);
	gtk_paned_pack1(GTK_PANED (rt->output_paned), rt->panes_outer_box, TRUE, TRUE);

	//mark the active pane
	e2_pane_flag_active ();

	WAIT_FOR_EVENTS; //make sure pane parameters are set, before the cb
	_e2_window_show_cb (NULL, rt);

	e2_alias_sync (&app.aliases);

	e2_pane_change_dir (&app.pane1, app.pane1.path);
	e2_pane_change_dir (&app.pane2, app.pane2.path);

	//set output visibility flag
	app.output.visible = (rt->output_paned_ratio != 1.0);

	e2_output_update_style (&app.output);
	//get rid of row 0 selected during the rebuild
//	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (curr_view->treeview));
//	gtk_tree_selection_unselect_all (selection);

#ifdef E2_INCLIST
	//FIXME
#else
	//re-select items
	GtkAdjustment *adj2;
	if (e2_fileview_reselect_names (curr_view, curr_selected_names))
	{	//there was a re-selection, so also restore old scroll position
		//the adjustments will have been recreated, with all parameters zeroed
		adj = gtk_tree_view_get_vadjustment
			(GTK_TREE_VIEW (curr_view->treeview));
		adj->upper = curr_uppery;
		gtk_adjustment_set_value (adj, curr_thumby);
		adj2 = gtk_tree_view_get_hadjustment
			(GTK_TREE_VIEW (curr_view->treeview));
		adj2->upper = curr_upperx;
		gtk_adjustment_set_value (adj2, curr_thumbx);

		gtk_adjustment_changed (adj);
		gtk_adjustment_value_changed (adj);
		gtk_adjustment_changed (adj2);
		gtk_adjustment_value_changed (adj2);
	}
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (other_view->treeview));
	gtk_tree_selection_unselect_all (selection);
	if (e2_fileview_reselect_names (other_view, other_selected_names))
	{
		adj = gtk_tree_view_get_vadjustment
			(GTK_TREE_VIEW (other_view->treeview));
		adj->upper = other_uppery;
		gtk_adjustment_set_value (adj, other_thumby);
		adj2 = gtk_tree_view_get_hadjustment
			(GTK_TREE_VIEW (other_view->treeview));
		adj2->upper = other_upperx;
		gtk_adjustment_set_value (adj2, other_thumbx);

		gtk_adjustment_changed (adj);
		gtk_adjustment_value_changed (adj);
		gtk_adjustment_changed (adj2);
		gtk_adjustment_value_changed (adj2);
	}
#endif //def E2_INCLIST
}
/**
 * @brief register main-window-related actions
 *
 * @return 
 **/
void e2_window_actions_register (void)
{	//(name strings freed as part of the registration process)
	gchar *action_name = g_strconcat(_A(8),".",_A(28),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM, 
		_e2_window_adjust_output_ratio_action, &app.window, TRUE);
	action_name = g_strconcat(_A(11),".",_A(28),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM, 
			_e2_window_adjust_pane_ratio_action, &app.window, TRUE);
	action_name = g_strconcat(_A(11),".",_A(59),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM, 
			e2_task_refresh, NULL, FALSE);
	action_name = g_strconcat(_A(11),".",_A(78),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM, 
			e2_filelist_disable_refresh, NULL, FALSE);
	action_name = g_strconcat(_A(11),".",_A(79),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM, 
			e2_filelist_enable_refresh, NULL, FALSE);
	action_name = g_strconcat(_A(11),".",_A(70),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
			e2_task_sync_dirs, NULL, FALSE);
	action_name = g_strconcat(_A(11),".",_A(71),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
			_e2_window_toggle_panes_direction, NULL, FALSE);
	action_name = g_strconcat(_A(11),".",_A(67),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
			e2_pane_activate_other, NULL, FALSE);
	action_name = g_strconcat(_A(8),".",_A(72),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_TOGGLE,
			_e2_window_toggle_visible_output, NULL, FALSE);
	action_name = g_strconcat(_A(8),".",_A(74),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_TOGGLE,
			_e2_window_toggle_full_output, NULL, FALSE);
	action_name = g_strconcat(_A(9),".",_A(74),NULL);
 	e2_action_register_simple (action_name, E2_ACTION_TYPE_TOGGLE,
			_e2_window_toggle_full_pane1, NULL, FALSE);
	action_name = g_strconcat(_A(10),".",_A(74),NULL);
 	e2_action_register_simple (action_name, E2_ACTION_TYPE_TOGGLE,
			_e2_window_toggle_full_pane2, NULL, FALSE);
}
