/* $Id: e2_menu.c 859 2005-06-09 23:38:30Z tpgww $
 * 
 * Copyright (C) 2004-2005 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file e2_menu.c 
 * @brief Menu utility functions
 *
 * This file contains utility and helper functions for menus.
 **/

#include "emelfm2.h"
#include <string.h>
#include "e2_menu.h"
#include "e2_plugins.h"

  /*******************/
 /***** callbacks *****/
/*******************/

/**
 * @brief callback to destroy a menu widget
 *
 * The menu widget @a menu is destroyed. This function might be used
 * as a callback for the @a "hide" signal of a menu.
 *
 * @param menu  the menu widget to destroy
 * @param data  this argument is ignored
 *
 * @return 
 **/
void e2_menu_destroy_cb (GtkWidget *menu, gpointer data)
{
	gtk_container_foreach (GTK_CONTAINER (menu), (GtkCallback) gtk_widget_destroy, NULL);
	gtk_widget_destroy (menu);
}

void e2_menu_control_cb (GtkWidget *menu_item, gpointer data)
{
	e2_option_connect (data, gtk_check_menu_item_get_active
		(GTK_CHECK_MENU_ITEM (menu_item)));
}

  /*****************/
 /***** public *****/
/****************/

/**
 * @brief add item to @a menu
 *
 * @param menu menu widget
 * @param label menu item text, optionally with mnemonic
 * @param icon custom iconfile path, or stock-icon identifier
 * @param tip tooltip string
 * @param func void* callback function for handling menu item selection, or NULL
 * @param data pointer to data to send to the callback
 *
 * @return the menu item widget
 **/
GtkWidget *e2_menu_add (GtkWidget *menu, gchar *label, gchar *icon,
	gchar *tip, void *func, gpointer data)
{
	GtkWidget *menu_item;
	if ((icon != NULL) && (e2_option_bool_get ("menu-show-icons")))
	{
		menu_item = gtk_image_menu_item_new_with_mnemonic 
			(label != NULL ? label : "");
		GtkWidget *image = e2_widget_get_icon (icon,
			e2_option_int_get ("menu-isize") + 1);
		if (image != NULL)
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), image);
	}
	else
		menu_item = gtk_menu_item_new_with_mnemonic (label);

	//turn on markup usage for menu items
	GList *children = gtk_container_get_children (GTK_CONTAINER (menu_item));
	GtkWidget *lab = children->data;
	gtk_label_set_use_markup (GTK_LABEL (lab), TRUE);
	
	if (func != NULL)
		g_signal_connect (G_OBJECT (menu_item), "activate", 
			G_CALLBACK (func), data);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show_all (menu_item);

	e2_widget_set_tooltip (NULL, menu_item, tip);
	return menu_item;
}
/**
 * @brief add action item to @a menu
 *
 * @param menu menu widget
 * @param label menu item text, optionally with mnemonic
 * @param icon custom iconfile path, or stock-icon identifier
 * @param tip tooltip string
 * @param action identifier of the form "_A(n)._A(m)"
 * @param arg argument(s) string for the action
 * @param free pointer to function that frees @a arg when the action is finished
 *
 * @return the menu item widget
 **/
GtkWidget *e2_menu_add_action (GtkWidget *menu, gchar *label, gchar *icon,
	gchar *tip, gchar *action, gchar *arg, gpointer free)
{
	E2_Action *act = e2_action_get (action);
	if (action == NULL)
		return NULL;
	E2_ActionRuntime *action_rt = e2_action_pack_runtime (act, arg, free);
	GtkWidget *menu_item = e2_menu_add (menu, label, icon, tip, e2_action_run,
		action_rt);
//	WHAT IS THIS FOR ?
//	g_object_set_data_full (G_OBJECT (menu_item), "free-callback-data",
//		action_rt, (GDestroyNotify) e2_action_free_runtime);
	return menu_item;
}
/**
 * @brief add check item to @a menu
 *
 * @param menu menu widget
 * @param label menu item text, optionally with mnemonic
 * @param state initial state of the created item, T/F
 * @param func void* callback function for handling menu item selection, or NULL
 * @param data pointer to data to send to the callback
 *
 * @return the menu item widget
 **/
GtkWidget *e2_menu_add_check (GtkWidget *menu,  gchar *label,
	gboolean state, void *func, gpointer data)
{
	GtkWidget *check = gtk_check_menu_item_new_with_mnemonic (label);
	if (menu != NULL)
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), check);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (check), state);
	if (func != NULL)
		g_signal_connect (G_OBJECT (check), "toggled", G_CALLBACK (func), data);
	gtk_widget_show (check);
	return check;
}
/**
 * @brief add radio item to @a menu
 *
 * @param menu menu widget
 * @param group pointer to gslist of radio group
 * @param label menu item text, optionally with mnemonic
 * @param state initial state of the created item, T/F
 * @param func void* callback function for handling menu-item selection, or NULL
 * @param data pointer to data to send to the callback
 *
 * @return the menu item widget
 **/
GtkWidget *e2_menu_add_radio (GtkWidget *menu, GSList **group, gchar *label,
	gboolean state, void *func, gpointer data)
{
	GtkWidget *radio = gtk_radio_menu_item_new_with_mnemonic (*group, label);
	*group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (radio));
	if (menu != NULL)
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), radio);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (radio), state);
	if (func != NULL)
		g_signal_connect (G_OBJECT (radio), "activate", G_CALLBACK (func), data);
	gtk_widget_show (radio);
	return radio;
}
/**
 * @brief add separator to @a menu
 *
 * @param menu menu widget
 *
 * @return the menu item widget
 **/
GtkWidget *e2_menu_add_separator (GtkWidget *menu)
{
	GtkWidget *menu_item = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show (menu_item);
	return menu_item;
}
/**
 * @brief add tear-off to @a menu
 *
 * @param menu menu widget
 *
 * @return the menu item widget
 **/
GtkWidget *e2_menu_add_tear_off (GtkWidget *menu)
{
	GtkWidget *menu_item = gtk_tearoff_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show (menu_item);
	return menu_item;
}
/**
 * @brief add sub-menu to @a menu
 *
 * @param menu menu widget
 * @param label_text string with the name to appear in the menu
 *
 * @return the sub-menu item widget
 **/
GtkWidget *e2_menu_add_submenu (GtkWidget *menu, gchar *label_text)
{
	GtkWidget *menu_item = gtk_menu_item_new_with_mnemonic (label_text);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	gtk_widget_show (menu_item);
	
	GtkWidget *submenu = gtk_menu_new ();
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item), submenu);
	gtk_widget_show (submenu);
	return submenu;
}
/**
 * @brief popup menu @a menu
 *
 * @param menu menu widget
 * @param button event button
 * @param time event time
 *
 * @return
 **/
void e2_menu_popup (GtkWidget *menu, gint button, guint32 time)
{
	g_signal_connect (G_OBJECT (menu), "selection-done",
		G_CALLBACK (e2_menu_destroy_cb), NULL);
	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, button, time);
}
/**
 * @brief
 *
 * @param menu menu widget
 * @param x UNUSED
 * @param y UNUSED
 * @param button event button
 * @param time event time
 *
 * @return
 **/
/* UNUSED
void _e2_menu_popup (GtkWidget *menu, gint x, gint y, gint button, guint32 time)
{
	g_signal_connect (G_OBJECT (menu), "selection-done",
		G_CALLBACK (e2_menu_destroy_cb), NULL);
	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, button, time);
}
*/
/**
 * @brief add items to bookmarks sub-menu, recursively if children exist
 *
 * @param action pointer to data for the pseudo-action being processed 
 * @param set pointer to bookmarks option data
 * @param menu widget to which the items will be added
 * @param iter pointer to iter to be used for interrogating the treemodel for @a set
 *
 * @return
 **/
void e2_menu_add_bookmarks_submenu (E2_Action *action,
	E2_OptionSet *set, GtkWidget *menu, GtkTreeIter *iter)
{
	//find which pane to use for the path, when adding marks
	gpointer pane = NULL;	//defualt = use current pane
	if (strstr (action->name, _A(9)) != NULL)
		pane = (GINT_TO_POINTER (1));
	else if (strstr (action->name, _A(10)) != NULL)
		pane = (GINT_TO_POINTER (2));

	do
	{
		gchar *name, *icon, *_tip, *_path;
		//don't support variables in label
//		gchar *_name, *icon, *_tip, *_path;
		gtk_tree_model_get (set->ex.tree.model, iter, 0, &name, 1, &icon, 2, &_tip, 3, &_path, -1);
//		gchar *name = e2_utils_replace_vars (_name);
		gchar *tip = e2_utils_replace_vars (_tip);
		gchar *path = e2_utils_replace_vars (_path);
//		g_free (_name);
		g_free (_tip);
		g_free (_path);

		GtkWidget *item;
		gboolean child = gtk_tree_model_iter_has_child (set->ex.tree.model, iter);
		if (child)
		{
			GtkWidget *menu2 = gtk_menu_new ();
			GtkTreeIter iter2;
			gtk_tree_model_iter_children (set->ex.tree.model, &iter2, iter);
			e2_menu_add_bookmarks_submenu (action, set, menu2, &iter2);	//recurse
			item = e2_menu_add (menu, name, icon, tip, NULL, NULL);
			gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu2);
			g_free (path);
		}
		else
		{
			item = e2_menu_add (menu, name, icon, tip, NULL, NULL);
		}
		g_signal_connect_data (G_OBJECT (item), "button-press-event",
			G_CALLBACK (e2_bookmark_click_cb),
			gtk_tree_model_get_path (set->ex.tree.model, iter),
			(GClosureNotify) gtk_tree_path_free, 0);
		//gtk sillyness fix, if we don't use connect_after for the activate signal
		//of the menuitem, the button-press-event will never occur
		if (!child)
		{
			E2_ActionRuntime *art = e2_action_pack_runtime (action, path, g_free);
			g_signal_connect_after (item, "activate", G_CALLBACK (e2_action_run), art);
			g_object_set_data_full (G_OBJECT (item), "free-callback-data-bookmark", art,
				(GDestroyNotify) e2_action_free_runtime);
		}
		g_object_set_data (G_OBJECT (item), "bookmark-path", path);
//		g_object_set_data_full (G_OBJECT (item), "bookmark-path", path,
//			(GDestroyNotify) g_free); //FIXME this causes crash, not sure why
		g_free (name);
		g_free (icon);
		g_free (tip);
		//signal which pane (or default) to use when adding an item
		g_object_set_data (G_OBJECT (item), "bookmark-pane", pane);
	} while ((gtk_tree_model_iter_next (set->ex.tree.model, iter)) == TRUE);
}

void e2_menu_create_plugins_menu (GtkWidget *menu)
{
	GList *tmp;

	for (tmp = cfg.plugins; tmp != NULL; tmp = tmp->next)
	{
		Plugin *p = tmp->data;
		if (p->show_in_menu)
			e2_menu_add (menu, p->menu_name, p->icon, p->description, e2_plugin_do_action, p);
	}
}
